//////////////////////////////////////////////////
//////////////////////////////////////////////////
// This source code is Copyright (2012) of      //
//  Jeppe R. Andersen and Jennifer M. Smillie   //
// and is distributed under the                 //
// Gnu Public License version 2                 //
// http://www.gnu.org/licenses/gpl-2.0.html     //
// You are allowed to distribute and alter the  //
// source under the conditions of the GPLv2     //
// as long as this copyright notice             //
// is unaltered and distributed with the source //
// Any use should comply with the               //
//             MCNET GUIDELINES                 //
//    for Event Generator Authors and Users     //
// as distributed with this source code         //
//////////////////////////////////////////////////
//////////////////////////////////////////////////
#include "Currents.h"

typedef COM current[4];

COM cdot(current j1, current j2)
{
  return j1[0]*j2[0]-j1[1]*j2[1]-j1[2]*j2[2]-j1[3]*j2[3];
}

void cmult(COM factor, current j1, current &cur)
{
  cur[0]=factor*j1[0];
  cur[1]=factor*j1[1];
  cur[2]=factor*j1[2];
  cur[3]=factor*j1[3];
}

void cadd(current j1, current j2, current j3, current j4, current j5, current &sum)
{
  sum[0]=j1[0]+j2[0]+j3[0]+j4[0]+j5[0];
  sum[1]=j1[1]+j2[1]+j3[1]+j4[1]+j5[1];
  sum[2]=j1[2]+j2[2]+j3[2]+j4[2]+j5[2];
  sum[3]=j1[3]+j2[3]+j3[3]+j4[3]+j5[3];
}

void cadd(current j1, current j2, current j3, current &sum)
{
  sum[0]=j1[0]+j2[0]+j3[0];
  sum[1]=j1[1]+j2[1]+j3[1];
  sum[2]=j1[2]+j2[2]+j3[2];
  sum[3]=j1[3]+j2[3]+j3[3];
}

double abs2(COM a)
{
  return (a*conj(a)).real();
}

void j (CLHEP::HepLorentzVector pout, bool helout, CLHEP::HepLorentzVector pin, bool helin,current &cur)
{
  
  cur[0]=0.;
  cur[1]=0.;
  cur[2]=0.;
  cur[3]=0.;
  
  double sqpop=sqrt(pout.plus());
  double sqpom=sqrt(pout.minus());
  COM poperp=pout.x()+COM(0,1)*pout.y();
  
  if (helout!=helin) {
    std::cerr<< "void j : Non-matching helicities\n";
  } else if (helout==false) { // negative helicity
    if (pin.plus()>pin.minus()) { // if forward
      double sqpip=sqrt(pin.plus());
      cur[0]=sqpop*sqpip;
      cur[1]=sqpom*sqpip*poperp/abs(poperp);
      cur[2]=-COM(0,1)*cur[1];
      cur[3]=cur[0];
    } else { // if backward
      double sqpim=sqrt(pin.minus());
      cur[0]=-sqpom*sqpim*poperp/abs(poperp);
      cur[1]=-sqpim*sqpop;
      cur[2]=COM(0,1)*cur[1];
      cur[3]=-cur[0];
    } 
  } else { // positive helicity
    if (pin.plus()>pin.minus()) { // if forward
      double sqpip=sqrt(pin.plus());
      cur[0]=sqpop*sqpip;
      cur[1]=sqpom*sqpip*conj(poperp)/abs(poperp);
      cur[2]=COM(0,1)*cur[1];
      cur[3]=cur[0];
    } else { // if backward
      double sqpim=sqrt(pin.minus());
      cur[0]=-sqpom*sqpim*conj(poperp)/abs(poperp);
      cur[1]=-sqpim*sqpop;
      cur[2]=-COM(0,1)*cur[1];
      cur[3]=-cur[0];
    } 
  }
}

void jio (CLHEP::HepLorentzVector pin, bool helin, CLHEP::HepLorentzVector pout, bool helout,current &cur)
{
  
  cur[0]=0.;
  cur[1]=0.;
  cur[2]=0.;
  cur[3]=0.;
  
  double sqpop=sqrt(pout.plus());
  double sqpom=sqrt(pout.minus());
  COM poperp=pout.x()+COM(0,1)*pout.y();
  
  if (helout!=helin) {
    std::cerr<< "void j : Non-matching helicities\n";
  } else if (helout==false) { // negative helicity
    if (pin.plus()>pin.minus()) { // if forward
      double sqpip=sqrt(pin.plus());
      cur[0]=sqpop*sqpip;
      cur[1]=sqpom*sqpip*conj(poperp)/abs(poperp);
      cur[2]=COM(0,1)*cur[1];
      cur[3]=cur[0];
    } else { // if backward
      double sqpim=sqrt(pin.minus());
      cur[0]=-sqpom*sqpim*conj(poperp)/abs(poperp);
      cur[1]=-sqpim*sqpop;
      cur[2]=-COM(0,1)*cur[1];
      cur[3]=-cur[0];
    } 
  } else { // positive helicity
    if (pin.plus()>pin.minus()) { // if forward
      double sqpip=sqrt(pin.plus());
      cur[0]=sqpop*sqpip;
      cur[1]=sqpom*sqpip*poperp/abs(poperp);
      cur[2]=-COM(0,1)*cur[1];
      cur[3]=cur[0];
    } else { // if backward
      double sqpim=sqrt(pin.minus());
      cur[0]=-sqpom*sqpim*poperp/abs(poperp);
      cur[1]=-sqpim*sqpop;
      cur[2]=COM(0,1)*cur[1];
      cur[3]=-cur[0];
    } 
  }
}

void joo (CLHEP::HepLorentzVector pi, bool heli, CLHEP::HepLorentzVector pj, bool helj, current &cur)
{
 
  if (heli!=helj) {
    std::cerr<< "void j : Non-matching helicities\n";
  } else if (heli==true) { // negative helicity
    CLHEP::HepLorentzVector dummy;
    dummy=pi;
    pi=pj;
    pj=dummy;
  }
  double sqpjp=sqrt(pj.plus());
  double sqpjm=sqrt(pj.minus());
  double sqpip=sqrt(pi.plus());
  double sqpim=sqrt(pi.minus());
  COM piperp=pi.x()+COM(0,1)*pi.y();
  COM pjperp=pj.x()+COM(0,1)*pj.y();
  COM phasei=piperp/abs(piperp);
  COM phasej=pjperp/abs(pjperp);

  cur[0]=sqpim*sqpjm*phasei*conj(phasej)+sqpip*sqpjp;
  cur[1]=sqpim*sqpjp*phasei+sqpip*sqpjm*conj(phasej);
  cur[2]=-COM(0,1)*(sqpim*sqpjp*phasei-sqpip*sqpjm*conj(phasej));
  cur[3]=-sqpim*sqpjm*phasei*conj(phasej)+sqpip*sqpjp;
}

double jM2 (CLHEP::HepLorentzVector p1out, bool hel1out, CLHEP::HepLorentzVector p1in, bool hel1in, CLHEP::HepLorentzVector p2out, bool hel2out, CLHEP::HepLorentzVector p2in, bool hel2in)
{
  CLHEP::HepLorentzVector q1=p1in-p1out;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);
  current C1,C2;
  j (p1out,hel1out,p1in,hel1in, C1);
  j (p2out,hel2out,p2in,hel2in, C2);
  
  std::cout << "# From Currents, C1 : ("<<C1[0]<<","<<C1[1]<<","<<C1[2]<<","<<C1[3]<<"\n";
  std::cout << "# From Currents, C2 : ("<<C2[0]<<","<<C2[1]<<","<<C2[2]<<","<<C2[3]<<"\n";

  COM M=cdot(C1,C2);

  return (M*conj(M)).real()/(q1.m2()*q2.m2());
}

void jW (CLHEP::HepLorentzVector pout, bool helout, CLHEP::HepLorentzVector pe, bool hele, CLHEP::HepLorentzVector pnu, bool helnu, CLHEP::HepLorentzVector pin, bool helin, current cur)
{
  // NOTA BENE: Conventions for W+ --> e+ nu, so that nu is lepton(6), e is anti-lepton(5)
  // Need to swap e and nu for events with W- --> e- nubar!
  if (helin==helout && hele==helnu) {
    CLHEP::HepLorentzVector qa=pout+pe+pnu;
    CLHEP::HepLorentzVector qb=pin-pe-pnu;
    double ta(qa.m2()),tb(qb.m2());
    
    current t65,vout,vin,temp2,temp3,temp5;
    joo(pnu,helnu,pe,hele,t65);
    vout[0]=pout.e();
    vout[1]=pout.x();
    vout[2]=pout.y();
    vout[3]=pout.z();
    vin[0]=pin.e();
    vin[1]=pin.x();
    vin[2]=pin.y();
    vin[3]=pin.z();
    
    COM brac615=cdot(t65,vout);
    COM brac645=cdot(t65,vin);
    
    // prod1565 and prod6465 are zero for Ws (not Zs)!!
    // noalias(temp)=prod(trans(CurrentOutOut(pout,helout,pnu,helout)),metric);
    joo(pout,helout,pnu,helout,temp2);
    // noalias(temp2)=prod(temp,ctemp);
    COM prod1665=cdot(temp2,t65);
    // noalias(temp)=prod(trans(Current(pe,helin,pin,helin)),metric);
    // noalias(temp2)=prod(temp,ctemp);
    j(pe,helin,pin,helin,temp3);
    COM prod5465=cdot(temp3,t65);
    // noalias(temp)=prod(trans(Current(pnu,helin,pin,helin)),metric);
    // noalias(temp2)=prod(temp,ctemp);

    joo(pout,helout,pe,helout,temp2);
    j(pnu,helnu,pin,helin,temp3);
    j(pout,helout,pin,helin,temp5);

    current term1,term2,term3,sum;
    cmult(2.*brac615/ta+2.*brac645/tb,temp5,term1);
    cmult(prod1665/ta,temp3,term2);
    cmult(-prod5465/tb,temp2,term3);

   //    cur=((2.*brac615*Current(pout,helout,pin,helin)+prod1565*Current(pe,helin,pin,helin)+prod1665*Current(pnu,helin,pin,helin))/ta + (2.*brac645*Current(pout,helout,pin,helin)-prod5465*CurrentOutOut(pout,helout,pe,helout)-prod6465*CurrentOutOut(pout,helout,pnu,helout))/tb);
    //    cur=((2.*brac615*temp5+prod1565*temp3+prod1665*temp4)/ta + (2.*brac645*temp5-prod5465*temp1-prod6465*temp2)/tb);
    cadd(term1,term2,term3,sum);
    //    std::cout<<"sum: ("<<sum[0]<<","<<sum[1]<<","<<sum[2]<<","<<sum[3]<<")\n";
    cur[0]=sum[0];
    cur[1]=sum[1];
    cur[2]=sum[2];
    cur[3]=sum[3];
  }
}

void jWbar (CLHEP::HepLorentzVector pout, bool helout, CLHEP::HepLorentzVector pe, bool hele, CLHEP::HepLorentzVector pnu, bool helnu, CLHEP::HepLorentzVector pin, bool helin, current cur)
{
  // NOTA BENE: Conventions for W+ --> e+ nu, so that nu is lepton(6), e is anti-lepton(5)
  // Need to swap e and nu for events with W- --> e- nubar!
  if (helin==helout && hele==helnu) {
    CLHEP::HepLorentzVector qa=pout+pe+pnu;
    CLHEP::HepLorentzVector qb=pin-pe-pnu;
    double ta(qa.m2()),tb(qb.m2());
    
    current t65,vout,vin,temp2,temp3,temp5;
    joo(pnu,helnu,pe,hele,t65);
    vout[0]=pout.e();
    vout[1]=pout.x();
    vout[2]=pout.y();
    vout[3]=pout.z();
    vin[0]=pin.e();
    vin[1]=pin.x();
    vin[2]=pin.y();
    vin[3]=pin.z();
    
    COM brac615=cdot(t65,vout);
    COM brac645=cdot(t65,vin);
    
    // prod1565 and prod6465 are zero for Ws (not Zs)!!
    joo(pe,helout,pout,helout,temp2);  //  temp2 is <5|alpha|1>
    COM prod5165=cdot(temp2,t65);
    jio(pin,helin,pnu,helin,temp3);      // temp3 is <4|alpha|6>
    COM prod4665=cdot(temp3,t65);

    joo(pnu,helout,pout,helout,temp2);  // temp2 is now <6|mu|1>
    jio(pin,helin,pe,helin,temp3);        // temp3 is now <4|mu|5>
    jio(pin,helin,pout,helout,temp5);     //  temp5 is <4|mu|1>

    current term1,term2,term3,sum;
    cmult(-2.*brac615/ta-2.*brac645/tb,temp5,term1);
    cmult(-prod5165/ta,temp3,term2);
    cmult(prod4665/tb,temp2,term3);

   //    cur=((2.*brac615*Current(pout,helout,pin,helin)+prod1565*Current(pe,helin,pin,helin)+prod1665*Current(pnu,helin,pin,helin))/ta + (2.*brac645*Current(pout,helout,pin,helin)-prod5465*CurrentOutOut(pout,helout,pe,helout)-prod6465*CurrentOutOut(pout,helout,pnu,helout))/tb);
    //    cur=((2.*brac615*temp5+prod1565*temp3+prod1665*temp4)/ta + (2.*brac645*temp5-prod5465*temp1-prod6465*temp2)/tb);
    cadd(term1,term2,term3,sum);
    //    std::cout<<"term1: ("<<temp5[0]<<"  "<<temp5[1]<<"  "<<temp5[2]<<"  "<<temp5[3]<<")"<<std::endl;
    //    std::cout<<"sum: ("<<sum[0]<<","<<sum[1]<<","<<sum[2]<<","<<sum[3]<<")\n";
    cur[0]=sum[0];
    cur[1]=sum[1];
    cur[2]=sum[2];
    cur[3]=sum[3];
  }
}

double jMWqQ (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe, CLHEP::HepLorentzVector pnu,CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
// Calculates the square of the current contractions for qQ->qenuQ scattering
// p1: quark (with W emittance)
// p2: Quark
{
  current mj1m,mj2p,mj2m;
  CLHEP::HepLorentzVector q1=p1in-p1out-pe-pnu;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);

  jW(p1out,false,pe,false,pnu,false,p1in,false,mj1m);
  j(p2out,true,p2in,true,mj2p);
  j(p2out,false,p2in,false,mj2m);
  
  // std::cout<<"jMW1: ("<<mj1m[0]<<","<<mj1m[1]<<","<<mj1m[2]<<","<<mj1m[3]<<")\n";
  // std::cout<<"jMW2: ("<<mj2p[0]<<","<<mj2p[1]<<","<<mj2p[2]<<","<<mj2p[3]<<")\n";
  // std::cout<<"jMW3: ("<<mj2m[0]<<","<<mj2m[1]<<","<<mj2m[2]<<","<<mj2m[3]<<")\n";

  // mj1m.mj2p 

  COM Mmp=cdot(mj1m,mj2p);

  // mj1m.mj2m
  COM Mmm=cdot(mj1m,mj2m);

  // sum of spinor strings ||^2
  double a2Mmp=abs2(Mmp);
  double a2Mmm=abs2(Mmm);

//   // Leave division by colour and Helicity avg until Tree files
  // Leave multi. of couplings to later
  // Multiply by Cf^2
  return (4./3.)*(4./3.)*(a2Mmp+a2Mmm)/(q1.m2()*q2.m2());

}

double jMWqQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe, CLHEP::HepLorentzVector pnu,CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
// Calculates the square of the current contractions for qQ->qenuQ scattering
// p1: quark (with W emittance)
// p2: Quark
{
  current mj1m,mj2p,mj2m;
  CLHEP::HepLorentzVector q1=p1in-p1out-pe-pnu;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);

  jW(p1out,false,pe,false,pnu,false,p1in,false,mj1m);
  jio(p2in,true,p2out,true,mj2p);
  jio(p2in,false,p2out,false,mj2m);
  
  // std::cout<<"jMW1: ("<<mj1m[0]<<","<<mj1m[1]<<","<<mj1m[2]<<","<<mj1m[3]<<")\n";
  // std::cout<<"jMW2: ("<<mj2p[0]<<","<<mj2p[1]<<","<<mj2p[2]<<","<<mj2p[3]<<")\n";
  // std::cout<<"jMW3: ("<<mj2m[0]<<","<<mj2m[1]<<","<<mj2m[2]<<","<<mj2m[3]<<")\n";

  // mj1m.mj2p 

  COM Mmp=cdot(mj1m,mj2p);

  // mj1m.mj2m
  COM Mmm=cdot(mj1m,mj2m);

  // sum of spinor strings ||^2
  double a2Mmp=abs2(Mmp);
  double a2Mmm=abs2(Mmm);

//   // Leave division by colour and Helicity avg until Tree files
  // Leave multi. of couplings to later
  // Multiply by Cf^2
  return (4./3.)*(4./3.)*(a2Mmp+a2Mmm)/(q1.m2()*q2.m2());

}

double jMWqbarQ (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe, CLHEP::HepLorentzVector pnu,CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
// Calculates the square of the current contractions for qQ->qenuQ scattering
// p1: quark (with W emittance)
// p2: Quark
{
  current mj1m,mj2p,mj2m;
  CLHEP::HepLorentzVector q1=p1in-p1out-pe-pnu;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);

  jWbar(p1out,false,pe,false,pnu,false,p1in,false,mj1m);
  j(p2out,true,p2in,true,mj2p);
  j(p2out,false,p2in,false,mj2m);
  
  // std::cout<<"jMW1: ("<<mj1m[0]<<","<<mj1m[1]<<","<<mj1m[2]<<","<<mj1m[3]<<")\n";
  // std::cout<<"jMW2: ("<<mj2p[0]<<","<<mj2p[1]<<","<<mj2p[2]<<","<<mj2p[3]<<")\n";
  // std::cout<<"jMW3: ("<<mj2m[0]<<","<<mj2m[1]<<","<<mj2m[2]<<","<<mj2m[3]<<")\n";

  // mj1m.mj2p 

  COM Mmp=cdot(mj1m,mj2p);

  // mj1m.mj2m
  COM Mmm=cdot(mj1m,mj2m);

  // sum of spinor strings ||^2
  double a2Mmp=abs2(Mmp);
  double a2Mmm=abs2(Mmm);

//   // Leave division by colour and Helicity avg until Tree files
  // Leave multi. of couplings to later
  // Multiply by Cf^2
  return (4./3.)*(4./3.)*(a2Mmp+a2Mmm)/(q1.m2()*q2.m2());

}

double jMWqbarQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe, CLHEP::HepLorentzVector pnu,CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
// Calculates the square of the current contractions for qQ->qenuQ scattering
// p1: quark (with W emittance)
// p2: Quark
{
  current mj1m,mj2p,mj2m;
  CLHEP::HepLorentzVector q1=p1in-p1out-pe-pnu;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);

  jWbar(p1out,false,pe,false,pnu,false,p1in,false,mj1m);
  jio(p2in,true,p2out,true,mj2p);
  jio(p2in,false,p2out,false,mj2m);
  
  // std::cout<<"jMW1: ("<<mj1m[0]<<","<<mj1m[1]<<","<<mj1m[2]<<","<<mj1m[3]<<")\n";
  // std::cout<<"jMW2: ("<<mj2p[0]<<","<<mj2p[1]<<","<<mj2p[2]<<","<<mj2p[3]<<")\n";
  // std::cout<<"jMW3: ("<<mj2m[0]<<","<<mj2m[1]<<","<<mj2m[2]<<","<<mj2m[3]<<")\n";

  // mj1m.mj2p 

  COM Mmp=cdot(mj1m,mj2p);

  // mj1m.mj2m
  COM Mmm=cdot(mj1m,mj2m);

  // sum of spinor strings ||^2
  double a2Mmp=abs2(Mmp);
  double a2Mmm=abs2(Mmm);

//   // Leave division by colour and Helicity avg until Tree files
  // Leave multi. of couplings to later
  // Multiply by Cf^2
  return (4./3.)*(4./3.)*(a2Mmp+a2Mmm)/(q1.m2()*q2.m2());

}

double jMWqg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe, CLHEP::HepLorentzVector pnu,CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
// Calculates the square of the current contractions for qg->qenug scattering
// p1: quark
// p2: gluon
{
  CLHEP::HepLorentzVector q1=p1in-p1out-pe-pnu;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);
  current mj1m,mj2p,mj2m;

  jW(p1out,false,pe,false,pnu,false,p1in,false,mj1m);

  j(p2out,true,p2in,true,mj2p);
  j(p2out,false,p2in,false,mj2m);
  
  // mj1m.mj2p 

  COM Mmp=cdot(mj1m,mj2p);

  // mj1m.mj2m
  COM Mmm=cdot(mj1m,mj2m);

  double ratio; // p2-/pb- in the notes
  // if (p2in.plus()>0) // if the gluon is the positive
  if (p2in.pz()>0.) // if the gluon is the positive
    ratio=p2out.plus()/p2in.plus();
  else // the gluon is the negative
    ratio=p2out.minus()/p2in.minus();
    

  double nonflipcolourmult=(1.-1./9.)/2.*(ratio+1./ratio)+1./9.;
  
  // sum of spinor strings ||^2
  double a2Mmp=abs2(Mmp);
  double a2Mmm=abs2(Mmm);
  double sst = nonflipcolourmult*(a2Mmp+a2Mmm);
  // double sstsave=sst;

//   // Leave division by colour and Helicity avg until Tree files
  // Leave multi. of couplings to later
  // Multiply by Cf*Ca=4
  return 4.*sst/(q1.m2()*q2.m2());

}

double jMWqbarg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector pe, CLHEP::HepLorentzVector pnu,CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
// Calculates the square of the current contractions for qg->qenug scattering
// p1: quark
// p2: gluon
{
  CLHEP::HepLorentzVector q1=p1in-p1out-pe-pnu;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);
  current mj1m,mj2p,mj2m;

  jWbar(p1out,false,pe,false,pnu,false,p1in,false,mj1m);

  j(p2out,true,p2in,true,mj2p);
  j(p2out,false,p2in,false,mj2m);
  
  // mj1m.mj2p 

  COM Mmp=cdot(mj1m,mj2p);

  // mj1m.mj2m
  COM Mmm=cdot(mj1m,mj2m);

  double ratio; // p2-/pb- in the notes
  // if (p2in.plus()>0) // if the gluon is the positive
  if (p2in.pz()>0.) // if the gluon is the positive
    ratio=p2out.plus()/p2in.plus();
  else // the gluon is the negative
    ratio=p2out.minus()/p2in.minus();
    

  double nonflipcolourmult=(1.-1./9.)/2.*(ratio+1./ratio)+1./9.;
  
  // sum of spinor strings ||^2
  double a2Mmp=abs2(Mmp);
  double a2Mmm=abs2(Mmm);
  double sst = nonflipcolourmult*(a2Mmp+a2Mmm);
  // double sstsave=sst;

//   // Leave division by colour and Helicity avg until Tree files
  // Leave multi. of couplings to later
  // Multiply by Cf*Ca=4
  return 4.*sst/(q1.m2()*q2.m2());

}

double jM2qQ (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
{
  //  std::cerr<<"Current: "<<p1out<<"  "<<p1in<<"  "<<p2out<<"  "<<p2in<<std::endl;

  CLHEP::HepLorentzVector q1=p1in-p1out;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);
  //  std::cerr<<"Current:  "<<q1.m2()<<"  "<<q2.m2()<<std::endl;
  current mj1m,mj1p,mj2m,mj2p;
  j(p1out,true,p1in,true,mj1p);
  j(p1out,false,p1in,false,mj1m);
  j(p2out,true,p2in,true,mj2p);
  j(p2out,false,p2in,false,mj2m);
  
  COM Mmp=cdot(mj1m,mj2p);
  COM Mmm=cdot(mj1m,mj2m);
  COM Mpp=cdot(mj1p,mj2p);
  COM Mpm=cdot(mj1p,mj2m);

  double sst=abs2(Mmm)+abs2(Mmp)+abs2(Mpp)+abs2(Mpm);

  // Multiply by Cf^2
  return (4./3.)*(4./3.)*(sst)/(q1.m2()*q2.m2());
}

double jM2qQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
{
  CLHEP::HepLorentzVector q1=p1in-p1out;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);
  current mj1m,mj1p,mj2m,mj2p;
  j(p1out,true,p1in,true,mj1p);
  j(p1out,false,p1in,false,mj1m);
  jio(p2in,true,p2out,true,mj2p);
  jio(p2in,false,p2out,false,mj2m);
  
  COM Mmp=cdot(mj1m,mj2p);
  COM Mmm=cdot(mj1m,mj2m);
  COM Mpp=cdot(mj1p,mj2p);
  COM Mpm=cdot(mj1p,mj2m);

  double sumsq=abs2(Mmm)+abs2(Mmp)+abs2(Mpp)+abs2(Mpm);

  // Multiply by Cf^2
  return (4./3.)*(4./3.)*(sumsq)/(q1.m2()*q2.m2());
}

double jM2qbarQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
{
  CLHEP::HepLorentzVector q1=p1in-p1out;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);
  current mj1m,mj1p,mj2m,mj2p;
  jio(p1in,true,p1out,true,mj1p);
  jio(p1in,false,p1out,false,mj1m);
  jio(p2in,true,p2out,true,mj2p);
  jio(p2in,false,p2out,false,mj2m);
  
  COM Mmp=cdot(mj1m,mj2p);
  COM Mmm=cdot(mj1m,mj2m);
  COM Mpp=cdot(mj1p,mj2p);
  COM Mpm=cdot(mj1p,mj2m);

  double sumsq=abs2(Mmm)+abs2(Mmp)+abs2(Mpp)+abs2(Mpm);

  // Multiply by Cf^2
  return (4./3.)*(4./3.)*(sumsq)/(q1.m2()*q2.m2());
}

double jM2qg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
// Calculates the square of the current contractions for qg scattering
// p1: quark
// p2: gluon
{
  CLHEP::HepLorentzVector q1=p1in-p1out;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);

  current mj1m,mj1p,mj2m,mj2p;
  j(p1out,true,p1in,true,mj1p);
  j(p1out,false,p1in,false,mj1m);
  j(p2out,true,p2in,true,mj2p);
  j(p2out,false,p2in,false,mj2m);

  COM Mmp=cdot(mj1m,mj2p);
  COM Mmm=cdot(mj1m,mj2m);
  COM Mpp=cdot(mj1p,mj2p);
  COM Mpm=cdot(mj1p,mj2m);

  double ratio; // p2-/pb- in the notes
  // if (p2in.plus()>0) // if the gluon is the positive
  if (p2in.pz()>0.) // if the gluon is the positive
    ratio=p2out.plus()/p2in.plus();
  else // the gluon is the negative
    ratio=p2out.minus()/p2in.minus();
    
  double nonflipcolourmult=(1.-1./9.)/2.*(ratio+1./ratio)+1./9.;
  
  // sum of spinor strings ||^2
  double a2Mmp=abs2(Mmp);
  double a2Mmm=abs2(Mmm);
  double a2Mpp=abs2(Mpp);
  double a2Mpm=abs2(Mpm);
  double sst = nonflipcolourmult*(a2Mpp+a2Mpm+a2Mmp+a2Mmm);
  // double sstsave=sst;

   // std::cout <<"ratio: "<<sst/sstsave<<std::endl;
  // Cf*Ca=4
  return 4.*sst/(q1.m2()*q2.m2());

}

double jM2qbarg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
// Calculates the square of the current contractions for qg scattering
// p1: quark
// p2: gluon
{
  CLHEP::HepLorentzVector q1=p1in-p1out;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);

  current mj1m,mj1p,mj2m,mj2p;
  jio(p1in,true,p1out,true,mj1p);
  jio(p1in,false,p1out,false,mj1m);
  j(p2out,true,p2in,true,mj2p);
  j(p2out,false,p2in,false,mj2m);

  COM Mmp=cdot(mj1m,mj2p);
  COM Mmm=cdot(mj1m,mj2m);
  COM Mpp=cdot(mj1p,mj2p);
  COM Mpm=cdot(mj1p,mj2m);

  double ratio; // p2-/pb- in the notes
  // if (p2in.plus()>0) // if the gluon is the positive
  if (p2in.pz()>0.) // if the gluon is the positive
    ratio=p2out.plus()/p2in.plus();
  else // the gluon is the negative
    ratio=p2out.minus()/p2in.minus();
    
  double nonflipcolourmult=(1.-1./9.)/2.*(ratio+1./ratio)+1./9.;
  
  // sum of spinor strings ||^2
  double a2Mmp=abs2(Mmp);
  double a2Mmm=abs2(Mmm);
  double a2Mpp=abs2(Mpp);
  double a2Mpm=abs2(Mpm);
  double sst = nonflipcolourmult*(a2Mpp+a2Mpm+a2Mmp+a2Mmm);
  // double sstsave=sst;

   // std::cout <<"ratio: "<<sst/sstsave<<std::endl;
  // Cf*Ca=4
  return 4.*sst/(q1.m2()*q2.m2());

}

double jM2gg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in)
// Calculates the square of the current contractions for gg scattering
// p1: gluon
// p2: gluon
{
  CLHEP::HepLorentzVector q1=p1in-p1out;
  CLHEP::HepLorentzVector q2=-(p2in-p2out);

  current mj1m,mj1p,mj2m,mj2p;
  j(p1out,true,p1in,true,mj1p);
  j(p1out,false,p1in,false,mj1m);
  j(p2out,true,p2in,true,mj2p);
  j(p2out,false,p2in,false,mj2m);

  COM Mmp=cdot(mj1m,mj2p);
  COM Mmm=cdot(mj1m,mj2m);
  COM Mpp=cdot(mj1p,mj2p);
  COM Mpm=cdot(mj1p,mj2m);

  double ratio1; // p1-/pa- in the notes
  // if (p1in.plus()>1.) // if the gluon is the positive. Should have been a
  // 		      // test against 0, but 1. is better
  if (p1in.pz()>0.) // a much better test
    ratio1=p1out.plus()/p1in.plus();
  else // the gluon is the negative
    ratio1=p1out.minus()/p1in.minus();
  double nonflipcolourmult1=(1.-1./9.)/2.*(ratio1+1./ratio1)+1./9.;
  
  double ratio2; // p2-/pb- in the notes
  if (p2in.pz()>0.) // a much better test
    ratio2=p2out.plus()/p2in.plus();
  else // the gluon is the negative
    ratio2=p2out.minus()/p2in.minus();
  double nonflipcolourmult2=(1.-1./9.)/2.*(ratio2+1./ratio2)+1./9.;
  
  // sum of spinor strings ||^2
  double a2Mmp=abs2(Mmp);
  double a2Mmm=abs2(Mmm);
  double a2Mpp=abs2(Mpp);
  double a2Mpm=abs2(Mpm);
  double sst = nonflipcolourmult1*nonflipcolourmult2*(a2Mpp+a2Mpm+a2Mmp+a2Mmm);
  // double sstsave=sst;
   // std::cout <<"ratio: "<<sst/sstsave<<std::endl;
  // Ca*Ca=9
  return 9.*sst/(q1.m2()*q2.m2());

}

double MH2helper(current C1, current C2, current q1, current q2)
{
  COM M;
  COM temp1,temp2;
  // First the C1.q2 * C2.q1 - part
  temp1=cdot(C1,q2);
  temp2=cdot(C2,q1);
  M=temp1*temp2;
  
  // Then the C1.C2 * q1.q2
  temp1=cdot(C1,C2);
  temp2=cdot(q1,q2);
  M-=temp1*temp2;

  return (M*conj(M)).real();
}

COM cHdot(current C1, current C2, current q1, current q2)
{
  return cdot(C1,C2)*cdot(q1,q2)-cdot(C1,q2)*cdot(C2,q1);
}

double MH2qQ (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector q2)
{
//   CLHEP::HepLorentzVector q1=p1in-p1out;
//   CLHEP::HepLorentzVector q2=-(p2in-p2out);
  current j1p,j1m,j2p,j2m,q1v,q2v;

  j (p1out,true,p1in,true,j1p);
  j (p1out,false,p1in,false,j1m);

  j (p2out,true,p2in,true,j2p);
  j (p2out,false,p2in,false,j2m);


  q1v[0]=q1.e();
  q1v[1]=q1.x();
  q1v[2]=q1.y();
  q1v[3]=q1.z();

  q2v[0]=q2.e();
  q2v[1]=q2.x();
  q2v[2]=q2.y();
  q2v[3]=q2.z();

  COM Mmp=cHdot(j1m,j2p,q1v,q2v);
  COM Mmm=cHdot(j1m,j2m,q1v,q2v);
  COM Mpp=cHdot(j1p,j2p,q1v,q2v);
  COM Mpm=cHdot(j1p,j2m,q1v,q2v);

  double sst=abs2(Mmp)+abs2(Mmm)+abs2(Mpp)+abs2(Mpm);
  // return (4./3.)*(4./3.)*sst/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
  return sst/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
}

double MH2qQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector q2)
{
//   CLHEP::HepLorentzVector q1=p1in-p1out;
//   CLHEP::HepLorentzVector q2=-(p2in-p2out);
  current j1p,j1m,j2p,j2m,q1v,q2v;

  j (p1out,true,p1in,true,j1p);
  j (p1out,false,p1in,false,j1m);

  jio (p2in,true,p2out,true,j2p);
  jio (p2in,false,p2out,false,j2m);


  q1v[0]=q1.e();
  q1v[1]=q1.x();
  q1v[2]=q1.y();
  q1v[3]=q1.z();

  q2v[0]=q2.e();
  q2v[1]=q2.x();
  q2v[2]=q2.y();
  q2v[3]=q2.z();

  COM Mmp=cHdot(j1m,j2p,q1v,q2v);
  COM Mmm=cHdot(j1m,j2m,q1v,q2v);
  COM Mpp=cHdot(j1p,j2p,q1v,q2v);
  COM Mpm=cHdot(j1p,j2m,q1v,q2v);

  double sst=abs2(Mmp)+abs2(Mmm)+abs2(Mpp)+abs2(Mpm);
  // return (4./3.)*(4./3.)*sst/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
  return sst/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
}

double MH2qbarQ (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector q2)
{
//   CLHEP::HepLorentzVector q1=p1in-p1out;
//   CLHEP::HepLorentzVector q2=-(p2in-p2out);
  current j1p,j1m,j2p,j2m,q1v,q2v;

  jio (p1in,true,p1out,true,j1p);
  jio (p1in,false,p1out,false,j1m);

  j (p2out,true,p2in,true,j2p);
  j (p2out,false,p2in,false,j2m);


  q1v[0]=q1.e();
  q1v[1]=q1.x();
  q1v[2]=q1.y();
  q1v[3]=q1.z();

  q2v[0]=q2.e();
  q2v[1]=q2.x();
  q2v[2]=q2.y();
  q2v[3]=q2.z();

  COM Mmp=cHdot(j1m,j2p,q1v,q2v);
  COM Mmm=cHdot(j1m,j2m,q1v,q2v);
  COM Mpp=cHdot(j1p,j2p,q1v,q2v);
  COM Mpm=cHdot(j1p,j2m,q1v,q2v);

  double sst=abs2(Mmp)+abs2(Mmm)+abs2(Mpp)+abs2(Mpm);
  // return (4./3.)*(4./3.)*sst/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
  return sst/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
}

double MH2qbarQbar (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector q2)
{
//   CLHEP::HepLorentzVector q1=p1in-p1out;
//   CLHEP::HepLorentzVector q2=-(p2in-p2out);
  current j1p,j1m,j2p,j2m,q1v,q2v;

  jio (p1in,true,p1out,true,j1p);
  jio (p1in,false,p1out,false,j1m);

  jio (p2in,true,p2out,true,j2p);
  jio (p2in,false,p2out,false,j2m);


  q1v[0]=q1.e();
  q1v[1]=q1.x();
  q1v[2]=q1.y();
  q1v[3]=q1.z();

  q2v[0]=q2.e();
  q2v[1]=q2.x();
  q2v[2]=q2.y();
  q2v[3]=q2.z();

  COM Mmp=cHdot(j1m,j2p,q1v,q2v);
  COM Mmm=cHdot(j1m,j2m,q1v,q2v);
  COM Mpp=cHdot(j1p,j2p,q1v,q2v);
  COM Mpm=cHdot(j1p,j2m,q1v,q2v);

  double sst=abs2(Mmp)+abs2(Mmm)+abs2(Mpp)+abs2(Mpm);
  // return (4./3.)*(4./3.)*sst/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
  return sst/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
}

double MH2qg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector q2)
// q~p1 g~p2 (i.e. ALWAYS p1 for quark, p2 for gluon)
// should be called with q1 meant to be contracted with p2 in first part of vertex 
// (i.e. if g is backward, q1 is forward)
{
  current j1p,j1m,j2p,j2m,q1v,q2v;
  
  j (p1out,true,p1in,true,j1p);
  j (p1out,false,p1in,false,j1m);
  
  j (p2out,true,p2in,true,j2p);
  j (p2out,false,p2in,false,j2m);
  
  q1v[0]=q1.e();
  q1v[1]=q1.x();
  q1v[2]=q1.y();
  q1v[3]=q1.z();
  
  q2v[0]=q2.e();
  q2v[1]=q2.x();
  q2v[2]=q2.y();
  q2v[3]=q2.z();
  
  // First, calculate the non-flipping amplitudes:
  
  COM Mpp=cHdot(j1p,j2p,q1v,q2v);
  COM Mpm=cHdot(j1p,j2m,q1v,q2v);
  COM Mmp=cHdot(j1m,j2p,q1v,q2v);
  COM Mmm=cHdot(j1m,j2m,q1v,q2v);

  double ratio; // p2-/pb- in the notes
  // if (p2in.plus()>0) // if the gluon is the positive
  if (p2in.pz()>0) // if the gluon is the positive
    ratio=p2out.plus()/p2in.plus();
  else // the gluon is the negative
    ratio=p2out.minus()/p2in.minus();

  double nonflipcolourmult=(1.-1./9.)/2.*(ratio+1./ratio)+1./9.;
  double sst=nonflipcolourmult*(abs2(Mmp)+abs2(Mmm)+abs2(Mpp)+abs2(Mpm));

  // Cf*Ca=4
  // return 4.*sst/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
  return sst/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
}

double MH2qbarg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector q2)
// qbar~p1 g~p2 (i.e. ALWAYS p1 for anti-quark, p2 for gluon)
// should be called with q1 meant to be contracted with p2 in first part of vertex 
// (i.e. if g is backward, q1 is forward)
{
  current j1p,j1m,j2p,j2m,q1v,q2v;
  
  jio (p1in,true,p1out,true,j1p);
  jio (p1in,false,p1out,false,j1m);
  
  j (p2out,true,p2in,true,j2p);
  j (p2out,false,p2in,false,j2m);
  
  q1v[0]=q1.e();
  q1v[1]=q1.x();
  q1v[2]=q1.y();
  q1v[3]=q1.z();
  
  q2v[0]=q2.e();
  q2v[1]=q2.x();
  q2v[2]=q2.y();
  q2v[3]=q2.z();
  
  // First, calculate the non-flipping amplitudes:
  
  COM amp,amm,apm,app;
  app=cHdot(j1p,j2p,q1v,q2v);
  apm=cHdot(j1p,j2m,q1v,q2v);
  amp=cHdot(j1m,j2p,q1v,q2v);
  amm=cHdot(j1m,j2m,q1v,q2v);

  double MH2sum;
  MH2sum=abs2(app);
  // std::cout << "MH2sum : "<<MH2sum<<std::endl;
  // then do plus on minus
  MH2sum+=abs2(amm);
  // etc...
  MH2sum+=abs2(apm);
  MH2sum+=abs2(amp);

  double ratio; // p2-/pb- in the notes
  // if (p2in.plus()>0) // if the gluon is the positive
  if (p2in.pz()>0) // if the gluon is the positive
    ratio=p2out.plus()/p2in.plus();
  else // the gluon is the negative
    ratio=p2out.minus()/p2in.minus();

  double nonflipcolourmult=(1.-1./9.)/2.*(ratio+1./ratio)+1./9.;
  MH2sum*=nonflipcolourmult;
  
  // Cf*Ca=4
  // return 4.*MH2sum/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
  return MH2sum/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
}

double MH2gg (CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in, CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in, CLHEP::HepLorentzVector q1, CLHEP::HepLorentzVector q2)
// g~p1 g~p2
// should be called with q1 meant to be contracted with p2 in first part of vertex 
// (i.e. if g is backward, q1 is forward)
{
  current j1p,j1m,j2p,j2m,q1v,q2v;
  
  j (p1out,true,p1in,true,j1p);
  j (p1out,false,p1in,false,j1m);
  
  j (p2out,true,p2in,true,j2p);
  j (p2out,false,p2in,false,j2m);
  
  q1v[0]=q1.e();
  q1v[1]=q1.x();
  q1v[2]=q1.y();
  q1v[3]=q1.z();
  
  q2v[0]=q2.e();
  q2v[1]=q2.x();
  q2v[2]=q2.y();
  q2v[3]=q2.z();
  
  // First, calculate the non-flipping amplitudes:
  
  COM amp,amm,apm,app;
  app=cHdot(j1p,j2p,q1v,q2v);
  apm=cHdot(j1p,j2m,q1v,q2v);
  amp=cHdot(j1m,j2p,q1v,q2v);
  amm=cHdot(j1m,j2m,q1v,q2v);

  double MH2sum;
  MH2sum=abs2(app);
  // std::cout << "MH2sum : "<<MH2sum<<std::endl;
  // then do plus on minus
  MH2sum+=abs2(amm);
  // etc...
  MH2sum+=abs2(apm);
  MH2sum+=abs2(amp);

  double ratio1; // p1-/pa- in the notes
  // if (p1in.plus()>0) // if the gluon is the positive
  if (p1in.pz()>0.) // if the gluon is the positive
    ratio1=p1out.plus()/p1in.plus();
  else // the gluon is the negative
    ratio1=p1out.minus()/p1in.minus();

  double nonflipcolourmult1=(1.-1./9.)/2.*(ratio1+1./ratio1)+1./9.;

  double ratio2; // p2-/pb- in the notes
  // if (p2in.plus()>0) // if the gluon is the positive
  if (p2in.pz()>0.) // if the gluon is the positive
    ratio2=p2out.plus()/p2in.plus();
  else // the gluon is the negative
    ratio2=p2out.minus()/p2in.minus();

  double nonflipcolourmult2=(1.-1./9.)/2.*(ratio2+1./ratio2)+1./9.;
  MH2sum*=nonflipcolourmult1*nonflipcolourmult2;
  
  // Ca*Ca=9
  // return 9.*MH2sum/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
  return MH2sum/((p1in-p1out).m2()*(p2in-p2out).m2()*q1.m2()*q2.m2());
}
