//////////////////////////////////////////////////
//////////////////////////////////////////////////
// This source code is Copyright (2012) of      //
//  Jeppe R. Andersen and Jennifer M. Smillie   //
// and is distributed under the                 //
// Gnu Public License version 2                 //
// http://www.gnu.org/licenses/gpl-2.0.html     //
// You are allowed to distribute and alter the  //
// source under the conditions of the GPLv2     //
// as long as this copyright notice             //
// is unaltered and distributed with the source //
// Any use should comply with the               //
//             MCNET GUIDELINES                 //
//    for Event Generator Authors and Users     //
// as distributed with this source code         //
//////////////////////////////////////////////////
//////////////////////////////////////////////////
#define USELHAPDF 0
#include "Hjets.h"
#include "Currents.h"
#include <fenv.h>
#if USELHAPDF
#include "LHAPDF/LHAPDF.h"
#endif
#define printout 0
#define oldfklphsp 0
#define optfklphsp 0 // use for VBF type cuts with ya*yb<0
#define old2jphsp 0
#define old3jphsp 0
#define SmartThing 1
#define SmartThingCL 1
#define printalphas 0
#define calcscaleunc 1
#define dijetLO 0
const double mus=1.;
const double cut=-2.;
const double clambda=0.2;
const double ccut=.2;
const double sf=.5;
const double cqtsqgen=25.*25.; // lambda=0.5: 50.*50.; lambda=0.2: 25.*25.

double invmass(std::vector<fastjet::PseudoJet> * jets)
{
  fastjet::PseudoJet sum;
  std::vector<fastjet::PseudoJet>::const_iterator itr = jets->begin();
  for( ; itr != jets->end() ; ++itr)
    sum+=(*itr);
  return sum.m();
}

double geometricmean(std::vector<fastjet::PseudoJet> * jets)
{
  double gm(1.);
  std::vector<fastjet::PseudoJet>::const_iterator itr = jets->begin();
  for( ; itr != jets->end() ; ++itr)
    gm*=(*itr).perp();
  gm=pow(gm,1./jets->size());
  return gm;
}

// The constructor
CHMultijet::CHMultijet(void) 
//  : parton("Grids/mstw2008nlo.00.dat"),
  :     ran(new Ranlux64Engine())
{
  // set the default values
  lambda=clambda;
  ymax=5.;// the maximum (and -minimum) rapidity generated for any particle
  std::cout << "************************************************************\n";
  std::cout << "*                                                          *\n";
  std::cout << "*                #     #  #######        #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #######  #####          #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #######   #####                  *\n";
  std::cout << "*                                                          *\n";
  std::cout << "* http://cern.ch/hej                      Version 1.4.4    *\n";
  std::cout << "* by Jeppe R. Andersen, jeppe.andersen@cern.ch             *\n";
  std::cout << "*    Jennifer M. Smillie, j.m.smillie@ed.ac.uk             *\n";
  std::cout << "* If you use this software, please cite the following      *\n";
  std::cout << "* papers:                                                  *\n";
  std::cout << "* arXiv:1101.5394 (JHEP 1106:010,2011)                     *\n";
  std::cout << "* arXiv:0910.5113 (Phys.Rev. D81 (2010) 114021)            *\n";
  std::cout << "* arXiv:0908.2786 (JHEP 1001:039,2010)                     *\n";
  std::cout << "************************************************************\n";
#if USELHAPDF
  LHAPDF::initPDFSet(GeneratorFlags.LHAPDFName, LHAPDF::LHGRID, GeneratorFlags.LHAPDFsubset);
#endif
  parton = new c_mstwpdf("Grids/mstw2008nlo.00.dat"); // default: warn=false, fatal=true
  // init alpha_s
  // init alpha_s
  double FR2 = 1e0, Q0 = 1e0, mTop = 1.e10;
  InitAlphaS(parton->alphaSorder,FR2,Q0,parton->alphaSQ0,parton->mCharm,parton->mBottom,mTop);
}

CHMultijet::CHMultijet(string filename) 
//  : parton("Grids/mstw2008nlo.00.dat"),
  :  ran(new Ranlux64Engine())
{
  // set the default values
  lambda=clambda;
  ymax=5.;// the maximum (and -minimum) rapidity generated for any particle
  std::cout << "************************************************************\n";
  std::cout << "*                                                          *\n";
  std::cout << "*                #     #  #######        #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #######  #####          #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #######   #####                  *\n";
  std::cout << "*                                                          *\n";
  std::cout << "* http://cern.ch/hej                      Version 1.4.4    *\n";
  std::cout << "* by Jeppe R. Andersen, jeppe.andersen@cern.ch             *\n";
  std::cout << "*    Jennifer M. Smillie, j.m.smillie@ed.ac.uk             *\n";
  std::cout << "* If you use this software, please cite the following      *\n";
  std::cout << "* papers:                                                  *\n";
  std::cout << "* arXiv:1101.5394 (JHEP 1106:010,2011)                     *\n";
  std::cout << "* arXiv:0910.5113 (Phys.Rev. D81 (2010) 114021)            *\n";
  std::cout << "* arXiv:0908.2786 (JHEP 1001:039,2010)                     *\n";
  std::cout << "************************************************************\n";
#if USELHAPDF
  LHAPDF::initPDFSet(GeneratorFlags.LHAPDFName, LHAPDF::LHGRID, GeneratorFlags.LHAPDFsubset);
#endif
  parton = new c_mstwpdf(filename.c_str()); // default: warn=false, fatal=true
  // init alpha_s
  // init alpha_s
  double FR2 = 1e0, Q0 = 1e0, mTop = 1.e10;
  InitAlphaS(parton->alphaSorder,FR2,Q0,parton->alphaSQ0,parton->mCharm,parton->mBottom,mTop);
  LHoutput.open(filename.c_str());
  LHoutput << "<LesHouchesEvents version=\"1.0\">\n";
  LHoutput << "<!--\n";
  LHoutput << "File generated with HEJ Jets vers. 1.4.4\n";
  LHoutput << "-->\n";
  LHoutput << "<init>\n";
  LHoutput << "  2212   2212  "<<flags.Ebeam<<" "<<flags.Ebeam<<"     0     0     7     7     4     1\n";
  LHoutput << "  1E+00  1E-02  1.000000E+00    1\n";
  LHoutput << "</init>\n";
}

CHMultijet::CHMultijet(SGeneratorFlags GeneratorFlagsa) 
//  :    parton(GeneratorFlagsa.mrstpdffile),
  :  ran(new Ranlux64Engine())
{
  flags=GeneratorFlagsa;
  // set the default values
  lambda=clambda;
  ymax=5.;// the maximum (and -minimum) rapidity generated for any particle
  std::cout << "************************************************************\n";
  std::cout << "*                                                          *\n";
  std::cout << "*                #     #  #######        #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #######  #####          #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #######   #####                  *\n";
  std::cout << "*                                                          *\n";
  std::cout << "* http://cern.ch/hej                      Version 1.4.4    *\n";
  std::cout << "* by Jeppe R. Andersen, jeppe.andersen@cern.ch             *\n";
  std::cout << "*    Jennifer M. Smillie, j.m.smillie@ed.ac.uk             *\n";
  std::cout << "* If you use this software, please cite the following      *\n";
  std::cout << "* papers:                                                  *\n";
  std::cout << "* arXiv:1101.5394 (JHEP 1106:010,2011)                     *\n";
  std::cout << "* arXiv:0910.5113 (Phys.Rev. D81 (2010) 114021)            *\n";
  std::cout << "* arXiv:0908.2786 (JHEP 1001:039,2010)                     *\n";
  std::cout << "************************************************************\n";

#if USELHAPDF
  LHAPDF::initPDFSet(GeneratorFlagsa.LHAPDFName, LHAPDF::LHGRID, GeneratorFlagsa.LHAPDFsubset);
#endif
  parton = new c_mstwpdf(GeneratorFlagsa.mstwpdffile); // default: warn=false, fatal=true
  // init alpha_s
  double FR2 = 1e0, Q0 = 1e0, mTop = 1.e10;
  InitAlphaS(parton->alphaSorder,FR2,Q0,parton->alphaSQ0,parton->mCharm,parton->mBottom,mTop);

  if (flags.raninitfile.length()!=0) {
    cout << "# Restoring Random Engine Status : \n";
    ran.restoreEngineStatus(flags.raninitfile.c_str());
  }

  maxweight[0]=0;
  maxweight[1]=1.5E4;
  maxweight[2]=5E3;
  maxweight[3]=1E4;
  xsec[0][0]=xsec[1][0]=xsec[2][0]=xsec[3][0]=0.;
  xsec[0][1]=xsec[1][1]=xsec[2][1]=xsec[3][1]=0.;
  if (flags.LHoutput) {
    LHoutput.open(flags.LHoutputfile.c_str());
    LHoutput << "<LesHouchesEvents version=\"1.0\">\n";
    LHoutput << "<!--\n";
    LHoutput << "File generated with HEJ Jets vers. 1.4.4\n";
    LHoutput << "-->\n";
    LHoutput << "<init>\n";
    LHoutput << "  ";
    if (flags.beam[0]==1)
      LHoutput<<"2212";
    else 
      LHoutput<<"-2212";
    LHoutput << "   ";
    if (flags.beam[1]==1)
      LHoutput<<"2212 ";
    else 
      LHoutput<<"-2212 ";

    LHoutput <<scientific<<setprecision(7)<<setw(12)<<flags.Ebeam<<"  "<<flags.Ebeam<<" 0     0     7     7     ";
    if (flags.unweight)
      LHoutput << "2     ";
    else
      LHoutput << "4     ";
    // The trick is that we can first write some arbitrary data to the file, 
    // save a pointer, and then return later to overwrite the fields! 
    // lean and mean!
    xsecpos=LHoutput.tellp();
    if (flags.match) { // if matching
      LHoutput <<"3\n"; // How many processes form part of the matching?
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 1\n"; 
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 2\n"; 
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 3\n"; 
    } else { // if no matching
      LHoutput <<"1\n";
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 1\n"; 
    }
    // 

    LHoutput << "</init>\n";
  }
}

CHMultijet::~CHMultijet(void)
  // The destructor
{
  if(LHoutput.is_open()) {
    LHoutput << "</LesHouchesEvents>\n";

    LHoutput.seekp(xsecpos);
    int processcount;
    if (flags.match) // if matching
      processcount=nprocesstype;
    else // if no matching
      processcount=1;
    LHoutput <<processcount<<"\n"; // How many processes form part of the matching?
    for (int tcount=1;tcount<=processcount;tcount++ )
      LHoutput << "  "<<setprecision(7)<<setw(12)<<xsec[tcount][0]<<" "<<sqrt(xsec[tcount][1])<<" "<<1<<" "<<tcount<<"\n"; 
    
    LHoutput.close();
  }
  delete(parton);
}

void CHMultijet::LHSaveEvent(void)
{
  LHoutput << "<event>\n";
  
  if (flags.unweight==false)
    LHoutput << " " << LHparticles.size()<< " 1 " <<scientific<<setprecision(10)<<setw(19)<<wt<<" "<<flags.scale<<" "<<1./128.<< " "<<Halphas(flags.scale)<<endl;
  else
    LHoutput << " " << LHparticles.size()<< " "<< getprocesstype()<<" " <<scientific<<setprecision(2)<<setw(4)<<1<<" "<<scientific<<setprecision(10)<<setw(19)<<flags.scale<<" "<<1./128.<< " "<<Halphas(flags.scale)<<"\n";

  for (unsigned int i=0;i<LHparticles.size();i++) {
    LHoutput << " " ;
    LHoutput << setw(3) << LHparticles.at(i).PID<< " " ;
    LHoutput << setw(2) << LHparticles.at(i).ISTUP<< " " ;
    LHoutput << setw(2) << LHparticles.at(i).MOT1<< " " ;
    LHoutput << setw(2) << LHparticles.at(i).MOT2<< " " ;
    LHoutput << setw(3) << LHparticles.at(i).COL1<< " " ;
    LHoutput << setw(3) << LHparticles.at(i).COL2<< " " ;
    LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.x()<< " " ;
    LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.y()<< " " ;
    LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.z()<< " " ;
    LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.e()<< " " ;
    if (LHparticles.at(i).PID!=25)
      LHoutput << scientific<<setprecision(12)<<setw(19) << 0. << " " ;
    else
      LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.m()<< " " ;
    LHoutput << "0. 9.\n" ;
  }
  LHoutput << "</event>\n";
}

int CHMultijet::getMstatus(void)
{
  return status;
}

int CHMultijet::getstatus(void)
{
  return status;
}

void CHMultijet::setstatus(int i)
{
  status=i;
}

double CHMultijet::getcmsenergy(void)
{
  return 2.*flags.Ebeam;
}

int CHMultijet::getnp(void)
{
  return partonmomenta.size(); // The total number of partons in final state
}

int CHMultijet::getprocesstype(void)
{
  return processtype;
}

HepLorentzVector CHMultijet::getpartvn(int ni)
{

  if ((ni<=int(partonmomenta.size()-1))&&(ni>=0))
    {
      return partonmomenta.at(ni);
    }
  else {
    cerr << "Parton number out of range : "<<ni<<endl;
    exit(0);
  }
}

double CHMultijet::pdfpt(int beam,double x, double uf, int pt)
{
  if (x<1e-5)
    return 0.;
  if (pt==21) // gluon
#if USELHAPDF
    {
      if (flags.useLHApdf)
	return max(0.,LHAPDF::xfx(x,uf,0));
      else
	return max(0.,CHMultijet::parton->parton(0,x,uf));
    }
#else
    return max(0.,CHMultijet::parton->parton(0,x,uf));
#endif
  else if (abs(pt)<7)
#if USELHAPDF
    {
      if (flags.useLHApdf)
	return max(0.,LHAPDF::xfx(x,uf,pt*flags.beam[beam]));
      else
	return max(0.,parton->parton(pt*flags.beam[beam],x,uf));
    }
#else
    return max(0.,parton->parton(pt*flags.beam[beam],x,uf));
#endif
  else {
    cout << "particle type unknown: "<<pt<<endl;
    return 0.;
  }
}

double CHMultijet::Halphas(double q)
{
  double as;
#if USELHAPDF
  if (flags.useLHApdf)
    as=LHAPDF::alphasPDF(q);
  else
    as=AlphaS(q);
#else
  as=AlphaS(q);
#endif
  if (isnan(as)||as>.5)
    as=0.5;
  return as;
}

double CHMultijet::GenerateFKLIncomingParticles(double xa,double ufa,double xb,double ufb)
{
  double r1,lwt(1.),sum;

//   parton->mrst(xa,ufa);
//   pdfa=fabs(parton->cont.glu)+4.0/9.0*(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));
//   pdfa=max(0.,parton->cont.glu)+4.0/9.0*(max(0.,parton->cont.upv+parton->cont.usea) + max(0.,parton->cont.dnv+parton->cont.dsea) + max(0.,parton->cont.usea)+max(0.,parton->cont.dsea) + 2.0*(max(0.,parton->cont.str)+max(0.,parton->cont.chm)+max(0.,parton->cont.bot)));
  pdfga=fabs(pdfpt(0,xa,ufa,21));
  pdfua=fabs(pdfpt(0,xa,ufa,2));
  pdfda=fabs(pdfpt(0,xa,ufa,1));
  pdfuxa=fabs(pdfpt(0,xa,ufa,-2));
  pdfdxa=fabs(pdfpt(0,xa,ufa,-1));
  pdfca=fabs(pdfpt(0,xa,ufa,4));
  pdfsa=fabs(pdfpt(0,xa,ufa,3));
  pdfsxa=fabs(pdfpt(0,xa,ufa,-3));
  pdfba=fabs(pdfpt(0,xa,ufa,5));

  pdfa=pdfga+4.0/9.0*(pdfua + pdfda + pdfuxa + pdfdxa +pdfsa +pdfsxa + 2.0*(pdfca +pdfba ));
  r1=pdfa*ran.flat();

#if 1
  if (r1<(sum=pdfga)) {
    aptype=21;
    lwt*=pdfa/pdfga;}
  else if (r1<(sum+=(4./9.)*pdfua)) {
    aptype=2;
    lwt*=pdfa/((4./9.)*pdfua); }
  else if (r1<(sum+=(4./9.)*pdfda)) {
    aptype=1;
    lwt*=pdfa/((4./9.)*pdfda); }
  else if (r1<(sum+=(4./9.)*pdfuxa)) {
    aptype=-2;
    lwt*=pdfa/((4./9.)*pdfuxa); }
  else if (r1<(sum+=(4./9.)*pdfdxa)) {
    aptype=-1;
    lwt*=pdfa/((4./9.)*pdfdxa); }
  else if (r1<(sum+=(4./9.)*pdfca)) {
    aptype=4;
    lwt*=pdfa/((4./9.)*pdfca); }
  else if (r1<(sum+=(4./9.)*pdfca)){
    aptype=-4;
    lwt*=pdfa/((4./9.)*pdfca); }
  else if (r1<(sum+=(4./9.)*pdfsa)) {
    aptype=3;
    lwt*=pdfa/((4./9.)*pdfsa); }
  else if (r1<(sum+=(4./9.)*pdfsxa)) {
    aptype=-3;
    lwt*=pdfa/((4./9.)*pdfsxa); }
  else if (r1<(sum+=(4./9.)*pdfba)) {
    aptype=5;
    lwt*=pdfa/((4./9.)*pdfba); }
  else if (r1<=(sum+=(4./9.)*pdfba)) {
    aptype=-5;
    lwt*=pdfa/((4./9.)*pdfba); }
  else {
    cout << "Error in choosing incoming parton a : "<<xa<<" "<<ufa<<" "<<sum<<" "<<pdfa<<" "<<r1;
    cout << " "<<pdfga+4./9.*(pdfua+pdfuxa+pdfda+pdfdxa+pdfsa+pdfsxa+2.*(pdfca+pdfba))<<endl;
    setstatus(15);
    return 0;
  }
#else
  aptype=1;
#endif

  pdfta=pdfpt(0,xa,ufa,aptype);
  // std::cout<<xa<<" "<<ufa<<" "<<aptype<<" "<<pdfta<<std::endl;

//   parton->mrst(xb,ufb);
//   pdfb=max(0.,parton->cont.glu)+4.0/9.0*(max(0.,parton->cont.upv+parton->cont.usea) + max(0.,parton->cont.dnv+parton->cont.dsea) + max(0.,parton->cont.usea)+max(0.,parton->cont.dsea) + 2.0*(max(0.,parton->cont.str)+max(0.,parton->cont.chm)+max(0.,parton->cont.bot)));
//   pdfb=fabs(parton->cont.glu)+4.0/9.0*(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));
  pdfgb=fabs(pdfpt(1,xb,ufb,0));
  pdfub=fabs(pdfpt(1,xb,ufb,2));
  pdfdb=fabs(pdfpt(1,xb,ufb,1));
  pdfuxb=fabs(pdfpt(1,xb,ufb,-2));
  pdfdxb=fabs(pdfpt(1,xb,ufb,-1));
  pdfcb=fabs(pdfpt(1,xb,ufb,4));
  pdfsb=fabs(pdfpt(1,xb,ufb,3));
  pdfsxb=fabs(pdfpt(1,xb,ufb,-3));
  pdfbb=fabs(pdfpt(1,xb,ufb,5));

  pdfb=pdfgb+4.0/9.0*(pdfub + pdfdb + pdfuxb + pdfdxb + pdfsb +pdfsxb +2.0*(pdfcb +pdfbb ));
  r1=pdfb*ran.flat();

#if 1
  if (r1<(sum=pdfgb)) {
    bptype=21;
    lwt*=pdfb/pdfgb; }
  else if (r1<(sum+=(4./9.)*pdfub)) {
    bptype=2;
    lwt*=pdfb/((4./9.)*pdfub); }
  else if (r1<(sum+=(4./9.)*pdfdb)) {
    bptype=1;
    lwt*=pdfb/((4./9.)*pdfdb); }
  else if (r1<(sum+=(4./9.)*pdfuxb)) {
    bptype=-2;
    lwt*=pdfb/((4./9.)*pdfuxb); }
  else if (r1<(sum+=(4./9.)*pdfdxb)) {
    bptype=-1;
    lwt*=pdfb/((4./9.)*pdfdxb); }
  else if (r1<(sum+=(4./9.)*pdfcb)) {
    bptype=4;
    lwt*=pdfb/((4./9.)*pdfcb); }
  else if (r1<(sum+=(4./9.)*pdfcb)){
    bptype=-4;
    lwt*=pdfb/((4./9.)*pdfcb); }
  else if (r1<(sum+=(4./9.)*pdfsb)) {
    bptype=3;
    lwt*=pdfb/((4./9.)*pdfsb); }
  else if (r1<(sum+=(4./9.)*pdfsxb)) {
    bptype=-3;
    lwt*=pdfb/((4./9.)*pdfsxb); }
  else if (r1<(sum+=(4./9.)*pdfbb)) {
    bptype=5;
    lwt*=pdfb/((4./9.)*pdfbb); }
  else if (r1<(sum+=(4./9.)*pdfbb)) {
    bptype=-5;
    lwt*=pdfb/((4./9.)*pdfbb); }
  else {
    cout << "Error in choosing incoming partons b : "<<xb<<" "<<ufb<<" "<<sum<<" "<<pdfb<<" "<<r1;
    cout << " "<<pdfgb+4./9.*(pdfub+pdfuxb+pdfdb+pdfdxb+pdfsb+pdfsxb+2.*(pdfcb+pdfbb))<<endl;
    setstatus(15);
    return 0;
  }

#else
 bptype=21;
#endif

  pdftb=pdfpt(1,xb,ufb,bptype);

  return lwt;
}

double CHMultijet::GenerateIncomingParticles(double xa,double ufa,double xb,double ufb)
{
  double r1,sum,lwt(1.);

//   parton->mrst(xa,ufa);
//   pdfa=fabs(parton->cont.glu)+(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));

  pdfga=fabs(pdfpt(0,xa,ufa,21));
  pdfua=fabs(pdfpt(0,xa,ufa,2));
  pdfda=fabs(pdfpt(0,xa,ufa,1));
  pdfuxa=fabs(pdfpt(0,xa,ufa,-2));
  pdfdxa=fabs(pdfpt(0,xa,ufa,-1));
  pdfca=fabs(pdfpt(0,xa,ufa,4));
  pdfsa=fabs(pdfpt(0,xa,ufa,3));  
  pdfsxa=fabs(pdfpt(0,xa,ufa,-3));  
  pdfba=fabs(pdfpt(0,xa,ufa,5));

  pdfa=pdfga+(pdfua + pdfda + pdfuxa + pdfdxa + pdfsa +pdfsxa +2.0*(pdfca +pdfba ));

#if 1
  r1=pdfa*ran.flat();

  if (r1<(sum=pdfga)) {
    aptype=21;
    lwt*=pdfa/pdfga;}
  else if (r1<(sum+=pdfua)) {
    aptype=2;
    lwt*=pdfa/(pdfua); }
  else if (r1<(sum+=pdfda)) {
    aptype=1;
    lwt*=pdfa/(pdfda); }
  else if (r1<(sum+=pdfuxa)) {
    aptype=-2;
    lwt*=pdfa/(pdfuxa); }
  else if (r1<(sum+=pdfdxa)) {
    aptype=-1;
    lwt*=pdfa/(pdfdxa); }
  else if (r1<(sum+=pdfca)) {
    aptype=4;
    lwt*=pdfa/(pdfca); }
  else if (r1<(sum+=pdfca)){
    aptype=-4;
    lwt*=pdfa/(pdfca); }
  else if (r1<(sum+=pdfsa)) {
    aptype=3;
    lwt*=pdfa/(pdfsa); }
  else if (r1<(sum+=pdfsa)) {
    aptype=-3;
    lwt*=pdfa/(pdfsa); }
  else if (r1<(sum+=pdfba)) {
    aptype=5;
    lwt*=pdfa/(pdfba); }
  else if (r1<(sum+=pdfba)) {
    aptype=-5;
    lwt*=pdfa/(pdfba); }
  else {
//     cout << "Error in choosing incoming partons: "<<sum<<" "<<pdfa<<" "<<r1;
//     cout << " "<<pdfga+4./9.*(pdfua+pdfuxa+pdfda+pdfdxa+2.*(pdfsa+pdfca+pdfba))<<endl;
    setstatus(15);
    return 0;
  }  
#else
  aptype=21;
#endif

  pdfta=pdfpt(0,xa,ufa,aptype);

//   parton->mrst(xb,ufb);

//   pdfb=fabs(parton->cont.glu)+(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));
  pdfgb=fabs(pdfpt(1,xb,ufb,21));
  pdfub=fabs(pdfpt(1,xb,ufb,2));
  pdfdb=fabs(pdfpt(1,xb,ufb,1));
  pdfuxb=fabs(pdfpt(1,xb,ufb,-2));
  pdfdxb=fabs(pdfpt(1,xb,ufb,-1));
  pdfcb=fabs(pdfpt(1,xb,ufb,4));
  pdfsb=fabs(pdfpt(1,xb,ufb,3));
  pdfsxb=fabs(pdfpt(1,xb,ufb,-3));
  pdfbb=fabs(pdfpt(1,xb,ufb,5));  
  
  pdfb=pdfgb+(pdfub + pdfdb + pdfuxb + pdfdxb + pdfsb +pdfsxb +2.0*(pdfcb +pdfbb ));

  r1=pdfb*ran.flat();

#if 1
  if (r1<(sum=pdfgb)) {
    bptype=21;
    lwt*=pdfb/pdfgb; }
  else if (r1<(sum+=pdfub)) {
    bptype=2;
    lwt*=pdfb/(pdfub); }
  else if (r1<(sum+=pdfdb)) {
    bptype=1;
    lwt*=pdfb/(pdfdb); }
  else if (r1<(sum+=pdfuxb)) {
    bptype=-2;
    lwt*=pdfb/(pdfuxb); }
  else if (r1<(sum+=pdfdxb)) {
    bptype=-1;
    lwt*=pdfb/(pdfdxb); }
  else if (r1<(sum+=pdfcb)) {
    bptype=4;
    lwt*=pdfb/(pdfcb); }
  else if (r1<(sum+=pdfcb)){
    bptype=-4;
    lwt*=pdfb/(pdfcb); }
  else if (r1<(sum+=pdfsb)) {
    bptype=3;
    lwt*=pdfb/(pdfsb); }
  else if (r1<(sum+=pdfsb)) {
    bptype=-3;
    lwt*=pdfb/(pdfsb); }
  else if (r1<(sum+=pdfbb)) {
    bptype=5;
    lwt*=pdfb/(pdfbb); }
  else if (r1<(sum+=pdfbb)) {
    bptype=-5;
    lwt*=pdfb/(pdfbb); }
  else {
//     cout << "Error in choosing incoming partons: "<<sum<<" "<<pdfb<<" "<<r1;
//     cout << " "<<pdfgb+4./9.*(pdfub+pdfuxb+pdfdb+pdfdxb+2.*(pdfsb+pdfcb+pdfbb))<<endl;
    setstatus(15);
    return 0;
  }

#else
  bptype=21;
#endif

  pdftb=pdfpt(1,xb,ufb,bptype);

  if (pdfta<1e-10||pdftb<1e-10) {
    setstatus(21);
  }

  return lwt;
}

std::vector<double> CHMultijet::GetWeights(void)
{
  return weights;
}

void CHMultijet::SetupLHInformation()
// 
{
  std::list<int> vecl,vecr;

  if (aptype>0) { // if "a" is a gluon or quark
    vecl.push_front(-1);
    vecr.push_back(0);
  } else { // if a is an anti-quark
    vecl.push_front(0); 
    vecr.push_back(-1);
  }

  for (unsigned int i=1;i<partonmomenta.size()-1;i++) { // don't deal with the last gluon
    if (ran.flat()<.5) 
      vecr.push_back(i);
    else vecl.push_front(i);
  }
  
  if (bptype==21) {
    if (ran.flat()<.5) {
      vecr.push_back(partonmomenta.size()-1);
      vecl.push_front(-2);
    } else {
      vecl.push_front(partonmomenta.size()-1);
      vecr.push_back(-2);
    } 
  } else if (bptype<0) { // if b is an anti-quark
    vecr.push_back(partonmomenta.size()-1);
    vecl.push_front(-2);
  } else { // if b is a quark
    vecl.push_front(partonmomenta.size()-1);
    vecr.push_back(-2);
  }
  
  // now generate the colour connection information to go in the LH records
  int ctypea(1); // ctype: the colour start type: anti- or colour?
                 // when we swap the side of particles/anti-particles we should
                 // always start with a colour line
  
  particle tparticle; // a temp. particle
  {
    // setup vector of particles
    tparticle.PID=0;
    tparticle.COL1=0;
    tparticle.COL2=0;
    tparticle.MOT1=0;
    tparticle.MOT2=0;
    tparticle.ISTUP=0;
    tparticle.p.set(0,0,0,0);
    for (unsigned int i=0;i<partonmomenta.size()+2;i++)
      LHparticles.push_back(tparticle);
  }
  
  int ctype,col(501);
  ctype=ctypea;    // which of the colour columns should we fill?

//   cout << "LHparticles size : "<<LHparticles.size()<<endl;
  int n=vecr.front()+2;
  int np1;
  vecr.pop_front();
  while (vecr.size()!=0) {
    int np1=vecr.front()+2;
    vecr.pop_front();
//     cerr << "n, np1, ctype : "<<n<< ", "<<np1<<" "<<ctype<<endl;
    if (ctype>0) {
      LHparticles.at(n).COL1=col;
      LHparticles.at(np1).COL2=col;
    } else {
      LHparticles.at(n).COL2=col;
      LHparticles.at(np1).COL1=col;
    }
    col++;
    n=np1;
//     ctype*=-1;
  }
  
  // Now we are at the end of the right hand side.
  // Should we connect to the left hand side?
  
//   cout << "  --\n";
  if (bptype==21) {
    np1=vecl.front()+2;
    vecl.pop_front();
//     cerr << "n, np1, ctype : "<<n<< ", "<<np1<<" "<<ctype<<endl;
    if (ctype>0) {
      LHparticles.at(n).COL1=col;
      LHparticles.at(np1).COL2=col;
    } else {
      LHparticles.at(n).COL2=col;
      LHparticles.at(np1).COL1=col;
    }
    col++;
    n=np1;
  } else {
    n=vecl.front()+2;
    vecl.pop_front();
  }
//   cout << "  --\n";
  while (vecl.size()!=0) {
    int np1=vecl.front()+2;
    vecl.pop_front();
//     cerr << "n, np1, ctype : "<<n<< ", "<<np1<<" "<<ctype<<endl;
    if (ctype>0) {
      LHparticles.at(n).COL1=col;
      LHparticles.at(np1).COL2=col;
    } else {
      LHparticles.at(n).COL2=col;
      LHparticles.at(np1).COL1=col;
    }
    col++;
    n=np1;
//     ctype*=-1;
  }
  
  // if particle a is gluon we need to make another connections
  if (aptype==21) {
    // we know that if a is a gluon, then we started with particle 0 
    // and have ended with particle -1. We also know that the last colour line 
    // is opposite that of ctype - so connect from -1 to 0 with a ctype!
    if (ctype>0) {
      LHparticles.at(1).COL1=col;
      LHparticles.at(2).COL2=col;
    } else {
      LHparticles.at(1).COL2=col;
      LHparticles.at(2).COL1=col;
    }
    col++;
    n=np1;
  }
  
  // swap colours for particles of incoming type
  {  
    int itemp1,itemp2;
    itemp1=LHparticles.at(0).COL1;
    itemp2=LHparticles.at(0).COL2;
    LHparticles.at(0).COL1=itemp2;
    LHparticles.at(0).COL2=itemp1;
    itemp1=LHparticles.at(1).COL1;
    itemp2=LHparticles.at(1).COL2;
    LHparticles.at(1).COL1=itemp2;
    LHparticles.at(1).COL2=itemp1;
  }

  // incoming particles
  LHparticles.at(0).p=pb;
  LHparticles.at(0).PID=bptype;
  LHparticles.at(0).MOT1=0;
  LHparticles.at(0).MOT2=0;
  LHparticles.at(0).ISTUP=-1;
  LHparticles.at(1).p=pa;
  LHparticles.at(1).PID=aptype;
  LHparticles.at(1).MOT1=0;
  LHparticles.at(1).MOT2=0;
  LHparticles.at(1).ISTUP=-1;
    
  // outgoing particles
  for (unsigned int i=2;i<partonmomenta.size()+2;i++) { 
    LHparticles.at(i).p=partonmomenta.at(i-2);
    if (i==2)
      LHparticles.at(i).PID=aptype;
    else if (i==partonmomenta.size()+1)
      LHparticles.at(i).PID=bptype;
    else LHparticles.at(i).PID=21;
    LHparticles.at(i).MOT1=1;
    LHparticles.at(i).MOT2=2;
    LHparticles.at(i).ISTUP=1;
  }

  // setup vector of particles
  tparticle.PID=25;// Higgs boson
  tparticle.COL1=0;
  tparticle.COL2=0;
  tparticle.MOT1=1;
  tparticle.MOT2=2;
  if (flags.DecayHiggs==false)
    tparticle.ISTUP=1; // stable final state particle
  else
    tparticle.ISTUP=2; // unstable final state particle whos daughters should
		       // keep fixed invariant mass after showering
  tparticle.p=getpHv();
  LHparticles.push_back(tparticle);

  if (flags.DecayHiggs==true) {
    tparticle.PID=22;
    tparticle.COL1=0;
    tparticle.COL2=0;
    tparticle.MOT1=LHparticles.size();
    tparticle.MOT2=LHparticles.size();
    tparticle.ISTUP=1; // stable final state particle
    tparticle.p=getPhoton(0);
    LHparticles.push_back(tparticle);

    tparticle.p=getPhoton(1);
    LHparticles.push_back(tparticle);
  }
}

HepLorentzVector CHMultijet::getpHv()
{
  return pHv;
}

HepLorentzVector CHMultijet::getPhoton(int n)
{
  return photonmomenta.at(n);
}

double CHMultijet::omega0(double q2, double lambda2)
{
  double result1(0.);
  if (flags.logcorrect==0) {
      result1=-Halphas(renscale)*3./M_PI*log(q2/lambda2);
      // renscale=PTMAX if necessary in weight() before Virtuals
  }
  else {   // flags.logcorrect==1
    //    result1=-Halphas(sqrt(sqrt(q2*lambda2))*flags.scalefactor)*3./M_PI*log(q2/lambda2);
    //  Instead, leave log explicitly
    result1=-Halphas(renscale)*3./M_PI*log(q2/lambda2)*(1.+Halphas(renscale)*beta0*log(pow(renscale,4)/(q2*lambda2))/(8.*M_PI));
#if printalphas
    std::cout<<"omega0: "<<Halphas(sqrt(sqrt(q2*lambda2)))<<"  "<<Halphas(renscale)*(1.+Halphas(renscale)*beta0*log(pow(renscale,4)/(q2*lambda2))/(8.*M_PI))<<"\n";
#endif
  }
  return result1;
}

double CHMultijet::C2LargeMt(HepLorentzVector q1, HepLorentzVector pH)
// The Large Mt limit of the g*g*H-vertex
{
  // static double as=Halphas(flags.scale);
  static double as=Halphas(flags.mH);
  static double A=as/(3.*M_PI*v);
  HepLorentzVector q2(q1-pH);
  //  HepLorentzVector q2(pH-q1);
  //  double temp=A*(pH.mt2()/2.-q1.dot(q2));
  //  double temp=A*(pH.mt2()/2.+q1.dot(q2));
//   double temp=A/2.*(pH.perp2()+q1.m2()+q2.m2()); // HE limit 
  double temp=A/2.*(pH.perp2()-q1.perp2()-q2.perp2()); // the full HE limit also of invariants
  //  double temp=A*((MH*MH+pH.perp2())/2.-q1.dot(q2));
  //  cout << pH.mt2()/2. << " "<<(MH*MH+pH.perp2())/2.<<endl;
  return temp*temp;
}

double CHMultijet::C2LipatovVDD(HepLorentzVector qav, HepLorentzVector qbv, HepLorentzVector p1, HepLorentzVector p2)
{
  double sa1,sb1,shat;
  HepLorentzVector p5=qav-qbv;
  HepLorentzVector temptrans=(qav+qbv);
  sa1=-2.*p1.dot(p5);
  sb1=-2.*p2.dot(p5);
  shat=2.*p1.dot(p2);
  temptrans[2]=temptrans[3]=0; // it is gauge invariant when these are not set to zero

  HepLorentzVector CL=temptrans-(sa1/shat+2.*qbv.m2()/sb1)*p2+(sb1/shat+2.*qav.m2()/sa1)*p1;
  //  cout << "#Should be negative : "<<temptrans.m2()<<endl;
//   cout << "#VDD qav.perp2() : "<< qav.perp2()<<endl;
//   cout << "#VDD qbv.perp2() : "<< qbv.perp2()<<endl;
#if printoutput
  cout << "#VDD qa : "<<qav<<" "<<qav.m2()<<" "<<-qav.perp2()<<endl;
  cout << "#VDD qb : "<<qbv<<" "<<qbv.m2()<<" "<<-qbv.perp2()<<endl;
  cout << "#VDD p1 : "<<p1<<endl;
  cout << "#VDD p2 : "<<p2<<endl;
  cout << "#VDD Gauge Check : "<< CL.dot(p5)<<endl;
  cout << "#VDD C2L : "<< -CL.dot(CL)<<" "<<-CL.dot(CL)/(qav.m2()*qbv.m2())/(4./p5.perp2())<<endl;
  cout << "#VDD p5 : "<<p5<<endl;
#endif
#if dump
  if (-CL.dot(CL)<0.) {
    idump=1;
    cout << "#VDD qa : "<<qav<<" "<<qav.m2()<<" "<<-qav.perp2()<<endl;
    cout << "#VDD qb : "<<qbv<<" "<<qbv.m2()<<" "<<-qbv.perp2()<<endl;
    cout << "#VDD p1 : "<<p1<<endl;
    cout << "#VDD p2 : "<<p2<<endl;
    cout << "#VDD Gauge Check : "<< CL.dot(p5)<<endl;
    cout << "#VDD C2L : "<< -CL.dot(CL)<<" "<<-CL.dot(CL)/(qav.m2()*qbv.m2())/(4./p5.perp2())<<endl;
    cout << "#VDD p5 : "<<p5<<endl;
    if (fabs(CL.dot(CL))<fabs(CL.dot(p5)))
      cout << "# Gauge check : FAIL\n";
    return 0.;
  } else {
    idump=0;
    return -CL.dot(CL);
  }
#else
  // Change behaviour: don't return zero, but its value whether or not it is positive
  // just check status() afterwards. Easier to implement lnR-matching this way

  //#if 0
//   CL2=-CL.dot(CL);
  if (-CL.dot(CL)<0.) //{
    //   if (fabs(CL.dot(p5))>fabs(CL.dot(CL))) // not sufficient!
    setstatus(8);
    //    return 0.;
  //  } else 
    return -CL.dot(CL);
  // #endif
//   return 4.*qav.m2()*qbv.m2()*shat/(sb1*sa1);
#endif
}

double CHMultijet::C2Lipatov(HepLorentzVector qav, HepLorentzVector qbv, HepLorentzVector p1, HepLorentzVector p2)
{
  HepLorentzVector temptrans=-(qav+qbv);
  HepLorentzVector p5=qav-qbv;
  HepLorentzVector CL=temptrans+(qav.m2()/p5.dot(p1)+2.*p5.dot(p2)/p1.dot(p2))*p1-p2*(qbv.m2()/p5.dot(p2)+2.*p5.dot(p1)/p1.dot(p2));
  
#if printoutput
  cout << "#Fadin qa : "<<qav<<endl;
  cout << "#Fadin qb : "<<qbv<<endl;
  cout << "#Fadin p1 : "<<p1<<endl;
  cout << "#Fadin p2 : "<<p2<<endl;
  cout << "#Fadin p5 : "<<p5<<endl;
  cout << "#Fadin Gauge Check : "<< CL.dot(p5)<<endl;
  cout << "#Fadin C2L : "<< -CL.dot(CL)<<" "<<-CL.dot(CL)/(qav.m2()*qbv.m2())/(4./p5.perp2())<<endl;
#endif
#if 0
  if (-CL.dot(CL)<0.)
    //   if (fabs(CL.dot(p5))>fabs(CL.dot(CL))) // not sufficient!
    return 0.;
  else 
#endif
    return -CL.dot(CL);
}

double CHMultijet::C2Lipatovots(CLHEP::HepLorentzVector qav, CLHEP::HepLorentzVector qbv, CLHEP::HepLorentzVector p1, CLHEP::HepLorentzVector p2)
{
  double kperp=(qav-qbv).perp();
  if (kperp>clambda)
    return C2Lipatov(qav, qbv, p1, p2)/(qav.m2()*qbv.m2());
  else {
    double Cls=(C2Lipatov(qav, qbv, p1, p2)/(qav.m2()*qbv.m2()));
    double temp=Cls-4./(kperp*kperp);
    // std::cout <<kperp <<" "<<temp<<" "<<4./(kperp*kperp)<<" "<<(C2Lipatov(qav, qbv, pa, pb, p1, p2)/(qav.m2()*qbv.m2()))<<std::endl;
    return temp;
  }
}

double CHMultijet::C2Lipatov(CLHEP::HepLorentzVector qav, CLHEP::HepLorentzVector qbv, CLHEP::HepLorentzVector pim, CLHEP::HepLorentzVector pip,CLHEP::HepLorentzVector pom, CLHEP::HepLorentzVector pop) // B
{
  CLHEP::HepLorentzVector temptrans=-(qav+qbv);
  CLHEP::HepLorentzVector p5=qav-qbv;
//   std::cout<<"#C2L, p2: "<<p5<<std::endl;
  CLHEP::HepLorentzVector CL=temptrans+qav.m2()*(1./p5.dot(pip)*pip+1./p5.dot(pop)*pop)/2.-qbv.m2()*(1./p5.dot(pim)*pim+1./p5.dot(pom)*pom)/2.+(pip*(p5.dot(pim)/pip.dot(pim)+p5.dot(pom)/pip.dot(pom)) + pop*(p5.dot(pim)/pop.dot(pim)+p5.dot(pom)/pop.dot(pom)) - pim*(p5.dot(pip)/pip.dot(pim) + p5.dot(pop)/pop.dot(pim)) - pom*(p5.dot(pip)/pip.dot(pom) + p5.dot(pop)/pop.dot(pom)) )/2.;
  //    CLHEP::HepLorentzVector CL=temptrans+qav.m2()*(1./p5.dot(pip)*pip+1./p5.dot(pop)*pop)/2.-qbv.m2()*(1./p5.dot(pim)*pim+1./p5.dot(pom)*pom)/2.+2.*(pip*(p5.dot(pim)/pip.dot(pim)) - pim*(p5.dot(pip)/pip.dot(pim) ) );
  
#if 0
  cout << "#Fadin pplus : "<<pip.plus()<<endl;
  cout << "#Fadin pminus : "<<pim.minus()<<endl;
  cout << "#Fadin pplus : "<<pop.plus()<<endl;
  cout << "#Fadin pminus : "<<pom.minus()<<endl;
  cout << "#Fadin qa : "<<qav<< " "<<qav.m2()<<std::endl;
  cout << "#Fadin qb : "<<qbv<<" "<<qbv.m2()<<std::endl;
//   cout << "#Fadin p1 : "<<p1<<endl;
//   cout << "#Fadin p2 : "<<p2<<endl;
  cout << "#Fadin p5 : "<<p5<<endl;
  cout << "#Fadin Gauge Check : "<< CL.dot(p5)<<endl;
  cout << "#Fadin C2L : "<< -CL.dot(CL)<<" "<<-CL.dot(CL)/(qav.m2()*qbv.m2())/(4./p5.perp2())<<endl;
#endif

    return -CL.dot(CL);
}

double CHMultijet::C2Lipatovots(HepLorentzVector qav, HepLorentzVector qbv, HepLorentzVector pim, HepLorentzVector pip, HepLorentzVector pom, HepLorentzVector pop)
{
  return C2Lipatov(qav, qbv, pim, pip, pom, pop)/(qav.m2()*qbv.m2());
}

void CHMultijet::AddParton(double p[4])
{
  HepLorentzVector pi(p[1],p[2],p[3],p[0]);
  partonmomenta.push_back(pi);
}

// void CHMultijet::AddParton1(double p[4])
// {
//   HepLorentzVector pi(p[1],p[2],p[3],p[0]);
//   gluons1.push_back(pi);
// }

// void CHMultijet::AddParton2(double p[4])
// {
//   HepLorentzVector pi(p[1],p[2],p[3],p[0]);
//   gluons2.push_back(pi);
// }

double CHMultijet::C2gHgm(HepLorentzVector p2, HepLorentzVector p1, HepLorentzVector pH)
{
  static double A=1./(3.*M_PI*v);
  // Implements Eq. (4.22) in hep-ph/0301013 with modifications to incoming plus momenta 
  double s12,p1p,p2p;
  COM p1perp,p3perp,phperp;
  // Determine first whether this is the case p1p\sim php>>p3p og the opposite
  s12=p1.invariantMass2(-p2);
  if (p2.pz()>0.) { // case considered in hep-ph/0301013
    p1p=p1.plus();
    p2p=p2.plus();
  } else { // opposite case
    p1p=p1.minus();
    p2p=p2.minus();
  }
  p1perp=p1.px()+COM(0,1)*p1.py();
  phperp=pH.px()+COM(0,1)*pH.py();
  p3perp=-(p1perp+phperp);

  COM temp=COM(0,1)*A/(2.*s12)*(p2p/p1p*conj(p1perp)*p3perp+p1p/p2p*p1perp*conj(p3perp));
  temp=temp*conj(temp);
  return temp.real();
}

double CHMultijet::C2gHgp(HepLorentzVector p2, HepLorentzVector p1, HepLorentzVector pH)
{
  static double A=1./(3.*M_PI*v);
  // Implements Eq. (4.23) in hep-ph/0301013
  double s12,php,p1p,phm;
  COM p1perp,p3perp,phperp;
  // Determine first whether this is the case p1p\sim php>>p3p og the opposite
  s12=p1.invariantMass2(-p2);
  if (p2.pz()>0.) { // case considered in hep-ph/0301013
    php=pH.plus();
    phm=pH.minus();
    p1p=p1.plus();
  } else { // opposite case
    php=pH.minus();
    phm=pH.plus();
    p1p=p1.minus();
  }
  p1perp=p1.px()+COM(0,1)*p1.py();
  phperp=pH.px()+COM(0,1)*pH.py();
  p3perp=-(p1perp+phperp);

  COM temp=-COM(0,1)*A/(2.*s12)*(conj(p1perp*p3perp)*pow(php/p1p,2)/(1.+php/p1p)+s12*(pow(conj(phperp),2)/(pow(abs(phperp),2)+p1p*phm)-pow(conj(p3perp)+(1.+php/p1p)*conj(p1perp),2)/((1.+php/p1p)*(pH.m2()+2.*p1.dot(pH)))));
  temp=temp*conj(temp);
  return temp.real();
}

double CHMultijet::C2qHqm(HepLorentzVector p2, HepLorentzVector p1, HepLorentzVector pH)
{
  static double A=1./(3.*M_PI*v);
  // Implements Eq. (4.22) in hep-ph/0301013
  double s12,p2p,p1p;
  COM p1perp,p3perp,phperp;
  // Determine first whether this is the case p1p\sim php>>p3p og the opposite
  s12=p1.invariantMass2(-p2);
  if (p2.pz()>0.) { // case considered in hep-ph/0301013
    p2p=p2.plus();
    p1p=p1.plus();
  } else { // opposite case
    p2p=p2.minus();
    p1p=p1.minus();
  }
  p1perp=p1.px()+COM(0,1)*p1.py();
  phperp=pH.px()+COM(0,1)*pH.py();
  p3perp=-(p1perp+phperp);

  COM temp=A/(2.*s12)*(sqrt(p2p/p1p)*p3perp*conj(p1perp)+sqrt(p1p/p2p)*p1perp*conj(p3perp));
  temp=temp*conj(temp);
  return temp.real();
}

double CHMultijet::FKLamplitude(std::vector<CLHEP::HepLorentzVector> emissions,CLHEP::HepLorentzVector ti,CLHEP::HepLorentzVector pa,CLHEP::HepLorentzVector pb,double ystart,double yend,fastjet::ClusterSequence * ev)
// assumed ystart<yend and the emission vector entries are ordered with increasing rapidity
// t is t-channel momentum at the low (in rapidity) end of the chain
{
  std::vector<fastjet::PseudoJet> jets;
  jets = sorted_by_rapidity((*ev).inclusive_jets(flags.jetptmin));

  double temp(1.);
  HepLorentzVector ta(ti),tb;
  HepLorentzVector g;
  // yim1=ystart;

  std::vector<CLHEP::HepLorentzVector>::iterator itemissions;

  for (itemissions=emissions.begin();itemissions!=emissions.end();itemissions++) {
    g=*(itemissions); // the momentum of the current emission
    // yi=g.rapidity();

    tb=ta-g; // the t-channel momentum AFTER emission

#if SmartThingCL
    static const double cut_par(.01); // of scale GeV^2
    bool jetmember(false);
    // Is the parton part of the jet of lowest rapidity?
    std::vector<fastjet::PseudoJet> constituents = (*ev).constituents(jets[0]);
    for (unsigned int ll=0;ll<constituents.size();ll++)
      // if (gcount==unsigned(constituents[ll].user_index())) 
      // if (g.invariantMass2(constituents[ll])<cut_par) 
      if (g.invariantMass2(CLHEP::HepLorentzVector(constituents[ll].px(),constituents[ll].py(),constituents[ll].pz(),constituents[ll].e()))<cut_par) 
	jetmember=true;
    // Is the parton part of the jet of highest rapidity?  A problem is the
    // construction of FKL-amplitude being called twice with a different set
    // of std::vector gluons1/2 with a numbering different to that of gluons
    // check instead on invariant mass (equal to zero) - slower!
    constituents = (*ev).constituents(jets[jets.size()-1]);      
    for (unsigned int ll=0;ll<constituents.size();ll++)
      // if (gcount==unsigned(constituents[ll].user_index())) 
      // if (g.invariantMass2(constituents[ll])<cut_par) 
      if (g.invariantMass2(CLHEP::HepLorentzVector(constituents[ll].px(),constituents[ll].py(),constituents[ll].pz(),constituents[ll].e()))<cut_par) 
	jetmember=true;

    // Define gs2 for logcorrect settings
    double gs2(0.);
    if (flags.logcorrect==0) {
      gs2=Halphas(renscale)*4.*M_PI;
    }
    else {
	gs2=4.*M_PI*Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(g.perp()/renscale)/(2.*M_PI));
    }
    if (jetmember)
      temp*=C2Lipatovots(ta,tb,pa,pb)*3.*gs2;
    else
      temp*=C2Lipatovots(-tb,-ta,pa,pb,partonmomenta.front(),partonmomenta.back())*3.*gs2;
#endif
//     if (getstatus()==8)
//       return 0.;

    //    temp*=exp(omega0(-ta.m2(),lambda*lambda,mu2)*fabs(yi-yim1)); // fabs() in case the gluons are not ordered in rapidity yi>yim1
    // yim1=yi;
    ta=tb; // the t-channel momentum BEFORE the NEXT emission
  }
  t2v=ta; // saves the final outgoing t-momentum in class variable t2

  return temp;
}

double CHMultijet::GeneratePHSPpoint()
// Generates a phase space point with unit weight
// one higgs, ng gluons, 2 extremal partons.
// then classifies event int nga gluons before, higgs, ngb gluons after (larger rapidity)
{
  int ng;
  double y,ysum,p[4],pt,r1,temp1,temp2,phi,yh,maxptsofar;
  double tscale,kt[2];
//   static const double ptmax(2000.);
//   static const double ptSQgen(40.*40.);
  double omega0n,qtSQgen,qtSQ,qtmod,pH[4];
//   double minv2after;
  double ptpar(40.);
  double sqrts;

  double pta[GLUEINT+2][2];
  double deltay[GLUEINT+2];

  lambda=clambda; // trouble is we don't know the value of lambda until after the whole generation

  double wt=1.;
  sqrts=2.*flags.Ebeam;

  // The parton of lowest rapidity:
#if 1
  // Generate the transverse momentum of gluon a
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  wt*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  Ht+=pt;
//   wtmatch*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;
  //  pt1=pt;
#else
  {
    double ptSQ;
    r1=ran.flat();
    ptSQ=ptSQgen/(exp(r1*log(1.+ptSQgen))-1.);
    wt*=2.*M_PI*(ptSQ)*(1.+(ptSQ)/ptSQgen)*log(1.+ptSQgen)/2.;
    pt=sqrt(ptSQ);
  }
#endif
  maxptsofar=pt;

  phi=2.*M_PI*ran.flat(); // Random angle in x,y-plane

  y=ysum=0.;
  

  p[1]=pt*cos(phi);
  p[2]=pt*sin(phi);
  pta[0][0]=pt*cos(phi);
  pta[0][1]=pt*sin(phi);
  deltay[0]=0.;

  kt[0]=-p[1];
  kt[1]=-p[2];

  tscale=kt[0]*kt[0]+kt[1]*kt[1];

  // Generate a rapidity span
  //  omega0n=0.6*min(omega0(tscale,lambda*lambda,renscale2),cut);
  //  if (flags.scalesetting==1||flags.scalesetting==2)
  renscale=maxptsofar*flags.scalefactor;   // This is corrected below, necessary for n
  omega0n=min(sf*omega0(tscale,lambda*lambda),cut);

  y=log(1.-ran.flat())/omega0n;
  wt*=-1./omega0n/exp(omega0n*y); // take out the exponential factor which was included 
                                  // in the generation of y - we want this in the 
                                  // FKL evaluation
  ysum+=y;
  deltay[1]=ysum;

  // Assume we have already chosen a regulater mu
  // then generate the gluon emissions
  
#if twojet
  ng=0;
#else
#if threejet
  ng=1; // a random int number of gluons between 0 and GLUEINT-1
#else  
  ng=int((GLUEINT)*ran.flat()); // a random int number of gluons between 0 and GLUEINT-1
  wt*=double(GLUEINT);
#endif
#endif
  for (int ii=1;ii<=ng;ii++) {
    
    // Generate a random (weighted) qti min: 0, max: ptmaxsq 
    r1=ran.flat();
    //      qtSQgen=pt1SQ;
//     qtSQgen=12.*12.;
//    qtSQgen=15.*15.;
//    qtSQgen=20.*20.;
//    qtSQgen=25.*25.;
    //    qtSQgen=30.*30.;
    //    qtSQgen=40.*40.;
    //    qtSQgen=50.*50.;
    qtSQgen=cqtsqgen;

    double t2=log(1.+qtSQgen);
    double t=exp(r1*t2);
    qtSQ=qtSQgen/(t-1.)-(1.-ccut*ccut);
    wt*=2.*M_PI*(qtSQ+1.-ccut*ccut)*(qtSQ+1.-ccut*ccut)*t*t2/qtSQgen/2.;

    wt/=16.*pow(M_PI,3);
    qtmod=sqrt(qtSQ);    
    Ht+=qtmod;
    if (qtmod>maxptsofar)
      maxptsofar=qtmod;

    // Generate random angle 
    phi=2.*M_PI*ran.flat();

    p[1]=pta[ii][0]=qtmod*cos(phi);
    p[2]=pta[ii][1]=qtmod*sin(phi);

    // keep track of the transverse t-channel scale in the evolution
    kt[0]-=p[1];
    kt[1]-=p[2];

    tscale=kt[0]*kt[0]+kt[1]*kt[1];

    // Generate a rapidity span after the emission
    //    omega0n=0.6*min(omega0(tscale,lambda*lambda,renscale2),cut);
    //    if (flags.scalesetting==1)
      renscale=maxptsofar*flags.scalefactor;  // This is corrected below, necessary for now.

    omega0n=min(sf*omega0(tscale,lambda*lambda),cut);
    y=log(1.-ran.flat())/omega0n;
    ysum+=y;
    deltay[ii+1]=ysum;
    wt*=-1./omega0n/exp(omega0n*y); // take out the exponential factor which was included 
                                    // in the generation of y - we want this in the 
                                    // FKL evaluation

  }
  // Generate the transverse momentum of gluon b
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  wt*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  Ht+=pt;

  phi=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  
  p[1]=pta[ng+1][0]=pt*cos(phi);
  p[2]=pta[ng+1][1]=pt*sin(phi);
  kt[0]-=p[1];
  kt[1]-=p[2];
  
  // Now we know the pt of the Higgs:
  pH[1]=kt[0];
  pH[2]=kt[1];
  pt=sqrt(pH[1]*pH[1]+pH[2]*pH[2]);
  if (!flags.DecayHiggs) // If the Higgs is treated as stable
    Ht+=pt;


  yh=-ymax+2.*ymax*ran.flat();
  wt*=2.*ymax;

  if (yh<-ymax) {
    setstatus(3);
    wt=0.;
    return wt;
  }
  if (yh>ymax) {
    setstatus(2);
    wt=0.;
    return wt;
  }

  r1=ran.flat();
  double sH=flags.mH*(flags.mH + flags.GammaH*tan((M_PI*r1)/2. + (-1. + r1)*atan(flags.mH/flags.GammaH)));
  double mHperp=sqrt(pt*pt+sH);
  pH[3]=mHperp*sinh(yh);
  pH[0]=mHperp*cosh(yh) ; // Energy component, massive H
  
  pHv.set(pH[1],pH[2],pH[3],pH[0]);
  xa+=(pHv.e()-pHv.z())/sqrts;
  xb+=(pHv.e()+pHv.z())/sqrts;

  if (flags.DecayHiggs) {
    CLHEP::HepLorentzVector g1v,g2v;
    double pgH[4];
    double pgHs=sqrt(sH)/2.;
    // Choose theta and phi
    double pptheta=2.*M_PI*ran.flat();
    double cosppphi=2.*ran.flat()-1.;  
    double sinphi=sqrt(1.-cosppphi*cosppphi);  // Know 0 < phi < pi

    // construct 4-vector in H rest frame
    pgH[0]=pgHs;
    pgH[1]=pgHs*cos(pptheta)*sinphi;
    pgH[2]=pgHs*sin(pptheta)*sinphi;
    pgH[3]=pgHs*cosppphi;
    g1v.set(pgH[1],pgH[2],pgH[3],pgH[0]);
    g2v.set(-pgH[1],-pgH[2],-pgH[3],pgH[0]);

    // translate to lab frame
    CLHEP::HepLorentzVector pp1=g1v.boost(pHv.boostVector());
    CLHEP::HepLorentzVector pp2=g2v.boost(pHv.boostVector());
  
    Ht+=pp1.perp()+pp2.perp(); // If Higgs boson decays, add transverse momenta
                               // of decay products to Ht
    photonmomenta.push_back(pp1);
    photonmomenta.push_back(pp2);
    wt*=flags.HBRgammagamma;
  }

  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double yoffset;
    double ycenter=ysum/(2.);
    {
      double lninvr1,r1,r2,temp,a(2.);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      wt=wt*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to gluons and xa,xb
    for (int ii=0;ii<=ng+1;ii++) {
      double yi=deltay[ii]+yoffset;
      if (fabs(yi)>ymax) {
	setstatus(2);
	wt=0.;
	return wt;
      }
      
      pt=sqrt(pta[ii][0]*pta[ii][0]+pta[ii][1]*pta[ii][1]);
      
      p[0]=pt*cosh(yi); // the energy component
      p[1]=pta[ii][0];
      p[2]=pta[ii][1];
      p[3]=pt*sinh(yi); // the z-component

      xa+=(p[0]-p[3])/sqrts;
      xb+=(p[0]+p[3])/sqrts;
      AddParton(p);

    }
    
  }
  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);

  
//   if (!(partonmomenta.front().rapidity()<pHv.rapidity()&&partonmomenta.back().rapidity()>pHv.rapidity())) {
//     wt=0.;
//     return wt;
//   }
//   if (((unsigned int)nga)!=gluons1.size()&&((unsigned int)ngb)!=gluons2.size())
//     cout << "Bummer!\n";
#if 0
  // Check of energy and momentum conservation
  std::vector<CLHEP::HepLorentzVector>::iterator itemissions;
  HepLorentzVector outsum(0.,0.,0.,0.);
  for (itemissions=partonmomenta.begin();itemissions!=partonmomenta.end();itemissions++) {
    outsum+=*itemissions;
  }
  outsum+=pHv;
  cout << outsum <<" "<<pa+pb<<" "<<outsum-pa-pb<<endl;
#endif

  return wt;
}

double CHMultijet::MEt2(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pH, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb,bool checkmomenta,fastjet::ClusterSequence * ev)
{
  double wt(0.);
  // double ura=flags.scale;  // scales now set to renscale
  // double urb=flags.scale;

  ////////////////////////////////////////////////////////////
  // Calculate the factors making up the cross section      //
  ////////////////////////////////////////////////////////////

    // Cluster into jets already at this stage, so MEt2 knows whether a
    // parton is inside an extremal jet
    // Count how many hard jets we have
    std::vector<CLHEP::HepLorentzVector>::iterator itpm,ijpm;
    std::vector<fastjet::PseudoJet>::iterator fijpm;
    
    if (checkmomenta) {
      if (jets.size()>=2) {
	bool jetmember1(false);
	// Check to see if the extremal partons are included in the extremal jets
	std::vector<fastjet::PseudoJet> constituents = (ev)->constituents(jets[0]);
	for (unsigned int ll=0;ll<constituents.size();ll++)
	  if (0==unsigned(constituents[ll].user_index())) 
	    jetmember1=true;
	
	bool jetmember2(false);
	// Check to see if the extremal partons are included in the extremal jets
	constituents = (ev)->constituents(jets[jets.size()-1]);
	for (unsigned int ll=0;ll<constituents.size();ll++)
	  if ((partons.size()-1)==unsigned(constituents[ll].user_index())) 
	    jetmember2=true;
	if (!(jetmember2&&jetmember1)) {
	  wt=0;
	  setstatus(18);
	  return wt;
	}
      } else { // we didn't even have two jets - reject outright!
	wt=0;
	setstatus(17);
	return wt;
      }
    }

    std::vector<CLHEP::HepLorentzVector> gluons1,gluons2;

    // Now setup the gluons1 and gluons2
    {
      std::vector<CLHEP::HepLorentzVector>::iterator itemissions,itemissions2;
      itemissions=itemissions2=partons.begin();
      itemissions++;
      itemissions2++;
      itemissions2++;
      while (itemissions2!=partons.end()) {
	if ((*itemissions).rapidity()<pH.rapidity()) {
	  gluons1.push_back(*itemissions);
	} else {
	  gluons2.push_back(*itemissions);
	}
	itemissions++;
	itemissions2++;
      }
    }

    double shat2=pow(pa.invariantMass2(pb),2);

  // the approximation to the matrix element squared
  if (partons.front().rapidity()<pH.rapidity()&&partons.back().rapidity()>pH.rapidity())
    {  // Higgs between jets
       HepLorentzVector pav,pbv,ptemp1v,ptemp2v;

      //    p2v.set(0,0,-xa*sqrts/2.,xa*sqrts/2.); // Incoming parton (neg. z momentum)
      //    p2v=pa;
      //    p4v.set(0,0,xb*sqrts/2.,xb*sqrts/2.);  // Incoming parton (pos. z momentum)
      //    p4v=pb;
      //    pav.set(pa[1],pa[2],pa[3],pa[0]);      // Outgoing parton with smallest rapidity
      pav=partons.front();
      //    pbv.set(pb[1],pb[2],pb[3],pb[0]);      // Outgoing parton with largest rapidity
      pbv=partons.back();

      ptemp1v=pa-pav;
      ptemp2v=-pb+pbv;

      t1v=ptemp1v;

      t1vsave=t1v;

      wt=FKLamplitude(gluons1,t1v,pa,pb,pav.rapidity(),pH.rapidity(),ev); // fudge:sets t2v, should really be implemented as a class;

      pHvi=t2v;

      t2v=pHvi-pH; // the t-channel incoming momentum for the second FKL amplitude
      t2vsave=t2v;

      wt*=FKLamplitude(gluons2,t2vsave,pa,pb,pH.rapidity(),pbv.rapidity(),ev); // fudge:sets t2v, should really be implemented as a class;

      double extcouplings=0.;
      if (flags.logcorrect==0) {
	extcouplings=Halphas(renscale)*Halphas(renscale);
      }
      else {
	extcouplings=pow(Halphas(renscale),2)*(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI));
      }
      wt*=pdfta*pdftb*extcouplings*pow(Halphas(flags.mH),2)*9./(128.*pow(M_PI,3)*shat2)/pow(6.*M_PI*v,2);

      if (aptype==21&&bptype==21) // gg initial state
	wt*=MH2gg(pbv,pb,pav,pa,-pHvi+pH,-pHvi);
      else if (aptype==21&&bptype!=21) {
	if (bptype > 0)
	  wt*=MH2qg(pbv,pb,pav,pa,-pHvi+pH,-pHvi)*4./9.;
	else
	  wt*=MH2qbarg(pbv,pb,pav,pa,-pHvi+pH,-pHvi)*4./9.;
      }	
      else if (bptype==21&&aptype!=21) {
	if (aptype > 0)
	  wt*=MH2qg(pav,pa,pbv,pb,-pHvi,-pHvi+pH)*4./9.;
	else
	  wt*=MH2qbarg(pav,pa,pbv,pb,-pHvi,-pHvi+pH)*4./9.;
      }
      else { // they are both quark 
	if (bptype>0) {
	  if (aptype>0)
	    wt*=MH2qQ(pbv,pb,pav,pa,-pHvi+pH,-pHvi)*4.*4./(9.*9.);
	  else
	    wt*=MH2qQbar(pbv,pb,pav,pa,-pHvi+pH,-pHvi)*4.*4./(9.*9.);
	}
	else {
	  if (aptype>0)
	    wt*=MH2qQbar(pav,pa,pbv,pb,-pHvi,-pHvi+pH)*4.*4./(9.*9.);
	  else
	    wt*=MH2qbarQbar(pbv,pb,pav,pa,-pHvi+pH,-pHvi)*4.*4./(9.*9.);
	}
      }
    } //higgs between jets end;
  else if (partons.front().rapidity()>pH.rapidity()) { // if Higgs boson rapidity is 
                                                       // smaller than rapidity of parton a
    double ans,t1,t2;
    HepLorentzVector pi1,po1,pi2,po2,ptemp1v,ptemp2v;
    //     cout <<pH.rapidity()<< " "<< partons.front().rapidity()<< " "<<partons.back().rapidity()<<endl;
    pi1=pa;
    po1=partons.front();
    ptemp1v=pi1-po1-pH;
    t1=ptemp1v.m2();
    pi2=pb;
    po2=partons.back();
    ptemp2v=po2-pi2;
    t2=ptemp2v.m2();

    wt=FKLamplitude(gluons2,ptemp1v,pa,pb,po1.rapidity(),po2.rapidity(),ev); // fudge:sets t2v, should really be implemented as a class;

    // For gluons:
    if (aptype==21) 
      ans=pdfta*pdftb*Halphas(renscale)*Halphas(renscale)*pow(Halphas(flags.mH),2)*9./(32.*pow(M_PI,3))*(C2gHgp(pi1,po1,pH)+C2gHgm(pi1,po1,pH))/(t1*t2);
    else
      ans=pdfta*pdftb*Halphas(renscale)*Halphas(renscale)*pow(Halphas(flags.mH),2)*9./(32.*pow(M_PI,3))*(C2qHqm(pi1,po1,pH))/(t1*t2);
    wt*=ans;
    if (aptype!=21)
      wt*=4./9.;
    if (bptype!=21)
      wt*=4./9.;
  } else {  // if Higgs boson rapidity is larger than rapidity of parton b
    double ans,t1,t2;
    HepLorentzVector pi1,po1,pi2,po2,ptemp1v,ptemp2v;

    pi1=pa;
    po1=partons.front();
    ptemp1v=pi1-po1;
    t1=ptemp1v.m2();
    pi2=pb;
    po2=partons.back();
    ptemp2v=pH+po2-pi2;
    t2=ptemp2v.m2();

    wt=FKLamplitude(gluons1,ptemp1v,pa,pb,po1.rapidity(),po2.rapidity(),ev); // fudge:sets t2v, should really be implemented as a class;

    // For gluons:
    if (bptype==21) 
      ans=pdftb*pdfta*Halphas(renscale)*Halphas(renscale)*pow(Halphas(flags.mH),2)*9./(32.*pow(M_PI,3))*(C2gHgp(pi2,po2,pH)+C2gHgm(pi2,po2,pH))/(t1*t2);
    else
      ans=pdftb*pdfta*Halphas(renscale)*Halphas(renscale)*pow(Halphas(flags.mH),2)*9./(32.*pow(M_PI,3))*(C2qHqm(pi2,po2,pH))/(t1*t2);
    wt*=ans;

    if (aptype!=21)
      wt*=4./9.;
    if (bptype!=21)
      wt*=4./9.;
  }
  return wt;
}

double CHMultijet::VirtualCorrections()
// oh - and to all orders
// well, at least the log piece

// The typical complication compared to the pure jet case: The Higgs momentum
// needs insertion
{
  double vt(0.0);
  HepLorentzVector tiv,p1,p2;
  double y1,y2;
  tiv=pa-partonmomenta.front();
  
  double yH=pHv.rapidity();
  bool Hsubtracted(false);
  if (yH<partonmomenta.front().rapidity()) {
    tiv-=pHv;
    Hsubtracted=true;
  }

  std::vector<CLHEP::HepLorentzVector>::iterator itpartonmomenta;
  
  itpartonmomenta=partonmomenta.begin();
  p1=*(itpartonmomenta);
  y1=p1.rapidity();
  itpartonmomenta++;
  for (;itpartonmomenta!=partonmomenta.end();itpartonmomenta++) {
    p2=*(itpartonmomenta);
    y2=p2.rapidity();
    if (Hsubtracted==true||yH>y2)
      // vt*=exp(omega0(tiv.perp2(),lambda*lambda)*(y2-y1));
      vt+=omega0(tiv.perp2(),lambda*lambda)*(y2-y1);
    else if (Hsubtracted==false&&(yH<y2&&yH>y1)) {
      // vt*=exp(omega0(tiv.perp2(),lambda*lambda)*(yH-y1));
      vt+=omega0(tiv.perp2(),lambda*lambda)*(yH-y1);
      tiv-=pHv;
      // vt*=exp(omega0(tiv.perp2(),lambda*lambda)*(y2-yH));
      vt+=omega0(tiv.perp2(),lambda*lambda)*(y2-yH);
      Hsubtracted=true;
    }
    tiv-=p2;
    p1=p2;
    y1=y2;
  }

  return exp(vt);
}

double CHMultijet::weight()
{
  //  double temp1,temp2,pttemp[2],ptHvertex[2],r1,pt1SQ,pt2SQ,pt1,pt2,q1SQ,q2SQ,t1,t2;
  double ufa,ufb,ura,urb; // factorisation scale
  
  //  double ptH,mHperp;
  //  double Smatch,shat2,HElimit;

#if printoutput
  cout << "--------------------\n";
#endif

  setstatus(0);
  wt=1;
  
#if dijetLO==1
  wt*=Generate2jphsppoint();
#else
  wt*=GeneratePHSPpoint();
#endif
  if (getstatus()!=0)
    return 0.;

  if (xa>1.0||xb>1.0) {
    setstatus(5);
    wt=0.;
    return wt;
  } 

  // Cluster into jets already at this stage to define PTMAX for scales
  // Count how many hard jets we have
  std::vector<CLHEP::HepLorentzVector>::iterator itpm,ijpm;
  std::vector<fastjet::PseudoJet>::iterator fijpm;
  
  std::vector<fastjet::PseudoJet> avec;
  
  unsigned int i(0);
  for (itpm=partonmomenta.begin();itpm!=partonmomenta.end();itpm++) {
    avec.push_back(*(itpm));
    avec.at(i).set_user_index(i); // keep track of the partons. These are now numbered 
    // according to their rapidity (increasing)
    i++;
  }
  
  /** fastjet find clusters */
  fastjet::ClusterSequence ev(avec, flags.jet_def);
  
  /** Retrieve the final state jets from fastjet Sorted by Rapidity*/
  jets = sorted_by_rapidity(ev.inclusive_jets(flags.jetptmin));
  //
  
  if (jets.size()>=2) {
    bool jetmember1(false);
    // Check to see if the extremal partons are included in the extremal jets
    std::vector<fastjet::PseudoJet> constituents = (ev).constituents(jets[0]);
    for (unsigned int ll=0;ll<constituents.size();ll++)
      if (0==unsigned(constituents[ll].user_index())) 
    	jetmember1=true;
    
    bool jetmember2(false);
    // Check to see if the extremal partons are included in the extremal jets
    constituents = (ev).constituents(jets[jets.size()-1]);
    for (unsigned int ll=0;ll<constituents.size();ll++)
      if ((partonmomenta.size()-1)==unsigned(constituents[ll].user_index())) 
    	jetmember2=true;
    if (!(jetmember2&&jetmember1)) {
      //      std::cout<<jetmember1<<"  "<<jetmember2<<"  ";
      wt=0;
      setstatus(18);
      return wt;
    }
    // if (flags.scalesetting==1) {//  Find and define PTMAX   //
    //   std::vector<fastjet::PseudoJet> ptjets = sorted_by_pt(jets);
    //   renscale=ptjets.at(0).perp()*flags.scalefactor;
    // }
    // else if (flags.scalesetting==2)
    //   renscale=geometricmean(&jets)*flags.scalefactor;
  } 
  else { // we didn't even have two jets - reject outright!
    wt=0;
    setstatus(17);
    return wt;
  }

  std::vector<fastjet::PseudoJet> ptjets = sorted_by_pt(jets);
#if calcscaleunc
  renscale=ptjets.at(0).perp()*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=flags.scale*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=Ht/2.*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=geometricmean(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=invmass(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=ptjets.at(0).perp()*flags.scalefactor;
#else
  if (flags.scalesetting==0) //max pt(jets)
    renscale=ptjets.at(0).perp()*flags.scalefactor;
  else if (flags.scalesetting==1) //fixed scale
    renscale=flags.scale;
  else if (flags.scalesetting==2) // Ht/2
    renscale=Ht/2.*flags.scalefactor;
  else if (flags.scalesetting==3) //geometric mean of pt(jets)
    renscale=geometricmean(&ptjets)*flags.scalefactor;
#endif

  ufa=ura=renscale;    // PTMAX set a few lines above if necessary
  ufb=urb=renscale;    // O'wise pre-set in Event()

  // Implement LH interface:
  // Choose incoming parton species and assign pdfs accordingly
  wt*=GenerateFKLIncomingParticles(xa,ufa,xb,ufb);
  if (status!=0) { // if we failed in picking a sensible incoming state
    wt=0.;
    return wt;
  }

  if (flags.LHoutput)
    SetupLHInformation();

  double MEsq=MEt2(partonmomenta,pHv,pa,pb,true,&ev); // true: use jet-member test

  wt*=MEsq;

#if calcscaleunc
    std::vector<double> MEs;
    std::vector<double> virtuals;
    std::vector<double> pdfs;
    {
      double orgscale=renscale;
      for (unsigned int ii=0;ii<scales.size();ii++) {
	renscale=scales.at(ii);
	MEs.push_back(MEt2(partonmomenta,pHv,pa,pb,true,&ev));
#if dijetLO==0&&trijetLO==0&&fourjetLO==0
	virtuals.push_back(VirtualCorrections());
#endif
	pdfs.push_back(pdfpt(0,xa,renscale,aptype)*pdfpt(1,xb,renscale,bptype));
      }
      renscale=orgscale; // put things back to normal
    }
#endif
#if dijetLO==0&&trijetLO==0&&fourjetLO==0
  double vt=VirtualCorrections();
  wt*=vt;
#endif
  if (flags.match) {
      
    int njets=int(jets.size());
    if (njets>1&&njets<4) { // if the right number of jets are detected, i.e. not too many

      // cluster all partons into the identified jets
      std::vector<CLHEP::HepLorentzVector> jetlist,fastjetlist,nojetlist;
      CLHEP::HepLorentzVector zerovec;
      zerovec.set(0.,0.,0.,0.);

      for (fijpm=jets.begin();fijpm!=jets.end();fijpm++) {
	jetlist.push_back(zerovec);
	CLHEP::HepLorentzVector tempfast;
	tempfast.set((*(fijpm)).px(),(*(fijpm)).py(),(*(fijpm)).pz(),(*(fijpm)).e());
	fastjetlist.push_back(tempfast);
      }

      // Add all the partons to either jet-list or no-jet-list

      for (unsigned int ii=0;ii<partonmomenta.size();ii++) {
	bool included(false);
	// check to see if ii is included in one of the already existing jets
	for (unsigned int jj=0;jj<jets.size();jj++) {
	  // get constituents:
	  vector<fastjet::PseudoJet> constituents = ev.constituents(jets[jj]);
	  for (unsigned int ll=0;ll<constituents.size();ll++) {
	    if (ii==unsigned(constituents[ll].user_index())) {
#if printout
	      CLHEP::HepLorentzVector tempfast;
	      tempfast.set(constituents[ll].px(),constituents[ll].py(),constituents[ll].pz(),constituents[ll].e());
	      std::cout << "Found one vector in jet : "<<tempfast<<std::endl;
#endif
	      // add to jet vector
	      jetlist[jj]+=partonmomenta[ii];
	      included=true;
	    }
	  }
	}
	if (!included) { // add to list of non-included particles
	  nojetlist.push_back(partonmomenta[ii]);
	}
      }

#if printout
      std::cout<< "Fastjet Jets: "<<jets.size()<<std::endl;
      std::cout<< "partonmomenta size : "<<partonmomenta.size()<<std::endl;
      std::cout<< "nojetlist size before: "<<nojetlist.size()<<std::endl;
#endif

      // Now cluster the rest according to min inv. mass compared to already existing jets
      while (nojetlist.size()>0) {
	// cycle through the list of unclustered partons and the list of jets to find 
	// the pair of minimum invariant mass
	double smin=1e10;
	unsigned int ii_save(10),jj_save(10);
	for (unsigned int ii=0;ii<nojetlist.size();ii++) {
	  for (unsigned int jj=0;jj<jetlist.size();jj++) {
	    double temp;
	    temp=jetlist[jj].invariantMass2(nojetlist[ii]);
	    if (temp<smin) {
	      smin=temp;
	      ii_save=ii;
	      jj_save=jj;
	    }
	  }
	} // now we have found the smallest smin, and saved info in ii_save, jj_save;
	jetlist.at(jj_save)+=nojetlist[ii_save];
	nojetlist.erase(nojetlist.begin()+ii_save);
      }
      
#if printout
      std::cout<< "nojetlist size after: "<<nojetlist.size()<<std::endl;
#endif

#if printout
      std::cout<<"FastJet versus OurJet"<<std::endl;
      std::cout<<"Length: "<<fastjetlist.size()<<"  "<<jetlist.size()<<std::endl;
      std::cout<<"Number of partons: "<< partonmomenta.size()<<std::endl;
#endif
      double xE,xz;
      xE=pHv.e();
      xz=pHv.pz();
      for (int icount=0;icount<njets;icount++) {
	CLHEP::HepLorentzVector temp=fastjetlist.at(icount);
	CLHEP::HepLorentzVector diff=temp-jetlist.at(icount);
#if printout
	double difflength=diff.x()*diff.x()+diff.y()*diff.y()+diff.z()*diff.z();
	double templength=temp.x()*temp.x()+temp.y()*temp.y()+temp.z()*temp.z();
	std::cout<<"Diff Length "<<icount<<": "<<difflength<<" ratio= "<<difflength/templength<<" "<<(ev.constituents(jets[icount])).size()<<" "<<" "<<jets[icount].m2()<<" "<<jetlist.at(icount).m2()<<std::endl;
#endif
	temp=jetlist.at(icount);
	// keep the rapidities fixed 
	double mperp=temp.perp();
	double rapidity=jets.at(icount).rapidity();
	jetlist.at(icount).set(temp.px(),temp.py(),mperp*sinh(rapidity),mperp*cosh(rapidity));
#if printout
	std::cout << "Our Jet "<<icount<<" : "<<jetlist.at(icount)<<std::endl;
	std::cout << "Fastjet Jet "<<icount<<" : "<<fastjetlist.at(icount)<<std::endl;
#endif
	//  Now we calculate new incoming particles
	xE+=(jetlist.at(icount)).e();
	xz+=(jetlist.at(icount)).z();
	  
      }
      CLHEP::HepLorentzVector npa,npb;
      npa.set(0,0,-(xE-xz)/2.,(xE-xz)/2.);
      npb.set(0,0,(xE+xz)/2.,(xE+xz)/2.);
#if printout
      std::cout<<"'pa's:"<<pa<<"  "<<npa<<std::endl;
      std::cout<<"'pb's:"<<pb<<"  "<<npb<<std::endl;
      HepLorentzVector outsum(0.,0.,0.,0.);
      for (unsigned int ii=0;ii<jetlist.size();ii++) {
	outsum+=CLHEP::HepLorentzVector(jetlist.at(ii));
      }
      outsum+=pHv;
      cout << outsum <<" "<<npa+npb<<" "<<outsum-npa-npb<<endl;
#endif
      
#if 1
      double matchreweight(0.);
#if printout
      std::cout << "jetlist.size() : "<<jetlist.size()<<std::endl;
#endif
      switch (jetlist.size()) {
      case 2 : 
#if printout
	std::cout<<"---"<<std::endl;
	std::cout<<MG(jetlist,pHv,npa,npb)<<std::endl;
	std::cout<<MEt2(partonmomenta,pHv,pa,pb,true,&ev)<<std::endl;
	std::cout<<MEt2(jetlist,pHv,npa,npb,false,&ev)<<std::endl;
	std::cout<<MG(jetlist,pHv,npa,npb)/MEt2(jetlist,pHv,npa,npb,false,&ev)<<std::endl; //ev never used
#endif
	//	if (jetlist.front().rapidity()<pHv.rapidity()&&jetlist.back().rapidity()>pHv.rapidity()) {
	matchreweight=MG(jetlist,pHv,npa,npb)/MEt2(jetlist,pHv,npa,npb,false,&ev); //ev never used
	wt*=matchreweight;
	//	}
	//	std::cout<<"case 2 match: "<<MG(jetlist,npa,npb,alepton.p,lepton.p)<<"\n";
	break;
      case 3 : 
	matchreweight=MG3(jetlist,pHv,npa,npb)/MEt2(jetlist,pHv,npa,npb,false,&ev); //ev never used
	wt*=matchreweight;
#if printout
	std::cout<<"3jet matching : "<<matchreweight<<std::endl;
#endif
      	break;
      default:
	break;
      }
#endif
    }
  }
  

  // GeV^(-2)->micro barn
  wt*=389.379292;
  wt*=1E6; // micro barn-> pico barn

//   const double tau = pHv.m2()/(4.*mt*mt);
//   const double series_t = 1. + tau*(7./30. + tau*(2./21. + tau*26./525.));
    
//   wt*=series_t*series_t;

  if (isnan(wt)) {
    setstatus(6);
  }

  if (wt<0) {
    setstatus(13);
  }

#if 0
//   if (fabs(wt)>1.6E5) {
//   if (fabs(wt)>5E8) {
//   if (fabs(wt)>4.4E6) {
  if (fabs(wt)>2.7E5) {  // Cuts off around 12.5
    setstatus(7);
  }
#endif  
  
#if calcscaleunc
  for (unsigned int ii=0;ii<scales.size();ii++) {
#if dijetLO==0&&trijetLO==0&&fourjetLO==0
    double temp=wt*MEs.at(ii)/MEsq*virtuals.at(ii)/vt*pdfs.at(ii)/(pdfta*pdftb);
#else
    double temp=wt*MEs.at(ii)/MEsq*pdfs.at(ii)/(pdfta*pdftb);
#endif
    weights.push_back(temp);
  }
#endif
  
  return wt;
}

double CHMultijet::Generate2jphsppoint()
{
  double temp1,temp2,p[4],pt,px,py,pz,E,r1;
  double weight=1.;
  HepLorentzVector p1,p2;
  double sqrts=2.*flags.Ebeam;
  double y;
#if old2jphsp
#else
  double pta[2][2];
  double deltay[2];
  double phia[2];
#endif

  // generate momenta

  // Generate the transverse momentum of p3
  temp1=atan((flags.partonptmax-flags.jetptmin)/40.0);
  r1=ran.flat();
  pt=flags.jetptmin+40.0*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;
  Ht+=pt;

#if old2jphsp
  {
    double lninvr1,r1,r2,temp,a;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    a=1.4;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    //    y=temp-2.;
    y=temp-3;
    weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
  }

  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  
  p[0]=pt*cosh(y); // the energy component
  p[1]=pt*cos(temp1);
  p[2]=pt*sin(temp1);
  p[3]=pt*sinh(y); // the z-component
    
  xa+=(p[0]-p[3])/sqrts;
  xb+=(p[0]+p[3])/sqrts;

  p1.set(p[1],p[2],p[3],p[0]);
#else
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  phia[0]=temp1;
  pta[0][0]=pt*cos(temp1);
  pta[0][1]=pt*sin(temp1);
  deltay[0]=0.;
#endif


  // Generate the transverse momentum of p4
  temp1=atan((flags.partonptmax-flags.jetptmin)/40.0);
  r1=ran.flat();
  pt=flags.jetptmin+40.0*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;
  Ht+=pt;
  
#if old2jphsp
  // Generate a rapidity difference to jet 2
  y=y+ran.flat()*12.; // 5-(-1) instead
  weight*=12.;

  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  
  p[0]=pt*cosh(y); // the energy component
  p[1]=pt*cos(temp1);
  p[2]=pt*sin(temp1);
  p[3]=pt*sinh(y); // the z-component
    
  xa+=(p[0]-p[3])/sqrts;
  xb+=(p[0]+p[3])/sqrts;
  p2.set(p[1],p[2],p[3],p[0]);
#else
  y=ran.flat()*10;
  weight*=10.;
  deltay[1]=y;
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  phia[1]=temp1;
  pta[1][0]=pt*cos(temp1);
  pta[1][1]=pt*sin(temp1);
#endif
  // The Higgs:
  // Generate a y Gaussian distributed around 0
 {
    double lninvr1,r1,r2,temp,a;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    a=1.6;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    y=temp;
    weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
 }
  r1=ran.flat();
  double sH=flags.mH*(flags.mH + flags.GammaH*tan((M_PI*r1)/2. + (-1. + r1)*atan(flags.mH/flags.GammaH)));

#if old2jphsp
  px=-(p1.x()+p2.x());
  py=-(p1.y()+p2.y());
#else
  px=-(pta[0][0]+pta[1][0]);
  py=-(pta[0][1]+pta[1][1]);
#endif
  double mHperp=sqrt(px*px+py*py+sH);
  pz=mHperp*sinh(y);
  E=mHperp*cosh(y);

  xa+=(E-pz)/sqrts;
  xb+=(E+pz)/sqrts;
  pHv.set(px,py,pz,E);

  if (flags.DecayHiggs) {
    CLHEP::HepLorentzVector g1v,g2v;
    double pgH[4];
    double pgHs=sqrt(sH)/2.;
    // Choose theta and phi
    double pptheta=2.*M_PI*ran.flat();
    double cosppphi=2.*ran.flat()-1.;  
    double sinphi=sqrt(1.-cosppphi*cosppphi);  // Know 0 < phi < pi

    // construct 4-vector in H rest frame
    pgH[0]=pgHs;
    pgH[1]=pgHs*cos(pptheta)*sinphi;
    pgH[2]=pgHs*sin(pptheta)*sinphi;
    pgH[3]=pgHs*cosppphi;
    g1v.set(pgH[1],pgH[2],pgH[3],pgH[0]);
    g2v.set(-pgH[1],-pgH[2],-pgH[3],pgH[0]);

    // translate to lab frame
    CLHEP::HepLorentzVector pp1=g1v.boost(pHv.boostVector());
    CLHEP::HepLorentzVector pp2=g2v.boost(pHv.boostVector());
  
    photonmomenta.push_back(pp1);
    photonmomenta.push_back(pp2);
    Ht+=pp1.perp()+pp2.perp();
    wt*=flags.HBRgammagamma;
  } else {
    Ht+=pHv.perp();
  }

#if old2jphsp
#else
  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double ycenter=deltay[1]/2.;
    double yoffset;
    
    {
      double lninvr1,r1,r2,temp,a(2.);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to gluons and xa,xb
    double yi=deltay[0]+yoffset;
    if (fabs(yi)>ymax) {
      setstatus(2);
      weight=0.;
      return weight;
    }
    pt=sqrt(pta[0][0]*pta[0][0]+pta[0][1]*pta[0][1]);
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pt*cos(phia[0]);
    p[2]=pt*sin(phia[0]);
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    p1.set(p[1],p[2],p[3],p[0]);

    yi=deltay[1]+yoffset;
    if (fabs(yi)>ymax) {
      setstatus(2);
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[1][0]*pta[1][0]+pta[1][1]*pta[1][1]);
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pt*cos(phia[1]);
    p[2]=pt*sin(phia[1]);
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    p2.set(p[1],p[2],p[3],p[0]);
  }
//     cout <<"--------------------\n";
//     cout << p1 << endl<<p2<<endl;
#endif


  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);
  
//   cout<< "----------\n" << pa<<"\n"<<pb<<"\n";
  partonmomenta.push_back(p1);
  partonmomenta.push_back(p2);

  // Check if it is a 2-jet phase space point:
  std::vector<fastjet::PseudoJet> avec;
  HepLorentzVector pi;
  
  for (int ii=0;ii<int(partonmomenta.size());ii++) {
    pi=getpartvn(ii);
    avec.push_back(pi);
  }
  
  // /**set fastjet flags */
  // double Rparam = 0.7;
  // fastjet::Strategy strategy = fastjet::Best;
  // fastjet::RecombinationScheme recomb_scheme = fastjet::E_scheme;
    
  // /** Define jet for fastjet*/
  // fastjet::JetDefinition jet_def(fastjet::kt_algorithm, Rparam, recomb_scheme, strategy);
    
  /** fastjet find clusters */
  //  fastjet::ClusterSequence ev(avec, jet_def);
  fastjet::ClusterSequence ev(avec, flags.jet_def);
    
  /** Retrieve the final state jets from fastjet Sorted by Rapidity*/
  std::vector<fastjet::PseudoJet> jets = sorted_by_pt(ev.inclusive_jets(flags.jetptmin));
    
  /** Save the number of final state jets */
  int njets=int(jets.size());

  if (njets!=2) {
    setstatus(11);
    weight=0;
    return weight;
  } 
    
#if calcscaleunc
  std::vector<fastjet::PseudoJet> ptjets = sorted_by_pt(jets);  
  renscale=ptjets.at(0).perp()*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=flags.scale*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=Ht/2.*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=geometricmean(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=invmass(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=ptjets.at(0).perp()*flags.scalefactor;
#else
  if (flags.scalesetting==0) //max pt(jets)
    renscale=ptjets.at(0).perp()*flags.scalefactor;
  else if (flags.scalesetting==1) //fixed scale
    renscale=flags.scale;
  else if (flags.scalesetting==2) // Ht/2
    renscale=Ht/2.*flags.scalefactor;
  else if (flags.scalesetting==3) //geometric mean of pt(jets)
    renscale=geometricmean(&ptjets)*flags.scalefactor;
#endif


  return weight;
}

double CHMultijet::weight1() // The two-jet cross section
{
  wt=1.;

  wt*=Generate2jphsppoint();
  if (status!=0) { // if we failed in generating a 2-jet configuration
    wt=0.;
    return wt;
  }
  if (xa>1.0||xb>1.0) {
    setstatus(5);
    wt=0.;
    return wt;
  } 
  wt*=Generate2jNonFKLflavourconf_and_eval();
#if calcscaleunc
  for (unsigned int ii=0;ii<scales.size();ii++) {
    double reweight=pdfpt(0,xa,scales.at(ii),aptype)*pdfpt(1,xb,scales.at(ii),bptype)/(pdfta*pdftb)*pow(Halphas(scales.at(ii))/Halphas(renscale),2);
    if (flags.logcorrect==1) {
      reweight*=(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.front().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.back().perp()/scales.at(ii))/(2.*M_PI));
      reweight/=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI));
    }
    weights.push_back(wt*reweight);
  }
#endif

  wt*=389.379292;
  wt*=1E6; // micro barn-> pico barn

  return wt;
}

double CHMultijet::Generate3jphsppoint()
{
  double temp1,temp2,p[4],pt,y,px,py,pz,E,r1;
  double weight(1.);
  HepLorentzVector p1,p2,p3;
  double sqrts=2.*flags.Ebeam;
  
#if old3jphsp
  const static double ylc(-2.8);
  const static double yspanc(5.);
#else
  double pta[3][2];
  double deltay[3];
  double phia[3];
#endif
  // generate momenta

  // Generate the transverse momentum of p3
  temp1=atan((flags.partonptmax-flags.jetptmin)/40.0);
  r1=ran.flat();
  pt=flags.jetptmin+40.0*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;

#if old3jphsp
  // Generate a y Gaussian distributed around -2.8
  {
    double lninvr1,r1,r2,temp;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    temp=sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    y=temp+ylc;
//     y=temp-2.;
//     y=temp-2.8;
    weight=weight*(exp(temp*temp/2.))*sqrt(2.*M_PI);
  }

  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  
  p[0]=pt*cosh(y); // the energy component
  p[1]=pt*cos(temp1);
  p[2]=pt*sin(temp1);
  p[3]=pt*sinh(y); // the z-component
    
  xa+=(p[0]-p[3])/sqrts;
  xb+=(p[0]+p[3])/sqrts;

  p1.set(p[1],p[2],p[3],p[0]);
#else
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane

  pta[0][0]=pt*cos(temp1);
  pta[0][1]=pt*sin(temp1);
  phia[0]=temp1;
  deltay[0]=0.;
#endif

  // Generate the transverse momentum of p4
  temp1=atan((flags.partonptmax-flags.jetptmin)/40.0);
  r1=ran.flat();
  pt=flags.jetptmin+40.0*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;
  
#if old3jphsp
  // Generate a y Gaussian distributed around 2.5
  {
    double lninvr1,r1,r2,temp;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    temp=sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    y=y+temp+yspanc;
//     y=temp+2.5;
    //     y=y+4.+temp+2.8;
    weight=weight*(exp(temp*temp/2.))*sqrt(2.*M_PI);
    if ((y-p1.rapidity())<0.) {
      weight=0.;
      setstatus(13);
      return 0.;}
  }

  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  
  p[0]=pt*cosh(y); // the energy component
  p[1]=pt*cos(temp1);
  p[2]=pt*sin(temp1);
  p[3]=pt*sinh(y); // the z-component
    
  xa+=(p[0]-p[3])/sqrts;
  xb+=(p[0]+p[3])/sqrts;

  p3.set(p[1],p[2],p[3],p[0]);
#else
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  pta[2][0]=pt*cos(temp1);
  pta[2][1]=pt*sin(temp1);
  phia[2]=temp1;
  deltay[2]=10.*ran.flat();
  weight*=10.;
#endif

  // Generate the transverse momentum of p2
  temp1=atan((flags.partonptmax-flags.jetptmin)/40.0);
  r1=ran.flat();
  pt=flags.jetptmin+40.0*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;
  weight/=16.*pow(M_PI,3);

#if old3jphsp
  // Generate a y flat between y3 and y4
  {
    double temp=p3.rapidity()-p1.rapidity();
    y=p1.rapidity()+temp*ran.flat();
    weight*=temp;
  }

  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  
  p[0]=pt*cosh(y); // the energy component
  p[1]=pt*cos(temp1);
  p[2]=pt*sin(temp1);
  p[3]=pt*sinh(y); // the z-component
    
  xa+=(p[0]-p[3])/sqrts;
  xb+=(p[0]+p[3])/sqrts;

  p2.set(p[1],p[2],p[3],p[0]);
#else
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  pta[1][0]=pt*cos(temp1);
  pta[1][1]=pt*sin(temp1);
  phia[1]=temp1;
  deltay[1]=deltay[2]*ran.flat();
  weight*=deltay[2];

  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double ycenter=deltay[2]/2.;
    double yoffset;
    
    {
      double lninvr1,r1,r2,temp,a(2.);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to gluons and xa,xb
    double yi=deltay[0]+yoffset;
    if (fabs(yi)>ymax) {
      setstatus(2);
      weight=0.;
      return weight;
    }
    pt=sqrt(pta[0][0]*pta[0][0]+pta[0][1]*pta[0][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pt*cos(phia[0]);
    p[2]=pt*sin(phia[0]);
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    p1.set(p[1],p[2],p[3],p[0]);

    yi=deltay[1]+yoffset;
    if (fabs(yi)>ymax) {
      setstatus(2);
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[1][0]*pta[1][0]+pta[1][1]*pta[1][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pt*cos(phia[1]);
    p[2]=pt*sin(phia[1]);
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    p2.set(p[1],p[2],p[3],p[0]);

    yi=deltay[2]+yoffset;
    if (fabs(yi)>ymax) {
      setstatus(2);
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[2][0]*pta[2][0]+pta[2][1]*pta[2][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pt*cos(phia[2]);
    p[2]=pt*sin(phia[2]);
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    p3.set(p[1],p[2],p[3],p[0]);
  }
//   cout <<"--------------------\n";
//   cout << p1<<endl<<p2<<endl<<p3<<endl;;
#endif

  // The Higgs:
  // Generate a y Gaussian distributed around 0
 {
    double lninvr1,r1,r2,temp,a;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    a=1.6;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    y=temp;
    weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
 }

  px=-(p1.x()+p2.x()+p3.x());
  py=-(p1.y()+p2.y()+p3.y());

  r1=ran.flat();
  double sH=flags.mH*(flags.mH + flags.GammaH*tan((M_PI*r1)/2. + (-1. + r1)*atan(flags.mH/flags.GammaH)));

  double mHperp=sqrt(px*px+py*py+sH);
  pz=mHperp*sinh(y);
  E=mHperp*cosh(y);

  xa+=(E-pz)/sqrts;
  xb+=(E+pz)/sqrts;
  pHv.set(px,py,pz,E);
//   cout << pHv<<endl;
  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);
  
  if (flags.DecayHiggs) {
    CLHEP::HepLorentzVector g1v,g2v;
    double pgH[4];
    double pgHs=sqrt(sH)/2.;
    // Choose theta and phi
    double pptheta=2.*M_PI*ran.flat();
    double cosppphi=2.*ran.flat()-1.;  
    double sinphi=sqrt(1.-cosppphi*cosppphi);  // Know 0 < phi < pi

    // construct 4-vector in H rest frame
    pgH[0]=pgHs;
    pgH[1]=pgHs*cos(pptheta)*sinphi;
    pgH[2]=pgHs*sin(pptheta)*sinphi;
    pgH[3]=pgHs*cosppphi;
    g1v.set(pgH[1],pgH[2],pgH[3],pgH[0]);
    g2v.set(-pgH[1],-pgH[2],-pgH[3],pgH[0]);

    // translate to lab frame
    CLHEP::HepLorentzVector pp1=g1v.boost(pHv.boostVector());
    CLHEP::HepLorentzVector pp2=g2v.boost(pHv.boostVector());
  
    photonmomenta.push_back(pp1);
    photonmomenta.push_back(pp2);
    Ht+=pp1.perp()+pp2.perp();
    wt*=flags.HBRgammagamma;
  } else {
    Ht+=pHv.perp();
  }

//   cout<< "----------\n" << pa<<"\n"<<pb<<"\n";
  partonmomenta.push_back(p1);
  partonmomenta.push_back(p2);
  partonmomenta.push_back(p3);

//   cout << pa+pb-p1-p2-p3-pHv<<endl;
  // Check if it is a 3-jet phase space point:
  std::vector<fastjet::PseudoJet> avec;
  HepLorentzVector pi;
  
  for (int ii=0;ii<int(partonmomenta.size());ii++) {
    pi=getpartvn(ii);
    avec.push_back(pi);
  }
  
  // /**set fastjet flags */
  // double Rparam = 0.7;
  // fastjet::Strategy strategy = fastjet::Best;
  // fastjet::RecombinationScheme recomb_scheme = fastjet::E_scheme;
    
  // /** Define jet for fastjet*/
  // fastjet::JetDefinition jet_def(fastjet::kt_algorithm, Rparam, recomb_scheme, strategy);
    
  /** fastjet find clusters */
  fastjet::ClusterSequence ev(avec, flags.jet_def);
    
  /** Retrieve the final state jets from fastjet Sorted by Rapidity*/
  std::vector<fastjet::PseudoJet> jets = sorted_by_pt(ev.inclusive_jets(flags.jetptmin));
    
  /** Save the number of final state jets */
  int njets=int(jets.size());

  if (njets!=3) {
    setstatus(12);
    weight=0;
    return weight;
  } 

#if calcscaleunc
  // NB jets have been sorted_by_pt here
  renscale=jets.at(0).perp()*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=flags.scale*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=Ht/2.*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=geometricmean(&jets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=invmass(&jets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=jets.at(0).perp()*flags.scalefactor;
#else
  if (flags.scalesetting==0) //max pt(jets)
    renscale=jets.at(0).perp()*flags.scalefactor;
  else if (flags.scalesetting==1) //fixed scale
    renscale=flags.scale;
  else if (flags.scalesetting==2) // Ht/2
    renscale=Ht/2.*flags.scalefactor;
  else if (flags.scalesetting==3) //geometric mean of pt(jets)
    renscale=geometricmean(&ptjets)*flags.scalefactor;
#endif

  return weight;
}




double CHMultijet::weight2() // The three-jet cross section
{
  wt=1.;

  wt*=Generate3jphsppoint();
  if (status!=0) { // if we failed in generating a 3-jet configuration
    wt=0.;
    return wt;
  }
  if (xa>1.0||xb>1.0) {
    setstatus(5);
    wt=0.;
    return wt;
  } 

  wt*=Generate3jNonFKLflavourconf_and_eval();

  wt*=389.379292;
  wt*=1E6; // micro barn-> pico barn
#if calcscaleunc
  for (unsigned int ii=0;ii<scales.size();ii++) {
    double reweight=pdfpt(0,xa,scales.at(ii),aptype)*pdfpt(1,xb,scales.at(ii),bptype)/(pdfta*pdftb)*pow(Halphas(scales.at(ii))/Halphas(renscale),3);
    if (flags.logcorrect==1) {
      reweight*=(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.front().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.back().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.at(1).perp()/scales.at(ii))/(2.*M_PI));
      reweight/=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI));
    }
    weights.push_back(wt*reweight);
  }
#endif
  return wt;
}

double CHMultijet::Event()
{

  static double processc[3]={0.,0.,0.}; // how often to choose the different type of processes (cummulative)
  static double procnorm[3]={1.,1.,1.};
  static bool initialisearray(true);
  static double sum(0.);
  // run only once:
  if (initialisearray) {
    if (flags.resum) {
      sum+=0.9;
    }
    processc[0]=sum;
    if (flags.nonFKLmatch2j) {
      sum+=0.05;
    }
    processc[1]=sum;
    if (flags.nonFKLmatch3j) {
      sum+=0.05;
    }
    processc[2]=sum;
    procnorm[0]=sum/processc[0];
    for (int ii=1;ii<3;ii++) {
      procnorm[ii]=sum/(processc[ii]-processc[ii-1]);
    }
    initialisearray=false; // don't do this again!
  }

  renscale=flags.scale*flags.scalefactor;  // if scalesetting==1, this is changed later
  setstatus(0);
  partonmomenta.clear();
  photonmomenta.clear();
  LHparticles.clear();
  jets.clear();
  scales.clear();
  weights.clear();
  xa=xb=0.;
  Ht=0.;
  aptype=bptype=0;

  // Clear Exceptions
  feclearexcept(FE_ALL_EXCEPT);

  static int test_excepts=(FE_DIVBYZERO & FE_INEXACT & FE_INVALID & FE_OVERFLOW & FE_UNDERFLOW);

  double rtemp=sum*ran.flat();
  double tempw;
  int set_excepts;

  if (rtemp<processc[0]) {
    processtype=1;
    tempw=weight();
    set_excepts = fetestexcept(test_excepts );
    if (set_excepts & test_excepts)
      setstatus(20);
  } else if (rtemp<processc[1]) {
    processtype=2;
    tempw=weight1();
    set_excepts = fetestexcept(test_excepts );
    if (set_excepts & test_excepts) {
      setstatus(21);
      cout << "Bugger, 2jet\n";
    }      
  } else {
    processtype=3;
    tempw=weight2();
    set_excepts = fetestexcept(test_excepts );
    if (set_excepts & test_excepts) {
      setstatus(22);
      cout << "Bugger, 3jet\n";
    }      
  } 
  tempw*=procnorm[processtype-1];
#if calcscaleunc
  for (unsigned int ii=0;ii<weights.size();++ii) {
    weights.at(ii)*=procnorm[processtype-1];
  }
#endif
    return tempw;
}
