#ifndef _HMULTIJET_H_INCLUDED_
#define _HMULTIJET_H_INCLUDED_
#include <math.h>
#include <complex>
#include <list>
#include "mstwpdf.h"
#include "LOjets.h"
#include "GeneratorFlags.h"
#include <CLHEP/Vector/LorentzVector.h>
#include <CLHEP/Random/Randomize.h>
#include <CLHEP/Random/RanluxEngine.h>
#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"

typedef std::complex<double> COM;

#define printoutput 0
using namespace std;
using namespace CLHEP;

double const v=246.; // The Higgs field vacuum expectation value in GeV
double const mt=174.;

struct particle {
  int PID;
  int COL1;
  int COL2;
  int MOT1;
  int MOT2;
  int ISTUP;
  CLHEP::HepLorentzVector p;
};

class CHMultijet
{
private:
  double Halphas(double q);

  // Generator flags:
  SGeneratorFlags GeneratorFlags;
  
  double wt; // The weight for the matching correction
/*   double sqrts; */
  //  double pt1,pt2;
  HepLorentzVector pHv;
  HepLorentzVector pHvi; //the t-channel momentum entering the Higgs vertex
  std::vector<CLHEP::HepLorentzVector> partonmomenta;
  std::vector<CLHEP::HepLorentzVector> photonmomenta;
  HepLorentzVector t1v,t2v,t1vsave,t2vsave;
  HepLorentzVector pa,pb;
  std::vector<particle> LHparticles;
  ofstream LHoutput;
  long xsecpos;
  /* static double const ptmin=35.; */
  static int const GLUEINT=23; // the max number of gluons -1
  //  static int const GLUEINT=2; // the max number of gluons -1
  //  static int const GLUEINT=12; // the max number of gluons -1
  //  static double const RAPINT=10.0; // the maxlength of the rapidity interval
  static double const beta0=11.-2.*5/3.;
  double lambda; // the phase space slicing regularisation scale
/*   double pdfa,pdfb,pdfda,pdfdb,pdfuxa,pdfuxb,pdfdxa,pdfdxb,pdfga,pdfba,pdfgb,pdfsa,pdfsb,pdfca,pdfcb,pdfua,pdfub,pdfbb,pdfta,pdftb; */
  int aptype,bptype;
  double xa,xb; // xa-, xb+
  /* double matchingweight,matchingweight2; */
  double ymax; // the maximum rapidity generated for any particle
  class c_mstwpdf * parton;
  double C2LargeMt(HepLorentzVector q1, HepLorentzVector pH);
  double C2gHgm(HepLorentzVector p2, HepLorentzVector p1, HepLorentzVector pH);
  double C2gHgp(HepLorentzVector p2, HepLorentzVector p1, HepLorentzVector pH);
  double C2qHqm(HepLorentzVector p2, HepLorentzVector p1, HepLorentzVector pH);
  void AddParton(double p[4]);
  void AddParton1(double p[4]);
  void AddParton2(double p[4]);
  /* double tildaa(double q2); */
  double omega0(double q2, double mu2);
  /* double beta0nc; */
  double GeneratePHSPpoint();
  double Generate2jphsppoint();
  double Generate3jphsppoint();
  double Generate2jNonFKLflavourconf_and_eval();
  double Generate3jNonFKLflavourconf_and_eval();
  double pdfpt(int beam, double a, double q,int t);
  double GenerateFKLIncomingParticles(double xa,double ufa,double xb,double ufb);
  double GenerateIncomingParticles(double xa,double ufa,double xb,double ufb);
  void SetupLHInformation();

  double C2LipatovVDD(HepLorentzVector qav, HepLorentzVector qbv, HepLorentzVector p1, HepLorentzVector p2);
  double C2Lipatov(HepLorentzVector qav, HepLorentzVector qbv, HepLorentzVector p1, HepLorentzVector p2);
  double C2Lipatov(CLHEP::HepLorentzVector qav, CLHEP::HepLorentzVector qbv, CLHEP::HepLorentzVector pim, CLHEP::HepLorentzVector pip,CLHEP::HepLorentzVector pom, CLHEP::HepLorentzVector pop); // B
  double C2Lipatovots(HepLorentzVector qav, HepLorentzVector qbv, HepLorentzVector p1, HepLorentzVector p2);
  double C2Lipatovots(HepLorentzVector qav, HepLorentzVector qbv, HepLorentzVector pim, HepLorentzVector pip, HepLorentzVector pom, HepLorentzVector pop);
  double FKLamplitude(std::vector<CLHEP::HepLorentzVector> emissions,CLHEP::HepLorentzVector ti,CLHEP::HepLorentzVector pa,CLHEP::HepLorentzVector pb,double ystart,double yend,fastjet::ClusterSequence * ev);
  /* double FKLamplitude2(std::vector<CLHEP::HepLorentzVector> emissions,CLHEP::HepLorentzVector ti,CLHEP::HepLorentzVector pa,CLHEP::HepLorentzVector pb,CLHEP::HepLorentzVector pstart,CLHEP::HepLorentzVector pend,int iswitch); */
  
  int status; // the return status of an event
  int processtype; // the process type of an event - matching, 2,3,... jets
  static const int nprocesstype=3; // resummed&matched, 2jet, 3jet
  double Ht; // The scalar sum of transverse momenta in the event
  double renscale;  // Globally accessible scale variable.

  // jets:
  std::vector<fastjet::PseudoJet> jets;

  std::vector<double> scales;
  std::vector<double> weights;

public:
  // constructor & destructor to be defined later
  virtual ~CHMultijet(void);
  CHMultijet(void);
  CHMultijet(string filename);
  CHMultijet(SGeneratorFlags GeneratorFlagsa);

  double pdfa,pdfb,pdfda,pdfdb,pdfuxa,pdfuxb,pdfdxa,pdfdxb,pdfga,pdfba,pdfgb,pdfsa,pdfsxa,pdfsb,pdfsxb,pdfca,pdfcb,pdfua,pdfub,pdfbb,pdfta,pdftb;
  SGeneratorFlags flags;
  
  double getcmsenergy();
  // Used for matching
  double MEt2(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pH, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb,bool checkmomenta,fastjet::ClusterSequence * ev);
  double MG(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pH, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb);
  double MG3(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pH, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb);
  double VirtualCorrections();
  double weight();  // resummed (and matched ?)
  double weight1(); // 2-jet
  double weight2(); // 3-jet
  double Event();
  int getnp();
  HepLorentzVector getpartvn(int n);
  void getHv(double A[4]);
  HepLorentzVector getpHv();
  HepLorentzVector getPhoton(int n);
  int getMstatus(void);
  HepRandom ran;
  int getstatus();
  int getprocesstype();
  void setstatus(int s);
  double GenerateEvent();
  void LHSaveEvent(void);
  double xsec[nprocesstype+1][2];
  double maxweight[nprocesstype+1];
  std::vector<double> GetWeights(void);
};
#endif
