//////////////////////////////////////////////////
//////////////////////////////////////////////////
// This source code is Copyright (2012) of      //
//  Jeppe R. Andersen                           //
//  Jennifer M. Smillie                         //
// and is distributed under the                 //
// Gnu Public License version 2                 //
// http://www.gnu.org/licenses/gpl-2.0.html     //
// You are allowed to distribute and alter the  //
// source under the conditions of the GPLv2     //
// as long as this copyright notice             //
// is unaltered and distributed with the source //
// Any use should comply with the               //
//             MCNET GUIDELINES                 //
//    for Event Generator Authors and Users     //
// as distributed with this source code         //
//////////////////////////////////////////////////
//////////////////////////////////////////////////
#include "WJetGen.h"
#define makeplots 0
#define usehepmc 0
#if usehepmc
#include "HepMC/GenEvent.h"
#include "HepMC/IO_GenEvent.h"
#endif

SGeneratorFlags Flags;
double jetrapmax(4.5);
double erapmax(2.5);
double eptmin(20.);
double missptmin(20.);
double mtwmin(0.);

bool HEPMCoutput(false);
string HEPMCoutputfile;

int init(int argn, char* argv[])
{
  ifstream input;
  istringstream sline;
  int iline(0);
  string label,temp,line;

  if (argn!=2) {
    cerr << "# Usage:\n./WJetGen [input.file]\n";
    cerr << "# where input.file is a file with parameters controlling the generation of events.\n";
    cerr << "# More information at http://cern.ch/HEJ\n";
    if (argn>2)
      return 1;
    else {
      cout << "# Using standard values for generator parameters\n";
      return 0;
    }
  } else {

    // Now open input file:
    input.open(argv[1], ios::in); // Open input file
    
  // Read parameters in from input file:
    while(getline(input,line)) {
      line.push_back(' ');
      sline.str(line);
      sline>>label;
      iline++;
      if(label=="Ebeam:") {
	if(!(sline>>Flags.Ebeam)) {
	  cout << "Error in input file at line " <<iline<<endl;
	  cout << "Value of Beam Energy invalid!"<<endl;
	  return 1;
	}
	if(Flags.Ebeam<0) {
	  cout << "Beam Energy<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
	Flags.partonptmax=Flags.Ebeam;
      }
      else if(label=="jetptmin:") {
	if(!(sline>>Flags.jetptmin)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of jetptmin invalid!"<<endl;
	  return 1;
	}
	if(Flags.jetptmin<0) {
	  cout <<"jetptmin<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="eptmin:") {
	if(!(sline>>eptmin)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of eptmin invalid!"<<endl;
	  return 1;
	}
	if(eptmin<0) {
	  cout<<"eptmin<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="missptmin:") {
	if(!(sline>>missptmin)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of missptmin invalid!"<<endl;
	  return 1;
	}
	if(missptmin<0) {
	  cout<<"missptmin<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="mtwmin:") {
	if(!(sline>>mtwmin)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of mtwmin invalid!"<<endl;
	  return 1;
	}
	if(mtwmin<0) {
	  cout<<"mtwmin<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="extpartonptmin:") {
	if(!(sline>>Flags.extpartonptmin)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of extpartonptmin invalid!"<<endl;
	  return 1;
	}
	if(Flags.extpartonptmin<0) {
	  cout <<"extpartonptmin<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="scale:") {
	if(!(sline>>Flags.scale)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of scale invalid!"<<endl;
	  return 1;
	}
	if(Flags.scale<0) {
	  cout<<"scale<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="scalefactor:") {
	if(!(sline>>Flags.scalefactor)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of scalefactor invalid!"<<endl;
	  return 1;
	}
	if(Flags.scalefactor<0) {
	  cout<<"scalefactor <0  at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="jetrapmax:") {
	if(!(sline>>jetrapmax)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of jetrapmax invalid!"<<endl;
	  return 1;
	}
	if(jetrapmax<0) {
	  cout<<"jetrapmax<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="erapmax:") {
	if(!(sline>>erapmax)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of erapmax invalid!"<<endl;
	  return 1;
	}
	if(erapmax<0) {
	  cout<<"erapmax<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="NEvents:") {
	if(!(sline>>Flags.NEvents)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of NEvents invalid!"<<endl;
	  return 1;
	}
	if(Flags.NEvents<0) {
	  cout<<"NEvents<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="scalesetting:") { //0:ptmax, 1:fixed, 2:ht/2, 3:ptmean
	if(!(sline>>Flags.scalesetting)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of scalesetting invalid!"<<endl;
	  return 1;
	}
	if(abs(Flags.scalesetting)>3) {
	  cout<<"scalesetting \notin {0,1,2,3}  at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="logcorrect:") { 
	if(!(sline>>Flags.logcorrect)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of logcorrect invalid!"<<endl;
	  return 1;
	}
	if(Flags.logcorrect!=0&&Flags.logcorrect!=1) {
	  cout<<"logcorrrect \notin {0,1}  at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="beam1:") {
	if(!(sline>>temp)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Beam 1 type invalid!"<<endl;
	  return 1;
	}
	if(temp=="P") Flags.beam[0]=1;
	else if(temp=="PBAR") Flags.beam[0]=-1;
	else {
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "Unrecognised beam type for beam 1!"<<endl;
	  return 1;
	}
      }
      else if(label=="beam2:") {
	if(!(sline>>temp)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Beam 2 type invalid!"<<endl;
	  return 1;
	}
	if(temp=="P") Flags.beam[1]=1;
	else if(temp=="PBAR") Flags.beam[1]=-1;
	else {
	  cout <<"Error in input file at line "<<iline<<endl;
	  cout <<"Unrecognised beam type for beam 2!"<<endl;
	  return 1;
	}
      }
      else if(label=="Rparam:") {
	if(!(sline>>Flags.Rparam)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of Rparam invalid!"<<endl;
	  return 1;
	}
	if(Flags.Rparam<0) {
	  cout<<"Rparam<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="jetalgo:") {
	if(!(sline>>temp)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"jetalgo type invalid!"<<endl;
	  return 1;
	}
	if(temp=="antikt") {
	  Flags.jetalgo=fastjet::antikt_algorithm;
	  fastjet::Strategy strategy = fastjet::Best;
	  fastjet::RecombinationScheme recomb_scheme = fastjet::E_scheme;
	  Flags.jet_def=fastjet::JetDefinition(Flags.jetalgo, Flags.Rparam, recomb_scheme, strategy);
	}
	else if(temp=="kt") {
	  Flags.jetalgo=fastjet::kt_algorithm;
	  fastjet::Strategy strategy = fastjet::Best;
	  fastjet::RecombinationScheme recomb_scheme = fastjet::E_scheme;
	  Flags.jet_def=fastjet::JetDefinition(Flags.jetalgo, Flags.Rparam, recomb_scheme, strategy);
	}
	else if(temp=="D0RunIICone") {
	  fastjet::JetDefinition::Plugin * plugin= new fastjet::D0RunIIConePlugin(Flags.Rparam,6.,0.5);
	  Flags.jet_def=plugin;
	}
	else {
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "Unrecognised beam type for beam 1!"<<endl;
	  return 1;
	}
      }
      else if(label=="LHoutput:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "LHoutput flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.LHoutput=true; else Flags.LHoutput=false;
      }
      else if(label=="LHoutputfile:") {
	if(!(sline>>Flags.LHoutputfile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "Eventfile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="HEPMCoutput:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "HEPMCoutput flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") HEPMCoutput=true; else HEPMCoutput=false;
      }
      else if(label=="HEPMCoutputfile:") {
	if(!(sline>>HEPMCoutputfile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "HEPMCEventfile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="rootfile:") {
	if(!(sline>>Flags.rootfile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "rootfile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="useLHApdf:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "useLHApdf flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.useLHApdf=true; else Flags.useLHApdf=false;
      }
      else if(label=="LHAPDFName:") {
	if(!(sline>>Flags.LHAPDFName)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "LHAPDFName name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="LHAPDFsubset:") {
	if(!(sline>>Flags.LHAPDFsubset)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of LHAPDFsubset invalid!"<<endl;
	  return 1;
	}
	if(Flags.LHAPDFsubset<0) {
	  cout<<"LHAPDFsubset<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="match:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.match=true; else Flags.match=false;
      }
      else if(label=="resum:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.resum=true; else Flags.resum=false;
      }
      else if(label=="nonFKLmatch2j:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.nonFKLmatch2j=true; else Flags.nonFKLmatch2j=false;
      }
      else if(label=="nonFKLmatch3j:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.nonFKLmatch3j=true; else Flags.nonFKLmatch3j=false;
      }
      else if(label=="nonFKLmatch4j:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.nonFKLmatch4j=true; else Flags.nonFKLmatch4j=false;
      }
      else if(label=="rootevents:") {
      	if(!(sline>>temp)){
      	  cout << "Error in input file at line "<<iline<<endl;
      	  cout << "rootevents flag invalid!"<<endl;
      	  return 1;
      	}
      	if(temp=="Y"||temp=="y") Flags.rootevents=true; else Flags.rootevents=false;
      }
      else if(label=="unweight:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "unweight flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.unweight=true; else Flags.unweight=false;
      }
      else if(label=="raninitfile:") {
	if(!(sline>>Flags.raninitfile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "raninitfile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label.at(0)==47){
      }
      else{
	cout<<"Unrecognised parameter at line "<<iline<<" in input file:"<<endl;
	cout<<line<<endl;
	return 1;
      }
    }
    
    input.close();
  }

  return 0;
}

void Fill2DHistograms(TH1D* h, TH1D* hN, TH1D *hwt,double x,double wt)
{
  double binwidth; // the binwidth
  binwidth=h->GetXaxis()->GetBinWidth(1);
  h->Fill(x,wt/binwidth);
  hwt->Fill(x,wt*wt/binwidth);
  hN->Fill(x);
}

int main(int argn, char *argv[])
{
  double wt,wts; // The event weight
  double ya,yb,yi;
  CLHEP::HepLorentzVector pH,pi;
  int np; // The number of final state partons
  int njetsb,njetst; // The number of final state jets
  CLHEP::HepLorentzVector ja,jb,jthird,jfourth;
  CLHEP::Ranlux64Engine rUnweight;
  
  // Float_t rpp[24][4]; // four vectors for parton momenta, for use with ROOT
  // Float_t rpH[4]; // the four vector for the Higgs boson, for use with ROOT

  if ((init(argn,argv))==1) 
    return(-1);

  // if (HEPMCoutput)
#if usehepmc
  HepMC::IO_GenEvent HEPMCio(HEPMCoutputfile.c_str());
#endif
  // Only simple analysis and self-consistency check:
  // Example histogram file:
  TFile hfile(Flags.rootfile.c_str(),"RECREATE","Output from WJetGen");
  TROOT simple("WJetGen","Output from WJetGen");
  TH1D *hmstatus,*hwt,*hwtwt;
  TH1D *htotal,*htotalN,*htotalwt2;
  htotal = new TH1D("total","total",1,-0.5,.5);
  htotal->Sumw2();
  htotalN = new TH1D("totalN","totalN",1,-0.5,.5);
  htotalwt2 = new TH1D("totalwt2","totalwt2",1,-0.5,.5);

#if makeplots
  TH1D *hya,*hyaN,*hyawt2;
  hya = new TH1D("ya","ya",40,-5.,5.);
  hya->Sumw2();
  hyaN = new TH1D("yaN","yaN",40,-5.,5.);
  hyawt2 = new TH1D("yawt2","yawt2",40,-5.,5.);

  TH1D *hyb,*hybN,*hybwt2;
  hyb = new TH1D("yb","yb",40,-5.,5.);
  hyb->Sumw2();
  hybN = new TH1D("ybN","ybN",40,-5.,5.);
  hybwt2 = new TH1D("ybwt2","ybwt2",40,-5.,5.);

  TH1D *hpt1,*hpt1N,*hpt1wt2;
  hpt1 = new TH1D("pt1","pt1",100,0.,1000.);
  hpt1->Sumw2();
  hpt1N = new TH1D("pt1N","pt1N",100,0.,1000.);
  hpt1wt2 = new TH1D("pt1wt2","pt1wt2",100,0.,1000.);

  TH1D *hpt2,*hpt2N,*hpt2wt2;
  hpt2 = new TH1D("pt2","pt2",100,0.,1000.);
  hpt2->Sumw2();
  hpt2N = new TH1D("pt2N","pt2N",100,0.,1000.);
  hpt2wt2 = new TH1D("pt2wt2","pt2wt2",100,0.,1000.);

  // TH1D *hptmax,*hptmaxN,*hptmaxwt2;
  // hptmax = new TH1D("ptmax","ptmax",100,0.,1000.);
  // hptmax->Sumw2();
  // hptmaxN = new TH1D("ptmaxN","ptmaxN",100,0.,1000.);
  // hptmaxwt2 = new TH1D("ptmaxwt2","ptmaxwt2",100,0.,1000.);

  // TH1D *hpt2max,*hpt2maxN,*hpt2maxwt2;
  // hpt2max = new TH1D("pt2max","pt2max",100,0.,1000.);
  // hpt2max->Sumw2();
  // hpt2maxN = new TH1D("pt2maxN","pt2maxN",100,0.,1000.);
  // hpt2maxwt2 = new TH1D("pt2maxwt2","pt2maxwt2",100,0.,1000.);

  TH1D *hktn,*hktnN,*hktnwt2;
  hktn = new TH1D("ktn","ktn",100,0.,1000.);
  hktn->Sumw2();
  hktnN = new TH1D("ktnN","ktnN",100,0.,1000.);
  hktnwt2 = new TH1D("ktnwt2","ktnwt2",100,0.,1000.);

  TH1D *hycentre,*hycentreN,*hycentrewt2;
  hycentre = new TH1D("ycentre","ycentre",40,-5.,5.);
  hycentre->Sumw2();
  hycentreN = new TH1D("ycentreN","ycentreN",40,-5.,5.);
  hycentrewt2 = new TH1D("ycentrewt2","ycentrewt2",40,-5.,5.);

  TH1D *hnjets,*hnjetsN,*hnjetswt2,*hnjetsp,*hnjetsm;//,*hnjetsmN;
  hnjets = new TH1D("njets","njets dist",10,0.5,10.5);
  hnjets->Sumw2();
  hnjetsN = new TH1D("njetsN","njetsN dist",10,0.5,10.5);
  hnjetswt2 = new TH1D("njetswt2","njetswt2 dist",10,0.5,10.5);
  hnjetsm = new TH1D("njetsm","njetsm dist",10,0.5,10.5);
  hnjetsm->Sumw2();
  // hnjetsmN = new TH1D("njetsmN","njetsmN dist",10,0.5,10.5);
  // hnjetsmwt2 = new TH1D("njetsmwt2","njetsmwt2 dist",10,0.5,10.5);
  hnjetsp = new TH1D("njetsp","njetsp dist",10,0.5,10.5);
  hnjetsp->Sumw2();
  // hnjetspN = new TH1D("njetspN","njetspN dist",10,0.5,10.5);
  // hnjetspwt2 = new TH1D("njetspwt2","njetspwt2 dist",10,0.5,10.5);

  TH1D *hnp,*hnpN,*hnpwt2;
  hnp = new TH1D("np","np dist",35,-.5,34.5);
  hnp->Sumw2();
  hnpN = new TH1D("npN","npN dist",35,-.5,34.5);
  hnpwt2 = new TH1D("npwt2","npwt2 dist",35,-.5,34.5);

  TH1D *hydij;
  hydij = new TH1D("ydij","ydij",40,0.,5.);
  hydij->Sumw2();
  // hydijN = new TH1D("ydijN","ydijN",40,0.,5.);
  // hydijwt2 = new TH1D("ydijwt2","ydijwt2",40,0.,5.);

  TH1D *havgnjets,*havgnjetsN,*havgnjetswt2;
  havgnjets = new TH1D("avgnjets","avgnjets",35,0.0,7.0);
  havgnjets->Sumw2();
  havgnjetsN = new TH1D("avgnjetsN","avgnjetsN",35,0.0,7.0);
  havgnjetswt2 = new TH1D("avgnjetswt2","avgnjetswt2",35,0.0,7.0);

  TH1D *havgnjets3,*havgnjets3N,*havgnjets3wt2;
  havgnjets3 = new TH1D("avgnjets3","avgnjets3",35,0.0,7.0);
  havgnjets3->Sumw2();
  havgnjets3N = new TH1D("avgnjets3N","avgnjets3N",35,0.0,7.0);
  havgnjets3wt2 = new TH1D("avgnjets3wt2","avgnjets3wt2",35,0.0,7.0);

  TH1D *havgnjets4,*havgnjets4N,*havgnjets4wt2;
  havgnjets4 = new TH1D("avgnjets4","avgnjets4",35,0.0,7.0);
  havgnjets4->Sumw2();
  havgnjets4N = new TH1D("avgnjets4N","avgnjets4N",35,0.0,7.0);
  havgnjets4wt2 = new TH1D("avgnjets4wt2","avgnjets4wt2",35,0.0,7.0);

  TH1D *havgnjets5,*havgnjets5N,*havgnjets5wt2;
  havgnjets5 = new TH1D("avgnjets5","avgnjets5",35,0.0,7.0);
  havgnjets5->Sumw2();
  havgnjets5N = new TH1D("avgnjets5N","avgnjets5N",35,0.0,7.0);
  havgnjets5wt2 = new TH1D("avgnjets5wt2","avgnjets5wt2",35,0.0,7.0);

  TH1D *havgnjets6,*havgnjets6N,*havgnjets6wt2;
  havgnjets6 = new TH1D("avgnjets6","avgnjets6",36,0.0,7.0);
  havgnjets6->Sumw2();
  havgnjets6N = new TH1D("avgnjets6N","avgnjets6N",36,0.0,7.0);
  havgnjets6wt2 = new TH1D("avgnjets6wt2","avgnjets6wt2",36,0.0,7.0);

  TH1D *havgnjetsnorm,*havgnjetsnormN,*havgnjetsnormwt2;
  havgnjetsnorm = new TH1D("avgnjetsnorm","avgnjetsnorm",35,0.0,7.0);
  havgnjetsnorm->Sumw2();
  havgnjetsnormN = new TH1D("avgnjetsnormN","avgnjetsnormN",35,0.0,7.0);
  havgnjetsnormwt2 = new TH1D("avgnjetsnormwt2","avgnjetsnormwt2",35,0.0,7.0);
#endif

  hmstatus = new TH1D("mstatus","mstatus",25,-.5,24.5);
  hwt = new TH1D("wt","wt",800,-30.,50.);
  hwtwt = new TH1D("wtwt","wtwt",800,-30.,50.);

#if makeplots
  TH1D *hydif,*hydifN,*hydifwt2;
  hydif = new TH1D("ydif","ydif",32,0.,8.);
  hydif->Sumw2();
  hydifN = new TH1D("ydifN","ydifN",32,0.,8.);
  hydifwt2 = new TH1D("ydifwt2","ydifwt2",32,0.,8.);

  TH1D *hydif3,*hydif3N,*hydif3wt2;
  hydif3 = new TH1D("ydif3","ydif3",32,0.,8.);
  hydif3->Sumw2();
  hydif3N = new TH1D("ydif3N","ydif3N",32,0.,8.);
  hydif3wt2 = new TH1D("ydif3wt2","ydif3wt2",32,0.,8.);

  TH1D *hydiftwojet,*hydiftwojetN,*hydiftwojetwt2;
  hydiftwojet = new TH1D("ydiftwojet","ydiftwojet",32,0.,8.);
  hydiftwojet->Sumw2();
  hydiftwojetN = new TH1D("ydiftwojetN","ydiftwojetN",32,0.,8.);
  hydiftwojetwt2 = new TH1D("ydiftwojetwt2","ydiftwojetwt2",32,0.,8.);

  TH1D *hydifthreejet,*hydifthreejetN,*hydifthreejetwt2;
  hydifthreejet = new TH1D("ydifthreejet","ydifthreejet",32,0.,8.);
  hydifthreejet->Sumw2();
  hydifthreejetN = new TH1D("ydifthreejetN","ydifthreejetN",32,0.,8.);
  hydifthreejetwt2 = new TH1D("ydifthreejetwt2","ydifthreejetwt2",32,0.,8.);

  TH1D *hydiffourjet,*hydiffourjetN,*hydiffourjetwt2;
  hydiffourjet = new TH1D("ydiffourjet","ydiffourjet",32,0.,8.);
  hydiffourjet->Sumw2();
  hydiffourjetN = new TH1D("ydiffourjetN","ydiffourjetN",32,0.,8.);
  hydiffourjetwt2 = new TH1D("ydiffourjetwt2","ydiffourjetwt2",32,0.,8.);

  TH1D *hydiffivejet,*hydiffivejetN,*hydiffivejetwt2;
  hydiffivejet = new TH1D("ydiffivejet","ydiffivejet",32,0.,8.);
  hydiffivejet->Sumw2();
  hydiffivejetN = new TH1D("ydiffivejetN","ydiffivejetN",32,0.,8.);
  hydiffivejetwt2 = new TH1D("ydiffivejetwt2","ydiffivejetwt2",32,0.,8.);

  TH1D *hydifsixjet,*hydifsixjetN,*hydifsixjetwt2;
  hydifsixjet = new TH1D("ydifsixjet","ydifsixjet",32,0.,8.);
  hydifsixjet->Sumw2();
  hydifsixjetN = new TH1D("ydifsixjetN","ydifsixjetN",32,0.,8.);
  hydifsixjetwt2 = new TH1D("ydifsixjetwt2","ydifsixjetwt2",32,0.,8.);

  TH1D *hydifsevenjet,*hydifsevenjetN,*hydifsevenjetwt2;
  hydifsevenjet = new TH1D("ydifsevenjet","ydifsevenjet",32,0.,8.);
  hydifsevenjet->Sumw2();
  hydifsevenjetN = new TH1D("ydifsevenjetN","ydifsevenjetN",32,0.,8.);
  hydifsevenjetwt2 = new TH1D("ydifsevenjetwt2","ydifsevenjetwt2",32,0.,8.);

  TH1D *hydifeightjet,*hydifeightjetN,*hydifeightjetwt2;
  hydifeightjet = new TH1D("ydifeightjet","ydifeightjet",32,0.,8.);
  hydifeightjet->Sumw2();
  hydifeightjetN = new TH1D("ydifeightjetN","ydifeightjetN",32,0.,8.);
  hydifeightjetwt2 = new TH1D("ydifeightjetwt2","ydifeightjetwt2",32,0.,8.);

  TH1D *hydifninejet,*hydifninejetN,*hydifninejetwt2;
  hydifninejet = new TH1D("ydifninejet","ydifninejet",32,0.,8.);
  hydifninejet->Sumw2();
  hydifninejetN = new TH1D("ydifninejetN","ydifninejetN",32,0.,8.);
  hydifninejetwt2 = new TH1D("ydifninejetwt2","ydifninejetwt2",32,0.,8.);

  TH1D *hptnu,*hptnuN,*hptnuwt2;
  hptnu = new TH1D("ptnu","ptnu",100,0.,1000.);
  hptnu->Sumw2();
  hptnuN = new TH1D("ptnuN","ptnuN",100,0.,1000.);
  hptnuwt2 = new TH1D("ptnuwt2","ptnuwt2",100,0.,1000.);

  TH1D *hptnup,*hptnupN,*hptnupwt2;
  hptnup = new TH1D("ptnup","ptnup",100,0.,1000.);
  hptnup->Sumw2();
  hptnupN = new TH1D("ptnupN","ptnupN",100,0.,1000.);
  hptnupwt2 = new TH1D("ptnupwt2","ptnupwt2",100,0.,1000.);

  TH1D *hptnum,*hptnumN,*hptnumwt2;
  hptnum = new TH1D("ptnum","ptnum",100,0.,1000.);
  hptnum->Sumw2();
  hptnumN = new TH1D("ptnumN","ptnumN",100,0.,1000.);
  hptnumwt2 = new TH1D("ptnumwt2","ptnumwt2",100,0.,1000.);

  TH1D *hpte,*hpteN,*hptewt2;
  hpte = new TH1D("pte","pte",100,0.,1000.);
  hpte->Sumw2();
  hpteN = new TH1D("pteN","pteN",100,0.,1000.);
  hptewt2 = new TH1D("ptewt2","ptewt2",100,0.,1000.);

  TH1D *hptep,*hptepN,*hptepwt2;
  hptep = new TH1D("ptep","ptep",100,0.,1000.);
  hptep->Sumw2();
  hptepN = new TH1D("ptepN","ptepN",100,0.,1000.);
  hptepwt2 = new TH1D("ptepwt2","ptepwt2",100,0.,1000.);

  TH1D *hptem,*hptemN,*hptemwt2;
  hptem = new TH1D("ptem","ptem",100,0.,1000.);
  hptem->Sumw2();
  hptemN = new TH1D("ptemN","ptemN",100,0.,1000.);
  hptemwt2 = new TH1D("ptemwt2","ptemwt2",100,0.,1000.);

  TH1D *hsenu,*hsenuN,*hsenuwt2;
  hsenu = new TH1D("senu","senu",150,70,100.);
  hsenu->Sumw2();
  hsenuN = new TH1D("senuN","senuN",150,70,100.);
  hsenuwt2 = new TH1D("senuwt2","senuwt2",150,70,100.);

  TH1D *hsij,*hsijN,*hsijwt2;
  hsij = new TH1D("sij","sij",150,50,1000.);
  hsij->Sumw2();
  hsijN = new TH1D("sijN","sijN",150,50,1000.);
  hsijwt2 = new TH1D("sijwt2","sijwt2",150,50,1000.);

  TH1D *hht,*hhtN,*hhtwt2;
  hht = new TH1D("ht","ht",100,0.,2000.);
  hht->Sumw2();
  hhtN = new TH1D("htN","htN",100,0.,2000.);
  hhtwt2 = new TH1D("htwt2", "htwt2", 100,0.,2000.);

  TH1D *hhtavgnjets,*hhtavgnjetsN,*hhtavgnjetswt2;
  hhtavgnjets = new TH1D("htavgnjets","htavgnjets",200,0.0,2000.);
  hhtavgnjets->Sumw2();
  hhtavgnjetsN = new TH1D("htavgnjetsN","htavgnjetsN",200,0.0,2000.);
  hhtavgnjetswt2 = new TH1D("htavgnjetswt2","htavgnjetswt2",200,0.0,2000.);

  TH1D *hhtavgnjets3,*hhtavgnjets3N,*hhtavgnjets3wt2;
  hhtavgnjets3 = new TH1D("htavgnjets3","htavgnjets3",200,0.0,2000.);
  hhtavgnjets3->Sumw2();
  hhtavgnjets3N = new TH1D("htavgnjets3N","htavgnjets3N",200,0.0,2000.);
  hhtavgnjets3wt2 = new TH1D("htavgnjets3wt2","htavgnjets3wt2",200,0.0,2000.);

  TH1D *hhtavgnjets4,*hhtavgnjets4N,*hhtavgnjets4wt2;
  hhtavgnjets4 = new TH1D("htavgnjets4","htavgnjets4",200,0.0,2000.);
  hhtavgnjets4->Sumw2();
  hhtavgnjets4N = new TH1D("htavgnjets4N","htavgnjets4N",200,0.0,2000.);
  hhtavgnjets4wt2 = new TH1D("htavgnjets4wt2","htavgnjets4wt2",200,0.0,2000.);

  TH1D *hhtavgnjets5,*hhtavgnjets5N,*hhtavgnjets5wt2;
  hhtavgnjets5 = new TH1D("htavgnjets5","htavgnjets5",200,0.0,2000.);
  hhtavgnjets5->Sumw2();
  hhtavgnjets5N = new TH1D("htavgnjets5N","htavgnjets5N",200,0.0,2000.);
  hhtavgnjets5wt2 = new TH1D("htavgnjets5wt2","htavgnjets5wt2",200,0.0,2000.);

  TH1D *hhtavgnjets6,*hhtavgnjets6N,*hhtavgnjets6wt2;
  hhtavgnjets6 = new TH1D("htavgnjets6","htavgnjets6",35,0.0,7.0);
  hhtavgnjets6->Sumw2();
  hhtavgnjets6N = new TH1D("htavgnjets6N","htavgnjets6N",35,0.0,7.0);
  hhtavgnjets6wt2 = new TH1D("htavgnjets6wt2","htavgnjets6wt2",35,0.0,7.0);

  TH1D *hhtavgnjetsnorm,*hhtavgnjetsnormN,*hhtavgnjetsnormwt2;
  hhtavgnjetsnorm = new TH1D("htavgnjetsnorm","htavgnjetsnorm",200,0.0,2000.);
  hhtavgnjetsnorm->Sumw2();
  hhtavgnjetsnormN = new TH1D("htavgnjetsnormN","htavgnjetsnormN",200,0.0,2000.);
  hhtavgnjetsnormwt2 = new TH1D("htavgnjetsnormwt2","htavgnjetsnormwt2",200,0.0,2000.);

  TH1D *hht3,*hht3N,*hht3wt2;
  hht3 = new TH1D("ht3","ht3",100,0.,2000.);
  hht3->Sumw2();
  hht3N = new TH1D("ht3N","ht3N",100,0.,2000.);
  hht3wt2 = new TH1D("ht3wt2", "ht3wt2", 100,0.,2000.);

  TH1D *hye,*hyeN,*hyewt2;
  hye = new TH1D("ye","ye",40,-5.,5.);
  hye->Sumw2();
  hyeN = new TH1D("yeN","yeN",40,-5.,5.);
  hyewt2 = new TH1D("yewt2","yewt2",40,-5.,5.);

  TH1D *hynu,*hynuN,*hynuwt2;
  hynu = new TH1D("ynu","ynu",40,-5.,5.);
  hynu->Sumw2();
  hynuN = new TH1D("ynuN","ynuN",40,-5.,5.);
  hynuwt2 = new TH1D("ynuwt2","ynuwt2",40,-5.,5.);

  TH1D *hyem,*hyemN,*hyemwt2;
  hyem = new TH1D("yem","yem",40,-5.,5.);
  hyem->Sumw2();
  hyemN = new TH1D("yemN","yemN",40,-5.,5.);
  hyemwt2 = new TH1D("yemwt2","yemwt2",40,-5.,5.);

  TH1D *hynum,*hynumN,*hynumwt2;
  hynum = new TH1D("ynum","ynum",40,-5.,5.);
  hynum->Sumw2();
  hynumN = new TH1D("ynumN","ynumN",40,-5.,5.);
  hynumwt2 = new TH1D("ynumwt2","ynumwt2",40,-5.,5.);

  TH1D *hyep,*hyepN,*hyepwt2;
  hyep = new TH1D("yep","yep",40,-5.,5.);
  hyep->Sumw2();
  hyepN = new TH1D("yepN","yepN",40,-5.,5.);
  hyepwt2 = new TH1D("yepwt2","yepwt2",40,-5.,5.);

  TH1D *hynup,*hynupN,*hynupwt2;
  hynup = new TH1D("ynup","ynup",40,-5.,5.);
  hynup->Sumw2();
  hynupN = new TH1D("ynupN","ynupN",40,-5.,5.);
  hynupwt2 = new TH1D("ynupwt2","ynupwt2",40,-5.,5.);

  TH1D *hyw,*hywN,*hywwt2;
  hyw = new TH1D("yw","yw",80,-10.,10.);
  hyw->Sumw2();
  hywN = new TH1D("ywN","ywN",80,-10.,10.);
  hywwt2 = new TH1D("ywwt2","ywwt2",80,-10.,10.);

  TH1D *hthetaenu;//,*hthetaenuN,*hthetaenuwt2;
  hthetaenu = new TH1D("thetaenu","thetaenu",60,0.,1.);
  hthetaenu->Sumw2();
  // hthetaenuN = new TH1D("thetaenuN","thetaenuN",60,0.,1.);
  // hthetaenuwt2 = new TH1D("thetaenuwt2","thetaenuwt2",60,0.,1.);

  TH1D *hcosphi,*hcosphiN,*hcosphiwt2;
  hcosphi = new TH1D("cosphi","Avg. cos(phijj)",52,0.0,13.0);
  hcosphi->Sumw2();
  hcosphiN = new TH1D("cosphiN","Avg. cos(phijj)",52,0.0,13.0);
  hcosphiwt2 = new TH1D("cosphiwt2","Avg. cos(phijj)",52,0.0,13.0);

  TH1D *hcosphinorm,*hcosphinormN,*hcosphinormwt2;
  hcosphinorm = new TH1D("cosphinorm","Avg. cos(phijj)",52,0.0,13.0);
  hcosphinorm->Sumw2();
  hcosphinormN = new TH1D("cosphinormN","Avg. cos(phijj)",52,0.0,13.0);
  hcosphinormwt2 = new TH1D("cosphinormwt2","Avg. cos(phijj)",52,0.0,13.0);

  TProfile * tpavgcosphi;
  tpavgcosphi = new TProfile("tpavgcosphi","Avg. cos phi",52,0.,13.,-1.,1.,"s");
  tpavgcosphi->Sumw2();

  TProfile * tpavgjets;
  tpavgjets = new TProfile("tpavgjets","Avg. jets",35,0.,7.,2.,20.,"s");
  tpavgjets->Sumw2();

  TProfile * tphtjets;
  tphtjets = new TProfile("tphtjets","Avg. HT",200,0.,2000.,2.,20.,"s");
  tphtjets->Sumw2();

  TProfile * tphtydif;
  tphtydif = new TProfile("tphtydif","Avg. HT",200,0.,2000.,0.,8.,"s");
  tphtydif->Sumw2();
#endif

  // Make tree of events:
  Float_t rpp[22][4]; // four vectors for parton momenta, for use with ROOT
  Float_t rpe[4]; // the four vector for the Higgs boson, for use with ROOT
  Float_t rpnu[4]; // the four vector for the Higgs boson, for use with ROOT
  // Double_t weights[12]; // weights, including scale variations
  double weights[12]; // weights, including scale variations
  double matchreweight; // The FKL matching reweight factor
  Int_t processtype;
  Int_t aptype,bptype;
  Double_t xa,xb,mura,murb,pdfta,pdftb;
  Bool_t iselectron;

  TTree * Events;
  Events = new TTree("Events","Events");
  Events->Branch("weight",&wts,"weight/D");
  Events->Branch("weights",&weights,"weights[12]/D");
  Events->Branch("matchreweight",&matchreweight,"matchreweight/D");
  Events->Branch("np",&np,"np/I");
  Events->Branch("Pfourv",&rpp,"Pfourv[np][4]/F");
  Events->Branch("Pe",&rpe,"Pe[4]/F");
  Events->Branch("Pnu",&rpnu,"Pnu[4]/F");
  Events->Branch("processtype",&processtype,"processtype/I");
  Events->Branch("aptype",&aptype,"aptype/I");
  Events->Branch("bptype",&bptype,"bptype/I");
  Events->Branch("xa",&xa,"xa/D");
  Events->Branch("xb",&xb,"xb/D");
  Events->Branch("mura",&mura,"mura/D");
  Events->Branch("murb",&murb,"murb/D");
  Events->Branch("pdfta",&pdfta,"pdfta/D");
  Events->Branch("pdftb",&pdftb,"pdftb/D");
  Events->Branch("iselectron",&iselectron,"iselectron/B");

  CWMultijet event(Flags); // use just standard settings

  // Events->Branch("pdfta",&event.pdfta,"pdfta/D");
  // Events->Branch("pdftb",&event.pdftb,"pdftb/D");
  // Generate the events:
  
  unsigned int evententries(0);
  for (unsigned int i=0;i<Flags.NEvents;i++) {
    wt=event.Event(); // in pico-barn
    //    std::cout<<"WJetGen: "<<wt<<"  "<<event.getMstatus()<<"\n";
    wts=wt/double(Flags.NEvents); // in pico-barn
    matchreweight=event.getmatchreweight();
    hmstatus->Fill(event.getMstatus());
    if (event.getMstatus()==0) { // If event was good
      // Number of partons in event:
      np=event.getnp();
      
      //      Fill2DHistograms(hthetaenu,hthetaenuN,hthetaenuwt2,event.getthetaenu(),wts);
      std::vector<fastjet::PseudoJet> avec;
      
//       std::cout<<"====================\n";
      for (int ii=0;ii<np;ii++) {
	pi=event.getpartvn(ii);
// 	cout << pi<<endl;
	avec.push_back(pi);
      }
//       std::cout<<"--------------------\n";
	
      pi=event.getpartvn(0);
      pi=event.getpartvn(np-1);

      /**set fastjet flags */
      // double Rparam = 0.5;
      // fastjet::Strategy strategy = fastjet::Best;
      // fastjet::RecombinationScheme recomb_scheme = fastjet::E_scheme;
      
      // /** Define jet for fastjet*/
      // fastjet::JetDefinition jet_def(fastjet::antikt_algorithm, Rparam, recomb_scheme, strategy);
      
      /** fastjet find clusters */
      fastjet::ClusterSequence ev(avec, Flags.jet_def);
      
      /** Retrieve the final state jets from fastjet Sorted by Rapidity*/
      std::vector<fastjet::PseudoJet> jets = sorted_by_rapidity(ev.inclusive_jets(Flags.jetptmin));
      
      // njets=int(jets.size());
      // Now find the tagging jets
      njetst=0;
      njetsb=0;
      ya=10;
      yb=-10;

      // Count how many of the fastjet jets have Et>energy_cut GeV 
      // and calculate largest distance between jets of more than ptcut2
      std::vector<fastjet::PseudoJet>::const_iterator itr = jets.begin();
      std::vector<fastjet::PseudoJet>::const_iterator jai,jbi;
      double ht=0.;
      //      std::cout<<"----------\n";
      for( ; itr != jets.end() ; ++itr) {
	if (((*itr).perp()>Flags.jetptmin)&&(fabs((*itr).rapidity())<jetrapmax)) {
	  njetsb++;
	  if ((*itr).perp()>Flags.jetptmin) {
	    njetst++;
	    ht+=(*itr).perp();
	    yi=(*itr).rapidity();
	    if (yi<ya) { ya=yi ; ja.set((*itr)[0],(*itr)[1],(*itr)[2],(*itr)[3]);jai=itr;};
	    if (yi>yb) { yb=yi ; jb.set((*itr)[0],(*itr)[1],(*itr)[2],(*itr)[3]);jbi=itr;};
	  }
	}
      }

      // Cut on electron and neutrino
      bool leptoncuts(false);
      particle l,al;
      HepLorentzVector * pe,*pnu;
      l=event.getlepton();
      al=event.getalepton();
      // find the electron and the neutrino
      
      if (l.PID==11) { // if the lepton is the negatively charged electron
	pe=&l.p;
	pnu=&al.p;
	iselectron=true;
      } else {
	pe=&al.p;
	pnu=&l.p;
	iselectron=false;
      }
      //      std::cout<<"WJG: "<<wt<<"  "<<pe->e()<<"  "<<pe->perp()<<"  "<<pnu->perp()<<"  "<<pe->rapidity()<<"\n";
      double MWperp=sqrt(2.*pe->perp()*pnu->perp()*(1.-cos(pe->phi()-pnu->phi())));
      if ((pe->perp()>eptmin)&&(fabs(pe->rapidity())<erapmax)&&(pnu->perp()>missptmin)&&(MWperp>mtwmin)) {
      	leptoncuts=true;
	ht+=pe->perp()+pnu->perp();
      }
      
      if (njetst>=2&&leptoncuts)
	// Perform the analysis
	{
	  event.xsec[event.getprocesstype()][0]+=wts;
	  event.xsec[event.getprocesstype()][1]+=wts*wts;
	  if (wt>event.maxweight[event.getprocesstype()])
	    event.maxweight[event.getprocesstype()]=wt;
	  if (Flags.LHoutput) {
	    if ((Flags.unweight&&(rUnweight.flat()<wt/event.maxweight[event.getprocesstype()]))||!Flags.unweight) {
	      event.LHSaveEvent();
	      //              hwt2->Fill(log(wt));
	    }
	  }
	  
	  if (Flags.rootevents) {
	    if ((Flags.unweight&&(rUnweight.flat()<wt/event.maxweight[event.getprocesstype()]))||!Flags.unweight) {
	      // if (np>22) std::cout << "Ouch!\n";
	      int ii=0;
	      std::vector<fastjet::PseudoJet>::const_iterator itr = jets.begin();
	      for( ; itr != jets.end() ; ++itr) {
	      	rpp[ii][0]=(*itr).e();
	      	rpp[ii][1]=(*itr).px();
	      	rpp[ii][2]=(*itr).py();
	      	rpp[ii][3]=(*itr).pz();
	      	++ii;
	      }
	      // for (ii=0;ii<np;ii++) {
	      // 	pi=event.getpartvn(ii);
	      // 	// 		  std::cout << pi<<endl;
	      // 	// 		  rpp[ii][0]=0.;
	      // 	// 		  rpp[ii][1]=1.;
	      // 	// 		  rpp[ii][2]=2.;
	      // 	// 		  rpp[ii][3]=3.;
	      // 	rpp[ii][0]=pi.e();
	      // 	rpp[ii][1]=pi.x();
	      // 	rpp[ii][2]=pi.y();
	      // 	rpp[ii][3]=pi.z();
	      // }
	      for (int iii=ii;iii<22;iii++) { // reset the rest
		rpp[iii][0]=0.;
		rpp[iii][1]=0.;
		rpp[iii][2]=0.;
		rpp[iii][3]=0.;
	      }
	      rpe[0]=pe->e();
	      rpe[1]=pe->x();
	      rpe[2]=pe->y();
	      rpe[3]=pe->z();
	      rpnu[0]=pnu->e();
	      rpnu[1]=pnu->x();
	      rpnu[2]=pnu->y();
	      rpnu[3]=pnu->z();
	      processtype=event.getprocesstype();
	      std::vector<double> vweights=event.GetWeights();
	      for (unsigned int ii=0;ii<vweights.size();++ii) {
		weights[ii]=vweights.at(ii)/double(Flags.NEvents);
		if (isnan(weights[ii])) {
			  std::cerr << "isnan : "<<ii<<std::endl;
			  std::cerr << "weight :" << wts<<std::endl;
			  std::cerr << "weights["<<ii<<"] :" << weights[ii]<<std::endl;
			  std::cerr << event.getpdfta()<<std::endl;
			  std::cerr << event.getpdftb()<<std::endl;
		}
	      }
	      // retrieve all the values and save in local variables to be written
	      aptype=event.getaptype();
	      bptype=event.getbptype();
	      xa=event.getxa();
	      xb=event.getxb();
	      mura=event.getQa();
	      murb=event.getQb();
	      pdfta=event.getpdfta();
	      pdftb=event.getpdftb();
	      // ...
	      Events->Fill();
	      ++evententries;
	    }
	    if (evententries % 1000000 == 0) Events->AutoSave();
	  }
	  
	  if (HEPMCoutput) {
#if usehepmc
	    HepMC::GenEvent evt;
	    evt.use_units(HepMC::Units::GEV, HepMC::Units::MM);
	    evt.weights().push_back(wts);
	    HepMC::GenVertex * v = new HepMC::GenVertex();
	    evt.add_vertex(v);

	    HepMC::GenParticle * b1 = new HepMC::GenParticle(HepMC::FourVector(0,0,Flags.Ebeam,Flags.Ebeam),(Flags.beam[1]==1?2212:-2212),4);
	    HepMC::GenParticle * b2 = new HepMC::GenParticle(HepMC::FourVector(0,0,-Flags.Ebeam,Flags.Ebeam),(Flags.beam[0]==1?2212:-2212),4);

	    v->add_particle_in(b1);
	    v->add_particle_in(b2);
	    evt.set_beam_particles(b1,b2);
	    for (int ii=0;ii<np;ii++) {
	      pi=event.getpartvn(ii);
	      HepMC::GenParticle * p;
	      p=new HepMC::GenParticle(HepMC::FourVector(pi.px(),pi.py(),pi.pz(),pi.e()),21,1);
	      p->set_generated_mass(0.);
	      v->add_particle_out(p);
	    }
	    HepMC::GenParticle * gpe= new HepMC::GenParticle(HepMC::FourVector(pe->px(),pe->py(),pe->pz(),pe->e()),(iselectron?11:-11),1);
	    gpe->set_generated_mass(0.);
	    v->add_particle_out(gpe);
	    HepMC::GenParticle * gpnu = new HepMC::GenParticle(HepMC::FourVector(pnu->px(),pnu->py(),pnu->pz(),pnu->e()),(iselectron?-12:12),1);
	    gpnu->set_generated_mass(0.);
	    v->add_particle_out(gpnu);
	    HepMC::GenCrossSection Gxsec;
	    // Gxsec.set_cross_section();
	    // evt.print();
	    HEPMCio.write_event(&evt);
#endif
	  }
	  Fill2DHistograms(htotal,htotalN,htotalwt2,0.,wts);
#if makeplots
	  // std::cout <<pe->perp() << " "<<pe->rapidity()<<" "<<pnu->perp()<<std::endl;
	  // htotal->Fill(0.,wts);
	  // hya->Fill(ya,wts*4.);
	  Fill2DHistograms(hya,hyaN,hyawt2,ya,wts);
	  // hyb->Fill(yb,wts*4.);
	  Fill2DHistograms(hyb,hybN,hybwt2,yb,wts);
	  // hydif->Fill(fabs(yb-ya),wts*4.);
	  Fill2DHistograms(hydif,hydifN,hydifwt2,fabs(yb-ya),wts);
	  // hye->Fill(pe->rapidity(),wts*4.);
	  if (njetst>=3) {
	    Fill2DHistograms(hydif3,hydif3N,hydif3wt2,fabs(yb-ya),wts);
	    Fill2DHistograms(hht3,hht3N,hht3wt2,ht,wts);
	  }
	  if (njetst==2)
	    Fill2DHistograms(hydiftwojet,hydiftwojetN,hydiftwojetwt2,fabs(yb-ya),wts);
	  if (njetst==3)
	    Fill2DHistograms(hydifthreejet,hydifthreejetN,hydifthreejetwt2,fabs(yb-ya),wts);
	  if (njetst==4)
	    Fill2DHistograms(hydiffourjet,hydiffourjetN,hydiffourjetwt2,fabs(yb-ya),wts);
	  if (njetst==5)
	    Fill2DHistograms(hydiffivejet,hydiffivejetN,hydiffivejetwt2,fabs(yb-ya),wts);
	  if (njetst==6)
	    Fill2DHistograms(hydifsixjet,hydifsixjetN,hydifsixjetwt2,fabs(yb-ya),wts);
	  if (njetst==7)
	    Fill2DHistograms(hydifsevenjet,hydifsevenjetN,hydifsevenjetwt2,fabs(yb-ya),wts);
	  if (njetst==8)
	    Fill2DHistograms(hydifeightjet,hydifeightjetN,hydifeightjetwt2,fabs(yb-ya),wts);
	  if (njetst==9)
	    Fill2DHistograms(hydifninejet,hydifninejetN,hydifninejetwt2,fabs(yb-ya),wts);
	  Fill2DHistograms(hye,hyeN,hyewt2,pe->rapidity(),wts);
	  if (l.PID==11) {
	    Fill2DHistograms(hyem,hyemN,hyemwt2,pe->rapidity(),wts);
	    Fill2DHistograms(hynum,hynumN,hynumwt2,pnu->rapidity(),wts);
	  }
	  else {
	    Fill2DHistograms(hyep,hyepN,hyepwt2,pe->rapidity(),wts);
	    Fill2DHistograms(hynup,hynupN,hynupwt2,pnu->rapidity(),wts);
	  }
	  // hynu->Fill(pnu->rapidity(),wts*4.);
	  Fill2DHistograms(hynu,hynuN,hynuwt2,pnu->rapidity(),wts);
	  // hpt1->Fill(ja.perp(),wts/10);
	  Fill2DHistograms(hyw,hywN,hywwt2,(*pe+*pnu).rapidity(),wts);
	  Fill2DHistograms(hpt1,hpt1N,hpt1wt2,ja.perp(),wts);
	  // hpt2->Fill(jb.perp(),wts/10);
	  Fill2DHistograms(hpt2,hpt2N,hpt2wt2,jb.perp(),wts);
	  // hnjetst->Fill(njetst,wts);
	  Fill2DHistograms(hnjets,hnjetsN,hnjetswt2,njetst,wts);
	  // hptnu->Fill(pnu->perp(),wts/10.);
	  Fill2DHistograms(hptnu,hptnuN,hptnuwt2,pnu->perp(),wts);
	  // hpte->Fill(pe->perp(),wts/10.);
	  Fill2DHistograms(hpte,hpteN,hptewt2,pe->perp(),wts);
	  if (l.PID==11) {
	    Fill2DHistograms(hptnum,hptnumN,hptnumwt2,pnu->perp(),wts);
	    Fill2DHistograms(hptem,hptemN,hptemwt2,pe->perp(),wts);
	  }
	  else {
	    Fill2DHistograms(hptnup,hptnupN,hptnupwt2,pnu->perp(),wts);
	    Fill2DHistograms(hptep,hptepN,hptepwt2,pe->perp(),wts);
	  }
	  // hsenu->Fill(pe->invariantMass(*pnu),wts*5.);
	  Fill2DHistograms(hsenu,hsenuN,hsenuwt2,pe->invariantMass(*pnu),wts);
	  // hnp->Fill(event.getnp(),wts);
	  Fill2DHistograms(hnp,hnpN,hnpwt2,event.getnp(),wts);
	  // havgnjets->Fill(yb-ya,njetst*wts*5.);
	  Fill2DHistograms(havgnjets,havgnjetsN,havgnjetswt2,fabs(yb-ya),njetst*wts);
	  if (njetst<=3)
	    Fill2DHistograms(havgnjets3,havgnjets3N,havgnjets3wt2,fabs(yb-ya),njetst*wts);
	  if (njetst<=4)
	    Fill2DHistograms(havgnjets4,havgnjets4N,havgnjets4wt2,fabs(yb-ya),njetst*wts);
	  if (njetst<=5)
	    Fill2DHistograms(havgnjets5,havgnjets5N,havgnjets5wt2,fabs(yb-ya),njetst*wts);
	  if (njetst<=6)
	    Fill2DHistograms(havgnjets6,havgnjets6N,havgnjets6wt2,fabs(yb-ya),njetst*wts);
	  tpavgjets->Fill(fabs(yb-ya),njetst,wts);
	  Fill2DHistograms(havgnjetsnorm,havgnjetsnormN,havgnjetsnormwt2,fabs(yb-ya),wts);
	  double cosphi=((ja.x()*jb.x()+ja.y()*jb.y())/(sqrt(ja.x()*ja.x()+ja.y()*ja.y())*sqrt(jb.x()*jb.x()+jb.y()*jb.y())));
	  // hcosphi->Fill(fabs(ya-yb),-4.*cosphi*wt);
	  Fill2DHistograms(hcosphi,hcosphiN,hcosphiwt2,fabs(ya-yb),-cosphi*wts);
	  // hcosphinorm->Fill(fabs(ya-yb),4.*wt);
	  Fill2DHistograms(hcosphinorm,hcosphinormN,hcosphinormwt2,fabs(ya-yb),wts);
	  tpavgcosphi->Fill(fabs(ya-yb),-cosphi,wts);

	  Fill2DHistograms(hht,hhtN,hhtwt2,ht,wts);
	  Fill2DHistograms(hhtavgnjets,hhtavgnjetsN,hhtavgnjetswt2,ht,njetst*wts);
	  if (njetst<=3)
	    Fill2DHistograms(hhtavgnjets3,hhtavgnjets3N,hhtavgnjets3wt2,ht,njetst*wts);
	  if (njetst<=4)
	    Fill2DHistograms(hhtavgnjets4,hhtavgnjets4N,hhtavgnjets4wt2,ht,njetst*wts);
	  if (njetst<=5)
	    Fill2DHistograms(hhtavgnjets5,hhtavgnjets5N,hhtavgnjets5wt2,ht,njetst*wts);
	  if (njetst<=6)
	    Fill2DHistograms(hhtavgnjets6,hhtavgnjets6N,hhtavgnjets6wt2,ht,njetst*wts);
	  Fill2DHistograms(hhtavgnjetsnorm,hhtavgnjetsnormN,hhtavgnjetsnormwt2,ht,wts);

	  tphtjets->Fill(ht,njetst,wts);
	  tphtydif->Fill(ht,fabs(ya-yb),wts);

	  std::vector<fastjet::PseudoJet> ptjets = sorted_by_pt(ev.inclusive_jets(Flags.jetptmin));
	  CLHEP::HepLorentzVector jha,jhb;
	  jha.set(ptjets.at(0).px(),ptjets.at(0).py(),ptjets.at(0).pz(),ptjets.at(0).e());
	  jhb.set(ptjets.at(1).px(),ptjets.at(1).py(),ptjets.at(1).pz(),ptjets.at(1).e());
	  double sij=2.*jha.dot(jhb);
	  Fill2DHistograms(hsij,hsijN,hsijwt2,sij,wts);
#endif
	  hwt->Fill(log(wt));
	  hwtwt->Fill(log(wt),wt);
#if makeplots
	  // Number of partons in event:
	  int np=event.getnp();
	  double ysum(0.),yim1(0.);
	  for (int ii=0;ii<np;ii++) {
	      
	    CLHEP::HepLorentzVector pi=event.getpartvn(ii);
	    if (ii>0)
	      hydij->Fill(fabs(pi.rapidity()-yim1),wt*10.);
	    yim1=pi.rapidity();
	    ysum+=pi.rapidity();
	    if ((ii>0)&&(ii<np-1))
	      // hktn->Fill(pi.perp(),wts/10.);
	      Fill2DHistograms(hktn,hktnN,hktnwt2,pi.perp(),wts);
	  }
	  // hycentre->Fill(ysum/np,wts*4.);
	  Fill2DHistograms(hycentre,hycentreN,hycentrewt2,ysum/np,wts);
#endif
	}
    }
  }
  // event.xsec[1][0]=htotal->GetBinContent(1);
  // event.xsec[1][1]=htotal->GetBinError(1);
  std::cout<<"Cross section passing cuts (pb): "<<htotal->GetBinContent(1)<<" +/- "<<htotal->GetBinError(1)<<std::endl;
  for (int count=1;count<=4;count++) 
    std::cout<<"Cross section passing cuts (pb): "<<event.xsec[count][0]<<" +/- "<<sqrt(event.xsec[count][1])<<std::endl;

  for (int count=1;count<=4;count++) 
    std::cout << event.maxweight[count] << " ";
  std::cout << endl;

  hfile.Write();
  hfile.Close();
  return 0;

}
