//////////////////////////////////////////////////
//////////////////////////////////////////////////
// This source code is Copyright (2012) of      //
//  Jeppe R. Andersen                           //
//  Jennifer M. Smillie                         //
// and is distributed under the                 //
// Gnu Public License version 2                 //
// http://www.gnu.org/licenses/gpl-2.0.html     //
// You are allowed to distribute and alter the  //
// source under the conditions of the GPLv2     //
// as long as this copyright notice             //
// is unaltered and distributed with the source //
// Any use should comply with the               //
//             MCNET GUIDELINES                 //
//    for Event Generator Authors and Users     //
// as distributed with this source code         //
//////////////////////////////////////////////////
//////////////////////////////////////////////////
// The current status of the code is (switches):
// (choice of pdfs)
#define USELHAPDF 0
#include "WJets.h"
#include "Currents.h"
#include "Tree2jenu.h"
#include "Tree3jenu.h"
#if USELHAPDF
#include "LHAPDF/LHAPDF.h"
#endif
#include <fenv.h>
const double cut=-2.0; // lambda=2: -1; lambda=.2: -2
const double clambda=0.2;
const double ccut=0.2;
const double sf=.5; // lambda=2: 0.25; lambda=.2: .5
// const double mus=1.;
// const double cut=-1.;
// const double clambda=0.2;
// const double ccut=0.2;

#define printout 0
#define NewPhSp 0
#define FKLphspone 1
#define BlackHatTest 0
#define SmartThingCL 1
#define MirrorTest 0
#define fix 0
#define nottreeonly 1   // 0 generates tree-level exact
#define dijetLO 0
#define trijetLO 0
#define fourjetLO 0
#define calcscaleunc 1
#define newreshuffle 1 // keep the rapidities of jets fixed, change pt and E

double geometricmean(std::vector<fastjet::PseudoJet> * jets)
{
  double gm(1.);
  std::vector<fastjet::PseudoJet>::const_iterator itr = jets->begin();
  for( ; itr != jets->end() ; ++itr)
    gm*=(*itr).perp();
  gm=pow(gm,1./jets->size());
  return gm;
}

double C2Lipatov(HepLorentzVector qav, HepLorentzVector qbv, HepLorentzVector p1, HepLorentzVector p2)
{
  HepLorentzVector temptrans=-(qav+qbv);
  HepLorentzVector p5=qav-qbv;
  HepLorentzVector CL=temptrans+(qav.m2()/p5.dot(p1)+2.*p5.dot(p2)/p1.dot(p2))*p1-p2*(qbv.m2()/p5.dot(p2)+2.*p5.dot(p1)/p1.dot(p2));
  
#if printoutput
  cout << "#Fadin qa : "<<qav<<endl;
  cout << "#Fadin qb : "<<qbv<<endl;
  cout << "#Fadin p1 : "<<p1<<endl;
  cout << "#Fadin p2 : "<<p2<<endl;
  cout << "#Fadin p5 : "<<p5<<endl;
  cout << "#Fadin Gauge Check : "<< CL.dot(p5)<<endl;
  cout << "#Fadin C2L : "<< -CL.dot(CL)<<" "<<-CL.dot(CL)/(qav.m2()*qbv.m2())/(4./p5.perp2())<<endl;
#endif
#if 0
  if (-CL.dot(CL)<0.)
    //   if (fabs(CL.dot(p5))>fabs(CL.dot(CL))) // not sufficient!
    return 0.;
  else 
#endif
    return -CL.dot(CL);
}

double C2Lipatovots(HepLorentzVector qav, HepLorentzVector qbv, HepLorentzVector p1, HepLorentzVector p2)
{
  double kperp=(qav-qbv).perp();
  if (kperp>clambda)
    return C2Lipatov(qav, qbv, p1, p2)/(qav.m2()*qbv.m2());
  else {
    double Cls=(C2Lipatov(qav, qbv, p1, p2)/(qav.m2()*qbv.m2()));
    double temp=Cls-4./(kperp*kperp);
    // std::cout <<kperp <<" "<<temp<<" "<<4./(kperp*kperp)<<" "<<(C2Lipatov(qav, qbv, pa, pb, p1, p2)/(qav.m2()*qbv.m2()))<<std::endl;
    return temp;
  }
}

double C2Lipatov(CLHEP::HepLorentzVector qav, CLHEP::HepLorentzVector qbv, CLHEP::HepLorentzVector pim, CLHEP::HepLorentzVector pip,CLHEP::HepLorentzVector pom, CLHEP::HepLorentzVector pop) // B
{
  CLHEP::HepLorentzVector temptrans=-(qav+qbv);
  CLHEP::HepLorentzVector p5=qav-qbv;
  CLHEP::HepLorentzVector CL=temptrans+qav.m2()*(1./p5.dot(pip)*pip+1./p5.dot(pop)*pop)/2.-qbv.m2()*(1./p5.dot(pim)*pim+1./p5.dot(pom)*pom)/2.+(pip*(p5.dot(pim)/pip.dot(pim)+p5.dot(pom)/pip.dot(pom)) + pop*(p5.dot(pim)/pop.dot(pim)+p5.dot(pom)/pop.dot(pom)) - pim*(p5.dot(pip)/pip.dot(pim) + p5.dot(pop)/pop.dot(pim)) - pom*(p5.dot(pip)/pip.dot(pom) + p5.dot(pop)/pop.dot(pom)) )/2.;

    return -CL.dot(CL);
}

double C2Lipatovots(CLHEP::HepLorentzVector qav, CLHEP::HepLorentzVector qbv, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb, CLHEP::HepLorentzVector p1, CLHEP::HepLorentzVector p2)
{
  double kperp=(qav-qbv).perp();
  if (kperp>clambda)
    return C2Lipatov(qav, qbv, pa, pb, p1, p2)/(qav.m2()*qbv.m2());
  else {
    double Cls=(C2Lipatov(qav, qbv, pa, pb, p1, p2)/(qav.m2()*qbv.m2()));
    double temp=Cls-4./(kperp*kperp);
    // std::cout <<kperp <<" "<<temp<<" "<<4./(kperp*kperp)<<" "<<(C2Lipatov(qav, qbv, pa, pb, p1, p2)/(qav.m2()*qbv.m2()))<<std::endl;
    return temp;
  }
}

// The constructor
CWMultijet::CWMultijet(void) 
  // : parton("Grids/mstw2008nlo.00.dat"),
  : ran(new Ranlux64Engine())
{
  // set the default values
  lambda=clambda;
  ymax=5.5;// the maximum (and -minimum) rapidity generated for any particle
  std::cout << "************************************************************\n";
  std::cout << "*                                                          *\n";
  std::cout << "*                #     #  #######        #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #######  #####          #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #######   #####                  *\n";
  std::cout << "*                                                          *\n";
  std::cout << "* http://cern.ch/hej                      Version 1.4.4    *\n";
  std::cout << "* by Jeppe R. Andersen, jeppe.andersen@cern.ch             *\n";
  std::cout << "*    Jennifer M. Smillie, j.m.smillie@ed.ac.uk             *\n";
  std::cout << "* If you use this software, please cite the following      *\n";
  std::cout << "* papers:                                                  *\n";
  std::cout << "* arXiv:1206.6763 (JHEP 1209 (2012) 047)                   *\n";
  std::cout << "* arXiv:0910.5113 (Phys.Rev. D81 (2010) 114021)            *\n";
  std::cout << "* arXiv:0908.2786 (JHEP 1001:039,2010)                     *\n";
  std::cout << "************************************************************\n";
  parton = new c_mstwpdf("Grids/mstw2008nlo.00.dat"); // default: warn=false, fatal=true
  // init alpha_s
  double FR2 = 1e0, Q0 = 1e0, mTop = 1.e10;
  InitAlphaS(parton->alphaSorder,FR2,Q0,parton->alphaSQ0,parton->mCharm,parton->mBottom,mTop); // 
  // Setup_Currents();
}

CWMultijet::CWMultijet(string filename) 
  // :  parton("Grids/mstw2008nlo.00.dat"),
  : ran(new Ranlux64Engine())
{
  // set the default values
  lambda=clambda;
  ymax=5.5;// the maximum (and -minimum) rapidity generated for any particle
  std::cout << "************************************************************\n";
  std::cout << "*                                                          *\n";
  std::cout << "*                #     #  #######        #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #######  #####          #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #######   #####                  *\n";
  std::cout << "*                                                          *\n";
  std::cout << "* http://cern.ch/hej                      Version 1.4.4    *\n";
  std::cout << "* by Jeppe R. Andersen, jeppe.andersen@cern.ch             *\n";
  std::cout << "*    Jennifer M. Smillie, j.m.smillie@ed.ac.uk             *\n";
  std::cout << "* If you use this software, please cite the following      *\n";
  std::cout << "* papers:                                                  *\n";
  std::cout << "* arXiv:1206.6763 (JHEP 1209 (2012) 047)                   *\n";
  std::cout << "* arXiv:0910.5113 (Phys.Rev. D81 (2010) 114021)            *\n";
  std::cout << "* arXiv:0908.2786 (JHEP 1001:039,2010)                     *\n";
  std::cout << "************************************************************\n";
  parton = new c_mstwpdf(filename.c_str()); // default: warn=false, fatal=true
  // init alpha_s
  double FR2 = 1e0, Q0 = 1e0, mTop = 1.e10;
  InitAlphaS(parton->alphaSorder,FR2,Q0,parton->alphaSQ0,parton->mCharm,parton->mBottom,mTop);
  // Setup_Currents();

  LHoutput.open(filename.c_str());
  LHoutput << "<LesHouchesEvents version=\"1.0\">\n";
  LHoutput << "<!--\n";
  LHoutput << "File generated with HEJ Jets vers. 1.4.4\n";
  LHoutput << "-->\n";
  LHoutput << "<init>\n";
  LHoutput << "  2212   2212  7.000000E+03  7.000000E+03     0     0     7     7     4     1\n";
  LHoutput << "  1E+00  1E-02  1.000000E+00    1\n";
  LHoutput << "</init>\n";
}

CWMultijet::CWMultijet(SGeneratorFlags GeneratorFlagsa) 
  // : parton(GeneratorFlagsa.mstwpdffile),
  : ran(new Ranlux64Engine())
{
  flags=GeneratorFlagsa;
  // set the default values
  lambda=clambda;
  ymax=5.5;// the maximum (and -minimum) rapidity generated for any particle

  std::cout << "************************************************************\n";
  std::cout << "*                                                          *\n";
  std::cout << "*                #     #  #######        #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #     #  #              #                 *\n";
  std::cout << "*                #######  #####          #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #        #     #                 *\n";
  std::cout << "*                #     #  #######   #####                  *\n";
  std::cout << "*                                                          *\n";
  std::cout << "* http://cern.ch/hej                      Version 1.4.4    *\n";
  std::cout << "* by Jeppe R. Andersen, jeppe.andersen@cern.ch             *\n";
  std::cout << "*    Jennifer M. Smillie, j.m.smillie@ed.ac.uk             *\n";
  std::cout << "* If you use this software, please cite the following      *\n";
  std::cout << "* papers:                                                  *\n";
  std::cout << "* arXiv:1206.6763 (JHEP 1209 (2012) 047)                   *\n";
  std::cout << "* arXiv:0910.5113 (Phys.Rev. D81 (2010) 114021)            *\n";
  std::cout << "* arXiv:0908.2786 (JHEP 1001:039,2010)                     *\n";
  std::cout << "************************************************************\n";
#if USELHAPDF
  LHAPDF::initPDFSet(GeneratorFlagsa.LHAPDFName, LHAPDF::LHGRID, GeneratorFlagsa.LHAPDFsubset);
#endif
  std::cout<<"# ";
  parton = new c_mstwpdf(GeneratorFlagsa.mstwpdffile); // default: warn=false, fatal=true

  // init alpha_s
  double FR2 = 1e0, Q0 = 1e0, mTop = 1.e10;
  InitAlphaS(parton->alphaSorder,FR2,Q0,parton->alphaSQ0,parton->mCharm,parton->mBottom,mTop);

  if (flags.raninitfile.length()!=0) {
    cout << "# Restoring Random Engine Status from : "<<flags.raninitfile<<std::endl;
    ran.restoreEngineStatus(flags.raninitfile.c_str());
  }
  maxweight[0]=0;
  maxweight[1]=1.5E4;
  maxweight[2]=5E3;
  maxweight[3]=1E4;
  maxweight[4]=1E4;
  xsec[0][0]=xsec[1][0]=xsec[2][0]=xsec[3][0]=xsec[4][0]=0.;
  xsec[0][1]=xsec[1][1]=xsec[2][1]=xsec[3][1]=xsec[4][1]=0.;
  if (flags.LHoutput) {
    LHoutput.open(flags.LHoutputfile.c_str());
    LHoutput << "<LesHouchesEvents version=\"1.0\">\n";
    LHoutput << "<!--\n";
    LHoutput << "File generated with HEJ Jets vers. 1.4.4\n";
    LHoutput << "-->\n";
    LHoutput << "<init>\n";
    LHoutput << "  ";
    if (flags.beam[0]==1)
      LHoutput<<"2212";
    else 
      LHoutput<<"-2212";
    LHoutput << "   ";
    if (flags.beam[1]==1)
      LHoutput<<"2212 ";
    else 
      LHoutput<<"-2212 ";

    LHoutput <<scientific<<setprecision(7)<<setw(12)<<flags.Ebeam<<"  "<<flags.Ebeam<<" 0     0     7     7     ";
    if (flags.unweight)
      LHoutput << "2     ";
    else
      LHoutput << "4     ";
    // The trick is that we can first write some arbitrary data to the file, 
    // save a pointer, and then return later to overwrite the fields! 
    // lean and mean!
    xsecpos=LHoutput.tellp();
    if (flags.match) { // if matching
      LHoutput <<"4\n"; // How many processes form part of the matching?
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 1\n"; 
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 2\n"; 
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 3\n"; 
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 4\n"; 
    } else { // if no matching
      LHoutput <<"1\n";
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 1\n"; 
    }
    // 

    LHoutput << "</init>\n";
  }
}

CWMultijet::~CWMultijet(void)
  // The destructor
{
  if(LHoutput.is_open()) {
    LHoutput << "</LesHouchesEvents>\n";

    LHoutput.seekp(xsecpos);
    int processcount;
    if (flags.match) // if matching
      processcount=nprocesstype;
    else // if no matching
      processcount=1;
    LHoutput <<processcount<<"\n"; // How many processes form part of the matching?
    for (int tcount=1;tcount<=processcount;tcount++ )
      LHoutput << "  "<<setprecision(7)<<setw(12)<<xsec[tcount][0]<<" "<<sqrt(xsec[tcount][1])<<" "<<1<<" "<<tcount<<"\n"; 
    
    LHoutput.close();
  }
  delete(parton);
}

double CWMultijet::getmatchreweight()
{
  return matchreweight;
}

double CWMultijet::getxa()
{
  return xa;
}

double CWMultijet::getxb()
{
  return xb;
}

double CWMultijet::getQa()
{
  return Qa;
}

double CWMultijet::getQb()
{
  return Qb;
}

double CWMultijet::getpdfta()
{
  return pdfta;
}

double CWMultijet::getpdftb()
{
  return pdftb;
}

int CWMultijet::getaptype()
{
  return aptype;
}

int CWMultijet::getbptype()
{
  return bptype;
}

void CWMultijet::LHSaveEvent(void)
{
  if(LHparticles.size()>0){
    LHoutput << "<event>\n";
  
    if (flags.unweight==false)
      LHoutput << " " << LHparticles.size()<< " 1 " <<scientific<<setprecision(10)<<setw(19)<<wt<<" "<<flags.scale<<" "<<1./128.<< " "<<AlphaS(flags.scale)<<endl;
    else
      LHoutput << " " << LHparticles.size()<< " "<< getprocesstype()<<" " <<scientific<<setprecision(2)<<setw(4)<<1<<" "<<scientific<<setprecision(10)<<setw(19)<<flags.scale<<" "<<1./128.<< " "<<AlphaS(flags.scale)<<"\n";
    
    for (unsigned int i=0;i<LHparticles.size();i++) {
      LHoutput << " " ;
      LHoutput << setw(3) << LHparticles.at(i).PID<< " " ;
      LHoutput << setw(2) << LHparticles.at(i).ISTUP<< " " ;
      LHoutput << setw(2) << LHparticles.at(i).MOT1<< " " ;
      LHoutput << setw(2) << LHparticles.at(i).MOT2<< " " ;
      LHoutput << setw(3) << LHparticles.at(i).COL1<< " " ;
      LHoutput << setw(3) << LHparticles.at(i).COL2<< " " ;
      LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.x()<< " " ;
      LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.y()<< " " ;
      LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.z()<< " " ;
      LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.e()<< " " ;
    if (LHparticles.at(i).PID!=24)
      LHoutput << scientific<<setprecision(12)<<setw(19) << 0. << " " ;
    else
      LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.m()<< " " ;
    LHoutput << "0. 9.\n" ;
    }
    LHoutput << "</event>\n";
  }
}

int CWMultijet::getMstatus(void)
{
  return status;
}

int CWMultijet::getstatus(void)
{
  return status;
}

double CWMultijet::getthetaenu(void)
{
  return thetaenu;
}

void CWMultijet::setstatus(int i)
{
  status=i;
}

double CWMultijet::getcmsenergy(void)
{
  return 2.*flags.Ebeam;
}

int CWMultijet::getnp(void)
{
  return partonmomenta.size(); // The total number of partons in final state
}

int CWMultijet::getprocesstype(void)
{
  return processtype;
}

particle CWMultijet::getlepton()
{
  return lepton;
}

particle CWMultijet::getalepton()
{
  return alepton;
}

COM CWMultijet::PW(double s) {
  return 1./(s-flags.MW*flags.MW+COM(0.,1.)*flags.GammaW*flags.MW);
}

HepLorentzVector CWMultijet::getpartvn(int ni)
{

  if ((ni<=int(partonmomenta.size()-1))&&(ni>=0))
    {
      return partonmomenta.at(ni);
    }
  else {
    cerr << "Parton number out of range : "<<ni<<endl;
    exit(0);
  }
}

double CWMultijet::pdfpt(int beam,double x, double uf, int pt)
{
  if (x<1e-5)
    return 0.;
  if (pt==21) // gluon
#if USELHAPDF
    {
      if (flags.useLHApdf)
	return max(0.,LHAPDF::xfx(x,uf,0));
      else
	return max(0.,CWMultijet::parton->parton(0,x,uf));
    }
#else
    return max(0.,CWMultijet::parton->parton(0,x,uf));
#endif
  else if (abs(pt)<7)
#if USELHAPDF
    {
      if (flags.useLHApdf)
	return max(0.,LHAPDF::xfx(x,uf,pt*flags.beam[beam]));
      else
	return max(0.,parton->parton(pt*flags.beam[beam],x,uf));
    }
#else
    return max(0.,parton->parton(pt*flags.beam[beam],x,uf));
#endif
  else {
    cout << "particle type unknown: "<<pt<<endl;
    return 0.;
  }
}

double CWMultijet::Halphas(double q)
{
  double as;
#if USELHAPDF
  if (flags.useLHApdf)
    as=LHAPDF::alphasPDF(q);
  else
    as=AlphaS(q);
#else
  as=AlphaS(q);
#endif
  if (isnan(as)||as>.5)
    as=0.5;
  return as;
}

double CWMultijet::GenerateFKLIncomingParticles(double xa,double ufa,double xb,double ufb)
{
  double r1,lwt(1.),sum;

//   parton->mrst(xa,ufa);
//   pdfa=fabs(parton->cont.glu)+4.0/9.0*(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));
//   pdfa=max(0.,parton->cont.glu)+4.0/9.0*(max(0.,parton->cont.upv+parton->cont.usea) + max(0.,parton->cont.dnv+parton->cont.dsea) + max(0.,parton->cont.usea)+max(0.,parton->cont.dsea) + 2.0*(max(0.,parton->cont.str)+max(0.,parton->cont.chm)+max(0.,parton->cont.bot)));
  pdfga=fabs(pdfpt(0,xa,ufa,21));
  pdfua=fabs(pdfpt(0,xa,ufa,2));
#if BlackHatTest
  pdfda=0.;
  pdfuxa=0.;
  pdfdxa=0.;
  pdfca=0.;
  pdfsa=0.;
  pdfsxa=0.;
  pdfba=0.;
#else
  pdfda=fabs(pdfpt(0,xa,ufa,1));
  pdfuxa=fabs(pdfpt(0,xa,ufa,-2));
  pdfdxa=fabs(pdfpt(0,xa,ufa,-1));
  pdfca=fabs(pdfpt(0,xa,ufa,4));
  pdfsa=fabs(pdfpt(0,xa,ufa,3));
  pdfsxa=fabs(pdfpt(0,xa,ufa,-3));
  pdfba=fabs(pdfpt(0,xa,ufa,5));
#endif

  pdfa=pdfga+4.0/9.0*(pdfua + pdfda + pdfuxa + pdfdxa +pdfsa +pdfsxa + 2.0*(pdfca +pdfba ));
  Qa=ufa;

  r1=pdfa*ran.flat();

#if 1
  if (r1<(sum=pdfga)) {
    aptype=21;
    lwt*=pdfa/pdfga;}
  else if (r1<(sum+=(4./9.)*pdfua)) {
    aptype=2;
    lwt*=pdfa/((4./9.)*pdfua); }
  else if (r1<(sum+=(4./9.)*pdfda)) {
    aptype=1;
    lwt*=pdfa/((4./9.)*pdfda); }
  else if (r1<(sum+=(4./9.)*pdfuxa)) {
    aptype=-2;
    lwt*=pdfa/((4./9.)*pdfuxa); }
  else if (r1<(sum+=(4./9.)*pdfdxa)) {
    aptype=-1;
    lwt*=pdfa/((4./9.)*pdfdxa); }
  else if (r1<(sum+=(4./9.)*pdfca)) {
    aptype=4;
    lwt*=pdfa/((4./9.)*pdfca); }
  else if (r1<(sum+=(4./9.)*pdfca)){
    aptype=-4;
    lwt*=pdfa/((4./9.)*pdfca); }
  else if (r1<(sum+=(4./9.)*pdfsa)) {
    aptype=3;
    lwt*=pdfa/((4./9.)*pdfsa); }
  else if (r1<(sum+=(4./9.)*pdfsxa)) {
    aptype=-3;
    lwt*=pdfa/((4./9.)*pdfsxa); }
  else if (r1<(sum+=(4./9.)*pdfba)) {
    aptype=5;
    lwt*=pdfa/((4./9.)*pdfba); }
  else if (r1<=(sum+=(4./9.)*pdfba)) {
    aptype=-5;
    lwt*=pdfa/((4./9.)*pdfba); }
  else {
    cout << "Error in choosing incoming parton a : "<<xa<<" "<<ufa<<" "<<sum<<" "<<pdfa<<" "<<r1;
    cout << " "<<pdfga+4./9.*(pdfua+pdfuxa+pdfda+pdfdxa+pdfsa+pdfsxa+2.*(pdfca+pdfba))<<endl;
    setstatus(15);
    return 0;
  }
#else
  aptype=1;
#endif

  pdfta=pdfpt(0,xa,ufa,aptype);
  // std::cout<<xa<<" "<<ufa<<" "<<aptype<<" "<<pdfta<<std::endl;

//   parton->mrst(xb,ufb);
//   pdfb=max(0.,parton->cont.glu)+4.0/9.0*(max(0.,parton->cont.upv+parton->cont.usea) + max(0.,parton->cont.dnv+parton->cont.dsea) + max(0.,parton->cont.usea)+max(0.,parton->cont.dsea) + 2.0*(max(0.,parton->cont.str)+max(0.,parton->cont.chm)+max(0.,parton->cont.bot)));
//   pdfb=fabs(parton->cont.glu)+4.0/9.0*(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));
  pdfgb=fabs(pdfpt(1,xb,ufb,21));
  pdfub=fabs(pdfpt(1,xb,ufb,2));
#if BlackHatTest
  pdfdb=0.;
  pdfuxb=0.;
  pdfdxb=0.;
  pdfcb=0.;
  pdfsb=0.;
  pdfsxb=0.;
  pdfbb=0.;
#else
  pdfdb=fabs(pdfpt(1,xb,ufb,1));
  pdfuxb=fabs(pdfpt(1,xb,ufb,-2));
  pdfdxb=fabs(pdfpt(1,xb,ufb,-1));
  pdfcb=fabs(pdfpt(1,xb,ufb,4));
  pdfsb=fabs(pdfpt(1,xb,ufb,3));
  pdfsxb=fabs(pdfpt(1,xb,ufb,-3));
  pdfbb=fabs(pdfpt(1,xb,ufb,5));
#endif  

  // avoid gluon b if a is a gluon
  if (aptype==21)
    pdfgb=0.;

  pdfb=pdfgb+4.0/9.0*(pdfub + pdfdb + pdfuxb + pdfdxb + pdfsb +pdfsxb +2.0*(pdfcb +pdfbb ));
  Qb=ufb;

  r1=pdfb*ran.flat();

#if 1
  if (r1<(sum=pdfgb)) {
    bptype=21;
    lwt*=pdfb/pdfgb; }
  else if (r1<(sum+=(4./9.)*pdfub)) {
    bptype=2;
    lwt*=pdfb/((4./9.)*pdfub); }
  else if (r1<(sum+=(4./9.)*pdfdb)) {
    bptype=1;
    lwt*=pdfb/((4./9.)*pdfdb); }
  else if (r1<(sum+=(4./9.)*pdfuxb)) {
    bptype=-2;
    lwt*=pdfb/((4./9.)*pdfuxb); }
  else if (r1<(sum+=(4./9.)*pdfdxb)) {
    bptype=-1;
    lwt*=pdfb/((4./9.)*pdfdxb); }
  else if (r1<(sum+=(4./9.)*pdfcb)) {
    bptype=4;
    lwt*=pdfb/((4./9.)*pdfcb); }
  else if (r1<(sum+=(4./9.)*pdfcb)){
    bptype=-4;
    lwt*=pdfb/((4./9.)*pdfcb); }
  else if (r1<(sum+=(4./9.)*pdfsb)) {
    bptype=3;
    lwt*=pdfb/((4./9.)*pdfsb); }
  else if (r1<(sum+=(4./9.)*pdfsxb)) {
    bptype=-3;
    lwt*=pdfb/((4./9.)*pdfsxb); }
  else if (r1<(sum+=(4./9.)*pdfbb)) {
    bptype=5;
    lwt*=pdfb/((4./9.)*pdfbb); }
  else if (r1<(sum+=(4./9.)*pdfbb)) {
    bptype=-5;
    lwt*=pdfb/((4./9.)*pdfbb); }
  else {
    cout << "Error in choosing incoming partons b : "<<xb<<" "<<ufb<<" "<<sum<<" "<<pdfb<<" "<<r1;
    cout << " "<<pdfgb+4./9.*(pdfub+pdfuxb+pdfdb+pdfdxb+pdfsb+pdfsxb+2.*(pdfcb+pdfbb))<<endl;
    setstatus(15);
    return 0;
  }

#else
 bptype=21;
#endif

  pdftb=pdfpt(1,xb,ufb,bptype);

  // std::cout << "aptype, bptype: "<<aptype<<", "<<bptype<<std::endl;
  // Reject if both incoming partons are gluons
  if (aptype==21&&bptype==21) {
    std::cout << "aptype, bptype: "<<aptype<<", "<<bptype<<std::endl;
    setstatus(16);
    return 0;
  }
  // std::cout<<pdfta<<" "<<pdftb<<std::endl;

  // if (pdfta<1e-20||pdftb<1e-20) {
  //   setstatus(21);
  // }
  return lwt;
}

double CWMultijet::GenerateIncomingParticles(double xa,double ufa,double xb,double ufb)
{
  double r1,sum,lwt(1.);

//   parton->mrst(xa,ufa);
//   pdfa=fabs(parton->cont.glu)+(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));

  pdfga=fabs(pdfpt(0,xa,ufa,21));
  pdfua=fabs(pdfpt(0,xa,ufa,2));
  pdfda=fabs(pdfpt(0,xa,ufa,1));
  pdfuxa=fabs(pdfpt(0,xa,ufa,-2));
  pdfdxa=fabs(pdfpt(0,xa,ufa,-1));
  pdfca=fabs(pdfpt(0,xa,ufa,4));
  pdfsa=fabs(pdfpt(0,xa,ufa,3));
  pdfsxa=fabs(pdfpt(0,xa,ufa,-3));  
  pdfba=fabs(pdfpt(0,xa,ufa,5));

  pdfa=pdfga+(pdfua + pdfda + pdfuxa + pdfdxa + pdfsa +pdfsxa +2.0*(pdfca +pdfba ));
  Qa=ufa;
#if 1
  r1=pdfa*ran.flat();

  if (r1<(sum=pdfga)) {
    aptype=21;
    lwt*=pdfa/pdfga;}
  else if (r1<(sum+=pdfua)) {
    aptype=2;
    lwt*=pdfa/(pdfua); }
  else if (r1<(sum+=pdfda)) {
    aptype=1;
    lwt*=pdfa/(pdfda); }
  else if (r1<(sum+=pdfuxa)) {
    aptype=-2;
    lwt*=pdfa/(pdfuxa); }
  else if (r1<(sum+=pdfdxa)) {
    aptype=-1;
    lwt*=pdfa/(pdfdxa); }
  else if (r1<(sum+=pdfca)) {
    aptype=4;
    lwt*=pdfa/(pdfca); }
  else if (r1<(sum+=pdfca)){
    aptype=-4;
    lwt*=pdfa/(pdfca); }
  else if (r1<(sum+=pdfsa)) {
    aptype=3;
    lwt*=pdfa/(pdfsa); }
  else if (r1<(sum+=pdfsxa)) {
    aptype=-3;
    lwt*=pdfa/(pdfsxa); }
  else if (r1<(sum+=pdfba)) {
    aptype=5;
    lwt*=pdfa/(pdfba); }
  else if (r1<(sum+=pdfba)) {
    aptype=-5;
    lwt*=pdfa/(pdfba); }
  else {
    cout << "Error in choosing incoming partons: "<<sum<<" "<<pdfa<<" "<<r1;
    cout << " "<<pdfga+4./9.*(pdfua+pdfuxa+pdfda+pdfdxa+pdfsa+pdfsxa+2.*(pdfca+pdfba))<<endl;
    setstatus(15);
    return 0;
  }  
#else
  aptype=21;
#endif

  pdfta=pdfpt(0,xa,ufa,aptype);

//   parton->mrst(xb,ufb);

//   pdfb=fabs(parton->cont.glu)+(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));
  pdfgb=fabs(pdfpt(1,xb,ufb,21));
  pdfub=fabs(pdfpt(1,xb,ufb,2));
  pdfdb=fabs(pdfpt(1,xb,ufb,1));
  pdfuxb=fabs(pdfpt(1,xb,ufb,-2));
  pdfdxb=fabs(pdfpt(1,xb,ufb,-1));
  pdfcb=fabs(pdfpt(1,xb,ufb,4));
  pdfsb=fabs(pdfpt(1,xb,ufb,3));
  pdfsxb=fabs(pdfpt(1,xb,ufb,-3));
  pdfbb=fabs(pdfpt(1,xb,ufb,5));  
  
  pdfb=pdfgb+(pdfub + pdfdb + pdfuxb + pdfdxb + pdfsb +pdfsxb +2.0*(pdfcb +pdfbb ));
  Qb=ufb;

  r1=pdfb*ran.flat();

#if 1
  if (r1<(sum=pdfgb)) {
    bptype=21;
    lwt*=pdfb/pdfgb; }
  else if (r1<(sum+=pdfub)) {
    bptype=2;
    lwt*=pdfb/(pdfub); }
  else if (r1<(sum+=pdfdb)) {
    bptype=1;
    lwt*=pdfb/(pdfdb); }
  else if (r1<(sum+=pdfuxb)) {
    bptype=-2;
    lwt*=pdfb/(pdfuxb); }
  else if (r1<(sum+=pdfdxb)) {
    bptype=-1;
    lwt*=pdfb/(pdfdxb); }
  else if (r1<(sum+=pdfcb)) {
    bptype=4;
    lwt*=pdfb/(pdfcb); }
  else if (r1<(sum+=pdfcb)){
    bptype=-4;
    lwt*=pdfb/(pdfcb); }
  else if (r1<(sum+=pdfsb)) {
    bptype=3;
    lwt*=pdfb/(pdfsb); }
  else if (r1<(sum+=pdfsxb)) {
    bptype=-3;
    lwt*=pdfb/(pdfsxb); }
  else if (r1<(sum+=pdfbb)) {
    bptype=5;
    lwt*=pdfb/(pdfbb); }
  else if (r1<(sum+=pdfbb)) {
    bptype=-5;
    lwt*=pdfb/(pdfbb); }
  else {
    cout << "Error in choosing incoming partons dd: "<<sum<<" "<<pdfb<<" "<<r1;
    cout << " "<<pdfgb+4./9.*(pdfub+pdfuxb+pdfdb+pdfdxb+pdfsb+pdfsxb+2.*(pdfcb+pdfbb))<<endl;
    setstatus(15);
    return 0;
  }

#else
  bptype=21;
#endif

  pdftb=pdfpt(1,xb,ufb,bptype);

  if (pdfta<1e-10||pdftb<1e-10) {
    setstatus(21);
  }

  return lwt;
}

std::vector<particle> CWMultijet::GetLHParticles(void)
{
  return LHparticles;
}

std::vector<double> CWMultijet::GetWeights(void)
{
  return weights;
}

void CWMultijet::SetupLHInformation()
// FIX:
// Does not yet change quark flavour of W-emitting line
// Does not yet include W and the decay products
//      This should read information from input file and only allow the tagged decay channels
{
  std::list<int> vecl,vecr;

  if (aptype>0) { // if "a" is a gluon or quark
    vecl.push_front(-1);
    vecr.push_back(3);
  } else { // if a is an anti-quark
    vecl.push_front(3); 
    vecr.push_back(-1);
  }

  for (unsigned int i=1;i<partonmomenta.size()-1;i++) { // don't deal with the last gluon
    if (ran.flat()<.5) 
      vecr.push_back(i+3);
    else vecl.push_front(i+3);
  }
  
  if (bptype==21) {
    if (ran.flat()<.5) {
      vecr.push_back(partonmomenta.size()+2);
      vecl.push_front(-2);
    } else {
      vecl.push_front(partonmomenta.size()+2);
      vecr.push_back(-2);
    } 
  } else if (bptype<0) { // if b is an anti-quark
    vecr.push_back(partonmomenta.size()+2);
    vecl.push_front(-2);
  } else { // if b is a quark
    vecl.push_front(partonmomenta.size()+2);
    vecr.push_back(-2);
  }
  
  // now generate the colour connection information to go in the LH records
  int ctypea(1); // ctype: the colour start type: anti- or colour?
                 // when we swap the side of particles/anti-particles we should
                 // always start with a colour line
  
  particle tparticle; // a temp. particle
  {
    // setup vector of particles
    tparticle.PID=0;
    tparticle.COL1=0;
    tparticle.COL2=0;
    tparticle.MOT1=0;
    tparticle.MOT2=0;
    tparticle.ISTUP=0;
    tparticle.p.set(0,0,0,0);
    for (unsigned int i=0;i<partonmomenta.size()+5;i++)
      LHparticles.push_back(tparticle);
  }
  
  int ctype,col(501);
  ctype=ctypea;    // which of the colour columns should we fill?

//   cout << "LHparticles size : "<<LHparticles.size()<<endl;
  int n=vecr.front()+2;
  int np1;
  vecr.pop_front();
  while (vecr.size()!=0) {
    int np1=vecr.front()+2;
    vecr.pop_front();
//     cerr << "n, np1, ctype : "<<n<< ", "<<np1<<" "<<ctype<<endl;
    if (ctype>0) {
      LHparticles.at(n).COL1=col;
      LHparticles.at(np1).COL2=col;
    } else {
      LHparticles.at(n).COL2=col;
      LHparticles.at(np1).COL1=col;
    }
    col++;
    n=np1;
//     ctype*=-1;
  }
  
  // Now we are at the end of the right hand side.
  // Should we connect to the left hand side?
  
//   cout << "  --\n";
  if (bptype==21) {
    np1=vecl.front()+2;
    vecl.pop_front();
//     cerr << "n, np1, ctype : "<<n<< ", "<<np1<<" "<<ctype<<endl;
    if (ctype>0) {
      LHparticles.at(n).COL1=col;
      LHparticles.at(np1).COL2=col;
    } else {
      LHparticles.at(n).COL2=col;
      LHparticles.at(np1).COL1=col;
    }
    col++;
    n=np1;
  } else {
    n=vecl.front()+2;
    vecl.pop_front();
  }
//   cout << "  --\n";
  while (vecl.size()!=0) {
    int np1=vecl.front()+2;
    vecl.pop_front();
//     cerr << "n, np1, ctype : "<<n<< ", "<<np1<<" "<<ctype<<endl;
    if (ctype>0) {
      LHparticles.at(n).COL1=col;
      LHparticles.at(np1).COL2=col;
    } else {
      LHparticles.at(n).COL2=col;
      LHparticles.at(np1).COL1=col;
    }
    col++;
    n=np1;
//     ctype*=-1;
  }
  
  // if particle a is gluon we need to make another connections
  if (aptype==21) {
    // we know that if a is a gluon, then we started with particle 0 
    // and have ended with particle -1. We also know that the last colour line 
    // is opposite that of ctype - so connect from -1 to 0 with a ctype!
    if (ctype>0) {
      LHparticles.at(1).COL1=col;
      LHparticles.at(5).COL2=col;
    } else {
      LHparticles.at(1).COL2=col;
      LHparticles.at(5).COL1=col;
    }
    col++;
    n=np1;
  }
  
  // swap colours for particles of incoming type
  {  
    int itemp1,itemp2;
    itemp1=LHparticles.at(0).COL1;
    itemp2=LHparticles.at(0).COL2;
    LHparticles.at(0).COL1=itemp2;
    LHparticles.at(0).COL2=itemp1;
    itemp1=LHparticles.at(1).COL1;
    itemp2=LHparticles.at(1).COL2;
    LHparticles.at(1).COL1=itemp2;
    LHparticles.at(1).COL2=itemp1;
  }

  // incoming particles
  LHparticles.at(0).p=pb;
  LHparticles.at(0).PID=bptype;
  LHparticles.at(0).MOT1=0;
  LHparticles.at(0).MOT2=0;
  LHparticles.at(0).ISTUP=-1;
  LHparticles.at(1).p=pa;
  LHparticles.at(1).PID=aptype;
  LHparticles.at(1).MOT1=0;
  LHparticles.at(1).MOT2=0;
  LHparticles.at(1).ISTUP=-1;
  LHparticles.at(2)=Wboson;
  LHparticles.at(3)=lepton;
  LHparticles.at(4)=alepton;
    
  // outgoing particles
  for (unsigned int i=5;i<partonmomenta.size()+5;i++) { 
    LHparticles.at(i).p=partonmomenta.at(i-5);
    if (i==5)
      LHparticles.at(i).PID=aoutptype;
    else if (i==partonmomenta.size()+4)
      LHparticles.at(i).PID=boutptype;
    else LHparticles.at(i).PID=21;
    LHparticles.at(i).MOT1=1;
    LHparticles.at(i).MOT2=2;
    LHparticles.at(i).ISTUP=1;
  }

  // // setup vector of particles
  // tparticle.PID=25;// Higgs boson
  // tparticle.COL1=0;
  // tparticle.COL2=0;
  // tparticle.MOT1=1;
  // tparticle.MOT2=2;
  // tparticle.ISTUP=1;
  // tparticle.p=getpHv();
  // LHparticles.push_back(tparticle);
}

double CWMultijet::omega0(double q2, double lambda2)
{
  double result1(0.);
  if (flags.logcorrect==0) {
      result1=-Halphas(renscale)*3./M_PI*log(q2/lambda2);
      // renscale=PTMAX if necessary in weight() before Virtuals
  }
  else {   // flags.logcorrect==1
    //    result1=-Halphas(sqrt(sqrt(q2*lambda2))*flags.scalefactor)*3./M_PI*log(q2/lambda2);
    //  Instead, leave log explicitly
    result1=-Halphas(renscale)*3./M_PI*log(q2/lambda2)*(1.+Halphas(renscale)*beta0*log(pow(renscale,4)/(q2*lambda2))/(8.*M_PI));
#if printalphas
    std::cout<<"omega0: "<<Halphas(sqrt(sqrt(q2*lambda2)))<<"  "<<Halphas(renscale)*(1.+Halphas(renscale)*beta0*log(pow(renscale,4)/(q2*lambda2))/(8.*M_PI))<<"\n";
#endif
  }
  return result1;
}

void CWMultijet::AddParton(double p[4])
{
  HepLorentzVector pi(p[1],p[2],p[3],p[0]);
  partonmomenta.push_back(pi);
}

double CWMultijet::GeneratePHSPpoint()
// Generates a phase space point with unit weight
// two W decay products, ng partonmomenta, 2 extremal partons.
{
  int ng;
  double y,ysum,p[4],pt,r1,temp1,temp2,phi,maxptsofar;
  double tscale,kt[2];
//   static const double ptmax(2000.);
//   static const double ptSQgen(40.*40.);
  double omega0n,qtSQgen,qtSQ,qtmod;
//   double minv2after;
  double ptpar(40.);
  double sqrts;

  double pta[GLUEINT+2][2];
  double deltay[GLUEINT+2];
  // double phia[GLUEINT+2];

  lambda=clambda; // trouble is we don't know the value of lambda until after the whole generation
  //  lambda=.25; // trouble is we don't know the value of lambda until after the whole generation

  double wt=1.;
  sqrts=2.*flags.Ebeam;

  // The parton of lowest rapidity:
#if 1
  // Generate the transverse momentum of parton a
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  thetaenu=r1;
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  wt*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
//   wtmatch*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;
  // renscale2=pt*pt;
  //  pt1=pt;
  Ht+=pt;
#else
  {
    double ptSQ;
    r1=ran.flat();
    ptSQ=ptSQgen/(exp(r1*log(1.+ptSQgen))-1.);
    wt*=2.*M_PI*(ptSQ)*(1.+(ptSQ)/ptSQgen)*log(1.+ptSQgen)/2.;
    pt=sqrt(ptSQ);
    // renscale2=ptSQ;
  }
#endif

  phi=2.*M_PI*ran.flat(); // Random angle in x,y-plane

  y=0.;
  ysum=0.;
  // phia[0]=phi;
  
  p[1]=pt*cos(phi);
  p[2]=pt*sin(phi);
  pta[0][0]=pt*cos(phi);
  pta[0][1]=pt*sin(phi);
  deltay[0]=0.;
  maxptsofar=pt;

  kt[0]=-p[1];
  kt[1]=-p[2];


  tscale=kt[0]*kt[0]+kt[1]*kt[1];

  // Generate a rapidity span
  //  omega0n=0.6*min(omega0(tscale,lambda*lambda,renscale2),cut);
  //  if (flags.scalesetting==1||flags.scalesetting==2||flags.scalesetting==3)
    renscale=maxptsofar*flags.scalefactor;   // This is corrected below, necessary for now.
  //  omega0n=min(1.25*omega0(tscale,lambda*lambda),cut);
  omega0n=min(sf*omega0(tscale,lambda*lambda),cut);
  // omega0n=-1.;
  y=log(1.-ran.flat())/omega0n;
  wt*=-1./omega0n/exp(omega0n*y); // take out the exponential factor which was included 
                                  // in the generation of y - we want this in the 
                                  // FKL evaluation
  ysum+=y;
  deltay[1]=ysum;

  // Assume we have already chosen a regulater mu
  // then generate the gluon emissions
  
#if nottreeonly
  ng=int((GLUEINT)*ran.flat()); // a random int number of partonmomenta between 0 and GLUEINT-1
  wt*=double(GLUEINT);
  //  ng=0;  // for nonFKL tests
#else
  ng=0;
#endif
  for (int ii=1;ii<=ng;ii++) {
    
    // Generate a random (weighted) qti min: 0, max: ptmaxsq 
    r1=ran.flat();
    //      qtSQgen=pt1SQ;
    //    qtSQgen=10.*10.;
    //    qtSQgen=12.*12.;
    qtSQgen=15.*15.;
//    qtSQgen=20.*20.;
//    qtSQgen=25.*25.;
//    qtSQgen=30.*30.;
//    qtSQgen=40.*40.;
//    qtSQgen=50.*50.;
    //    qtSQgen=60.*60.;

    // qtSQ=qtSQgen/(exp(r1*log(1.+qtSQgen))-1.)-1.;
    // wt*=2.*M_PI*(qtSQ+1.)*(1.+(qtSQ+1.)/qtSQgen)*log(1.+qtSQgen)/2.;
    double t2=log(1.+qtSQgen);
    double t=exp(r1*t2);
    qtSQ=qtSQgen/(t-1.)-(1.-ccut*ccut);
    wt*=2.*M_PI*(qtSQ+1.-ccut*ccut)*(qtSQ+1.-ccut*ccut)*t*t2/qtSQgen/2.;

    // double t2=log(1.+qtSQgen);
    // double t=exp(r1*t2);
    // qtSQ=qtSQgen/(t-1.)-(1.-ccut*ccut);
    // wt*=2.*M_PI*(qtSQ+1.-ccut*ccut)*(qtSQ+1.-ccut*ccut)*t*t2/qtSQgen/2.;
 
    wt/=16.*pow(M_PI,3);
    qtmod=sqrt(qtSQ);    
    Ht+=qtmod;
    if (qtmod>maxptsofar)
      maxptsofar=qtmod;

    // Generate random angle 
    phi=2.*M_PI*ran.flat();
    // phia[ii]=phi;

    p[1]=pta[ii][0]=qtmod*cos(phi);
    p[2]=pta[ii][1]=qtmod*sin(phi);

    // keep track of the transverse t-channel scale in the evolution
    kt[0]-=p[1];
    kt[1]-=p[2];

    tscale=kt[0]*kt[0]+kt[1]*kt[1];

    // Generate a rapidity span after the emission
    //    omega0n=0.6*min(omega0(tscale,lambda*lambda,renscale2),cut);
    //    omega0n=min(1.25*omega0(tscale,lambda*lambda),cut);
    renscale=maxptsofar*flags.scalefactor;  // This is corrected below, necessary for now
    omega0n=min(sf*omega0(tscale,lambda*lambda),cut);
    // omega0n=-1.;
    y=log(1.-ran.flat())/omega0n;
    ysum+=y;
    deltay[ii+1]=ysum;
    wt*=-1./omega0n/exp(omega0n*y); // take out the exponential factor which was included 
                                    // in the generation of y - we want this in the 
                                    // FKL evaluation
  }

  // Generate the transverse momentum of parton b
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  wt*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  //   wtmatch*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;
  
  phi=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  Ht+=pt;
  
  pta[ng+1][0]=pt*cos(phi);
  pta[ng+1][1]=pt*sin(phi);
  
  kt[0]+=-pta[ng+1][0];
  kt[1]+=-pta[ng+1][1];
  
  // Now we know the transverse momentum of the W, given by the array kt
  double pW[4];
  // Choose its mass according to Breit-Wigner
  r1=ran.flat();
  double sW=flags.MW*(flags.MW + flags.GammaW*tan((M_PI*r1)/2. + (-1. + r1)*atan(flags.MW/flags.GammaW)));
  // Multiply by derivate (d sap)/(d r)
  {
    static double temp=atan(flags.MW/flags.GammaW);
    wt*=(flags.GammaW*flags.MW*(M_PI+2.*temp))/(1.+cos(M_PI*r1+2.*(-1.+r1)*temp));
  }  
  pW[1]=kt[0];
  pW[2]=kt[1];
  double EtW=sqrt(pW[1]*pW[1]+pW[2]*pW[2]+sW);
  
  double yW;
#if 1
  // Generate a yW Gaussian distributed around 0
  {
    double lninvr1,r1,r2,temp,a;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    a=1.;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    yW=temp;
    wt=wt*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
  }
#else
  // choose a rapidity
  yW=-5.+10.*ran.flat();
  wt*=10;
#endif
  pW[0]=cosh(yW)*EtW;
  pW[3]=sinh(yW)*EtW;
  CLHEP::HepLorentzVector pWv;
  pWv.set(pW[1],pW[2],pW[3],pW[0]);
  
  double ppW[4],ppWs;
  CLHEP::HepLorentzVector ppWv,appWv;
  ppWs=sqrt(sW)/2.;
  // Choose theta and phi
  double pptheta=2.*M_PI*ran.flat();
  // ppphi=M_PI*ran.flat();
  // wt*=2.*pow(M_PI,2);
  double cosppphi=2.*ran.flat()-1.;  
  wt*=2.*M_PI*2.; 
  //  double sinphi=sin(ppphi);
  double sinphi=sqrt(1.-cosppphi*cosppphi);  // Know 0 < phi < pi
  //  wt*=1./(pow(2.*M_PI,3)*2.)*sinphi/(4.);   // Divide by 8, *2, see WPhaseSpace.tex 
  wt*=1./(pow(2.*M_PI,3)*2.)/(4.);   // Divide by 8, *2, see WPhaseSpace.tex 
                                     // (phi -> cosphi now so no sinphi in J)
  // construct 4-vector in W rest frame
  ppW[0]=ppWs;
  ppW[1]=ppWs*cos(pptheta)*sinphi;
  ppW[2]=ppWs*sin(pptheta)*sinphi;
  ppW[3]=ppWs*cosppphi;
  ppWv.set(ppW[1],ppW[2],ppW[3],ppW[0]);
  appWv.set(-ppW[1],-ppW[2],-ppW[3],ppW[0]);
  
  // translate to lab frame
  pd1=ppWv.boost(pWv.boostVector());
  pd2=appWv.boost(pWv.boostVector());
  // pd2=pWv-pd1;
  Ht+=(pd1+pd2).perp();
  // Ht+=pd2.perp();

  // The result of generation W instead of FS particle momenta:
  // wt*=pWv.e()/pd2.e();
  
  // std::cout << "Momentum Check: "<<pWv << " "<<pd1+pd2<<std::endl;
  { //double t;
    // calculate rapidity first
    // double y=log((pd1.e()+pd1.pz())/(pd1.e()-pd1.pz()));
    // if (fabs(y)>14.) {
    //   setstatus(22);
    //   wt=0;
    //   return wt;
    // }

    // t=exp(-pd1.rapidity());
    // // t=exp(-y);
    // double t1=(pd1.e()-pd1.pz())/sqrts;
    // double t2=pd1.perp()/sqrts*t;

    // std::cout<< "check : "<<t1<< " "<<t2<<std::endl;
    xa+=(pd1.e()-pd1.pz())/sqrts;//pd1.perp()/sqrts*t;
    xb+=(pd1.e()+pd1.pz())/sqrts;//pd1.perp()/sqrts/t;
  }
  { // double t;
    // t=exp(-pd2.rapidity());
    // xa+=pd2.perp()/sqrts*t;
    // xb+=pd2.perp()/sqrts/t;
    xa+=(pd2.e()-pd2.pz())/sqrts;
    xb+=(pd2.e()+pd2.pz())/sqrts;
  }

  // For the non-NewPhSp bit, see rev 1184 or before

  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double yoffset;
    double ycenter=ysum/(2.);
    {
      double lninvr1,r1,r2,temp,a(1.2);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      //      thetaenu=temp;
      yoffset=(temp-ycenter);
      wt=wt*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to partonmomenta and xa,xb
    for (int ii=0;ii<=ng+1;ii++) {
      double yi=deltay[ii]+yoffset;
      if (fabs(yi)>ymax) {
	setstatus(2);
	wt=0.;
	return wt;
      }
      
      pt=sqrt(pta[ii][0]*pta[ii][0]+pta[ii][1]*pta[ii][1]);
      
      p[0]=pt*cosh(yi); // the energy component
      p[1]=pta[ii][0];
      p[2]=pta[ii][1];
      p[3]=pt*sinh(yi); // the z-component

      // {double t=exp(-yi);
      // xa+=pt/sqrts*t;
      // xb+=pt/sqrts/t;
      // }
      xa+=(p[0]-p[3])/sqrts;
      xb+=(p[0]+p[3])/sqrts;
      AddParton(p);
    }
    
  }
  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);
  
#if 0
  // Check of energy and momentum conservation
  std::vector<CLHEP::HepLorentzVector>::iterator itemissions;
  HepLorentzVector outsum(0.,0.,0.,0.);
  for (itemissions=partonmomenta.begin();itemissions!=partonmomenta.end();itemissions++) {
    outsum+=*itemissions;
  }
  outsum+=pd1+pd2;
  cout << outsum <<" "<<pa+pb<<" "<<outsum-pa-pb<<endl;
#endif
  return wt;
}

double CWMultijet::MEt2(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb,bool checkmomenta,fastjet::ClusterSequence * ev)
//double CWMultijet::MEt2(std::vector<CLHEP::HepLorentzVector> partons)
{
  double wt(0.);
    // std::cout<<"Momenta in MEt2: \n";
    // std::cout<<pb<<"   "<<pa<<std::endl;
    // std::cout<<alepton.p<<"   "<<lepton.p<<std::endl;
    // std::cout<<partons.back()<<"   "<<partons.front()<<std::endl;

  if (aptype==21&&bptype!=21) {// we know they are not both gluons
    if (bptype > 0)
      wt=jMWqg(partons.back(),alepton.p,lepton.p,pb,partons.front(),pa);
    else
      wt=jMWqbarg(partons.back(),alepton.p,lepton.p,pb,partons.front(),pa);
  }
  else if (bptype==21&&aptype!=21) { // ----- || -----
    if (aptype > 0) 
      wt=jMWqg(partons.front(),alepton.p,lepton.p,pa,partons.back(),pb);
    else
      wt=jMWqbarg(partons.front(),alepton.p,lepton.p,pa,partons.back(),pb);
  }
  else { // they are both quark
    if (wc==true) {// emission off b, i.e. b is first current
      if (bptype>0) {
	if (aptype>0)
	  wt=jMWqQ(partons.back(),alepton.p,lepton.p,pb,partons.front(),pa);
	else
	  wt=jMWqQbar(partons.back(),alepton.p,lepton.p,pb,partons.front(),pa);
      }
      else {
	if (aptype>0)
	  wt=jMWqbarQ(partons.back(),alepton.p,lepton.p,pb,partons.front(),pa);
	else
	  wt=jMWqbarQbar(partons.back(),alepton.p,lepton.p,pb,partons.front(),pa);
      }
    }
    else {// emission off a, i.e. a is first current
      if (aptype > 0) {
	if (bptype > 0)
	  wt=jMWqQ(partons.front(),alepton.p,lepton.p,pa,partons.back(),pb);
	else
	  wt=jMWqQbar(partons.front(),alepton.p,lepton.p,pa,partons.back(),pb);
      }
      else {  // a is anti-quark
	if (bptype > 0)
	  wt=jMWqbarQ(partons.front(),alepton.p,lepton.p,pa,partons.back(),pb);
	else
	  wt=jMWqbarQbar(partons.front(),alepton.p,lepton.p,pa,partons.back(),pb);
      }
    }
  }

  // Multiply by couplings etc.
  COM PWs=PW((pd1+pd2).m2());
  double PropFactor=(PWs*conj(PWs)).real();
  wt*=PropFactor/(4.*8.);    // (Nc^2-1)*numb_in_hel_states
  // double gs(1.21771578),gw(0.653232911),couplings(gs*gs*gs*gs*gw*gw*gw*gw/4);  
  static const double gw(0.653232911);  
  double gs2=0.;
  
  gs2=4.*M_PI*Halphas(renscale);
  
  double couplings=pow(gs2,2)*gw*gw*gw*gw/4.;
  wt*=couplings;
  if (flags.logcorrect==1)
    wt*=(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI));
    
  // Now multiply by the number of Lipatov vertices and couplings
  CLHEP::HepLorentzVector qi, qip1,g;
  if (wc==true)
    qi=pb-partons.back()-alepton.p-lepton.p;
  else
    qi=pb-partons.back();
  for (unsigned gcount=partons.size()-2;gcount>0;gcount--) {
    g=partons.at(gcount);
    qip1=qi-g;
#if SmartThingCL
    if (checkmomenta) {
      bool jetmember(false);
      // Is the parton part of the jet of lowest rapidity?
      vector<fastjet::PseudoJet> constituents = (*ev).constituents(jets[0]);
      for (unsigned int ll=0;ll<constituents.size();ll++)
	if (gcount==unsigned(constituents[ll].user_index())) 
	  jetmember=true;
      // Is the parton part of the jet of highest rapidity?
      constituents = (*ev).constituents(jets[jets.size()-1]);      
      for (unsigned int ll=0;ll<constituents.size();ll++)
	if (gcount==unsigned(constituents[ll].user_index())) 
	  jetmember=true;
      if (jetmember)
	wt*=C2Lipatovots(qi,qip1,pa,pb)*3.*gs2;   // *Ca
      else
	wt*=C2Lipatovots(qi,qip1,pa,pb,partons.front(),partons.back())*3.*gs2;   // *Ca
    } else
      wt*=C2Lipatovots(qi,qip1,pa,pb,partons.front(),partons.back())*3.*gs2;   // *Ca
#else
    wt*=C2Lipatovots(qi,qip1,pa,pb)*3.*gs2;   // *Ca
#endif
    if (flags.logcorrect==1) {// include extra logs
      wt*=(1.-Halphas(renscale)*beta0*log(g.perp()/renscale)/(2.*M_PI));
    }
    qi=qip1;
  }

  return wt;
}

double CWMultijet::VirtualCorrections()
// oh - and to all orders
// well, at least the log piece
{
  double vt(0.0);
  
  HepLorentzVector tiv,p1,p2;
  double y1,y2;
  if (wc==false) //emission off a
    tiv=pa-partonmomenta.front()-lepton.p-alepton.p;
  else // emission off b
    tiv=pa-partonmomenta.front();
  
  std::vector<CLHEP::HepLorentzVector>::iterator itpartonmomenta;
  
  itpartonmomenta=partonmomenta.begin();
  p1=*(itpartonmomenta);
  y1=p1.rapidity();
  itpartonmomenta++;
  for (;itpartonmomenta!=partonmomenta.end();itpartonmomenta++) {
    p2=*(itpartonmomenta);
    y2=p2.rapidity();
    vt+=omega0(tiv.perp2(),lambda*lambda)*(y2-y1);
    tiv-=p2;
    p1=p2;
    y1=y2;
  }

  return exp(vt);
}


double CWMultijet::weight()
{
  //  double temp1,temp2,pttemp[2],ptHvertex[2],r1,pt1SQ,pt2SQ,pt1,pt2,q1SQ,q2SQ,t1,t2;
  double ufa(0.),ufb(0.),ura(0.),urb(0.); // factorisation scale
  
  //  double ptH,mHperp;
  //  double Smatch,shat2,HElimit;
  double shat2;//,wtmatch;

#if printoutput
  cout << "--------------------\n";
#endif

  setstatus(0);
  wt=1;

#if dijetLO==1
  wt*=Generate2jphsppoint();
#elif trijetLO==1
  wt*=Generate3jphsppoint();
#elif fourjetLO==1
  wt*=Generate4jphsppoint();
#else
  wt*=GeneratePHSPpoint();
#endif

  if (getstatus()!=0)
    return 0.;

  shat2=xa*xa*xb*xb*pow(getcmsenergy(),4);

  if (xa>1.0||xb>1.0) {
    setstatus(5);
    wt=0.;
    return wt;
  } 

  // Cluster into jets already at this stage, so MEt2 knows whether a
  // parton is inside an extremal jet
  // Count how many hard jets we have
  std::vector<CLHEP::HepLorentzVector>::iterator itpm,ijpm;
  std::vector<fastjet::PseudoJet>::iterator fijpm;
  
  std::vector<fastjet::PseudoJet> avec;
  
  unsigned int i(0);
  for (itpm=partonmomenta.begin();itpm!=partonmomenta.end();itpm++) {
    avec.push_back(*(itpm));
    avec.at(i).set_user_index(i); // keep track of the partons. These are now numbered 
    // according to their rapidity (increasing)
    i++;
  }
  
  /** fastjet find clusters */
  fastjet::ClusterSequence ev(avec, flags.jet_def);
  
  /** Retrieve the final state jets from fastjet Sorted by Rapidity*/
  jets = sorted_by_rapidity(ev.inclusive_jets(flags.jetptmin));
  
  if (jets.size()>=2) {
    bool jetmember1(false);
    // Check to see if the extremal partons are included in the extremal jets
    std::vector<fastjet::PseudoJet> constituents = (ev).constituents(jets[0]);
    for (unsigned int ll=0;ll<constituents.size();ll++)
      if (0==unsigned(constituents[ll].user_index())) 
	jetmember1=true;
    
    bool jetmember2(false);
    // Check to see if the extremal partons are included in the extremal jets
    constituents = (ev).constituents(jets[jets.size()-1]);
    for (unsigned int ll=0;ll<constituents.size();ll++)
      if ((partonmomenta.size()-1)==unsigned(constituents[ll].user_index())) 
	jetmember2=true;
    if (!(jetmember2&&jetmember1)) {
      wt=0;
      setstatus(18);
      return wt;
    }
  } 
  else { // we didn't even have two jets - reject outright!
    wt=0;
    setstatus(17);
    return wt;
  }
  
  // std::vector<fastjet::PseudoJet> ptjets = sorted_by_pt(ev.inclusive_jets(flags.jetptmin));

  std::vector<fastjet::PseudoJet> ptjets = sorted_by_pt(jets);  
#if calcscaleunc
  renscale=ptjets.at(0).perp()*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=flags.scale*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=Ht/2.*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=geometricmean(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=ptjets.at(0).perp()*flags.scalefactor;
#else
  if (flags.scalesetting==0) //max pt(jets)
    renscale=ptjets.at(0).perp()*flags.scalefactor;
  else if (flags.scalesetting==1) //fixed scale
    renscale=flags.scale;
  else if (flags.scalesetting==2) // Ht/2
    renscale=Ht/2.*flags.scalefactor;
  else if (flags.scalesetting==3) //geometric mean of pt(jets)
    renscale=geometricmean(&ptjets)*flags.scalefactor;
#endif
  // std::cout<<"------------ FKL ----------------\n";
  // std::cout<<"final scale: "<<renscale<<"  "<<Ht/2.<<"\n";
  // double tempperp(0.);
  // for (int kk; kk<partonmomenta.size(); kk++) {
  //   tempperp+=partonmomenta.at(kk).perp();
  //   std::cout<<partonmomenta.at(kk).perp()<<"  ";
  // }
  // std::cout<<"\n"<<(pd1+pd2).perp()<<"\n";
  // tempperp+=(pd1+pd2).perp();
  // std::cout<<"Total perps/2: "<<tempperp/2.<<"\n";
  ufa=ura=ufb=urb=renscale;
    
    // Implement LH interface:
    // Choose incoming parton species and assign pdfs accordingly
    wt*=GenerateFKLIncomingParticles(xa,ufa,xb,ufb);
    if (status!=0) { // if we failed in picking a sensible incoming state
      wt=0.;
      return wt;
    }

  // Which parton should the W be emitted from? Only a problem if both
  // incomding quarks are same sign. In that case, choose the closest (in
  // rapidity) to the rapidity average of the e,nu.  What sign is the W?
  // Choose according to the size of the incoming quark pdfs (maybe times the
  // coupling to the W?  

  // Do all this in one go!?
  // parton flux has nothing to do with the prob. of emitting a W. Once the pdfs are chosen, 
  // just choose the emitting quark with probabibility depending on proximity to W

#if 1
  if (aptype==21) { // emit off the b
    wc=true;
  }
  else if (bptype==21) // we know both can't be gluons (this is rejected as non-FKL state)
    wc=false;
  else { // if neither a nor b is a gluon
    HepLorentzVector pW=pd1+pd2;
    double da,db;
    da=exp(-fabs(partonmomenta.front().rapidity()-pW.rapidity()));
    db=exp(-fabs(partonmomenta.back().rapidity()-pW.rapidity()));
    double sum(da+db);
    double r1;
    r1=sum*ran.flat();
    if (r1<da) { // emit the W off particle a
      wc=false;
      wt*=sum/da;
    } else { // emit the W off particle b
      wc=true;
      wt*=sum/db;
    }
  }
#else
  wc=true; // emit off b, forward moving parton
  //  wc=false; // emit off a, backward moving parton
#endif
  // std::cout << aptype <<std::endl;
  if (!wc) { //emit off the a
    if (aptype==1||aptype==-2||aptype==3||aptype==-4||aptype==5) { // if particle type a is negatively charged
      Wboson.PID=-24;//W- Boson as intermediate particle
      Wboson.COL1=0;
      Wboson.COL2=0;
      Wboson.MOT1=1;
      Wboson.MOT2=2;
      Wboson.ISTUP=2;
      Wboson.p=pd1+pd2;

      lepton.PID=11; // then we emit an electron (or whatever else is allowed)
      lepton.COL1=0;
      lepton.COL2=0;
      lepton.MOT1=3; 
      lepton.MOT2=3;
      lepton.ISTUP=1;
      lepton.p=pd1;
      
      alepton.PID=-12; // anti-electron-neutrino
      alepton.COL1=0;
      alepton.COL2=0;
      alepton.MOT1=3; 
      alepton.MOT2=3;
      alepton.ISTUP=1;
      alepton.p=pd2;
      
    } else { // if particle type a is positively charged
      Wboson.PID=24; //W+ Boson as intermediate particle
      Wboson.COL1=0;
      Wboson.COL2=0;
      Wboson.MOT1=1; 
      Wboson.MOT2=2;
      Wboson.ISTUP=2;
      Wboson.p=pd1+pd2;
      
      lepton.PID=12; // then we emit an electron-neutrino
      lepton.COL1=0;
      lepton.COL2=0;
      lepton.MOT1=3; 
      lepton.MOT2=3;
      lepton.ISTUP=1;
      lepton.p=pd1;
      
      alepton.PID=-11; // anti-electron
      alepton.COL1=0;
      alepton.COL2=0;
      alepton.MOT1=3; 
      alepton.MOT2=3;
      alepton.ISTUP=1;
      alepton.p=pd2;
    }
    double rflav = ran.flat();
    switch(aptype){// change quark flavouring, top quarks not included yet
    case 1:
      if((rflav*(vud*vud+vcd*vcd))<vud*vud){ 
	aoutptype = 2;
      }
      else {
	aoutptype = 4;
      }
      break;
    case 2:
      if((rflav*(vud*vud+vus*vus+vub*vub))<vud*vud){ 
	aoutptype = 1;
      }
      else if((rflav*(vud*vud+vus*vus+vub*vub))<(vud*vud+vus*vus))  {
	aoutptype = 3;
      }
      else {
	aoutptype = 5;
      }
      break;
    case 3:
      if((rflav*(vus*vus+vcs*vcs))<vus*vus){ 
	aoutptype = 2;
      }
      else {
	aoutptype = 4;
      }
      break;
    case 4:
      if((rflav*(vcd*vcd+vcs*vcs+vcb*vcb))<vcd*vcd){ 
	aoutptype = 1;
      }
      else if((rflav*(vcd*vcd+vcs*vcs+vcb*vcb))<(vcd*vcd+vcs*vcs))  {
	aoutptype = 3;
      }
      else {
	aoutptype = 5;
      }
      break;
    case 5:
      if((rflav*(vub*vub+vcb*vcb))<vub*vub){ 
	aoutptype = 2;
      }
      else {
	aoutptype = 4;
      }
      break;
    case 6:
      aoutptype = 5;
      break;
    case -1:
      if((rflav*(vud*vud+vcd*vcd))<vud*vud){ 
	aoutptype = -2;
      }
      else {
	aoutptype = -4;
      }
      break;
    case -2:
      if((rflav*(vud*vud+vus*vus+vub*vub))<vud*vud){ 
	aoutptype = -1;
      }
      else if((rflav*(vud*vud+vus*vus+vub*vub))<(vud*vud+vus*vus))  {
	aoutptype = -3;
      }
      else {
	aoutptype = -5;
      }
      break;
    case -3:
      if((rflav*(vus*vus+vcs*vcs))<vus*vus){ 
	   aoutptype = -2;
      }
      else {
	aoutptype = -4;
      }
      break;
    case -4:
      if((rflav*(vcd*vcd+vcs*vcs+vcb*vcb))<vcd*vcd){ 
	aoutptype = -1;
      }
      else if((rflav*(vcd*vcd+vcs*vcs+vcb*vcb))<(vcd*vcd+vcs*vcs))  {
	aoutptype = -3;
      }
      else {
	aoutptype = -5;
      }
      break;
    case -5:
      if((rflav*(vub*vub+vcb*vcb))<vub*vub){ 
	aoutptype = -2;
      }
      else {
	aoutptype = -4;
      }
      break;
    case -6:
      aoutptype = -5;
      break;
    default: cout << "unidentified incoming particle, please check" << aptype << endl;
    }
    boutptype = bptype;

  } else { // emit off particle b
    if (bptype==1||bptype==-2||bptype==3||bptype==-4||bptype==5) { // if particle type b is negatively charged
      Wboson.PID=-24;//W- Boson as intermediate particle
      Wboson.COL1=0;
      Wboson.COL2=0;
      Wboson.MOT1=1;
      Wboson.MOT2=2;
      Wboson.ISTUP=2;
      Wboson.p=pd1+pd2;

      lepton.PID=11; // then we emit an electron (or whatever else is allowed)
      lepton.COL1=0;
      lepton.COL2=0;
      lepton.MOT1=3; 
      lepton.MOT2=3;
      lepton.ISTUP=1;
      lepton.p=pd1;
      
      alepton.PID=-12; // anti-electron-neutrino
      alepton.COL1=0;
      alepton.COL2=0;
      alepton.MOT1=3; 
      alepton.MOT2=3;
      alepton.ISTUP=1;
      alepton.p=pd2;
    } else { // if particle type b is positively charged
      Wboson.PID=24; //W+ Boson as intermediate particle
      Wboson.COL1=0;
      Wboson.COL2=0;
      Wboson.MOT1=1; 
      Wboson.MOT2=2;
      Wboson.ISTUP=2;
      Wboson.p=pd1+pd2;

      lepton.PID=12; // then we emit an electron-neutrino (or whatever else is allowed)
      lepton.COL1=0;
      lepton.COL2=0;
      lepton.MOT1=3; 
      lepton.MOT2=3;
      lepton.ISTUP=1;
      lepton.p=pd1;
      
      alepton.PID=-11; // anti-electron
      alepton.COL1=0;
      alepton.COL2=0;
      alepton.MOT1=3; 
      alepton.MOT2=3;
      alepton.ISTUP=1;
      alepton.p=pd2;
    }
    double rflav = ran.flat();
    switch(bptype){// change quark flavouring, includes mixing and excludes the top quark
    case 1:
      if((rflav*(vud*vud+vcd*vcd))<vud*vud){ 
	boutptype = 2;
      }
      else {
	boutptype = 4;
      }
      break;
    case 2:
      if((rflav*(vud*vud+vus*vus+vub*vub))<vud*vud){ 
	boutptype = 1;
      }
      else if((rflav*(vud*vud+vus*vus+vub*vub))<(vud*vud+vus*vus))  {
	boutptype = 3;
      }
      else {
	boutptype = 5;
	      }
      break;
    case 3:
      if((rflav*(vus*vus+vcs*vcs))<vus*vus){ 
	boutptype = 2;
      }
      else {
	boutptype = 4;
      }
      break;
    case 4:
      if((rflav*(vcd*vcd+vcs*vcs+vcb*vcb))<vcd*vcd){ 
	boutptype = 1;
      }
      else if((rflav*(vcd*vcd+vcs*vcs+vcb*vcb))<(vcd*vcd+vcs*vcs))  {
	boutptype = 3;
      }
      else {
	boutptype = 5;
      }
      break;
    case 5:
      if((rflav*(vub*vub+vcb*vcb))<vub*vub){ 
	boutptype = 2;
      }
      else {
	boutptype = 4;
	
      }
      break;
    case 6:
      boutptype = 5;
      break;
    case -1:
      if((rflav*(vud*vud+vcd*vcd))<vud*vud){ 
	boutptype = -2;
      }
      else {
	boutptype = -4;
      }
      break;
    case -2:
      if((rflav*(vud*vud+vus*vus+vub*vub))<vud*vud){ 
	boutptype = -1;
      }
      else if((rflav*(vud*vud+vus*vus+vub*vub))<(vud*vud+vus*vus))  {
	boutptype = -3;
      }
      else {
	boutptype = -5;
      }
      break;
    case -3:
      if((rflav*(vus*vus+vcs*vcs))<vus*vus){ 
	boutptype = -2;
      }
      else {
	boutptype = -4;
      }
      break;
    case -4:
      if((rflav*(vcd*vcd+vcs*vcs+vcb*vcb))<vcd*vcd){ 
	boutptype = -1;
      }
      else if((rflav*(vcd*vcd+vcs*vcs+vcb*vcb))<(vcd*vcd+vcs*vcs))  {
	boutptype = -3;
      }
      else {
	boutptype = -5;
      }
      break;
    case -5:
      if((rflav*(vub*vub+vcb*vcb))<vub*vub){ 
	boutptype = -2;
      }
      else {
	boutptype = -4;
      }
      break;
    case -6:
      boutptype = -5;
      break;
    default: cout << "wrong incoming particle, please check" << aptype << endl;
    }
    aoutptype = aptype;
    
  }    
  // wtmatch=wt;
  if (flags.LHoutput)
    SetupLHInformation();

  //////////////////////////////////////////////////
  // Choose lambda based on q_perps               //
  // reject phase space point if k_perp_i<lambda  //
  //////////////////////////////////////////////////
#if 0
  // choose lambda
  {
    HepLorentzVector tiv,g;
    if (wc==false) //emission off a
      tiv=pb-partonmomenta.back();
    else // emission off b
      tiv=pb-partonmomenta.back()-lepton.p-alepton.p;
    
    double minqp2=tiv.perp2();
    for (int gcount=partonmomenta.size()-2;gcount>0;gcount--) {
      g=partonmomenta.at(gcount);
      tiv=tiv-g; // the t-channel momentum AFTER emission
      double temp=tiv.perp2();
      if (temp<minqp2)
	minqp2=temp;
    }
    {
      double lambda_temp;
      lambda=.5;
      //      lambda=.25;
      lambda_temp=sqrt(minqp2)/20.;
      
      if (lambda_temp<lambda) {
	lambda=lambda_temp;
	// cout << "lambda : "<<lambda<<endl;
      }
    }
  }
#endif

  // Reject if k_perp_i<lambda; the extremal partons have to be yet harder
  // and are thus not checked here
  for (int gcount=partonmomenta.size()-2;gcount>0;gcount--) {
    if(partonmomenta.at(gcount).perp()<ccut) {
      setstatus(9);
      wt=0.;
      return wt;
    }
  }
  ////////////////////////////////////////////////////////////
  // Calculate the factors making up the cross section      //
  ////////////////////////////////////////////////////////////


#if nottreeonly    
    double MEsq=MEt2(partonmomenta,pa,pb,true,&ev); // true: use jet-member test
#else
#if 1    // 2j tree
    double MEsq=MG(partonmomenta,pa,pb,alepton.p,lepton.p);
    //    std::cout<<aptype<<"  "<<bptype<<"  "<<MEsq<<"  "<<MEt2(partonmomenta,pa,pb,true,&ev)<<"\n";
#else    // 3j tree  - remember to fix ng!
    double MEsq=MG3(partonmomenta,pa,pb,alepton.p,lepton.p);
#endif
#endif

#if calcscaleunc
    std::vector<double> MEs;
    {
      double orgscale=renscale;
      for (unsigned int ii=0;ii<scales.size();ii++) {
	renscale=scales.at(ii);
	MEs.push_back(MEt2(partonmomenta,pa,pb,true,&ev));
      }
      renscale=orgscale; // put things back to normal
    }
#endif

#if MirrorTest
  std::cout<<"--------------------\n";
  std::cout<<"ng,aptype,bptype : "<<partonmomenta.size()<<","<<aptype<<","<<bptype<<std::endl;
  std::cout<<MEsq<<std::endl;

  // Mirror momenta:
  {
    std::vector<CLHEP::HepLorentzVector> partonmomenta2;
    CLHEP::HepLorentzVector pa2,pb2,*t;
    int typesave;
    wc=!wc;
    typesave=aptype;
    aptype=bptype;
    bptype=typesave;
    pa2.set(pb.x(),pb.y(),-pb.z(),pb.e());
    pb2.set(pa.x(),pa.y(),-pa.z(),pa.e());
  
    std::vector<CLHEP::HepLorentzVector>::iterator itemissions;
    HepLorentzVector g,vtemp;
    for (int gcount=partonmomenta.size()-1;gcount>=0;gcount--) {
      g=partonmomenta.at(gcount);
      vtemp.set(g.x(),g.y(),-g.z(),g.e());
      partonmomenta2.push_back(vtemp);
    }
    // the leptons too:
    t=&lepton.p;
    lepton.p.set(t->x(),t->y(),-(t->z()),t->e());
    t=&alepton.p;
    alepton.p.set(t->x(),t->y(),-(t->z()),t->e());


    std::cout<<MEt2(partonmomenta2,pa2,pb2)<<std::endl;
  
    // Put things back to normal again:
    wc=!wc;
    typesave=aptype;
    aptype=bptype;
    bptype=typesave;
  }
  
#endif
  //  double gs2t=4.*M_PI*AlphaS(flags.scale);

  wt*=MEsq/(256.*shat2*pow(M_PI,5));

  // if (aptype<21 && bptype<21 && aptype<0 && bptype<0) {
  // std::cout<<"Compare: "<<JEWEL<<"  "<<"  "<<Tree3jenu(0.5,0.5,pa,pb,partonmomenta.front(),partonmomenta.at(1),partonmomenta.back(),alepton.p,lepton.p)*gs2t*gs2t*gs2t<<"  "<<MG3(partonmomenta,pa,pb,alepton.p,lepton.p)<<std::endl;  
  // std::cout<<"Compare: "<<JEWEL<<"  "<<aptype<<"  "<<bptype<<"  "<<wc<<"  "<<MG3(partonmomenta,pa,pb,alepton.p,lepton.p)<<std::endl;  
  // }
#if dijetLO==0&&trijetLO==0&&fourjetLO==0
#if calcscaleunc
  std::vector<double> virtuals;
  {
    double orgscale=renscale;
    for (unsigned int ii=0;ii<scales.size();ii++) {
      renscale=scales.at(ii);
      virtuals.push_back(VirtualCorrections());
    }
    renscale=orgscale;
  }
#endif
  double vt=VirtualCorrections();
  wt*=vt;
#endif
  //////////////////////////////////////////////////

  if (flags.match) {
      
    int njets=int(jets.size());
    if (njets>1&&njets<5) { // if the right number of jets are detected, i.e. not too many

      // cluster all partons into the identified jets
      std::vector<CLHEP::HepLorentzVector> jetlist,fastjetlist,nojetlist;
      CLHEP::HepLorentzVector zerovec;
      zerovec.set(0.,0.,0.,0.);

      for (fijpm=jets.begin();fijpm!=jets.end();fijpm++) {
	jetlist.push_back(zerovec);
	CLHEP::HepLorentzVector tempfast;
	tempfast.set(fijpm->px(),fijpm->py(),fijpm->pz(),fijpm->e());
	fastjetlist.push_back(tempfast);
      }

      // Add all the partons to either jet-list or no-jet-list

      for (unsigned int ii=0;ii<partonmomenta.size();ii++) {
	bool included(false);
	// check to see if ii is included in one of the already existing jets
	for (unsigned int jj=0;jj<jets.size();jj++) {
	  // get constituents:
	  vector<fastjet::PseudoJet> constituents = ev.constituents(jets[jj]);
	  for (unsigned int ll=0;ll<constituents.size();ll++) {
	    if (ii==unsigned(constituents[ll].user_index())) {
#if printout
	      CLHEP::HepLorentzVector tempfast;
	      tempfast.set(constituents[ll].px(),constituents[ll].py(),constituents[ll].pz(),constituents[ll].e());
	      std::cout << "Found one vector in jet : "<<tempfast<<std::endl;
#endif
	      // add to jet vector
	      jetlist[jj]+=partonmomenta[ii];
	      included=true;
	    }
	  }
	}
	if (!included) { // add to list of non-included particles
	  nojetlist.push_back(partonmomenta[ii]);
	}
      }

#if printout
      std::cout<< "Fastjet Jets: "<<jets.size()<<std::endl;
      std::cout<< "partonmomenta size : "<<partonmomenta.size()<<std::endl;
      std::cout<< "nojetlist size before: "<<nojetlist.size()<<std::endl;
#endif

      // Now cluster the rest according to min inv. mass compared to already existing jets
      while (nojetlist.size()>0) {
	// cycle through the list of unclustered partons and the list of jets to find 
	// the pair of minimum invariant mass
	double smin=1e10;
	unsigned int ii_save(10),jj_save(10);
	for (unsigned int ii=0;ii<nojetlist.size();ii++) {
	  for (unsigned int jj=0;jj<jetlist.size();jj++) {
	    double temp;
	    temp=jetlist[jj].invariantMass2(nojetlist[ii]);
	    if (temp<smin) {
	      smin=temp;
	      ii_save=ii;
	      jj_save=jj;
	    }
	  }
	} // now we have found the smallest smin, and saved info in ii_save, jj_save;
	jetlist.at(jj_save)+=nojetlist[ii_save];
	nojetlist.erase(nojetlist.begin()+ii_save);
      }
      
#if printout
      std::cout<< "nojetlist size after: "<<nojetlist.size()<<std::endl;
#endif
      // for (itpm=partonmomenta.begin();itpm!=partonmomenta.end();itpm++) {
      // 	double smin=10000.;
      // 	int jetplace=0;
      // 	int jettemp=0;
      // 	for (ijpm=fastjetlist.begin();ijpm!=fastjetlist.end();ijpm++) {
      // 	  double stemp = (*(ijpm)).invariantMass2(*(itpm));
      // 	  if (stemp<smin) {
      // 	    smin=stemp;
      // 	    jetplace=jettemp;
      // 	  }
      // 	  jettemp=jettemp+1;
      // 	}
      // 	jetlist[jetplace]+=*(itpm);
      // }

#if printout
      std::cout<<"FastJet versus OurJet"<<std::endl;
      std::cout<<"Length: "<<fastjetlist.size()<<"  "<<jetlist.size()<<std::endl;
      std::cout<<"Number of partons: "<< partonmomenta.size()<<std::endl;
#endif
      double xE,xz;
      xE=(alepton.p).e()+(lepton.p).e();
      xz=(alepton.p).z()+(lepton.p).z();
      for (int icount=0;icount<njets;icount++) {
	CLHEP::HepLorentzVector temp=fastjetlist.at(icount);
	CLHEP::HepLorentzVector diff=temp-jetlist.at(icount);
#if printout
	double difflength=diff.x()*diff.x()+diff.y()*diff.y()+diff.z()*diff.z();
	double templength=temp.x()*temp.x()+temp.y()*temp.y()+temp.z()*temp.z();
	std::cout<<"Diff Length "<<icount<<": "<<difflength<<" ratio= "<<difflength/templength<<" "<<(ev.constituents(jets[icount])).size()<<" "<<" "<<jets[icount].m2()<<" "<<jetlist.at(icount).m2()<<std::endl;
#endif
	temp=jetlist.at(icount);
#if newreshuffle
	// keep the rapidities fixed 
	double mperp=temp.perp();
	double rapidity=jets.at(icount).rapidity();
	jetlist.at(icount).set(temp.px(),temp.py(),mperp*sinh(rapidity),mperp*cosh(rapidity));
#else
	//  Put our jets onshell
	jetlist.at(icount).set(temp.x(),temp.y(),temp.z(),sqrt(temp.x()*temp.x()+temp.y()*temp.y()+temp.z()*temp.z()));
#endif
#if printout
	std::cout << "Our Jet "<<icount<<" : "<<jetlist.at(icount)<<std::endl;
	std::cout << "Fastjet Jet "<<icount<<" : "<<fastjetlist.at(icount)<<std::endl;
#endif
	//  Now we calculate new incoming particles
	xE+=(jetlist.at(icount)).e();
	xz+=(jetlist.at(icount)).z();
	  
      }
      CLHEP::HepLorentzVector npa,npb;
      npa.set(0,0,-(xE-xz)/2.,(xE-xz)/2.);
      npb.set(0,0,(xE+xz)/2.,(xE+xz)/2.);
#if printout
      std::cout<<"'pa's:"<<pa<<"  "<<npa<<std::endl;
      std::cout<<"'pb's:"<<pb<<"  "<<npb<<std::endl;
      
      // std::cout<<"New multiplier: "<<MG(jetlist,npa,npb,alepton.p,lepton.p)/MEt2(jetlist,npa,npb)<<" "<<MG(jetlist,npa,npb,alepton.p,lepton.p)<<" "<< MEt2(jetlist,npa,npb)<<" "<<JEWEL<<std::endl;
#endif
      
      switch (jetlist.size()) {
      case 2 : 
	matchreweight=MG(jetlist,npa,npb,alepton.p,lepton.p)/MEt2(jetlist,npa,npb,false,&ev); //ev never used
	wt*=matchreweight;
	//	std::cout<<"case 2 match: "<<MG(jetlist,npa,npb,alepton.p,lepton.p)<<"\n";
	break;
      case 3 : 
	matchreweight=MG3(jetlist,npa,npb,alepton.p,lepton.p)/MEt2(jetlist,npa,npb,false,&ev); //ev never used
      	wt*=matchreweight;
      	break;
      case 4 : 
	matchreweight=MG4(jetlist,npa,npb,alepton.p,lepton.p)/MEt2(jetlist,npa,npb,false,&ev); //ev never used
      	wt*=matchreweight;
      	break;
      default:
	break;
      }
    }
  }
  double pdf1=pdfta*pdftb;
  wt*=pdf1;
#if calcscaleunc
  std::vector<double> pdfs;
  {
    double orgscale=renscale;
    for (unsigned int ii=0;ii<scales.size();ii++) {
      renscale=scales.at(ii);
      pdfs.push_back(pdfpt(0,xa,renscale,aptype)*pdfpt(1,xb,renscale,bptype));
    }
    renscale=orgscale;
  }
#endif
  // GeV^(-2)->micro barn
  wt*=389.379292;
  wt*=1E6; // micro barn-> pico barn

  // if (isnan(wt)) {
  //   setstatus(6);
  // }

  // if (wt<0) {
  //   setstatus(13);
  // }

#if 1
  //if (fabs(wt)>5.18E21) { // cuts off at 50
  if (fabs(wt)>4.85E8) { // cuts off at 20
    //  if (fabs(wt)>6.56E7) { // cuts off at 18
  // if (fabs(wt)>8.89E6) { // cuts off at 16
  // if (fabs(wt)>3.27E6) { // cuts off at 15
    setstatus(7);
  }
#endif  

#if dijetLO==0&&trijetLO==0&&fourjetLO==0
#if calcscaleunc
  for (unsigned int ii=0;ii<scales.size();ii++) {
    double temp=wt*MEs.at(ii)/MEsq*pdfs.at(ii)/pdf1*virtuals.at(ii)/vt;
    weights.push_back(temp);
  }
#endif
#else
#if calcscaleunc
  for (unsigned int ii=0;ii<scales.size();ii++) {
    double temp=wt*MEs.at(ii)/MEsq*pdfs.at(ii)/pdf1;
    weights.push_back(temp);
  }
#endif
#endif
  return wt;
}

double CWMultijet::Generate2jphsppoint()
{
  double p[4];
  double sqrts=2.*flags.Ebeam;
#if 1       //  1: Generate as normal, 0: MG Standalone fixed point

#if NewPhSp
  double y,ysum,pt,r1,temp1,temp2,phi;
  double tscale,kt[2];
  double ptpar(40.);

  double pta[2][2];
  double deltay[2];
  // double phia[2];

  lambda=clambda; // trouble is we don't know the value of lambda until after the whole generation

  double wt=1.;
  sqrts=2.*flags.Ebeam;

  // The parton of lowest rapidity:
#if 1
  // Generate the transverse momentum of parton a
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  wt*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
//   wtmatch*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;
  // renscale2=pt*pt;
  //  pt1=pt;
  Ht+=pt;
#else
  {
    double ptSQ;
    r1=ran.flat();
    ptSQ=ptSQgen/(exp(r1*log(1.+ptSQgen))-1.);
    wt*=2.*M_PI*(ptSQ)*(1.+(ptSQ)/ptSQgen)*log(1.+ptSQgen)/2.;
    pt=sqrt(ptSQ);
    // renscale2=ptSQ;
  }
#endif

  phi=2.*M_PI*ran.flat(); // Random angle in x,y-plane

  y=0.;
  ysum=0.;
  // phia[0]=phi;
  
  p[1]=pt*cos(phi);
  p[2]=pt*sin(phi);
  pta[0][0]=pt*cos(phi);
  pta[0][1]=pt*sin(phi);
  deltay[0]=0.;

  kt[0]=-p[1];
  kt[1]=-p[2];


  tscale=kt[0]*kt[0]+kt[1]*kt[1];


  //  We will use omega0n=-1.;
  double a(1.3);
  y=-a*log(1.-ran.flat());
  wt*=a*exp(y/a);

  //  y=-log(1.-ran.flat());
  //  wt*=1./exp(-y); // take out the exponential factor which was included 
                  // in the generation of y - we want this in the 
                  // FKL evaluation
  ysum+=y;
  deltay[1]=ysum;

  // Generate the transverse momentum of parton b
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  wt*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  //   wtmatch*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;
  
  phi=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  Ht+=pt;
  pta[1][0]=pt*cos(phi);
  pta[1][1]=pt*sin(phi);
  
  kt[0]+=-pta[1][0];
  kt[1]+=-pta[1][1];
  
  // Now we know the transverse momentum of the W, given by the array kt
  double pW[4];
  // Choose its mass according to Breit-Wigner
  r1=ran.flat();
  double sW=flags.MW*(flags.MW + flags.GammaW*tan((M_PI*r1)/2. + (-1. + r1)*atan(flags.MW/flags.GammaW)));
  // Multiply by derivate (d sap)/(d r)
  {
    static double temp=atan(flags.MW/flags.GammaW);
    wt*=(flags.GammaW*flags.MW*(M_PI+2.*temp))/(1.+cos(M_PI*r1+2.*(-1.+r1)*temp));
  }  
  pW[1]=kt[0];
  pW[2]=kt[1];
  double EtW=sqrt(pW[1]*pW[1]+pW[2]*pW[2]+sW);
  
  double yW;
#if 1
  // Generate a yW Gaussian distributed around 0
  {
    double lninvr1,r1,r2,temp,a;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    a=0.7;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    yW=temp;
    wt=wt*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
  }
#else
  // choose a rapidity
  yW=-5.+10.*ran.flat();
  wt*=10;
#endif
  pW[0]=cosh(yW)*EtW;
  pW[3]=sinh(yW)*EtW;
  CLHEP::HepLorentzVector pWv;
  pWv.set(pW[1],pW[2],pW[3],pW[0]);
  
  double ppW[4],ppWs;
  CLHEP::HepLorentzVector ppWv,appWv;
  ppWs=sqrt(sW)/2.;
  // Choose theta and phi
  double pptheta=2.*M_PI*ran.flat();
  // ppphi=M_PI*ran.flat();
  // wt*=2.*pow(M_PI,2);
  double cosppphi=2.*ran.flat()-1.;  
  wt*=2.*M_PI*2.; 
  //  double sinphi=sin(ppphi);
  double sinphi=sqrt(1.-cosppphi*cosppphi);  // Know 0 < phi < pi
  //  wt*=1./(pow(2.*M_PI,3)*2.)*sinphi/(4.);   // Divide by 8, *2, see WPhaseSpace.tex 
  wt*=1./(pow(2.*M_PI,3)*2.)/(4.);   // Divide by 8, *2, see WPhaseSpace.tex 
                                     // (phi -> cosphi now so no sinphi in J)
  // construct 4-vector in W rest frame
  ppW[0]=ppWs;
  ppW[1]=ppWs*cos(pptheta)*sinphi;
  ppW[2]=ppWs*sin(pptheta)*sinphi;
  ppW[3]=ppWs*cosppphi;
  ppWv.set(ppW[1],ppW[2],ppW[3],ppW[0]);
  appWv.set(-ppW[1],-ppW[2],-ppW[3],ppW[0]);
  
  // translate to lab frame
  pd1=ppWv.boost(pWv.boostVector());
  pd2=appWv.boost(pWv.boostVector());
  // pd2=pWv-pd1;
  Ht+=(pd1+pd2).perp();
  // Ht+=pd2.perp();
  // The result of generation W instead of FS particle momenta:
  // wt*=pWv.e()/pd2.e();
  
  xa+=(pd1.e()-pd1.pz())/sqrts;//pd1.perp()/sqrts*t;
  xb+=(pd1.e()+pd1.pz())/sqrts;//pd1.perp()/sqrts/t;
  xa+=(pd2.e()-pd2.pz())/sqrts;
  xb+=(pd2.e()+pd2.pz())/sqrts;

  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double yoffset;
    double ycenter=ysum/(2.);
    {
      double lninvr1,r1,r2,temp,a(1.2);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      //      thetaenu=temp;
      yoffset=(temp-ycenter);
      wt=wt*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to partonmomenta and xa,xb
    for (int ii=0;ii<=1;ii++)  {
      double yi=deltay[ii]+yoffset;
      if (fabs(yi)>ymax) {
	setstatus(2);
	wt=0.;
	return wt;
      }
      
      pt=sqrt(pta[ii][0]*pta[ii][0]+pta[ii][1]*pta[ii][1]);
      
      p[0]=pt*cosh(yi); // the energy component
      p[1]=pta[ii][0];
      p[2]=pta[ii][1];
      p[3]=pt*sinh(yi); // the z-component

      // {double t=exp(-yi);
      // xa+=pt/sqrts*t;
      // xb+=pt/sqrts/t;
      // }
      xa+=(p[0]-p[3])/sqrts;
      xb+=(p[0]+p[3])/sqrts;
      AddParton(p);
    }
    
  }
  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);
  
#else  // Original CGenerator2jenu from FixedOrder
  double temp1,temp2,pea[4],pn[4],pt,r1,pte,sne,ptn;
  double wt(1.);
  double y,ye,yn;
  double pta[2][2];
  double deltay[2];
  double phie,phin;
  double ptpar(40.);

  CLHEP::HepLorentzVector p1,p2;

  xa=xb=0.;
  //  partons.clear();
  status=0;

  // generate momenta
  // Generate the transverse momentum of p3
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
  r1=ran.flat();
  pt=flags.jetptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  wt*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;

  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  pta[0][0]=pt*cos(temp1);
  pta[0][1]=pt*sin(temp1);
  deltay[0]=0.;
  // Generate the vectors for the electron and neutrino
  // The e-momenta:

  static const double pptpar(50.0);
  // The Transverse Momentum:
  temp1=atan((flags.pptmax-flags.pptmin)/pptpar);
  r1=ran.flat();
  pte=flags.pptmin+pptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  wt*=2.0*M_PI*pte*pptpar*temp1/temp2/temp2;
#if 1
  // The e-Rapidity:
  // Generate a ye Gaussian distributed around 0
  {
    double lninvr1,r1,r2,temp,a;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    a=1.6;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    ye=temp;
    wt=wt*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
  }
#else
  ye=-2.5+5.*ran.flat();
  wt*=5;
#endif 
  // Generate the four vector:
  phie=2.*M_PI*ran.flat(); // a random transverse angle for the e
  pea[0]=pte*cosh(ye);
  pea[1]=pte*cos(phie);
  pea[2]=pte*sin(phie);
  pea[3]=pte*sinh(ye);

  pd1.set(pea[1],pea[2],pea[3],pea[0]);
  xa+=pd1.perp()/sqrts*exp(-pd1.rapidity());
  xb+=pd1.perp()/sqrts*exp(pd1.rapidity());

  // The nu-momenta:
  // The smarter method: Generate s_ne according to the Breit-Wigner W shape
  r1=ran.flat();
  sne=flags.MW*(flags.MW + flags.GammaW*tan((M_PI*r1)/2. + (-1. + r1)*atan(flags.MW/flags.GammaW)));
  
  // Multiply by derivate (d sne)/(d r)
  wt*=(flags.GammaW*flags.MW*(M_PI+2.*atan(flags.MW/flags.GammaW)))/(1.+cos(M_PI*r1+2.*(-1.+r1)*atan(flags.MW/flags.GammaW)));
  // Generate phi for the neutrino
  phin=2.*M_PI*ran.flat(); // a random transverse angle for the nu
  wt*=2.*M_PI;
  // Generate the rapidity of the neutrino 
  // Generate a yn Gaussian distributed around ye
#if 1
  {
    double lninvr1,r1,r2,temp,a;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    a=1.6;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    yn=temp+ye;
    wt=wt*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
  }
#else
  yn=-5.+10.*ran.flat();
  wt*=10.;
#endif
  // Calculate the resulting transverse momentum of the neutrino
  ptn=sne/(2.*pte*(cosh(ye-yn)-cos(phie-phin)));

  //Multiply by ptn times the Jacobian (ptnu*dptnu/dsenu)
  wt*=ptn*ptn/sne;
  // Generate the four-vector
  pn[0]=ptn*cosh(yn);
  pn[1]=ptn*cos(phin);
  pn[2]=ptn*sin(phin);
  pn[3]=ptn*sinh(yn);
  wt/=4.*M_PI*pow(2.*M_PI,2);
  pd2.set(pn[1],pn[2],pn[3],pn[0]);
  Ht+=(pd1+pd2).perp();
  xa+=pd2.perp()/sqrts*exp(-pd2.rapidity());
  xb+=pd2.perp()/sqrts*exp(pd2.rapidity());
  
  // Calculate the transverse momentum of p4
  pta[1][0]=-(pta[0][0]+pea[1]+pn[1]);
  pta[1][1]=-(pta[0][1]+pea[2]+pn[2]);

  y=ran.flat()*10;
  wt*=10.;
  deltay[1]=y;
  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double ycenter=deltay[1]/2.;
    double yoffset;
    
    {
      double lninvr1,r1,r2,temp,a(2.);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      wt=wt*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to gluons and xa,xb
    double yi=deltay[0]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      wt=0.;
      return wt;
    }
    pt=sqrt(pta[0][0]*pta[0][0]+pta[0][1]*pta[0][1]);
    Ht+=pt;
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[0][0];
    p[2]=pta[0][1];
    p[3]=pt*sinh(yi); // the z-component

    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    AddParton(p);
    //    p1.set(p[1],p[2],p[3],p[0]);

    yi=deltay[1]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      wt=0.;
      return wt;
    } 
    pt=sqrt(pta[1][0]*pta[1][0]+pta[1][1]*pta[1][1]);
    Ht+=pt;
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[1][0];
    p[2]=pta[1][1];
    p[3]=pt*sinh(yi); // the z-component
    
    AddParton(p);
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    //    p2.set(p[1],p[2],p[3],p[0]);
  }
  //  Ht=p1.perp()+p2.perp()+(pd1+pd2).perp();
#endif /// Redoing generation ///

#else      // Fix momenta the same for every event  (MG Standalone ug point)
  pa.set(0,0,-500.,500.);
  pb.set(0,0,500.,500.);

  // u g --> e+ ve d g in that order
  // Call with aptype=21, bptype=2 => pd1=lepton=ve=4 in MG, pd2=alepton=e+=3 in MG
  pd1.set(-0.1038496E+03,-0.3019338E+03,0.7649492E+02,0.3283294E+03);
  pd2.set(-0.2210069E+02,0.4008035E+02,-0.7580543E+02,0.8855133E+02);

  // d g --> e- vebar u g in that order
  // Call with aptype=21, bptype=1 => pd1=lepton=e-=3 in MG, pd2=alepton=vebar=4 in MG
  // pd1.set(-0.2210069E+02,0.4008035E+02,-0.7580543E+02,0.8855133E+02);
  // pd2.set(-0.1038496E+03,-0.3019338E+03,0.7649492E+02,0.3283294E+03);

  // p1.set(0.2318313E+03 ,0.3595630E+03 ,-0.5023788E+02,0.4307611E+03);
  p[0]=0.4307611E+03;
  p[1]=0.2318313E+03;
  p[2]=0.3595630E+03;
  p[3]=-0.5023788E+02;
  AddParton(p);
  // p2.set(-0.1058810E+03,-0.9770964E+02, 0.4954839E+02,0.1523581E+03);
  p[0]=0.1523581E+03;
  p[1]=-0.1058810E+03;
  p[2]=-0.9770964E+02;
  p[3]=0.4954839E+02;
  AddParton(p);

  xa=500.*2./sqrts;
  xb=xa;
  wt=1.;
#endif


  /** fastjet find clusters */
  fastjet::ClusterSequence ev(partonmomenta, flags.jet_def);
  
  /** Retrieve the final state jets from fastjet Sorted by Rapidity*/
  std::vector<fastjet::PseudoJet> jets = sorted_by_pt(ev.inclusive_jets(flags.jetptmin));
  
  int njets=int(jets.size());
  //  std::cout << "Phase space njets : "<<njets<<std::endl;
  if (njets!=2) {
    status=11;
    wt=0;
    return wt;
  } 
    

#if calcscaleunc
  renscale=jets.at(0).perp()*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=flags.scale*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=Ht/2.*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=geometricmean(&jets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=jets.at(0).perp()*flags.scalefactor;
#else
  if (flags.scalesetting==0) //max pt(jets)
    renscale=ptjets.at(0).perp()*flags.scalefactor;
  else if (flags.scalesetting==1) //fixed scale
    renscale=flags.scale;
  else if (flags.scalesetting==2) // Ht/2
    renscale=Ht/2.*flags.scalefactor;
  else if (flags.scalesetting==3) //geometric mean of pt(jets)
    renscale=geometricmean(&ptjets)*flags.scalefactor;
#endif

  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);
 // cout<< "----------\n" << pa<<"\n"<<pb<<"\n"<<pd1<<"\n"<<pd2<<"\n"<<p1<<"\n"<<p2<<"\n";
  //  cout<< "Total: "<< pa+pb-pd1-pd2-p1-p2<<"\n";
  return wt;
}

double CWMultijet::weight1() // The two-jet FKL cross section
{
  wt=1.;

  wt*=Generate2jphsppoint();
  //  std::cout<<"t1: "<<wt<<"\n";
  if (status!=0) { // if we failed in generating a 2-jet configuration
    wt=0.;
    return wt;
  }
  double shat2=xa*xa*xb*xb*pow(getcmsenergy(),4);

  // now evaluate the process and setup the flavour and colour info
  if (xa>1.||xb>1.) {
    setstatus(5);
    wt=0.;
    return 0.;
  }
  // choose incoming flavours based on pdfs only
  double ufa(0.),ufb(0.);
  // std::cout<<"----------- Non-FKL -----------------\n";
  // std::cout<<"final scale: "<<renscale<<"  "<<Ht/2.<<"\n";
  // double tempperp(0.);
  // for (int kk; kk<partonmomenta.size(); kk++) {
  //   tempperp+=partonmomenta.at(kk).perp();
  //   std::cout<<partonmomenta.at(kk).perp()<<"  ";
  // }
  // std::cout<<"\n"<<(pd1+pd2).perp()<<"\n";
  // tempperp+=(pd1+pd2).perp();
  // std::cout<<"Total perps/2: "<<tempperp/2.<<"\n";
  ufa=ufb=renscale;

  // if (flags.scalesetting==0) {
  //   ufa=flags.scale;
  //   ufb=flags.scale;
  // } else if (flags.scalesetting==1) {
  //   ufa=Ht;
  //   ufb=Ht;
  // }

  wt*=GenerateIncomingParticles(xa,ufa,xb,ufb);
  //  std::cout<<"t2: "<<wt<<"\n";
  double pmg[8][4];
  pmg[0][1]=pb.x();
  pmg[0][2]=pb.y();
  pmg[0][3]=pb.z();
  pmg[0][0]=pb.e();

  pmg[1][1]=pa.x();
  pmg[1][2]=pa.y();
  pmg[1][3]=pa.z();
  pmg[1][0]=pa.e();

  pmg[2][1]=pd2.x(); // anti-lepton
  pmg[2][2]=pd2.y();
  pmg[2][3]=pd2.z();
  pmg[2][0]=pd2.e();

  pmg[3][1]=pd1.x(); // lepton
  pmg[3][2]=pd1.y();
  pmg[3][3]=pd1.z();
  pmg[3][0]=pd1.e();

  for (int ii=0;ii<2;ii++) {
    pmg[ii+4][1]=partonmomenta.at(1-ii).x();
    pmg[ii+4][2]=partonmomenta.at(1-ii).y();
    pmg[ii+4][3]=partonmomenta.at(1-ii).z();
    pmg[ii+4][0]=partonmomenta.at(1-ii).e();
  }

  wt*=Generate2jNonFKLflavourconf_and_eval(pmg);

  //  std::cout<<"t3: "<<wt<<"\n";
  wt/=(256.*shat2*pow(M_PI,5));
  double pdf1=pdfta*pdftb;
  wt*=pdf1;
  //  std::cout<<"t4: "<<wt<<"  "<<pdfta<<"  "<<pdftb<<"\n";
  wt*=389.379292;
  wt*=1E6; // micro barn-> pico barn
  //  std::cout<<"t5: "<<wt<<"\n";
#if calcscaleunc
  for (unsigned int ii=0;ii<scales.size();ii++) {
    double reweight=pdfpt(0,xa,scales.at(ii),aptype)*pdfpt(1,xb,scales.at(ii),bptype)/pdf1*pow(Halphas(scales.at(ii))/Halphas(renscale),2);
    if (flags.logcorrect==1) {
      reweight*=(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.front().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.back().perp()/scales.at(ii))/(2.*M_PI));
      reweight/=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI));
    }
    weights.push_back(wt*reweight);
  }
  // std::cout<<std::endl;
#endif
  //std::cout<<wt<<std::endl;
  return wt;
}

double CWMultijet::Generate3jphsppoint()
{
  double sqrts=2.*flags.Ebeam;
  double p[4],weight(1.);

  CLHEP::HepLorentzVector p2;
  // xa=xb=0.;
  // status=0;

#if 1    // Generate

#if NewPhSp
  double y,ysum,pt,r1,temp1,temp2,phi;
  double kt[2];
  double ptpar(40.);

  double pta[3][2];
  double deltay[3];
  // double phia[3];

  // The parton of lowest rapidity:
#if 1
  // Generate the transverse momentum of parton a
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
#else
  {
    double ptSQ;
    r1=ran.flat();
    ptSQ=ptSQgen/(exp(r1*log(1.+ptSQgen))-1.);
    weight*=2.*M_PI*(ptSQ)*(1.+(ptSQ)/ptSQgen)*log(1.+ptSQgen)/2.;
    pt=sqrt(ptSQ);
  }
#endif
  Ht+=pt;

  phi=2.*M_PI*ran.flat(); // Random angle in x,y-plane

  y=0.;
  ysum=0.;
  // phia[0]=phi;
  
  p[1]=pt*cos(phi);
  p[2]=pt*sin(phi);
  pta[0][0]=pt*cos(phi);
  pta[0][1]=pt*sin(phi);
  deltay[0]=0.;

  kt[0]=-p[1];
  kt[1]=-p[2];

  // Generate a rapidity span
  y=-log(1.-ran.flat());
  weight*=exp(y);
  ysum+=y;
  deltay[1]=ysum;

  //  Transverse momenta of p2
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
 
  weight/=16.*pow(M_PI,3);
  Ht+=pt;

  // Generate random angle 
  phi=2.*M_PI*ran.flat();
  // phia[1]=phi;

  p[1]=pta[1][0]=pt*cos(phi);
  p[2]=pta[1][1]=pt*sin(phi);

  // keep track of the transverse t-channel scale in the evolution
  kt[0]-=p[1];
  kt[1]-=p[2];

  y=-log(1.-ran.flat());
  ysum+=y;
  deltay[2]=ysum;
  weight*=exp(y);
  // Generate the transverse momentum of parton b
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  //   wtmatch*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;
  
  phi=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  Ht+=pt;
  
  pta[2][0]=pt*cos(phi);
  pta[2][1]=pt*sin(phi);
  
  kt[0]+=-pta[2][0];
  kt[1]+=-pta[2][1];
  
  // Now we know the transverse momentum of the W, given by the array kt
  double pW[4];
  // Choose its mass according to Breit-Wigner
  r1=ran.flat();
  double sW=flags.MW*(flags.MW + flags.GammaW*tan((M_PI*r1)/2. + (-1. + r1)*atan(flags.MW/flags.GammaW)));
  // Multiply by derivate (d sap)/(d r)
  {
    static double temp=atan(flags.MW/flags.GammaW);
    weight*=(flags.GammaW*flags.MW*(M_PI+2.*temp))/(1.+cos(M_PI*r1+2.*(-1.+r1)*temp));
  }  
  pW[1]=kt[0];
  pW[2]=kt[1];
  double EtW=sqrt(pW[1]*pW[1]+pW[2]*pW[2]+sW);
  
  double yW;
#if 1
  // Generate a yW Gaussian distributed around 0
  {
    double lninvr1,r1,r2,temp,a;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    a=1.;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    yW=temp;
    weight*=(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
  }
#else
  // choose a rapidity
  yW=-5.+10.*ran.flat();
  weight*=10;
#endif
  pW[0]=cosh(yW)*EtW;
  pW[3]=sinh(yW)*EtW;
  CLHEP::HepLorentzVector pWv;
  pWv.set(pW[1],pW[2],pW[3],pW[0]);
  
  double ppW[4],ppWs;
  CLHEP::HepLorentzVector ppWv,appWv;
  ppWs=sqrt(sW)/2.;
  // Choose theta and phi
  double pptheta=2.*M_PI*ran.flat();
  // ppphi=M_PI*ran.flat();
  // weight*=2.*pow(M_PI,2);
  double cosppphi=2.*ran.flat()-1.;  
  weight*=2.*M_PI*2.; 
  //  double sinphi=sin(ppphi);
  double sinphi=sqrt(1.-cosppphi*cosppphi);  // Know 0 < phi < pi
  //  weight*=1./(pow(2.*M_PI,3)*2.)*sinphi/(4.);   // Divide by 8, *2, see WPhaseSpace.tex 
  weight*=1./(pow(2.*M_PI,3)*2.)/(4.);   // Divide by 8, *2, see WPhaseSpace.tex 
                                     // (phi -> cosphi now so no sinphi in J)
  // construct 4-vector in W rest frame
  ppW[0]=ppWs;
  ppW[1]=ppWs*cos(pptheta)*sinphi;
  ppW[2]=ppWs*sin(pptheta)*sinphi;
  ppW[3]=ppWs*cosppphi;
  ppWv.set(ppW[1],ppW[2],ppW[3],ppW[0]);
  appWv.set(-ppW[1],-ppW[2],-ppW[3],ppW[0]);
  
  // translate to lab frame
  pd1=ppWv.boost(pWv.boostVector());
  pd2=appWv.boost(pWv.boostVector());
  // pd2=pWv-pd1;
  Ht+=(pd1+pd2).perp();
  // Ht+=pd2.perp();

  xa+=(pd1.e()-pd1.pz())/sqrts;//pd1.perp()/sqrts*t;
  xb+=(pd1.e()+pd1.pz())/sqrts;//pd1.perp()/sqrts/t;
  xa+=(pd2.e()-pd2.pz())/sqrts;
  xb+=(pd2.e()+pd2.pz())/sqrts;

  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double yoffset;
    double ycenter=ysum/(2.);
    {
      double lninvr1,r1,r2,temp,a(1.2);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      weight*=(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to partonmomenta and xa,xb
    for (int ii=0;ii<=2;ii++) {
      double yi=deltay[ii]+yoffset;
      if (fabs(yi)>ymax) {
	setstatus(2);
	weight=0.;
	return weight;
      }
      
      pt=sqrt(pta[ii][0]*pta[ii][0]+pta[ii][1]*pta[ii][1]);
      
      p[0]=pt*cosh(yi); // the energy component
      p[1]=pta[ii][0];
      p[2]=pta[ii][1];
      p[3]=pt*sinh(yi); // the z-component

      // {double t=exp(-yi);
      // xa+=pt/sqrts*t;
      // xb+=pt/sqrts/t;
      // }
      xa+=(p[0]-p[3])/sqrts;
      xb+=(p[0]+p[3])/sqrts;
      AddParton(p);
    }

  }
  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);

///////////////////////////////////////////////////////////////////////////////////
#else        //  Not using nice new phase space.
  double temp1,temp2,pea[4],pn[4],pt,r1,pte,sne,ptn;
  double y,ye,yn;
  double pta[3][2];
  double deltay[3];
  double phie,phin;
  double ptpar(40.);

  // generate momenta
  // Generate the transverse momentum of p1
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
// #if fix
//   r1=0.24;
// #else
  r1=ran.flat();
  //#endif
  pt=flags.jetptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  
#if fix
  temp1=2.*M_PI*0.56;
#else
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
#endif
  pta[0][0]=pt*cos(temp1);
  pta[0][1]=pt*sin(temp1);
  deltay[0]=0.;

  // Generate the transverse momentum of p3
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
#if fix
  r1=0.96;
#else
  r1=ran.flat();
#endif
  pt=flags.jetptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;

#if fix
  y=0.321;
#else
  y=ran.flat()*10;
#endif
  weight*=10.;
  deltay[2]=y;
#if fix
  temp1=2.*M_PI*0.359;
#else
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
#endif
  pta[2][0]=pt*cos(temp1);
  pta[2][1]=pt*sin(temp1);

  // Generate the vectors for the electron and neutrino
  // The e-momenta:
  // The Transverse Momentum:
  static const double pptpar(50.0);
  temp1=atan((flags.pptmax-flags.pptmin)/pptpar);
#if fix
  r1=0.582;
#else
  r1=ran.flat();
#endif
  pte=flags.pptmin+pptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pte*pptpar*temp1/temp2/temp2;
  
#if 1
  // The e-Rapidity:
  // Generate a ye Gaussian distributed around 0
  {
    double lninvr1,r1,r2,temp,a;
#if fix
    r1=0.4441;
    r2=0.229;
#else
    r1=ran.flat();
    r2=ran.flat();
#endif
    lninvr1=-log(r1);
    a=1.6;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    ye=temp;
    weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
  }
#else
#if fix
  ye=-2.5+5.*0.8001;
#else
  ye=-2.5+5.*ran.flat();
#endif
  weight*=5;
#endif 
  // Generate the four vector:
#if fix
  phie=2.*M_PI*0.772;
#else
  phie=2.*M_PI*ran.flat(); // a random transverse angle for the e
#endif
  pea[0]=pte*cosh(ye);
  pea[1]=pte*cos(phie);
  pea[2]=pte*sin(phie);
  pea[3]=pte*sinh(ye);

  pd1.set(pea[1],pea[2],pea[3],pea[0]);
  xa+=pd1.perp()/sqrts*exp(-pd1.rapidity());
  xb+=pd1.perp()/sqrts*exp(pd1.rapidity());
  //  std::cout<<"pd1: ("<<pd1.e()<<","<<pd1.x()<<","<<pd1.y()<<","<<pd1.z()<<")\n";

  // The nu-momenta:
  // The smarter method: Generate s_ne according to the Breit-Wigner W shape
#if fix
  r1=0.1;
#else  
  r1=ran.flat();
#endif
  sne=flags.MW*(flags.MW + flags.GammaW*tan((M_PI*r1)/2. + (-1. + r1)*atan(flags.MW/flags.GammaW)));
  
  // Multiply by derivate (d sne)/(d r)
  weight*=(flags.GammaW*flags.MW*(M_PI+2.*atan(flags.MW/flags.GammaW)))/(1.+cos(M_PI*r1+2.*(-1.+r1)*atan(flags.MW/flags.GammaW)));
  // Generate phi for the neutrino
#if fix
  phin=2.*M_PI*0.6402;
#else
  phin=2.*M_PI*ran.flat(); // a random transverse angle for the nu
#endif
  weight*=2.*M_PI;
  // Generate the rapidity of the neutrino 
  // Generate a yn Gaussian distributed around ye
#if 1
  {
    double lninvr1,r1,r2,temp,a;
#if fix
    r1=0.661;
    r2=0.0004;
#else
    r1=ran.flat();
    r2=ran.flat();
#endif
    lninvr1=-log(r1);
    a=1.6;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    yn=temp+ye;
    weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
  }
#else
#if fix
  yn=-5+10.*0.487;
#else
  yn=-5.+10.*ran.flat();
#endif
  weight*=10.;
#endif
  // Calculate the resulting transverse momentum of the neutrino
  ptn=sne/(2.*pte*(cosh(ye-yn)-cos(phie-phin)));

  //Multiply by ptn times the Jacobian (ptnu*dptnu/dsenu)
  weight*=ptn*ptn/sne;
  // Generate the four-vector
  pn[0]=ptn*cosh(yn);
  pn[1]=ptn*cos(phin);
  pn[2]=ptn*sin(phin);
  pn[3]=ptn*sinh(yn);
  weight/=4.*M_PI*pow(2.*M_PI,2);
  pd2.set(pn[1],pn[2],pn[3],pn[0]);
  //  std::cout<<"pd2: ("<<pd2.e()<<","<<pd2.x()<<","<<pd2.y()<<","<<pd2.z()<<")\n";
  xa+=pd2.perp()/sqrts*exp(-pd2.rapidity());
  xb+=pd2.perp()/sqrts*exp(pd2.rapidity());
  
  // Calculate the transverse momentum of p2
  pta[1][0]=-(pta[0][0]+pta[2][0]+pea[1]+pn[1]);
  pta[1][1]=-(pta[0][1]+pta[2][1]+pea[2]+pn[2]);
  weight/=16.*pow(M_PI,3);

#if fix
  y=0.892*deltay[2];
#else
  y=ran.flat()*deltay[2];
#endif
  weight*=deltay[2];
  deltay[1]=y;


  //  Ht=pd1.perp()+pd2.perp();
  Ht=(pd1+pd2).perp();
  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double ycenter=deltay[2]/2.;
    double yoffset;
    
    {
      double lninvr1,r1,r2,temp,a(2.);
#if fix
      r1=0.949;
      r2=0.294;
#else
      r1=ran.flat();
      r2=ran.flat();
#endif
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to gluons and xa,xb
    double yi=deltay[0]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    }
    pt=sqrt(pta[0][0]*pta[0][0]+pta[0][1]*pta[0][1]);

    Ht+=pt;
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[0][0];
    p[2]=pta[0][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    AddParton(p);
    //    std::cout<<"p1: ("<<p[0]<<","<<p[1]<<","<<p[2]<<","<<p[3]<<")\n";

    yi=deltay[1]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[1][0]*pta[1][0]+pta[1][1]*pta[1][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[1][0];
    p[2]=pta[1][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    p2.set(p[1],p[2],p[3],p[0]);
    //    std::cout<<"p2: ("<<p1.e()<<","<<p1.x()<<","<<p1.y()<<","<<p1.z()<<")\n";
    AddParton(p);

    yi=deltay[2]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[2][0]*pta[2][0]+pta[2][1]*pta[2][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[2][0];
    p[2]=pta[2][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;

    AddParton(p);
    //    std::cout<<"p3: ("<<p[0]<<","<<p[1]<<","<<p[2]<<","<<p[3]<<")\n";

  }

  if (p2.perp()<flags.jetptmin) {  
    status=3;
    weight=0.;
    return weight;
  }

#endif   // Use (or not) new generation

  if (xa>1.||xb>1.) {
    status=1;
    weight=0.;
    return 0.;
  }

  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);
  // std::cout<<"pa: ("<<pa.e()<<","<<pa.x()<<","<<pa.y()<<","<<pa.z()<<")\n";
  // std::cout<<"pb: ("<<pb.e()<<","<<pb.x()<<","<<pb.y()<<","<<pb.z()<<")\n";
  // std::cout<<"xa: "<<xa<<" xb: "<<xb<<"\n";
#else  // Fix momenta
  // pd1: (41.5717,28.0465,30.502,-3.35089)
  // pd2: (130.998,2.81434,74.8213,107.491)
  // p1: (2064.47,109.82,23.3078,-2061.42)
  // p2: (458.783,-92.7539,-100.02,-438.035)
  // p3: (173.197,-47.9267,-28.6108,163.956)
  // pa: (2550.19,0,0,-2550.19)
  // pb: (318.832,0,0,318.832)

// #if 1  // Agrees with CGenerator3jenu when fixed
//   pa.set(0,0,-2550.19,2550.19);
//   xa=0.364313;
//   pb.set(0,0,318.832,318.832);
//   xb=0.0455474;
//   pd1.set(28.0465,30.502,-3.35089,41.5717);
//   pd2.set(2.81434,74.8213,107.491,130.998);
  
//   p[0]=2064.47;
//   p[1]=109.82;
//   p[2]=23.3078;
//   p[3]=-2061.42;
//   AddParton(p);

//   p[0]=458.783;
//   p[1]=-92.7539;
//   p[2]=-100.02;
//   p[3]=-438.035;
//   AddParton(p);
  
//   p[0]=173.197;
//   p[1]=-47.9267;
//   p[2]=-28.6108;
//   p[3]=163.956;
//   AddParton(p);
// #else
//   pb.set(0,0,-2550.19,2550.19);
//   xb=0.364313;
//   pa.set(0,0,318.832,318.832);
//   xa=0.0455474;
//   pd1.set(28.0465,30.502,-3.35089,41.5717);
//   pd2.set(2.81434,74.8213,107.491,130.998);
  
//   p[0]=173.197;
//   p[1]=-47.9267;
//   p[2]=-28.6108;
//   p[3]=163.956;
//   AddParton(p);

//   p[0]=458.783;
//   p[1]=-92.7539;
//   p[2]=-100.02;
//   p[3]=-438.035;
//   AddParton(p);
  
//   p[0]=2064.47;
//   p[1]=109.82;
//   p[2]=23.3078;
//   p[3]=-2061.42;
//   AddParton(p);
// #endif
//   weight=1.;
#endif

  /** fastjet find clusters */
  fastjet::ClusterSequence ev(partonmomenta, flags.jet_def);
    
  /** Retrieve the final state jets from fastjet Sorted by Rapidity */
  std::vector<fastjet::PseudoJet> jets = sorted_by_pt(ev.inclusive_jets(flags.jetptmin));

  /** Save the number of final state jets */
  int njets=int(jets.size());

  //  std::cout << "Phase space njets : "<<njets<<std::endl;
  if (njets!=3) {
    status=11;
    weight=0;
    return weight;
  } 

#if calcscaleunc
  renscale=jets.at(0).perp()*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=flags.scale*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=Ht/2.*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=geometricmean(&jets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=jets.at(0).perp()*flags.scalefactor;
#else
  if (flags.scalesetting==0) //max pt(jets)
    renscale=ptjets.at(0).perp()*flags.scalefactor;
  else if (flags.scalesetting==1) //fixed scale
    renscale=flags.scale;
  else if (flags.scalesetting==2) // Ht/2
    renscale=Ht/2.*flags.scalefactor;
  else if (flags.scalesetting==3) //geometric mean of pt(jets)
    renscale=geometricmean(&ptjets)*flags.scalefactor;
#endif
    
  //   std::cout << "Phase space weight : "<<weight<<std::endl;
  return weight; 
}

double CWMultijet::weight2() // The three-jet cross section
{
  wt=1.;
  wt*=Generate3jphsppoint();
  if (status!=0) { // if we failed in generating a 3-jet configuration
    wt=0.;
    return wt;
  }

  double shat2=xa*xa*xb*xb*pow(getcmsenergy(),4);
  // now evaluate the process and setup the flavour and colour info
  if (xa>1.||xb>1.) {
    setstatus(5);
    wt=0.;
    return 0.;
  }
  // choose incoming flavours based on pdfs only
  double ufa(0.),ufb(0.);
  // if (flags.scalesetting==0) {
  //   ufa=flags.scale;
  //   ufb=flags.scale;
  // } else if (flags.scalesetting==1) {
  //   ufa=Ht;
  //   ufb=Ht;
  // }
  ufa=ufb=renscale;
  wt*=GenerateIncomingParticles(xa,ufa,xb,ufb);

  double pmg[7][4];
  pmg[0][1]=pb.x();
  pmg[0][2]=pb.y();
  pmg[0][3]=pb.z();
  pmg[0][0]=pb.e();

  pmg[1][1]=pa.x();
  pmg[1][2]=pa.y();
  pmg[1][3]=pa.z();
  pmg[1][0]=pa.e();

  pmg[2][1]=pd2.x(); // anti-lepton
  pmg[2][2]=pd2.y();
  pmg[2][3]=pd2.z();
  pmg[2][0]=pd2.e();

  pmg[3][1]=pd1.x(); // lepton
  pmg[3][2]=pd1.y();
  pmg[3][3]=pd1.z();
  pmg[3][0]=pd1.e();

  for (int ii=0;ii<3;ii++) {
    pmg[ii+4][1]=partonmomenta.at(2-ii).x();
    pmg[ii+4][2]=partonmomenta.at(2-ii).y();
    pmg[ii+4][3]=partonmomenta.at(2-ii).z();
    pmg[ii+4][0]=partonmomenta.at(2-ii).e();
  }

  wt*=Generate3jNonFKLflavourconf_and_eval(pmg);

  wt/=(256.*shat2*pow(M_PI,5));
  double pdf1=pdfta*pdftb;
  wt*=pdf1;
  wt*=389.379292;
  wt*=1E6; // micro barn-> pico barn
#if calcscaleunc
  for (unsigned int ii=0;ii<scales.size();ii++) {
    double reweight=pdfpt(0,xa,scales.at(ii),aptype)*pdfpt(1,xb,scales.at(ii),bptype)/pdf1*pow(Halphas(scales.at(ii))/Halphas(renscale),3);
    if (flags.logcorrect==1) {
      reweight*=(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.front().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.back().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.at(1).perp()/scales.at(ii))/(2.*M_PI));
      reweight/=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI));
    }
    weights.push_back(wt*reweight);
  }
#endif
  return wt;
}

double CWMultijet::Generate4jphsppoint()
{
  double sqrts=2.*flags.Ebeam;
  double p[4],weight(1.);
#if 1 // generate

#if NewPhSp

  double y,ysum,pt,r1,temp1,temp2,phi;
  double kt[2];
  double ptpar(40.);

  double pta[4][2];
  double deltay[4];
  // double phia[4];

  // The parton of lowest rapidity:
#if 1
  // Generate the transverse momentum of parton a
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
#else
  {
    double ptSQ;
    r1=ran.flat();
    ptSQ=ptSQgen/(exp(r1*log(1.+ptSQgen))-1.);
    weight*=2.*M_PI*(ptSQ)*(1.+(ptSQ)/ptSQgen)*log(1.+ptSQgen)/2.;
    pt=sqrt(ptSQ);
  }
#endif
  Ht+=pt;

  phi=2.*M_PI*ran.flat(); // Random angle in x,y-plane

  y=0.;
  ysum=0.;
  // phia[0]=phi;
  
  p[1]=pt*cos(phi);
  p[2]=pt*sin(phi);
  pta[0][0]=pt*cos(phi);
  pta[0][1]=pt*sin(phi);
  deltay[0]=0.;

  kt[0]=-p[1];
  kt[1]=-p[2];

  // Generate a rapidity span
  double a(0.8);
  double r=ran.flat();
  y=-a*log(1.-r);
  //  thetaenu=r;
  weight*=a*exp(y/a);
  ysum+=y;
  deltay[1]=ysum;

  //  Transverse momenta of p2
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
 
  weight/=16.*pow(M_PI,3);
  Ht+=pt;

  // Generate random angle 
  phi=2.*M_PI*ran.flat();
  // phia[1]=phi;

  p[1]=pta[1][0]=pt*cos(phi);
  p[2]=pta[1][1]=pt*sin(phi);

  // keep track of the transverse t-channel scale in the evolution
  kt[0]-=p[1];
  kt[1]-=p[2];

  //  Now for p3
  // Generate a rapidity span
  y=-a*log(1.-ran.flat());
  weight*=a*exp(y/a);
  // y=-log(1.-ran.flat());
  // weight*=exp(y);
  ysum+=y;
  deltay[2]=ysum;

  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
 
  weight/=16.*pow(M_PI,3);
  Ht+=pt;

  // Generate random angle 
  phi=2.*M_PI*ran.flat();
  // phia[2]=phi;

  p[1]=pta[2][0]=pt*cos(phi);
  p[2]=pta[2][1]=pt*sin(phi);

  // keep track of the transverse t-channel scale in the evolution
  kt[0]-=p[1];
  kt[1]-=p[2];

  y=-a*log(1.-ran.flat());
  weight*=a*exp(y/a);
  //  y=-log(1.-ran.flat());
  ysum+=y;
  deltay[3]=ysum;
  //  weight*=exp(y);
  // Generate the transverse momentum of parton b
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  //   wtmatch*=2.0*M_PI*pt*40.0*temp1/temp2/temp2;
  
  phi=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  Ht+=pt;
  
  pta[3][0]=pt*cos(phi);
  pta[3][1]=pt*sin(phi);
  
  kt[0]+=-pta[3][0];
  kt[1]+=-pta[3][1];
  
  // Now we know the transverse momentum of the W, given by the array kt
  double pW[4];
  // Choose its mass according to Breit-Wigner
  r1=ran.flat();
  double sW=flags.MW*(flags.MW + flags.GammaW*tan((M_PI*r1)/2. + (-1. + r1)*atan(flags.MW/flags.GammaW)));
  // Multiply by derivate (d sap)/(d r)
  {
    static double temp=atan(flags.MW/flags.GammaW);
    weight*=(flags.GammaW*flags.MW*(M_PI+2.*temp))/(1.+cos(M_PI*r1+2.*(-1.+r1)*temp));
  }  
  pW[1]=kt[0];
  pW[2]=kt[1];
  double EtW=sqrt(pW[1]*pW[1]+pW[2]*pW[2]+sW);
  
  double yW;
#if 1
  // Generate a yW Gaussian distributed around 0
  {
    double lninvr1,r1,r2,temp,a;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    a=1.;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    yW=temp;
    weight*=(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
  }
#else
  // choose a rapidity
  yW=-5.+10.*ran.flat();
  weight*=10;
#endif
  pW[0]=cosh(yW)*EtW;
  pW[3]=sinh(yW)*EtW;
  CLHEP::HepLorentzVector pWv;
  pWv.set(pW[1],pW[2],pW[3],pW[0]);
  
  double ppW[4],ppWs;
  CLHEP::HepLorentzVector ppWv,appWv;
  ppWs=sqrt(sW)/2.;
  // Choose theta and phi
  double pptheta=2.*M_PI*ran.flat();
  // ppphi=M_PI*ran.flat();
  // weight*=2.*pow(M_PI,2);
  double cosppphi=2.*ran.flat()-1.;  
  weight*=2.*M_PI*2.; 
  //  double sinphi=sin(ppphi);
  double sinphi=sqrt(1.-cosppphi*cosppphi);  // Know 0 < phi < pi
  //  weight*=1./(pow(2.*M_PI,3)*2.)*sinphi/(4.);   // Divide by 8, *2, see WPhaseSpace.tex 
  weight*=1./(pow(2.*M_PI,3)*2.)/(4.);   // Divide by 8, *2, see WPhaseSpace.tex 
                                     // (phi -> cosphi now so no sinphi in J)
  // construct 4-vector in W rest frame
  ppW[0]=ppWs;
  ppW[1]=ppWs*cos(pptheta)*sinphi;
  ppW[2]=ppWs*sin(pptheta)*sinphi;
  ppW[3]=ppWs*cosppphi;
  ppWv.set(ppW[1],ppW[2],ppW[3],ppW[0]);
  appWv.set(-ppW[1],-ppW[2],-ppW[3],ppW[0]);
  
  // translate to lab frame
  pd1=ppWv.boost(pWv.boostVector());
  pd2=appWv.boost(pWv.boostVector());
  // pd2=pWv-pd1;
  Ht+=(pd1+pd2).perp();
  // Ht+=pd2.perp();

  xa+=(pd1.e()-pd1.pz())/sqrts;//pd1.perp()/sqrts*t;
  xb+=(pd1.e()+pd1.pz())/sqrts;//pd1.perp()/sqrts/t;
  xa+=(pd2.e()-pd2.pz())/sqrts;
  xb+=(pd2.e()+pd2.pz())/sqrts;

  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double yoffset;
    double ycenter=ysum/(2.);
    {
      double lninvr1,r1,r2,temp,a(1.2);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      weight*=(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to partonmomenta and xa,xb
    for (int ii=0;ii<=3;ii++) {
      double yi=deltay[ii]+yoffset;
      if (fabs(yi)>ymax) {
	setstatus(2);
	weight=0.;
	return weight;
      }
      
      pt=sqrt(pta[ii][0]*pta[ii][0]+pta[ii][1]*pta[ii][1]);
      
      p[0]=pt*cosh(yi); // the energy component
      p[1]=pta[ii][0];
      p[2]=pta[ii][1];
      p[3]=pt*sinh(yi); // the z-component

      // {double t=exp(-yi);
      // xa+=pt/sqrts*t;
      // xb+=pt/sqrts/t;
      // }
      xa+=(p[0]-p[3])/sqrts;
      xb+=(p[0]+p[3])/sqrts;
      AddParton(p);
    }

  }

///////////////////////////////////////////////////////////////////////////////////
#else                 //  Not using nice new phase space.
  double temp1,temp2,pea[4],pn[4],sne,pt,r1,pte,ptn;
  double y,ye,yn;
  double pta[4][2];
  double deltay[4];
  double phie,phin,ptpar(40.);//phia[4]

  // generate momenta
  // Generate the transverse momentum of p1
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
  r1=ran.flat();
  pt=flags.jetptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;

  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  // phia[0]=temp1;
  pta[0][0]=pt*cos(temp1);
  pta[0][1]=pt*sin(temp1);
  deltay[0]=0.;

  // Generate the transverse momentum of p4
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
  r1=ran.flat();
  pt=flags.jetptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  
  y=ran.flat()*10;
  weight*=10.;
  deltay[3]=y;
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  // phia[3]=temp1;
  pta[3][0]=pt*cos(temp1);
  pta[3][1]=pt*sin(temp1);

  // Generate the transverse momentum of p3
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
  r1=ran.flat();
  pt=flags.jetptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  weight/=16.*pow(M_PI,3);
  
  y=deltay[3]*ran.flat(); // generate rapidity evenly distributed between the two other jets
  weight*=deltay[3];
  deltay[2]=y;
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  // phia[2]=temp1;
  pta[2][0]=pt*cos(temp1);
  pta[2][1]=pt*sin(temp1);

  // ***************** Now it's time for e and nu: ***************** //

  // Generate the vectors for the electron and neutrino
  // The e-momenta:
  // The Transverse Momentum:
  static const double pptpar(40.0);
  temp1=atan((flags.pptmax-flags.pptmin)/pptpar);
  r1=ran.flat();
  pte=flags.pptmin+pptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pte*pptpar*temp1/temp2/temp2;
  
#if 1
  // The e-Rapidity:
  // Generate a ye Gaussian distributed around 0
  {
    double lninvr1,r1,r2,temp,a;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    a=1.6;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    ye=temp;
    weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
  }
#else
  ye=-2.5+5.*ran.flat();
  weight*=5;
#endif 
  // Generate the four vector:
  phie=2.*M_PI*ran.flat(); // a random transverse angle for the e
  //  The weight*=2 Pi for this is in the line above immediately above #if 1
  pea[0]=pte*cosh(ye);
  pea[1]=pte*cos(phie);
  pea[2]=pte*sin(phie);
  pea[3]=pte*sinh(ye);

  pd1.set(pea[1],pea[2],pea[3],pea[0]);
  xa+=pd1.perp()/sqrts*exp(-pd1.rapidity());
  xb+=pd1.perp()/sqrts*exp(pd1.rapidity());

  // The nu-momenta:
  // The smarter method: Generate s_ne according to the Breit-Wigner W shape
  r1=ran.flat();
  sne=flags.MW*(flags.MW + flags.GammaW*tan((M_PI*r1)/2. + (-1. + r1)*atan(flags.MW/flags.GammaW)));
  
  // Multiply by derivate (d sne)/(d r)
  weight*=(flags.GammaW*flags.MW*(M_PI+2.*atan(flags.MW/flags.GammaW)))/(1.+cos(M_PI*r1+2.*(-1.+r1)*atan(flags.MW/flags.GammaW)));
  // Generate phi for the neutrino
  phin=2.*M_PI*ran.flat(); // a random transverse angle for the nu
  weight*=2.*M_PI;
  // Generate the rapidity of the neutrino 
  // Generate a yn Gaussian distributed around ye
#if 1
  {
    double lninvr1,r1,r2,temp,a;
    r1=ran.flat();
    r2=ran.flat();
    lninvr1=-log(r1);
    a=1.6;
    temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
    yn=temp+ye;
    weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
  }
#else
  yn=-5.+10.*ran.flat();
  weight*=10.;
#endif
  // Calculate the resulting transverse momentum of the neutrino
  ptn=sne/(2.*pte*(cosh(ye-yn)-cos(phie-phin)));

  //Multiply by ptn times the Jacobian (ptnu*dptnu/dsenu)
  weight*=ptn*ptn/sne;
  // Generate the four-vector
  pn[0]=ptn*cosh(yn);
  pn[1]=ptn*cos(phin);
  pn[2]=ptn*sin(phin);
  pn[3]=ptn*sinh(yn);
  weight/=4.*M_PI*pow(2.*M_PI,2);
  pd2.set(pn[1],pn[2],pn[3],pn[0]);
  xa+=pd2.perp()/sqrts*exp(-pd2.rapidity());
  xb+=pd2.perp()/sqrts*exp(pd2.rapidity());

  // ****************  End of e nu bit *************** //

  // Calculate the transverse momentum of p2
  pta[1][0]=-(pta[0][0]+pta[2][0]+pta[3][0]+pea[1]+pn[1]);
  pta[1][1]=-(pta[0][1]+pta[2][1]+pta[3][1]+pea[2]+pn[2]);
  weight/=16.*pow(M_PI,3);
  
  y=deltay[2]*ran.flat(); // generate rapidity evenly distributed between the two other jets
  weight*=deltay[2];
  deltay[1]=y;

  //  Ht=pd1.perp()+pd2.perp();
  Ht=(pd1+pd2).perp();

  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double ycenter=deltay[3]/2.;
    double yoffset;
    
    {
      double lninvr1,r1,r2,temp,a(2.);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to gluons and xa,xb
    double yi=deltay[0]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    }
    pt=sqrt(pta[0][0]*pta[0][0]+pta[0][1]*pta[0][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[0][0];
    p[2]=pta[0][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    AddParton(p);

    yi=deltay[1]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[1][0]*pta[1][0]+pta[1][1]*pta[1][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[1][0];
    p[2]=pta[1][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    AddParton(p);
    
    yi=deltay[2]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[2][0]*pta[2][0]+pta[2][1]*pta[2][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[2][0];
    p[2]=pta[2][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;

    AddParton(p);

    yi=deltay[3]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[3][0]*pta[3][0]+pta[3][1]*pta[3][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[3][0];
    p[2]=pta[3][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    AddParton(p);
  }
  //     cout <<"--------------------\n";
  //     cout << p1 << endl<<p2<<endl;
#endif  
  // //   std::cout << "Phase space xa, xb : "<<xa<<" , "<<xb<<std::endl;
  if (xa>1.||xb>1.) {
    status=1;
    weight=0.;
    return 0.;
  }

  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);
    //    cout<< "----------\n" << pa<<"\n"<<pb<<"\n"<<pW<<"\n"<<p1<<"\n"<<p2<<"\n";

  // Check if it is a 4-jet phase space point:
    
  /** fastjet find clusters */
  fastjet::ClusterSequence ev(partonmomenta, flags.jet_def);
    
  /** Retrieve the final state jets from fastjet Sorted by Pt*/
  std::vector<fastjet::PseudoJet> jets = sorted_by_pt(ev.inclusive_jets(flags.jetptmin));
    
  /** Save the number of final state jets */
  int njets=int(jets.size());

//   std::cout << "Phase space njets : "<<njets<<std::endl;
  if (njets!=4) {
    status=11;
    weight=0;
    return weight;
  } 

#if calcscaleunc
  renscale=jets.at(0).perp()*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=flags.scale*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=Ht/2.*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=geometricmean(&jets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=jets.at(0).perp()*flags.scalefactor;
#else
  if (flags.scalesetting==0) //max pt(jets)
    renscale=ptjets.at(0).perp()*flags.scalefactor;
  else if (flags.scalesetting==1) //fixed scale
    renscale=flags.scale;
  else if (flags.scalesetting==2) // Ht/2
    renscale=Ht/2.*flags.scalefactor;
  else if (flags.scalesetting==3) //geometric mean of pt(jets)
    renscale=geometricmean(&ptjets)*flags.scalefactor;
#endif
#else
  pa.set(0,0,-2204.85,2204.85);
  xa=0.629956;
  pb.set(0,0,776.215,776.215);
  xb=0.221776;
  pd2.set(-10.5659,-20.6489,38.4815,44.9315); // Jeppe's anti-lepton is 2, see 3086
  pd1.set(-19.1413,53.4114,216.247,223.567);
  //  (33.5632,-56.3247,-1952.89;1953.99)
  p[0]=1953.99;
  p[1]=33.5632;
  p[2]=-56.3247;
  p[3]=-1952.89;
  AddParton(p);
  //  (-47.218,26.4258,-227.433;233.781)
  p[0]=233.781;
  p[1]=-47.218;
  p[2]=26.4258;
  p[3]=-227.433;
  AddParton(p);
  //  (69.6585,-33.4644,103.009;128.775)
  p[0]=128.775;
  p[1]=69.6585;
  p[2]=-33.4644;
  p[3]=103.009;
  AddParton(p);
  //  (-26.2964,30.6009,393.954;396.015)
  p[0]=396.015;
  p[1]=-26.2964;
  p[2]=30.6009;
  p[3]=393.954;
  AddParton(p);

  weight=1.;
#endif
  return weight;
}

double CWMultijet::weight3() // The four-jet cross section
{
  wt=1.;
  wt*=Generate4jphsppoint();
  if (status!=0) { // if we failed in generating a 4-jet configuration
    wt=0.;
    return wt;
  }

  double shat2=xa*xa*xb*xb*pow(getcmsenergy(),4);

  // now evaluate the process and setup the flavour and colour info
  if (xa>1.||xb>1.) {
    setstatus(5);
    wt=0.;
    return 0.;
  }
  // choose incoming flavours based on pdfs only
  double ufa(0.),ufb(0.); // factorisation scale
  // if (flags.scalesetting==0) {
  //   ufa=flags.scale;
  //   ufb=flags.scale;
  // } else if (flags.scalesetting==1) {
  //   ufa=Ht;
  //   ufb=Ht;
  // }
  ufa=ufb=renscale;
  wt*=GenerateIncomingParticles(xa,ufa,xb,ufb);

  double pmg[8][4];
  pmg[0][1]=pb.x();
  pmg[0][2]=pb.y();
  pmg[0][3]=pb.z();
  pmg[0][0]=pb.e();

  pmg[1][1]=pa.x();
  pmg[1][2]=pa.y();
  pmg[1][3]=pa.z();
  pmg[1][0]=pa.e();

  pmg[2][1]=pd2.x(); // anti-lepton
  pmg[2][2]=pd2.y();
  pmg[2][3]=pd2.z();
  pmg[2][0]=pd2.e();

  pmg[3][1]=pd1.x(); // lepton
  pmg[3][2]=pd1.y();
  pmg[3][3]=pd1.z();
  pmg[3][0]=pd1.e();

  for (int ii=0;ii<4;ii++) {
    pmg[ii+4][1]=partonmomenta.at(3-ii).x();
    pmg[ii+4][2]=partonmomenta.at(3-ii).y();
    pmg[ii+4][3]=partonmomenta.at(3-ii).z();
    pmg[ii+4][0]=partonmomenta.at(3-ii).e();
  }

  wt*=Generate4jNonFKLflavourconf_and_eval(pmg);

  wt/=(256.*shat2*pow(M_PI,5));
  double pdf1=pdfta*pdftb;
  wt*=pdf1;
  wt*=389.379292;
  wt*=1E6; // micro barn-> pico barn
#if calcscaleunc
  for (unsigned int ii=0;ii<scales.size();ii++) {
    double reweight=pdfpt(0,xa,scales.at(ii),aptype)*pdfpt(1,xb,scales.at(ii),bptype)/pdf1*pow(Halphas(scales.at(ii))/Halphas(renscale),4);
    if (flags.logcorrect==1) {
      reweight*=(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.front().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.back().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.at(1).perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.at(2).perp()/scales.at(ii))/(2.*M_PI));
      reweight/=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(2).perp()/renscale)/(2.*M_PI));
    }
    weights.push_back(wt*reweight);
  }
#endif

  return wt;
}

double CWMultijet::Event()
{
  static double processc[4]={0.,0.,0.,0.}; // how often to choose the different type of processes (cummulative)
  static double procnorm[4]={1.,1.,1.,1.};

  static bool initialisearray(true);
  static double sum(0.);
  // run only once:
  if (initialisearray) {
    if (flags.resum) {
      sum+=0.9;
    }
    processc[0]=sum;
    if (flags.nonFKLmatch2j) {
      sum+=0.05;
    }
    processc[1]=sum;
    if (flags.nonFKLmatch3j) {
      sum+=0.05;
    }
    processc[2]=sum;
    if (flags.nonFKLmatch4j) {
      sum+=0.05;
    }
    processc[3]=sum;
    procnorm[0]=sum/processc[0];
    for (int ii=1;ii<4;ii++) {
      procnorm[ii]=sum/(processc[ii]-processc[ii-1]);
    }
    initialisearray=false; // don't do this again!
  }

  renscale=flags.scale; // will be reset dynamically if needed
  setstatus(0);
  partonmomenta.clear();
  LHparticles.clear();
  jets.clear();
  scales.clear();
  weights.clear();
  xa=xb=0.;
  Ht=0.;
  aptype=bptype=0;
  matchreweight=0.;
  // Clear Exceptions
  feclearexcept(FE_ALL_EXCEPT);

  static int test_excepts=(FE_DIVBYZERO & FE_INEXACT & FE_INVALID & FE_OVERFLOW & FE_UNDERFLOW);
  double rtemp=sum*ran.flat();
  double tempw;
  int set_excepts;

  if (rtemp<processc[0]) {
    processtype=1;
    tempw=weight();
    set_excepts = fetestexcept(test_excepts );
    if (set_excepts & test_excepts)
      setstatus(20);
  } else if (rtemp<processc[1]) {
    processtype=2;
    tempw=weight1();
    set_excepts = fetestexcept(test_excepts );
    if (set_excepts & test_excepts) {
      setstatus(21);
      cout << "Bugger, 2jet\n";
    }      
  } else if (rtemp<processc[2]) {
    processtype=3;
    tempw=weight2();
    set_excepts = fetestexcept(test_excepts );
    if (set_excepts & test_excepts) {
      setstatus(22);
      cout << "Bugger, 3jet\n";
    }      
  } else { // remember to add test when 4j-nonFKL matching is included
    processtype=4;
    tempw=weight3();
    
    set_excepts = fetestexcept(test_excepts );
    if (set_excepts & test_excepts) {
      setstatus(23);
      cout << "Bugger, 4jet\n";
    }
  }
  tempw*=procnorm[processtype-1];
#if calcscaleunc
  for (unsigned int ii=0;ii<weights.size();++ii) {
    weights.at(ii)*=procnorm[processtype-1];
  }
#endif
  return tempw;
}
