c====================================================================
c
c  Define common block containing all coupling constants and masses
c  which are used in the HELAS routines.
c
c  These include masses, widths and real/complex couplings.
c
c  This file can be built automatically from particles.dat and
c  interactions.dat
c
c====================================================================
c
c     QCD 
c
      double complex       gg(2)
      double precision     g
      common /COUPL_QCD/   g,gg   
c
c    kinematical masses
c
      double precision     hmass, wmass, zmass, amass,
     &                     tmass, bmass, lmass, cmass
      common /COUPL_MASS/  hmass, wmass, zmass, amass,
     &                     tmass, bmass, lmass, cmass
c
c   widths
c
      double precision     hwidth, wwidth, zwidth, 
     &                     twidth, lwidth, awidth
      common /COUPL_WIDTH/ hwidth, wwidth, zwidth, 
     &                     twidth, lwidth, awidth
c
c  couplings in the feynman rules
c
      double complex       gal(2), gad(2), gau(2), gwf(2),
     &                     gzn(2), gzl(2), gzd(2), gzu(2)
      double precision     gw, gwwa, gwwz
      common /COUPL_GAUGE/ gal   , gad   , gau   , gwf   ,
     &                     gzn   , gzl   , gzd   , gzu   ,
     &                     gw, gwwa, gwwz
c
      double complex       gwfc(2),  gwfs(2), gwfm(2)
      common /coupl_ckm/   gwfc,     gwfs   , gwfm	
c
      double complex       gwwh, gzzh, gwwhh, gzzhh, ghhh, ghhhh
      common /COUPL_SCAL/  gwwh, gzzh, gwwhh, gzzhh, ghhh, ghhhh
c
      double complex      ghtop(2), ghbot(2), ghtau(2), ghcha(2),
     &                    gh3top(2),gh3bot(2),gh3tau(2),gh3cha(2)
      common /COUPL_YUK/  ghtop   , ghbot   , ghtau   , ghcha,
     &                    gh3top  , gh3bot  , gh3tau  , gh3cha

c higgs effective theory couplings

      double complex      gh(2) , ga(2) , gphph(2)
      double precision    ghh(2), gha(2), gaa(2), dumm
      common /COUPL_HEFT/ gh    , ga    , gphph ,
     &                    ghh   , gha   , gaa   , dumm

c     /COUPL_QCD/
      gg(1)=  (-3.54490770182,0.)
      gg(2)= (-3.54490770182,0.)
      g=   3.54490770182

c     /COUPL_MASS/
      hmass=   126.
      wmass=   80.2673592
      zmass=   91.5445065
      amass=   0.
      tmass=   174.3
      bmass=   4.2
      lmass=   1.777
      cmass=   1.4

c     /COUPL_WIDTH/
      hwidth=   0.00421
      wwidth=   2.0353557
      zwidth=   2.44639985
      twidth=   1.5101349
      lwidth=   0.
      awidth=   0.

c      /COUPL_GAUGE/
      gal(1)=  (0.313451,0.)
      gal(2)=(0.313451,0.)
      gad(1)=  (0.104483667,0.)
      gad(2)=(0.104483667,0.)
      gau(1)=  (-0.208967333,0.)
      gau(2)=(-0.208967333,0.)
      gwf(1)=  (-0.460957353,0.)
      gwf(2)=(0.,0.)
      gzn(1)=  (-0.371739801,0.)
      gzn(2)=(0.,0.)
      gzl(1)=  (0.199847317,0.)
      gzl(2)=(-0.171892484,0.)
      gzd(1)=  (0.314442306,0.)
      gzd(2)=(-0.0572974946,0.)
      gzu(1)=  (-0.257144812,0.)
      gzu(2)=(0.114594989,0.)
      gw=   0.65189214
      gwwa=   0.313451
      gwwz=   0.571587118
      
c      /coupl_ckm/
      gwfc(1)=  (-0.449433419,0.)
      gwfc(2)=(0.,0.)
      gwfs(1)=  (-0.102426964,0.)
      gwfs(2)=(0.,0.)
      gwfm(1)=  (0.102426964,0.)
      gwfm(2)=(0.,0.)
      
c      /COUPL_SCAL/
      gwwh=  (52.3256606,0.)
      gzzh=  (68.0614732,0.)
      gwwhh=  (0.212481681,0.)
      gzzhh=  (0.276380959,0.)
      ghhh=  (-121.822646,0.)
      ghhhh=  (-0.494691905,0.)
      
c      /COUPL_YUK/
      ghtop(1)=  (-0.707789575,0.)
      ghtop(2)=(-0.707789575,0.)
      ghbot(1)=  (-0.0170551705,0.)
      ghbot(2)=(-0.0170551705,0.)
      ghtau(1)=  (-0.00721596141,0.)
      ghtau(2)=(-0.00721596141,0.)
      ghcha(1)=  (-0.0170551705,0.)
      ghcha(2)=(-0.0170551705,0.)
      gh3top(1)=  (0.,-0.707789575)
      gh3top(2)=(0.,0.707789575)
      gh3bot(1)=  (0.,-0.0170551705)
      gh3bot(2)=(0.,0.0170551705)
      gh3tau(1)=  (0.,-0.00721596141)
      gh3tau(2)=(0.,0.00721596141)
      gh3cha(1)=  (0.,-0.0170551705)
      gh3cha(2)=(0.,0.0170551705)
      
c      /COUPL_HEFT/
      gh(1)=  (4.31314208921E-04,0.)
      gh(2)=(0.,0.)
      ga(1)=  (0.,0.)
      ga(2)=(7.84506521E-05,0.)
      gphph(1)=  (-2.98145309E-05,0.)
      gphph(2)=(0.,0.)
      ghh(1)=   0. 
      ghh(2)=   0.
      gha(1)=   0.
      gha(2)=  0.
      gaa(1)=   0.
      gaa(2)=  0.
      dumm=   0.
      
