#include "WJets.h"
#include "Tree2jenu.h"
#include "MomentumDef.h"
//#include "Tree3jenu.h"

double CWMultijet::MG(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb, CLHEP::HepLorentzVector aleptonp, CLHEP::HepLorentzVector leptonp)
{
  double wt(0.);
  FREAL meout(0.),p12[6][4],p21[6][4],(*pchoice)[4];
  void (*smatrixchoice)(FREAL * P1, FREAL * ANS);

  // p12 = config where partons are ordered as in name
  // p21 = config where partons are reversed relative to name

  p12[0][0]=pb.e();    // forward parton
  p12[0][1]=pb.x();
  p12[0][2]=pb.y();
  p12[0][3]=pb.z();
  
  p12[1][0]=pa.e();   //  backward parton
  p12[1][1]=pa.x();
  p12[1][2]=pa.y();
  p12[1][3]=pa.z();
  
  p12[2][0]=aleptonp.e();   // anti-lepton
  p12[2][1]=aleptonp.x();
  p12[2][2]=aleptonp.y();
  p12[2][3]=aleptonp.z();
  
  p12[3][0]=leptonp.e();   // lepton
  p12[3][1]=leptonp.x();
  p12[3][2]=leptonp.y();
  p12[3][3]=leptonp.z();
  
  p12[4][0]=(partons.back()).e();   // more forward jet
  p12[4][1]=(partons.back()).x();
  p12[4][2]=(partons.back()).y();
  p12[4][3]=(partons.back()).z();
  
  p12[5][0]=(partons.front()).e();  // more backward jet
  p12[5][1]=(partons.front()).x();
  p12[5][2]=(partons.front()).y();
  p12[5][3]=(partons.front()).z();

  for (int ii=0; ii<4; ii++) {
    p21[0][ii]=p12[1][ii];
    p21[1][ii]=p12[0][ii];
    p21[2][ii]=p12[2][ii];
    p21[3][ii]=p12[3][ii];
    p21[4][ii]=p12[5][ii];
    p21[5][ii]=p12[4][ii];
  }

  double idpartfact=1.;     //  Factor to take account of identical final states

  if (aptype==21) {// a=g, we know they are not both gluons so b!= g
    if (bptype>0) { 
      smatrixchoice=smatrixugtepvedg_;
      pchoice=&p12[0];
      //      std::cout<<"Yes!"<<std::endl;
    }
    else {
      smatrixchoice=smatrixdxgtepveuxg_;
      pchoice=&p12[0];
    }
  }
  else if (bptype==21) {  // b=g, a!=g
    if (aptype>0) {
      smatrixchoice=smatrixugtepvedg_;
      pchoice=&p21[0];
    }
    else {
      smatrixchoice=smatrixdxgtepveuxg_;
      pchoice=&p21[0];
    }
  }
  else {  //  two (anti-)quarks to deal with!
    if (aptype==bptype) {
      if (aptype > 0) {
	smatrixchoice=smatrixuutepveud_;
	if (wc==true)   // emit off b
	  pchoice=&p21[0];
	else            // emit off a so swap momenta
	  pchoice=&p12[0];
      }
      else {
	smatrixchoice=smatrixdxdxtepveuxdx_;
	if (wc==true)   // emit off b
	  pchoice=&p12[0];
	else            // emit off a so swap momenta
	  pchoice=&p21[0];
      }
    }
    else if (aptype==-bptype) {
      if (aptype > 0) {  // backward particle is q, forward qbar
	pchoice=&p12[0];
	if (wc==true) 
	  smatrixchoice=smatrixdxdtepveuxd_;
	else 
	  smatrixchoice=smatrixuxutepveuxd_;
      }
      else {            // forward (b) particle is q, backward (a) qbar
	pchoice=&p21[0];
	if (wc==true) 
	  smatrixchoice=smatrixuxutepveuxd_;
	else 
	  smatrixchoice=smatrixdxdtepveuxd_;
      }
    }
    else if (aptype==1) {    // a=d, b!=d,dbar
      if (bptype==2) {  // b=u
	smatrixchoice=smatrixudtepvedd_;
	if (wc==true) 
	  pchoice=&p12[0];
	else 
	  pchoice=&p21[0];
	idpartfact*=2.;
      }
      else if (bptype>2) {
	smatrixchoice=smatrixustepveds_;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype==-2) {
	pchoice=&p21[0];
	if (wc==true) 
	  smatrixchoice=smatrixudxtepveuux_;
	else 
	  smatrixchoice=smatrixudxtepveddx_;
      }
      else {
	if (wc==true) {
	  smatrixchoice=smatrixdxstepveuxs_;
	  pchoice=&p12[0];
	}
	else {
	  smatrixchoice=smatrixusxtepvedsx_;
	  pchoice=&p21[0];
	}
      }
    }
    else if (aptype==2) { // a=u
      if (bptype==1) {  // b=d
	smatrixchoice=smatrixudtepvedd_;
	idpartfact*=2.;
	if (wc==true) 
	  pchoice=&p12[0];
	else                 //  Not sure if this is necessary
	  pchoice=&p21[0];
      }
      else if (bptype>2) {    // b = s,c,b
	smatrixchoice=smatrixustepveds_;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype==-1) {  // b = dbar
	pchoice=&p21[0];
	if (wc==true) 
	  smatrixchoice=smatrixudxtepveuux_;
	else 
	  smatrixchoice=smatrixudxtepveddx_;
      }
      else {            // b = sbar, cbar,bbar
	if (wc==true) {
	  smatrixchoice=smatrixdxstepveuxs_;
	  pchoice=&p12[0];
	}
	else {
	  smatrixchoice=smatrixusxtepvedsx_;
	  pchoice=&p21[0];
	}
      }
    }
    else if (aptype==3) { // a=s
      if (bptype==4) {  // b=c
	smatrixchoice=smatrixudtepvedd_;
	idpartfact*=2.;
	if (wc==true) 
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype>0) {    // b = u,d,b
	smatrixchoice=smatrixustepveds_;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype==-4) {  // b = cbar
	pchoice=&p21[0];
	if (wc==true) 
	  smatrixchoice=smatrixudxtepveuux_;
	else 
	  smatrixchoice=smatrixudxtepveddx_;
      }
      else {            // b = ubar, dbar,bbar
	if (wc==true) {
	  smatrixchoice=smatrixdxstepveuxs_;
	  pchoice=&p12[0];
	}
	else {
	  smatrixchoice=smatrixusxtepvedsx_;
	  pchoice=&p21[0];
	}
      }
    }
    else if (aptype==4) { // a=c
      if (bptype==3) {  // b=s
	smatrixchoice=smatrixudtepvedd_;
	idpartfact*=2.;
	if (wc==true) 
	  pchoice=&p12[0];
	else                 //  Not sure if this is necessary
	  pchoice=&p21[0];
      }
      else if (bptype>0) {    // b = u,d,b
	smatrixchoice=smatrixustepveds_;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype==-3) {  // b = cbar
	pchoice=&p21[0];
	if (wc==true) 
	  smatrixchoice=smatrixudxtepveuux_;
	else 
	  smatrixchoice=smatrixudxtepveddx_;
      }
      else {            // b = ubar, dbar,bbar
	if (wc==true) {
	  smatrixchoice=smatrixdxstepveuxs_;
	  pchoice=&p12[0];
	}
	else {
	  smatrixchoice=smatrixusxtepvedsx_;
	  pchoice=&p21[0];
	}
      }
    }
    else if (aptype>0) { // a=b
      if (bptype>0) {    // b = u,d,c,s
	smatrixchoice=smatrixustepveds_;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else {            // b = ubar, dbar, sbar, cbar
	if (wc==true) {
	  smatrixchoice=smatrixdxstepveuxs_;
	  pchoice=&p12[0];
	}
	else {
	  smatrixchoice=smatrixusxtepvedsx_;
	  pchoice=&p21[0];
	}
      }
    }
    ///  Now go for a = anti-quark
    else if (aptype==-1) {    // a=dbar, b!=d
      if (bptype==-2) {  // b=ubar
	smatrixchoice=smatrixdxuxtepveuxux_;
	idpartfact*=2.;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype<-2) {
	smatrixchoice=smatrixdxsxtepveuxsx_;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype==2) {
	pchoice=&p12[0];
	if (wc==true) 
	  smatrixchoice=smatrixudxtepveddx_;
	else 
	  smatrixchoice=smatrixudxtepveuux_;
      }
      else {  // a=dbar, b= s,c,t
	if (wc==true) {
	  smatrixchoice=smatrixusxtepvedsx_;
	  pchoice=&p12[0];
	}
	else {
	  smatrixchoice=smatrixdxstepveuxs_;
	  pchoice=&p21[0];
	}
      }
    }
    else if (aptype==-2) {    // a=ubar, b!=u
      if (bptype==-1) {  // b=dbar
	smatrixchoice=smatrixdxuxtepveuxux_;
	idpartfact*=2.;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype<-2) {  // b = sbar,cbar,bbar
	smatrixchoice=smatrixdxsxtepveuxsx_;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype==1) {  // a=ubar, b=d
	pchoice=&p12[0];
	if (wc==true) 
	  smatrixchoice=smatrixudxtepveddx_;
	else 
	  smatrixchoice=smatrixudxtepveuux_;
      }
      else {  // a=ubar, b= s,c,t
	if (wc==true) {
	  smatrixchoice=smatrixusxtepvedsx_;
	  pchoice=&p12[0];
	}
	else {
	  smatrixchoice=smatrixdxstepveuxs_;
	  pchoice=&p21[0];
	}
      }
    }
    else if (aptype==-3) {    // a=sbar, b!=u
      if (bptype==-4) {  // b=cbar
	smatrixchoice=smatrixdxuxtepveuxux_;
	idpartfact*=2.;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype<0) {  // b = ubar,dbar,bbar
	smatrixchoice=smatrixdxsxtepveuxsx_;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype==4) {  // a=ubar, b=c
	pchoice=&p12[0];
	if (wc==true) 
	  smatrixchoice=smatrixudxtepveddx_;
	else 
	  smatrixchoice=smatrixudxtepveuux_;
      }
      else {  // a=sbar, b= u,d,t
	if (wc==true) {
	  smatrixchoice=smatrixusxtepvedsx_;
	  pchoice=&p12[0];
	}
	else {
	  smatrixchoice=smatrixdxstepveuxs_;
	  pchoice=&p21[0];
	}
      }
    }
    else if (aptype==-4) {    // a=cbar, b!=u
      if (bptype==-3) {  // b=sbar
	smatrixchoice=smatrixdxuxtepveuxux_;
	idpartfact*=2.;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype<0) {  // b = ubar,dbar,bbar
	smatrixchoice=smatrixdxsxtepveuxsx_;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else if (bptype==3) {  // a=ubar, b=s
	pchoice=&p12[0];
	if (wc==true) 
	  smatrixchoice=smatrixudxtepveddx_;
	else 
	  smatrixchoice=smatrixudxtepveuux_;
      }
      else {  // a=cbar, b= u,d,t
	if (wc==true) {
	  smatrixchoice=smatrixusxtepvedsx_;
	  pchoice=&p12[0];
	}
	else {
	  smatrixchoice=smatrixdxstepveuxs_;
	  pchoice=&p21[0];
	}
      }
    }
    else if (aptype==-5) {    // a=bbar, b!=b
      if (bptype<0) {  // b = ubar,dbar,sbar, cbar
	smatrixchoice=smatrixdxsxtepveuxsx_;
	if (wc==true)
	  pchoice=&p12[0];
	else
	  pchoice=&p21[0];
      }
      else {  // a=bbar, b= u,d,s,c
	if (wc==true) {
	  smatrixchoice=smatrixusxtepvedsx_;
	  pchoice=&p12[0];
	}
	else {
	  smatrixchoice=smatrixdxstepveuxs_;
	  pchoice=&p21[0];
	}
      }
    }
    else  {
      std::cerr<<"Watch out - unanticipated parton configuration!"<<std::endl;
      smatrixchoice=smatrixdxdxtepveuxdx_;
      pchoice=&p12[0];
    }
  }

  smatrixchoice(pchoice[0],&meout);
  wt=meout*idpartfact;
  // smatrixuutepveud_(p12[0],&temp);
  // std::cout<<"Compare: "<<temp;
  // smatrixuutepveud_(p21[0],&temp);
  // std::cout<<" and "<<temp<<std::endl;
  // double gs2=0.;
  // if (flags.scalesetting==0) //fixed scale
  //   gs2=4.*M_PI*AlphaS(flags.scale);
  // else if (flags.scalesetting==1)
  //   gs2=4.*M_PI*AlphaS(Ht);
  double gs2=4.*M_PI*Halphas(renscale);

  // double gs2=1.2177158*1.2177158;
  // std::cout<<"------------------------\n";
  // std::cout<<pchoice[0][3]<<"  "<<pchoice[1][3]<<"  "<<pchoice[2][3]<<"  "<<pchoice[3][3]<<"  "<<pchoice[4][3]<<"  "<<pchoice[5][3]<<"\n";

  wt*=pow(gs2,2);
  if (flags.logcorrect==1)
    wt*=(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI));
  return wt;
}
#if new2
double CWMultijet::Generate2jNonFKLflavourconf_and_eval(double pmg[6][4])
#else
double CWMultijet::Generate2jNonFKLflavourconf_and_eval()
#endif
{
  //  std::cout<<"Yep, 2j non FKL: ";
  double weight=1.;

  if (status!=0) { // if we failed in picking a sensible incoming state
    wt=0.;
    return wt;
  }

#if new2   
  if (aptype==21) {// a=g 
    if (bptype==21) {
      static std::vector<sPMchoice> ggProcesses;
      if (ggProcesses.size()==0) 
	{
	  { // first gg process
	    struct sPMchoice PMtemp;
	    // tpmomshuffle tpmomtemp;
	    PMtemp.MGfunc=smatrixggtepveuxd_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=2;  // 2 flavours
	    
	    PMtemp.MomShuffleVector.push_back(0);
	    PMtemp.MomShuffleVector.push_back(1);
	    
	    ggProcesses.push_back(PMtemp);
	    PMtemp.ptype=false; // process type p
	    ggProcesses.push_back(PMtemp);
	  }
	}
      weight*=PickAndEvalProcAndMom2j(ggProcesses,pmg,&lepton,&alepton);
    }
    else {    //a=g,  b!=g
      if (fabs(bptype)==5)  {
	weight*=0.;
	setstatus(24);
      }
      else if (bptype==2 || bptype==4) {
	static std::vector<sPMchoice> guProcesses;
	if (guProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixugtepvedg_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      guProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(guProcesses,pmg,&lepton,&alepton);
      }  
      else if (bptype>0) {
	static std::vector<sPMchoice> gdProcesses;
	if (gdProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixugtepvedg_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      gdProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(gdProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1 || bptype==-3) {
	static std::vector<sPMchoice> gdxProcesses;
	if (gdxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxgtepveuxg_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      gdxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(gdxProcesses,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> guxProcesses;
	if (guxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxgtepveuxg_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      guxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(guxProcesses,pmg,&lepton,&alepton);
      }
    }
  }
  else if (bptype==21) {  // b=g, a!=g
    if (fabs(aptype)==5) {
      setstatus(24);
      return 0.;
    }
    else if (aptype==2 || aptype==4) { 
      static std::vector<sPMchoice> ugProcesses;
      if (ugProcesses.size()==0) 
	{
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixugtepvedg_;
	    PMtemp.ptype=false; // process type p
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(3);
	    ugProcesses.push_back(PMtemp);
	  }
	}
      weight*=NoPickButEval2j(ugProcesses,pmg,&lepton,&alepton);
    }
    else if (aptype>0) {
      static std::vector<sPMchoice> dgProcesses;
      if (dgProcesses.size()==0) 
	{
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixugtepvedg_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(3);
	    dgProcesses.push_back(PMtemp);
	  }
	}
      weight*=NoPickButEval2j(dgProcesses,pmg,&lepton,&alepton);
    }
    else if (aptype==-1 || aptype==-3) {
      static std::vector<sPMchoice> dxgProcesses;
      if (dxgProcesses.size()==0) 
	{
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxgtepveuxg_;
	    PMtemp.ptype=false; // process type p
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(3);
	    dxgProcesses.push_back(PMtemp);
	  }
	}
      weight*=NoPickButEval2j(dxgProcesses,pmg,&lepton,&alepton);
    }
    else {
      static std::vector<sPMchoice> uxgProcesses;
      if (uxgProcesses.size()==0) 
	{
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxgtepveuxg_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(3);
	    uxgProcesses.push_back(PMtemp);
	  }
	}
      weight*=NoPickButEval2j(uxgProcesses,pmg,&lepton,&alepton);
    }
  }
  else {  //  MARKER - two (anti-)quarks to deal with! 
    if (aptype==bptype) {
      setstatus(24);
      return 0.;       // All configs are in FKL bit
    }
    else if (aptype==-bptype) {    /// MARKER - q qbar same flavour /// 
      if (fabs(aptype)==5) {   //  ME is symmetric under s<-->sx
	static std::vector<sPMchoice> bbxProcesses;
	if (bbxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixssxtepveuxd_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2; // 2 flavours (u dx, c sx) and (ux d, cx s)
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(1);
	      bbxProcesses.push_back(PMtemp);
	      
	      PMtemp.ptype=true; // process type q
	      bbxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom2j(bbxProcesses,pmg,&lepton,&alepton);
      }
      else {                    //  not b bbar so uux-->dux, udx, scx, csx
	if (aptype==2 || aptype==4) {
	  static std::vector<sPMchoice> uuxProcesses;
	  if (uuxProcesses.size()==0) 
	    {
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixuxutepveuxd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(1);
		uuxProcesses.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixdxdtepveuxd_;
		PMtemp.ptype=true; // process type q
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(1);
		uuxProcesses.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixssxtepveuxd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(2);
		PMtemp.MomShuffleVector.push_back(3);
		uuxProcesses.push_back(PMtemp);

		PMtemp.ptype=true; // process type q
		uuxProcesses.push_back(PMtemp);
	      }
	    }
	  weight*=PickAndEvalProcAndMom2j(uuxProcesses,pmg,&lepton,&alepton);
	}
	else if (aptype==1 || aptype==3) { 
	  static std::vector<sPMchoice> ddxProcesses;
	  if (ddxProcesses.size()==0) 
	    {
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixuxutepveuxd_;
		PMtemp.ptype=true; // process type q
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(1);
		ddxProcesses.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixdxdtepveuxd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(1);
		ddxProcesses.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixssxtepveuxd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(2);
		PMtemp.MomShuffleVector.push_back(3);
		ddxProcesses.push_back(PMtemp);

		PMtemp.ptype=true; // process type q
		ddxProcesses.push_back(PMtemp);
	      }
	    }
	  weight*=PickAndEvalProcAndMom2j(ddxProcesses,pmg,&lepton,&alepton);
	}
	else if (aptype==-1 || aptype==-3) {
	  static std::vector<sPMchoice> dxdProcesses;
	  if (dxdProcesses.size()==0) 
	    {
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixuxutepveuxd_;
		PMtemp.ptype=true; // process type q
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(3);
		dxdProcesses.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixdxdtepveuxd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(3);
		dxdProcesses.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixssxtepveuxd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(0);
		PMtemp.MomShuffleVector.push_back(1);
		dxdProcesses.push_back(PMtemp);

		PMtemp.ptype=true; // process type q
		dxdProcesses.push_back(PMtemp);
	      }
	    }
	  weight*=PickAndEvalProcAndMom2j(dxdProcesses,pmg,&lepton,&alepton);
	}
	else {  // aptype==-2, -4
	  static std::vector<sPMchoice> uxuProcesses;
	  if (uxuProcesses.size()==0) 
	    {
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixuxutepveuxd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(3);
		uxuProcesses.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixdxdtepveuxd_;
		PMtemp.ptype=true; // process type q
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(3);
		uxuProcesses.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixssxtepveuxd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(0);
		PMtemp.MomShuffleVector.push_back(1);
		uxuProcesses.push_back(PMtemp);

		PMtemp.ptype=true; // process type q
		uxuProcesses.push_back(PMtemp);
	      }
	    }
	  weight*=PickAndEvalProcAndMom2j(uxuProcesses,pmg,&lepton,&alepton);
	}
      }
    }
    else if (aptype==1) {    // a=d, b!=d,dbar   ///  MARKER a = down //
      if (bptype==2) {  // b=u         
    	weight*=0.;  //  All in FKL bit
	setstatus(24);
      }
      else if (bptype==5) {   
	static std::vector<sPMchoice> dbProcesses;
	if (dbProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dbProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(dbProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==3) {
	static std::vector<sPMchoice> dsProcesses;
	if (dsProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // ds --> us
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dsProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  // ds --> dc
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      dsProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(dsProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==4) { // emit from c
	static std::vector<sPMchoice> dcProcesses;
	if (dcProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // dc --> uc
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dcProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  // dc --> ds
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      dcProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(dcProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-2) {  // dux --> uux, ddx, ssx, gg
	static std::vector<sPMchoice> duxProcesses;
	if (duxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // dux --> uux
	      PMtemp.MGfunc=smatrixudxtepveddx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      duxProcesses.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveuux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      duxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvessx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(1);
	      duxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvegg_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2; 
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      duxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom2j(duxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-5) {
	static std::vector<sPMchoice> dbxProcesses;
	if (dbxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dbxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(dbxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) {    // b is sx 
	static std::vector<sPMchoice> dsxProcesses;
	if (dsxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dsxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      dsxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(dsxProcesses,pmg,&lepton,&alepton);
      }
      else {    // b is cx 
	static std::vector<sPMchoice> dcxProcesses;
	if (dcxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dcxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      dcxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(dcxProcesses,pmg,&lepton,&alepton);
      }
    }   
    else if (aptype==2) { // MARKER a=u //
      if (bptype==1) {  // b=d         
    	weight*=0.;  //  All in FKL bit
	setstatus(24);
      }
      else if (bptype==5) {   
	static std::vector<sPMchoice> ubProcesses;
	if (ubProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      ubProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(ubProcesses,pmg,&lepton,&alepton);
      } 
      else if (bptype==3) {
	static std::vector<sPMchoice> usProcesses;
	if (usProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      usProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      usProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(usProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==4) { // emit from c
	static std::vector<sPMchoice> ucProcesses;
	if (ucProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      ucProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      ucProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(ucProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) {  // udx --> ddx, uux, ssx, gg
	static std::vector<sPMchoice> udxProcesses;
	if (udxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepveddx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      udxProcesses.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveuux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      udxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvessx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(1);
	      udxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvegg_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2; 
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      udxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom2j(udxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-5) {
	static std::vector<sPMchoice> ubxProcesses;
	if (ubxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      ubxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(ubxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) {    // b is sx 
	static std::vector<sPMchoice> usxProcesses;
	if (usxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      usxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      usxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(usxProcesses,pmg,&lepton,&alepton);
      }
      else {    // b is cx 
	static std::vector<sPMchoice> ucxProcesses;
	if (ucxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      ucxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      ucxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(ucxProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==3) { // a=s  
      if (bptype==4) {  // b=u         
    	weight*=0.;  //  All in FKL bit
	setstatus(24);
      }
      else if (bptype==5) {   
	static std::vector<sPMchoice> sbProcesses;
	if (sbProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      sbProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(sbProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==1) {
	static std::vector<sPMchoice> sdProcesses;
	if (sdProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // sd --> cd
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      sdProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  // sd --> su
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      sdProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(sdProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype>0) { // bptype = 2
	static std::vector<sPMchoice> suProcesses;
	if (suProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // su --> cu
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      suProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  // su --> sd
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      suProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(suProcesses,pmg,&lepton,&alepton);
      }  
      else if (bptype==-4) {  // scx --> ccx, ssx, ddx, gg
	static std::vector<sPMchoice> scxProcesses;
	if (scxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // dux --> uux
	      PMtemp.MGfunc=smatrixudxtepveddx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      scxProcesses.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveuux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      scxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvessx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(1);
	      scxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvegg_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2; 
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      scxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom2j(scxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-5) {
	static std::vector<sPMchoice> sbxProcesses;
	if (sbxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      sbxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(sbxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) {    // b is dx 
	static std::vector<sPMchoice> sdxProcesses;
	if (sdxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      sdxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      sdxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(sdxProcesses,pmg,&lepton,&alepton);
      }
      else {    // b is ux 
	static std::vector<sPMchoice> suxProcesses;
	if (suxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      suxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      suxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(suxProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==4) { // a=c 
      if (bptype==3) {  // b=s         
    	weight*=0.;  //  All in FKL bit
	setstatus(24);
      }
      else if (bptype==5) {     
	static std::vector<sPMchoice> cbProcesses;
	if (cbProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cbProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(cbProcesses,pmg,&lepton,&alepton);
      } 
      else if (bptype==1) {
	static std::vector<sPMchoice> cdProcesses;
	if (cdProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cdProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      cdProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(cdProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==2) { 
	static std::vector<sPMchoice> cuProcesses;
	if (cuProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cuProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      cuProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(cuProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) {  // csx --> ssx, ccx, ddx, gg
	static std::vector<sPMchoice> csxProcesses;
	if (csxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepveddx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      csxProcesses.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveuux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      csxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvessx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(1);
	      csxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvegg_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2; 
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      csxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom2j(csxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-5) {
	static std::vector<sPMchoice> cbxProcesses;
	if (cbxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cbxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(cbxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) {    // b is dx 
	static std::vector<sPMchoice> cdxProcesses;
	if (cdxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cdxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      cdxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(cdxProcesses,pmg,&lepton,&alepton);
      }
      else {    // b is ux 
	static std::vector<sPMchoice> cuxProcesses;
	if (cuxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cuxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      cuxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(cuxProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype>0) { // a=b, b \ne b
      if (bptype==2 || bptype==4) {
	static std::vector<sPMchoice> buProcesses;
	if (buProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      buProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(buProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype >0 ) {
	static std::vector<sPMchoice> bdProcesses;
	if (bdProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixustepveds_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      bdProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(bdProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1 || bptype==-3) {
	static std::vector<sPMchoice> bdxProcesses;
	if (bdxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      bdxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(bdxProcesses,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> buxProcesses;
	if (buxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      buxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(buxProcesses,pmg,&lepton,&alepton);
      }
    }
    // ///  Now go for a = anti-quark   --- MARKER a=qbar  ///  
    else if (aptype==-1) {    // a=dbar, b!=d   
      if (bptype==-2) {  // b=ux         
    	weight*=0.;  //  All in FKL bit
	setstatus(24);
      }
      else if (bptype==-5) {   
	static std::vector<sPMchoice> dxbxProcesses;
	if (dxbxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dxbxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(dxbxProcesses,pmg,&lepton,&alepton);
      } 
      else if (bptype==-4) {
	static std::vector<sPMchoice> dxcxProcesses;
	if (dxcxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dxcxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      dxcxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(dxcxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) { 
	static std::vector<sPMchoice> dxsxProcesses;
	if (dxsxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dxsxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      dxsxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(dxsxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==2) {  // dxu --> ddx, uux, ssx, gg
	static std::vector<sPMchoice> dxuProcesses;
	if (dxuProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepveddx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      dxuProcesses.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveuux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      dxuProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvessx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      dxuProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvegg_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2; 
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      dxuProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom2j(dxuProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==5) {
	static std::vector<sPMchoice> dxbProcesses;
	if (dxbProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dxbProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(dxbProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==4) {    // b is c
	static std::vector<sPMchoice> dxcProcesses;
	if (dxcProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      dxcProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dxcProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(dxcProcesses,pmg,&lepton,&alepton);
      }
      else {    // b is s 
	static std::vector<sPMchoice> dxsProcesses;
	if (dxsProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      dxsProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      dxsProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(dxsProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==-2) {    // a=ubar, b!=u  
      if (bptype==-1) {  // b=dx         
    	weight*=0.;  //  All in FKL bit
	setstatus(24);
      }
      else if (bptype==-5) {   
	static std::vector<sPMchoice> uxbxProcesses;
	if (uxbxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      uxbxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(uxbxProcesses,pmg,&lepton,&alepton);
      } 
      else if (bptype==-4) {
	static std::vector<sPMchoice> uxcxProcesses;
	if (uxcxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      uxcxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      uxcxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(uxcxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) { 
	static std::vector<sPMchoice> uxsxProcesses;
	if (uxsxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      uxsxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      uxsxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(uxsxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==1) {  // uxd --> uux, ddx, ssx, gg
	static std::vector<sPMchoice> uxdProcesses;
	if (uxdProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepveddx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      uxdProcesses.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveuux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      uxdProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvessx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      uxdProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvegg_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2; 
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      uxdProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom2j(uxdProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==5) {
	static std::vector<sPMchoice> uxbProcesses;
	if (uxbProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      uxbProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(uxbProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==4) {    // b is c
	static std::vector<sPMchoice> uxcProcesses;
	if (uxcProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      uxcProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      uxcProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(uxcProcesses,pmg,&lepton,&alepton);
      }
      else {    // b is s 
	static std::vector<sPMchoice> uxsProcesses;
	if (uxsProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      uxsProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      uxsProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(uxsProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==-3) {    // a=sbar, b!=u   
      if (bptype==-4) {  // b=cx         
    	weight*=0.;  //  All in FKL bit
	setstatus(24);
      }
      else if (bptype==-5) {   
	static std::vector<sPMchoice> sxbxProcesses;
	if (sxbxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      sxbxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(sxbxProcesses,pmg,&lepton,&alepton);
      } 
      else if (bptype==-2) {
	static std::vector<sPMchoice> sxuxProcesses;
	if (sxuxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      sxuxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      sxuxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(sxuxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) { 
	static std::vector<sPMchoice> sxdxProcesses;
	if (sxdxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      sxdxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      sxdxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(sxdxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==4) {  // sxc --> ssx, ccx, ddx, gg
	static std::vector<sPMchoice> sxcProcesses;
	if (sxcProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepveddx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      sxcProcesses.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveuux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      sxcProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvessx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      sxcProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvegg_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2; 
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      sxcProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom2j(sxcProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==5) {
	static std::vector<sPMchoice> sxbProcesses;
	if (sxbProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      sxbProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(sxbProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==2) {    // b is u
	static std::vector<sPMchoice> sxuProcesses;
	if (sxuProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      sxuProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      sxuProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(sxuProcesses,pmg,&lepton,&alepton);
      }
      else {    // b is d 
	static std::vector<sPMchoice> sxdProcesses;
	if (sxdProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      sxdProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      sxdProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(sxdProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==-4) {    // a=cbar, b!=u
      if (bptype==-3) {  // b=sx
    	weight*=0.;  //  All in FKL bit
	setstatus(24);
      }
      else if (bptype==-5) {   
	static std::vector<sPMchoice> cxbxProcesses;
	if (cxbxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cxbxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(cxbxProcesses,pmg,&lepton,&alepton);
      } 
      else if (bptype==-2) {
	static std::vector<sPMchoice> cxuxProcesses;
	if (cxuxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cxuxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      cxuxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(cxuxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) {  // b = dx
	static std::vector<sPMchoice> cxdxProcesses;
	if (cxdxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cxdxProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      cxdxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(cxdxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==3) {  // cxs --> ccx, ssx, ddx, gg
	static std::vector<sPMchoice> cxsProcesses;
	if (cxsProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepveddx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      cxsProcesses.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveuux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      cxsProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvessx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      cxsProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvegg_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2; 
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      cxsProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom2j(cxsProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==5) {
	static std::vector<sPMchoice> cxbProcesses;
	if (cxbProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cxbProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(cxbProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==2) {    // b is u
	static std::vector<sPMchoice> cxuProcesses;
	if (cxuProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      cxuProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cxuProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(cxuProcesses,pmg,&lepton,&alepton);
      }
      else {    // b is d 
	static std::vector<sPMchoice> cxdProcesses;
	if (cxdProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      cxdProcesses.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(3);
	      cxdProcesses.push_back(PMtemp);
	    }
	  }
	weight*=PickProcAndEval2j(cxdProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==-5) {    // a=bbar, b!=b
      if (bptype==2 || bptype==4) {
	static std::vector<sPMchoice> bxuProcesses;
	if (bxuProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      bxuProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(bxuProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype >0 ) {
	static std::vector<sPMchoice> bxdProcesses;
	if (bxdProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      bxdProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(bxdProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1 || bptype==-3) {
	static std::vector<sPMchoice> bxdxProcesses;
	if (bxdxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      bxdxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(bxdxProcesses,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> bxuxProcesses;
	if (bxuxProcesses.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxsxtepveuxsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      bxuxProcesses.push_back(PMtemp);
	    }
	  }
	weight*=NoPickButEval2j(bxuxProcesses,pmg,&lepton,&alepton);
      }
    }
    else  {
      std::cerr<<"Watch out - unanticipated parton configuration!"<<std::endl;
    }
  }
#else
  // setup arrays and LH information:
  FREAL meout(0.),p[6][4],p12[6][4],p22[6][4],p32[6][4],q[6][4],q12[6][4],q22[6][4],q32[6][4];//,(*pchoice)[4];
  //  void (*smatrixchoice)(FREAL * P1, FREAL * ANS);

  p[0][0]=pb.e();
  p[0][1]=pb.x();
  p[0][2]=pb.y();
  p[0][3]=pb.z();
  
  p[1][0]=pa.e();
  p[1][1]=pa.x();
  p[1][2]=pa.y();
  p[1][3]=pa.z();
  
  p[2][0]=pd1.e();   // Charged lepton
  p[2][1]=pd1.x();
  p[2][2]=pd1.y();
  p[2][3]=pd1.z();
  
  p[3][0]=pd2.e();   // (anti) neutrino
  p[3][1]=pd2.x();
  p[3][2]=pd2.y();
  p[3][3]=pd2.z();
  
  p[4][0]=(partonmomenta.back()).e();   // more forward jet
  p[4][1]=(partonmomenta.back()).x();
  p[4][2]=(partonmomenta.back()).y();
  p[4][3]=(partonmomenta.back()).z();
  		 
  p[5][0]=(partonmomenta.front()).e();  // more backward jet
  p[5][1]=(partonmomenta.front()).x();
  p[5][2]=(partonmomenta.front()).y();
  p[5][3]=(partonmomenta.front()).z();
  

  // Useful array of momenta ordered according to rapidity configuration
  for (int ii=0;ii<4;ii++) {
    q[0][ii]=p[0][ii]; 
    q[1][ii]=p[1][ii]; 
    q[2][ii]=p[3][ii]; 
    q[3][ii]=p[2][ii]; 
    q[4][ii]=p[4][ii]; 
    q[5][ii]=p[5][ii]; 
    
    p12[0][ii]=p[1][ii];
    p12[1][ii]=p[0][ii];
    p12[2][ii]=p[2][ii];
    p12[3][ii]=p[3][ii];
    p12[4][ii]=p[5][ii];
    p12[5][ii]=p[4][ii];
    
    q12[0][ii]=p12[0][ii]; 
    q12[1][ii]=p12[1][ii]; 
    q12[2][ii]=p12[3][ii]; 
    q12[3][ii]=p12[2][ii]; 
    q12[4][ii]=p12[4][ii]; 
    q12[5][ii]=p12[5][ii]; 
    
    p22[0][ii]=p[0][ii];
    p22[1][ii]=p[1][ii];
    p22[2][ii]=p[2][ii];
    p22[3][ii]=p[3][ii];
    p22[4][ii]=p[5][ii];
    p22[5][ii]=p[4][ii];
    
    q22[0][ii]=p22[0][ii]; 
    q22[1][ii]=p22[1][ii]; 
    q22[2][ii]=p22[3][ii]; 
    q22[3][ii]=p22[2][ii]; 
    q22[4][ii]=p22[4][ii]; 
    q22[5][ii]=p22[5][ii]; 
    
    p32[0][ii]=p[1][ii];
    p32[1][ii]=p[0][ii];
    p32[2][ii]=p[2][ii];
    p32[3][ii]=p[3][ii];
    p32[4][ii]=p[4][ii];
    p32[5][ii]=p[5][ii];	

    q32[0][ii]=p32[0][ii]; 
    q32[1][ii]=p32[1][ii]; 
    q32[2][ii]=p32[3][ii]; 
    q32[3][ii]=p32[2][ii]; 
    q32[4][ii]=p32[4][ii]; 
    q32[5][ii]=p32[5][ii]; 
  }

  // p22, p32, q22 and q32 are not FKL.
  // p   = b(+) a(-) --> e nu + -,   p12 = a(-) b(+) --> e nu - +
  // p22 = b(+) a(-) --> e nu - +,   p32 = a(-) b(+) --> e nu + -
  // q   = b(+) a(-) --> nu e + -,   q12 = a(-) b(+) --> nu e - +
  // q22 = b(+) a(-) --> nu e - +,   q32 = a(-) b(+) --> nu e + -

  // choose a matrix element to calculate and set the relevant flavours

  //  std::cout<<"Particles:  "<<aptype<<" and "<<bptype<<std::endl;

  double idpartfact=1.;     //  Factor to take account of identical final states

  if (aptype==21) {// a=g 
    if (bptype==21) {
      double process=ran.flat();   // 2 momentum configs
      idpartfact*=2.;    //  This is 1/(1/2) for the length of interval in selection
      double flavcomb=ran.flat();
      idpartfact*=4.;    //  2 types, 2 states in each (ux d, cx s) and (u dx,c sx)
      if (process<0.5) {
	if (flavcomb<0.5) {
	  smatrixggtepveuxd_(p22[0],&meout);
	  alepton.PID=-11;
	  alepton.p=pd1;
	  lepton.PID=12;
	  lepton.p=pd2;
	}
	else {
	  smatrixggtepveuxd_(q22[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
	}
      }
      else {
	if (flavcomb<0.5) {
	  smatrixggtepveuxd_(p32[0],&meout);
	  alepton.PID=-11;
	  alepton.p=pd1;
	  lepton.PID=12;
	  lepton.p=pd2;
	}
	else {
	  smatrixggtepveuxd_(q32[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
	}
      }
    }
    else {    //a=g,  b!=g
      if (fabs(bptype)==5)  
	meout=0.;
      else if (bptype==2 || bptype==4) {
	smatrixugtepvedg_(p22[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else if (bptype>0) {
	smatrixugtepvedg_(q22[0],&meout);
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
      else if (bptype==-1 || bptype==-3) {
	smatrixdxgtepveuxg_(p22[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else {
	smatrixdxgtepveuxg_(q22[0],&meout);
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
    }
  }
  else if (bptype==21) {  // b=g, a!=g
    if (fabs(aptype)==5) {
      setstatus(24);
      return 0.;
    }
    else if (aptype==2 || aptype==4) { 
      smatrixugtepvedg_(p32[0],&meout);
      alepton.PID=-11;
      alepton.p=pd1;
      lepton.PID=12;
      lepton.p=pd2;
    }
    else if (aptype>0) {
      smatrixugtepvedg_(q32[0],&meout);
      alepton.PID=-12;
      alepton.p=pd2;
      lepton.PID=11;
      lepton.p=pd1;
    }
    else if (aptype==-1 || aptype==-3) {
      smatrixdxgtepveuxg_(p32[0],&meout);
      alepton.PID=-11;
      alepton.p=pd1;
      lepton.PID=12;
      lepton.p=pd2;
    }
    else {
      smatrixdxgtepveuxg_(q32[0],&meout);
      alepton.PID=-12;
      alepton.p=pd2;
      lepton.PID=11;
      lepton.p=pd1;
    }
  }
  else {  //  MARKER - two (anti-)quarks to deal with!
    if (aptype==bptype) {
      setstatus(24);
      return 0.;       // All configs are in FKL bit
    }
    else if (aptype==-bptype) {    /// MARKER - q qbar same flavour ///
      if (fabs(aptype)==5) {   //  ME is symmetric under s<-->sx
    	idpartfact*=2.;  //  2 in each flavour group: (u dx, c sx) and (ux d, cx s)
    	double picklepton=ran.flat();    // 50% e+ nu, 50% e- nubar, picks flav group
    	double process=ran.flat();       // Picks momentum config
    	idpartfact*=4.; // *2 for picking process, *2 for picking lepton
    	if (picklepton<0.5) {
	  alepton.PID=-11;
	  alepton.p=pd1;
	  lepton.PID=12;
	  lepton.p=pd2;
    	  if (process<0.5)
    	    smatrixssxtepveuxd_(p12[0],&meout);
    	  else
    	    smatrixssxtepveuxd_(p32[0],&meout);
	  //	  cout<<"r1:"<<meout<<"\n";
    	}
    	else {
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
    	  if (process<0.5)
    	    smatrixssxtepveuxd_(q12[0],&meout);
    	  else
    	    smatrixssxtepveuxd_(q32[0],&meout);
	  //	  cout<<"r2:"<<meout<<"\n";
    	}
      }
      else {                    //  not b bbar so uux-->dux, udx, scx, csx
    	double process=ran.flat();  
    	idpartfact*=4.;
    	if (process<0.5) {   // same flavour outgoing
    	  if (aptype==2 || aptype==4) {  // backward particle is q, forward qbar
    	    if (process<0.25) {
    	      smatrixuxutepveuxd_(p22[0],&meout);  // uux->e+ nu dux
	      alepton.PID=-11;
	      alepton.p=pd1;
	      lepton.PID=12;
	      lepton.p=pd2;
	    }
    	    else {
    	      smatrixdxdtepveuxd_(q22[0],&meout);  // uux->e- nux udx
	      alepton.PID=-12;
	      alepton.p=pd2;
	      lepton.PID=11;
	      lepton.p=pd1;
	    }
    	  }
    	  else if (aptype==1 || aptype==3) {
    	    if (process<0.25) {
    	      smatrixuxutepveuxd_(q22[0],&meout);  // ddx->e- nux udx
	      alepton.PID=-12;
	      alepton.p=pd2;
	      lepton.PID=11;
	      lepton.p=pd1;
	    }
    	    else {
    	      smatrixdxdtepveuxd_(p22[0],&meout);  // ddx->e+ nu dux
	      alepton.PID=-11;
	      alepton.p=pd1;
	      lepton.PID=12;
	      lepton.p=pd2;
	    }
    	  }
    	  else if (aptype==-1 || aptype==-3) {
    	    if (process<0.25) {
    	      smatrixuxutepveuxd_(q32[0],&meout);  // dxd->e- nux udx
	      alepton.PID=-12;
	      alepton.p=pd2;
	      lepton.PID=11;
	      lepton.p=pd1;
	    }
    	    else {
    	      smatrixdxdtepveuxd_(p32[0],&meout);  // dxd->e+ nu dux
	      alepton.PID=-11;
	      alepton.p=pd1;
	      lepton.PID=12;
	      lepton.p=pd2;
	    }
    	  }
    	  else {  // aptype==-2, -4
    	    if (process<0.25) {
    	      smatrixuxutepveuxd_(p32[0],&meout);  // uxu->e+ nu dux
	      alepton.PID=-11;
	      alepton.p=pd1;
	      lepton.PID=12;
	      lepton.p=pd2;
	    }
    	    else {
    	      smatrixdxdtepveuxd_(q32[0],&meout);  // dxu->e- nux dux
	      alepton.PID=-12;
	      alepton.p=pd2;
	      lepton.PID=11;
	      lepton.p=pd1;
	    }
    	  }
    	}
    	else {
    	  double mom=ran.flat();
    	  idpartfact*=2.;
    	  if (process<0.75) { // uux->e+ nu scx
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	    if (mom <0.5) 
    	      smatrixssxtepveuxd_(p32[0],&meout);  
    	    else
    	      smatrixssxtepveuxd_(p12[0],&meout);
    	  }
    	  else {   // uux -> e- nux c sx
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
    	    if (mom <0.5) 
    	      smatrixssxtepveuxd_(q32[0],&meout);  
    	    else
    	      smatrixssxtepveuxd_(q12[0],&meout);
    	  }
    	}
      }
    }
    else if (aptype==1) {    // a=d, b!=d,dbar   ///  MARKER a = down //
      if (bptype==2) {  // b=u
    	meout=0.;  //  All in FKL bit
      }
      else if (bptype==5) {
    	smatrixustepveds_(q32[0],&meout);   // Don't emit from b
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
      else if (bptype>2) {
    	double r1=ran.flat();
    	idpartfact*=2.;
    	if (r1<0.5) {   // emit from d
    	  smatrixustepveds_(q32[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
    	}
    	else {   
    	  if (bptype==4) { // emit from c
    	    smatrixustepveds_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else         {   // emit from s 
    	    smatrixustepveds_(q22[0],&meout);
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
    	}
      }
      else if (bptype==-2) {
    	int process=int(4.*ran.flat())+1; // 1, 2 or 3 -- missing configs
    	idpartfact*=4.;
    	if (process==1) {
    	  smatrixudxtepveddx_(q32[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
	}
    	else if (process==2) {
    	  smatrixudxtepveuux_(q32[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
	}
    	else if (process==3) {
    	  idpartfact*=6.;    // *2 for 2 momm configs, *3 flavours
    	  smatrixdxutepvessx_(q22[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
    	}
    	else {
    	  idpartfact*=2.;    // *2 for gg final state
    	  smatrixdxutepvegg_(q22[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
    	}
      }
      else if (bptype==-5) {
    	smatrixusxtepvedsx_(q32[0],&meout);
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
      else {    // b is sx or cx
    	double r2=ran.flat();
    	idpartfact*=2.;
    	if (r2<0.5) {    // emit from d
    	  smatrixusxtepvedsx_(q32[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
	}
    	else {
    	  if (bptype==-3) {
    	    smatrixdxstepveuxs_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {
    	    smatrixdxstepveuxs_(q22[0],&meout);
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
    	}
      }
    }
    else if (aptype==2) { // MARKER a=u //
      if (bptype==1) {  // b=d
    	meout=0.;  //  All in FKL bit
      }
      else if (bptype==5) {
    	smatrixustepveds_(p32[0],&meout);   // Don't emit from b
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else if (bptype>2) {
    	double r1=ran.flat();    //  Choose where emit from
    	idpartfact*=2.;
    	if (r1<0.5) {   // emit from u
    	  smatrixustepveds_(p32[0],&meout);
	  alepton.PID=-11;
	  alepton.p=pd1;
	  lepton.PID=12;
	  lepton.p=pd2;
    	}
    	else {   
    	  if (bptype==4)  { // emit from c
    	    smatrixustepveds_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {            // emit from s 
    	    smatrixustepveds_(q22[0],&meout);
	    alepton.p=pd2;
	    alepton.PID=-12;
	    lepton.p=pd1;
	    lepton.PID=11;
	  }
    	}
      }
      else if (bptype==-1) {
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
    	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4 -- missing configs
    	idpartfact*=4.;
    	if (process==1) 
    	    smatrixudxtepveddx_(p32[0],&meout);
	else if (process==2) 
    	    smatrixudxtepveuux_(p32[0],&meout);
    	else if (process==3) {
    	  idpartfact*=6.;    // *2 for 2 momm configs, *3 flavours
    	  smatrixdxutepvessx_(p22[0],&meout);
	}
    	else {
    	  idpartfact*=2.;    // *2 for gg final state
    	  smatrixdxutepvegg_(p22[0],&meout);
    	}
      }
      else if (bptype==-5) {
    	smatrixusxtepvedsx_(p32[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else {    // b is sx or cx
    	double r2=ran.flat();
    	idpartfact*=2.;
    	if (r2<0.5) {    // emit from u
    	  smatrixusxtepvedsx_(p32[0],&meout);
	  alepton.PID=-11;
	  alepton.p=pd1;
	  lepton.PID=12;
	  lepton.p=pd2;
    	}
    	else {
    	  if (bptype==-3) {
    	    smatrixdxstepveuxs_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {
    	    smatrixdxstepveuxs_(q22[0],&meout);
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
	}
      }
    }
    else if (aptype==3) { // a=s
      if (bptype==4) {  // b=c
    	meout=0.;  //  All in FKL bit
      }
      else if (bptype==5)  {
    	smatrixustepveds_(q32[0],&meout);   // Don't emit from b
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
      else if (bptype>0) {          // b = d,u
    	double r1=ran.flat();
    	idpartfact*=2.;
    	if (r1<0.5) {   // emit from s
    	  smatrixustepveds_(q32[0],&meout);
	  alepton.p=pd2;
	  alepton.PID=-12;
	  lepton.p=pd1;
	  lepton.PID=11;
    	}
    	else {   
    	  if (bptype==2)  { // emit from u
    	    smatrixustepveds_(p22[0],&meout);
	    alepton.p=pd1;
	    alepton.PID=-11;
	    lepton.p=pd2;
	    lepton.PID=12;
	  }
    	  else       {     // emit from d 
    	    smatrixustepveds_(q22[0],&meout);
	    alepton.p=pd2;
	    alepton.PID=-12;
	    lepton.p=pd1;
	    lepton.PID=11;
	  }
    	}
      }
      else if (bptype==-4) {
	alepton.p=pd2;
	alepton.PID=-12;
	lepton.p=pd1;
	lepton.PID=11;
    	int process=int(4.*ran.flat())+1; // 1, 2 or 3 -- missing configs
    	idpartfact*=4.;
    	if (process==1) 
	  smatrixudxtepveddx_(q32[0],&meout);
    	else if (process==2)
	  smatrixudxtepveuux_(q32[0],&meout);
    	else if (process==3) {
    	  idpartfact*=6.;    // *2 for 2 momm configs, *3 flavours
    	  smatrixdxutepvessx_(q22[0],&meout);
    	}
    	else {
    	  idpartfact*=2.;    // *2 for gg final state
    	  smatrixdxutepvegg_(q22[0],&meout);
    	}
      }
      else if (bptype==-5) {
	alepton.p=pd2;
	alepton.PID=-12;
	lepton.p=pd1;
	lepton.PID=11;
    	smatrixusxtepvedsx_(q32[0],&meout);
      }
      else {    // b is dx or ux
    	double r2=ran.flat();
    	idpartfact*=2.;
    	if (r2<0.5) {    // emit from s
    	  smatrixusxtepvedsx_(q32[0],&meout);
	  alepton.p=pd2;
	  alepton.PID=-12;
	  lepton.p=pd1;
	  lepton.PID=11;
    	}
    	else {
    	  if (bptype==-1) {
    	    smatrixdxstepveuxs_(p22[0],&meout);
	    alepton.p=pd1;
	    alepton.PID=-11;
	    lepton.p=pd2;
	    lepton.PID=12;
	  }
    	  else {
    	    smatrixdxstepveuxs_(q22[0],&meout);
	    alepton.p=pd2;
	    alepton.PID=-12;
	    lepton.p=pd1;
	    lepton.PID=11;
	  }
	}
      }
    }
    else if (aptype==4) { // a=c
      if (bptype==3) {  // b=s
    	meout=0.;  //  All in FKL bit
      }
      else if (bptype==5) {
    	smatrixustepveds_(p32[0],&meout);   // Don't emit from b
	alepton.p=pd1;
	alepton.PID=-11;
	lepton.p=pd2;
	lepton.PID=12;
      }
      else if (bptype>0) {
    	double r1=ran.flat();    //  Choose where emit from
    	idpartfact*=2.;
    	if (r1<0.5) {   // emit from c
    	  smatrixustepveds_(p32[0],&meout);
	  //	  cout<<" c: "<<meout<<"\n";
	  alepton.PID=-11;
	  alepton.p=pd1;
	  lepton.PID=12;
	  lepton.p=pd2;
    	}
    	else {   
    	  if (bptype==2)  {// emit from u
    	    smatrixustepveds_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else         {   // emit from d 
    	    smatrixustepveds_(q22[0],&meout);
	    //	    cout<<" d: "<<meout<<"\n";
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
    	}
      }
      else if (bptype==-3) {
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
    	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4 -- missing configs
    	idpartfact*=4.;
    	if (process==1)
    	    smatrixudxtepveddx_(p32[0],&meout);
    	else if (process==2)
    	    smatrixudxtepveuux_(p32[0],&meout);
    	else if (process==3) {
    	  idpartfact*=6.;    // *2 for 2 momm configs, *3 flavours
    	  smatrixdxutepvessx_(p22[0],&meout);
    	}
    	else {
    	  idpartfact*=2.;    // *2 for gg final state
    	  smatrixdxutepvegg_(p22[0],&meout);
    	}
      }
      else if (bptype==-5)  {
    	smatrixusxtepvedsx_(p32[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else {    // b is dx or ux
    	double r2=ran.flat();
    	idpartfact*=2.;
    	if (r2<0.5) {    // emit from c
    	  smatrixusxtepvedsx_(p32[0],&meout);
	  alepton.PID=-11;
	  alepton.p=pd1;
	  lepton.PID=12;
	  lepton.p=pd2;
    	}
    	else {
    	  if (bptype==-1) {
    	    smatrixdxstepveuxs_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {
    	    smatrixdxstepveuxs_(q22[0],&meout);
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
	}
      }
    }
    else if (aptype>0) { // a=b
      if (bptype==2 || bptype==4)   {  // b = u,c
    	smatrixustepveds_(p22[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else if (bptype>0) {
    	smatrixustepveds_(q22[0],&meout);
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
      else if (bptype==-1 || bptype==-3) {      // b = dbar, sbar
    	smatrixdxstepveuxs_(p22[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else {
    	smatrixdxstepveuxs_(q22[0],&meout);
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
    }
    // ///  Now go for a = anti-quark   --- MARKER a=qbar  ///
    else if (aptype==-1) {    // a=dbar, b!=d
      if (bptype==-2)   // b=ubar
    	meout=0.;  // All in FKL
      else if (bptype==-5) {
    	smatrixdxsxtepveuxsx_(p32[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else if (bptype<-2) {
    	double r1=ran.flat();
    	idpartfact*=2.;
    	if (r1<0.5)  {      // Emit from dbar
    	  smatrixdxsxtepveuxsx_(p32[0],&meout);
	  alepton.PID=-11;
	  alepton.p=pd1;
	  lepton.PID=12;
	  lepton.p=pd2;
	}
    	else {
    	  if (bptype==-3) {
    	    smatrixdxsxtepveuxsx_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {
    	    smatrixdxsxtepveuxsx_(q22[0],&meout);
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
    	}
      }
      else if (bptype==2) {
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
    	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4 -- missing configs
    	idpartfact*=4.;
    	if (process==1) 
    	  smatrixudxtepveddx_(p22[0],&meout);
    	else if (process==2)
    	  smatrixudxtepveuux_(p22[0],&meout);
    	else if (process==3) {
    	  idpartfact*=6.;   // 2 final state momm configs, 3 flavours
    	  smatrixdxutepvessx_(p32[0],&meout);//ME symmetric under interchange of s <--> sx
    	}
    	else {
    	  idpartfact*=2.;   // *2 for gg final state
    	  smatrixdxutepvegg_(p32[0],&meout);
    	}
      }
      else if (bptype==5) {
    	smatrixdxstepveuxs_(p32[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else {  // a=dbar, b= s,c
    	double r1=ran.flat();
    	idpartfact*=2.;
    	if (r1<0.5) {
    	  smatrixdxstepveuxs_(p32[0],&meout);
	  alepton.PID=-11;
	  alepton.p=pd1;
	  lepton.PID=12;
	  lepton.p=pd2;
	}
    	else {
    	  if (bptype==4) {
    	    smatrixusxtepvedsx_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {
    	    smatrixusxtepvedsx_(q22[0],&meout);
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
    	}
      }
    }
    else if (aptype==-2) {    // a=ubar, b!=u
      if (bptype==-1)   // b=dbar
    	meout=0.;  // All in FKL
      else if (bptype==-5) {
    	smatrixdxsxtepveuxsx_(q32[0],&meout);
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
      else if (bptype<-2) {
    	double r1=ran.flat();
    	idpartfact*=2.;
    	if (r1<0.5)    {    // Emit from ubar
    	  smatrixdxsxtepveuxsx_(q32[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
	}
    	else {
    	  if (bptype==-3) {
    	    smatrixdxsxtepveuxsx_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {
    	    smatrixdxsxtepveuxsx_(q22[0],&meout);
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
    	}
      }
      else if (bptype==1) {
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
    	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4 -- missing configs
    	idpartfact*=4.;
    	if (process==1) 
    	  smatrixudxtepveddx_(q22[0],&meout);
    	else if (process==2)
    	  smatrixudxtepveuux_(q22[0],&meout);
    	else if (process==3) {
    	  idpartfact*=6.;   // 2 final state momm configs, 3 flavours
    	  smatrixdxutepvessx_(q32[0],&meout);//ME symmetric under interchange of s <--> sx
    	}
    	else {
    	  idpartfact*=2.;   // *2 for gg final state
    	  smatrixdxutepvegg_(q32[0],&meout);
    	}
      }
      else if (bptype==5) {
    	smatrixdxstepveuxs_(q32[0],&meout);
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
      else {  // a=ubar, b= s,c
    	double r1=ran.flat();
    	idpartfact*=2.;
    	if (r1<0.5) {// emit from a
    	  smatrixdxstepveuxs_(q32[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
	}
    	else {
    	  if (bptype==4) {
    	    smatrixusxtepvedsx_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {
    	    smatrixusxtepvedsx_(q22[0],&meout);
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
    	}
      }
    }
    else if (aptype==-3) {    // a=sbar, b!=u
      if (bptype==-4)   // b=cbar
    	meout=0.;  // All in FKL
      else if (bptype==-5) {
    	smatrixdxsxtepveuxsx_(p32[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else if (bptype<0) {
    	double r1=ran.flat();
    	idpartfact*=2.;
    	if (r1<0.5)   {     // Emit from sbar
    	  smatrixdxsxtepveuxsx_(p32[0],&meout);
	  alepton.PID=-11;
	  alepton.p=pd1;
	  lepton.PID=12;
	  lepton.p=pd2;
	}
    	else {
    	  if (bptype==-1) {
    	    smatrixdxsxtepveuxsx_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {
    	    smatrixdxsxtepveuxsx_(q22[0],&meout);
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
    	}
      }
      else if (bptype==4) {
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
    	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4 -- missing configs
    	idpartfact*=4.;
    	if (process==1) 
    	  smatrixudxtepveddx_(p22[0],&meout);
    	else if (process==2)
    	  smatrixudxtepveuux_(p22[0],&meout);
    	else if (process==3) {
    	  idpartfact*=6.;   // 2 final state momm configs, 3 flavours
    	  smatrixdxutepvessx_(p32[0],&meout);//ME symmetric under interchange of s <--> sx
    	}
    	else {
    	  idpartfact*=2.;   // *2 for gg final state
    	  smatrixdxutepvegg_(p32[0],&meout);
    	}
      }
      else if (bptype==5) {
    	smatrixdxstepveuxs_(p32[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else {  // a=sbar, b= d,u
    	double r1=ran.flat();
    	idpartfact*=2.;
    	if (r1<0.5) {
    	  smatrixdxstepveuxs_(p32[0],&meout);
	  alepton.PID=-11;
	  alepton.p=pd1;
	  lepton.PID=12;
	  lepton.p=pd2;
	}
    	else {
    	  if (bptype==2) {
    	    smatrixusxtepvedsx_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {
    	    smatrixusxtepvedsx_(q22[0],&meout);
	    alepton.p=pd2;
	    alepton.PID=-12;
	    lepton.p=pd1;
	    lepton.PID=11;
	  }
    	}
      }
    }
    else if (aptype==-4) {    // a=cbar, b!=u
      if (bptype==-3)   // b=sbar
    	meout=0.;  // All in FKL
      else if (bptype==-5) {
    	smatrixdxsxtepveuxsx_(q32[0],&meout);
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
      else if (bptype<0) {
    	double r1=ran.flat();
    	idpartfact*=2.;
    	if (r1<0.5)    {    // Emit from ubar
    	  smatrixdxsxtepveuxsx_(q32[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
	}
    	else {
    	  if (bptype==-1) {
    	    smatrixdxsxtepveuxsx_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {
    	    smatrixdxsxtepveuxsx_(q22[0],&meout);
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
    	}
      }
      else if (bptype==3) {
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
    	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4 -- missing configs
    	idpartfact*=4.;
    	if (process==1) 
    	  smatrixudxtepveddx_(q22[0],&meout);
    	else if (process==2)
    	  smatrixudxtepveuux_(q22[0],&meout);
    	else if (process==3) {
    	  idpartfact*=6.;   // 2 final state momm configs, 3 flavours
    	  smatrixdxutepvessx_(q32[0],&meout);//ME symmetric under interchange of s <--> sx
    	}
    	else {
    	  idpartfact*=2.;   // *2 for gg final state
    	  smatrixdxutepvegg_(q32[0],&meout);
    	}
      }
      else if (bptype==5) {
    	smatrixdxstepveuxs_(q32[0],&meout);
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
      else {  // a=ubar, b= d,u
    	double r1=ran.flat();
    	idpartfact*=2.;
    	if (r1<0.5) {// emit from a
    	  smatrixdxstepveuxs_(q32[0],&meout);
	  alepton.PID=-12;
	  alepton.p=pd2;
	  lepton.PID=11;
	  lepton.p=pd1;
	}
    	else {
    	  if (bptype==2) {
    	    smatrixusxtepvedsx_(p22[0],&meout);
	    alepton.PID=-11;
	    alepton.p=pd1;
	    lepton.PID=12;
	    lepton.p=pd2;
	  }
    	  else {
    	    smatrixusxtepvedsx_(q22[0],&meout);
	    alepton.PID=-12;
	    alepton.p=pd2;
	    lepton.PID=11;
	    lepton.p=pd1;
	  }
    	}
      }
    }
    else if (aptype==-5) {    // a=bbar, b!=b
      if (bptype==-1 || bptype==-3)   {// b = dbar,sbar
    	smatrixdxsxtepveuxsx_(p22[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else if (bptype==-2 || bptype==-4) {
    	smatrixdxsxtepveuxsx_(q22[0],&meout);
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
      else if (bptype==2 || bptype==4) {
     	smatrixusxtepvedsx_(p22[0],&meout);
	alepton.PID=-11;
	alepton.p=pd1;
	lepton.PID=12;
	lepton.p=pd2;
      }
      else { 
     	smatrixusxtepvedsx_(q22[0],&meout);
	alepton.PID=-12;
	alepton.p=pd2;
	lepton.PID=11;
	lepton.p=pd1;
      }
    }
    else  {
      std::cerr<<"Watch out - unanticipated parton configuration!"<<std::endl;
    }
  }

  //  std::cout<<"pointer gives: "<<meout<<std::endl;
  
  // if (isnan(meout)) {
  //   std::cout<<"Getting nan: "<<std::endl;
  // }
  weight*=meout*idpartfact;
#endif  
  //    std::cout<<"s1 "<<idpartfact<<"  "<<meout<<"  "<<weight<<"  "<<pow(4.*M_PI*AlphaS(flags.scale),2)<<std::endl;
  // if (flags.scalesetting==0) //fixed scale
  //   weight*=pow(4.*M_PI*AlphaS(flags.scale),2);
  // else if (flags.scalesetting==1)
  //   weight*=pow(4.*M_PI*AlphaS(Ht),2);

  weight*=pow(4.*M_PI*Halphas(renscale),2);
  if (flags.logcorrect==1)
    weight*=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI));
  //  std::cout<<"s2: "<<weight<<"\n";
  return weight;
}

double CWMultijet::MG3(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb, CLHEP::HepLorentzVector aleptonp, CLHEP::HepLorentzVector leptonp)
{
  double wt(0.);
  FREAL meout(0.),p12[7][4],p21[7][4];
  // p12 = config where partons are ordered as in name
  // p21 = config where partons are reversed relative to name

  p12[0][0]=pb.e();    // forward parton
  p12[0][1]=pb.x();
  p12[0][2]=pb.y();
  p12[0][3]=pb.z();
  
  p12[1][0]=pa.e();   //  backward parton
  p12[1][1]=pa.x();
  p12[1][2]=pa.y();
  p12[1][3]=pa.z();
  
  p12[2][0]=aleptonp.e();   // anti-lepton
  p12[2][1]=aleptonp.x();
  p12[2][2]=aleptonp.y();
  p12[2][3]=aleptonp.z();
  
  p12[3][0]=leptonp.e();   // lepton
  p12[3][1]=leptonp.x();
  p12[3][2]=leptonp.y();
  p12[3][3]=leptonp.z();
  
  p12[4][0]=(partons.back()).e();   // most forward jet
  p12[4][1]=(partons.back()).x();
  p12[4][2]=(partons.back()).y();
  p12[4][3]=(partons.back()).z();
  
  p12[5][0]=(partons.at(1)).e();  // middle jet
  p12[5][1]=(partons.at(1)).x();
  p12[5][2]=(partons.at(1)).y();
  p12[5][3]=(partons.at(1)).z();

  p12[6][0]=(partons.front()).e();  // most backward jet
  p12[6][1]=(partons.front()).x();
  p12[6][2]=(partons.front()).y();
  p12[6][3]=(partons.front()).z();

  for (int ii=0; ii<4; ii++) {
    p21[0][ii]=p12[1][ii];
    p21[1][ii]=p12[0][ii];
    p21[2][ii]=p12[2][ii];
    p21[3][ii]=p12[3][ii];
    p21[4][ii]=p12[6][ii];
    p21[5][ii]=p12[5][ii];
    p21[6][ii]=p12[4][ii];
  }

  if (aptype==21) {// a=g, we know they are not both gluons so b!= g
    if (bptype>0)
      smatrixugtepvedgg_(p12[0],&meout);
    //      std::cout<<"Yes!"<<std::endl;
    else
      smatrixdxgtepveuxgg_(p12[0],&meout);
    meout*=2.;
  }
  else if (bptype==21) {  // b=g, a!=g
    if (aptype>0) 
      smatrixugtepvedgg_(p21[0],&meout);
    else 
      smatrixdxgtepveuxgg_(p21[0],&meout);
    meout*=2.;
  }
  else {  //  two (anti-)quarks to deal with!
    if (aptype==bptype) {
      if (aptype > 0) {
	if (wc==true)   // emit off b
	  smatrixuutepveugd_(p21[0],&meout);
	else            // emit off a so swap momenta
	  smatrixuutepveugd_(p12[0],&meout);
      }
      else {
	if (wc==true)   // emit off b
	  smatrixdxdxtepveuxgdx_(p12[0],&meout);
	else            // emit off a so swap momenta
	  smatrixdxdxtepveuxgdx_(p21[0],&meout);
      }
    }
    else if (aptype==-bptype) {
      if (aptype > 0) {  // backward particle is q, forward qbar
	if (wc==true) 
	  smatrixdxdtepveuxgd_(p12[0],&meout);
	else 
	  smatrixuxutepveuxgd_(p12[0],&meout);
      }
      else {            // forward (b) particle is q, backward (a) qbar
	if (wc==true) 
	  smatrixuxutepveuxgd_(p21[0],&meout);
	else 
	  smatrixdxdtepveuxgd_(p21[0],&meout);
      }
    }
    else if (aptype==1) {    // a=d, b!=d,dbar
      if (bptype==2) {  // b=u
    	if (wc==true) 
    	  smatrixudtepvedgd_(p12[0],&meout);
    	else 
    	  smatrixudtepvedgd_(p21[0],&meout);
    	meout*=2.;
      }
      else if (bptype>2) {
    	if (wc==true)
    	  smatrixustepvedgs_(p12[0],&meout);
    	else
    	  smatrixustepvedgs_(p21[0],&meout);
      }
      else if (bptype==-2) {
    	if (wc==true) 
    	  smatrixudxtepveugux_(p21[0],&meout);
    	else 
    	  smatrixudxtepvedgdx_(p21[0],&meout);
      }
      else {
    	if (wc==true)
    	  smatrixdxstepveuxgs_(p12[0],&meout);
    	else
    	  smatrixusxtepvedgsx_(p21[0],&meout);
      }
    }
    else if (aptype==2) { // a=u
      if (bptype==1) {  // b=d
    	if (wc==true) 
    	  smatrixudtepvedgd_(p12[0],&meout);
    	else                 //  Not sure if this is necessary
    	  smatrixudtepvedgd_(p21[0],&meout);
    	meout*=2.;
      }
      else if (bptype>2) {    // b = s,c,b
    	if (wc==true)
    	  smatrixustepvedgs_(p12[0],&meout);
    	else
    	  smatrixustepvedgs_(p21[0],&meout);
      }
      else if (bptype==-1) {  // b = dbar
    	if (wc==true) 
    	  smatrixudxtepveugux_(p21[0],&meout);
    	else 
    	  smatrixudxtepvedgdx_(p21[0],&meout);
      }
      else {            // b = sbar, cbar,bbar
    	if (wc==true)
    	  smatrixdxstepveuxgs_(p12[0],&meout);
    	else 
    	  smatrixusxtepvedgsx_(p21[0],&meout);
      }
    }
    else if (aptype==3) { // a=s
      if (bptype==4) {  // b=c
    	if (wc==true) 
    	  smatrixudtepvedgd_(p12[0],&meout);
    	else
    	  smatrixudtepvedgd_(p21[0],&meout);
    	meout*=2.;
      }
      else if (bptype>0) {    // b = u,d,b
    	if (wc==true)
    	  smatrixustepvedgs_(p12[0],&meout);
    	else
    	  smatrixustepvedgs_(p21[0],&meout);
      }
      else if (bptype==-4) {  // b = cbar
    	if (wc==true) 
    	  smatrixudxtepveugux_(p21[0],&meout);
    	else 
    	  smatrixudxtepvedgdx_(p21[0],&meout);
      }
      else {            // b = ubar, dbar,bbar
    	if (wc==true)
    	  smatrixdxstepveuxgs_(p12[0],&meout);
    	else
    	  smatrixusxtepvedgsx_(p21[0],&meout);
      }
    }
    else if (aptype==4) { // a=c
      if (bptype==3) {  // b=s
    	if (wc==true) 
    	  smatrixudtepvedgd_(p12[0],&meout);
    	else                 //  Not sure if this is necessary
    	  smatrixudtepvedgd_(p21[0],&meout);
    	meout*=2.;
      }
      else if (bptype>0) {    // b = u,d,b
    	if (wc==true)
    	  smatrixustepvedgs_(p12[0],&meout);
    	else
    	  smatrixustepvedgs_(p21[0],&meout);
      }
      else if (bptype==-3) {  // b = cbar
    	if (wc==true) 
    	  smatrixudxtepveugux_(p21[0],&meout);
    	else 
    	  smatrixudxtepvedgdx_(p21[0],&meout);
      }
      else {            // b = ubar, dbar,bbar
    	if (wc==true)
    	  smatrixdxstepveuxgs_(p12[0],&meout);
    	else
    	  smatrixusxtepvedgsx_(p21[0],&meout);
      }
    }
    else if (aptype>0) { // a=b
      if (bptype>0) {    // b = u,d,c,s
    	if (wc==true)
    	  smatrixustepvedgs_(p12[0],&meout);
    	else
    	  smatrixustepvedgs_(p21[0],&meout);
      }
      else {            // b = ubar, dbar, sbar, cbar
    	if (wc==true)
    	  smatrixdxstepveuxgs_(p12[0],&meout);
    	else
    	  smatrixusxtepvedgsx_(p21[0],&meout);
      }
    }
    ///  Now go for a = anti-quark
    else if (aptype==-1) {    // a=dbar, b!=d
      if (bptype==-2) {  // b=ubar
    	if (wc==true)
    	  smatrixdxuxtepveuxgux_(p12[0],&meout);
    	else
    	  smatrixdxuxtepveuxgux_(p21[0],&meout);
    	meout*=2.;
      }
      else if (bptype<-2) {
    	if (wc==true)
    	  smatrixdxsxtepveuxgsx_(p12[0],&meout);
    	else
    	  smatrixdxsxtepveuxgsx_(p21[0],&meout);
      }
      else if (bptype==2) {
    	if (wc==true) 
    	  smatrixudxtepvedgdx_(p12[0],&meout);
    	else 
    	  smatrixudxtepveugux_(p12[0],&meout);
      }
      else {  // a=dbar, b= s,c,t
    	if (wc==true)
    	  smatrixusxtepvedgsx_(p12[0],&meout);
    	else
    	  smatrixdxstepveuxgs_(p21[0],&meout);
      }
    }
    else if (aptype==-2) {    // a=ubar, b!=u
      if (bptype==-1) {  // b=dbar
    	if (wc==true)
    	  smatrixdxuxtepveuxgux_(p12[0],&meout);
    	else
    	  smatrixdxuxtepveuxgux_(p21[0],&meout);
    	meout*=2.;
      }
      else if (bptype<-2) {  // b = sbar,cbar,bbar
    	if (wc==true)
    	  smatrixdxsxtepveuxgsx_(p12[0],&meout);
    	else
    	  smatrixdxsxtepveuxgsx_(p21[0],&meout);
      }
      else if (bptype==1) {  // a=ubar, b=d
    	if (wc==true) 
    	  smatrixudxtepvedgdx_(p12[0],&meout);
    	else 
    	  smatrixudxtepveugux_(p12[0],&meout);
      }
      else {  // a=ubar, b= s,c,t
    	if (wc==true) 
    	  smatrixusxtepvedgsx_(p12[0],&meout);
    	else 
    	  smatrixdxstepveuxgs_(p21[0],&meout);
      }
    }
    else if (aptype==-3) {    // a=sbar, b!=u
      if (bptype==-4) {  // b=cbar
    	if (wc==true)
    	  smatrixdxuxtepveuxgux_(p12[0],&meout);
    	else
    	  smatrixdxuxtepveuxgux_(p21[0],&meout);
    	meout*=2.;
      }
      else if (bptype<0) {  // b = ubar,dbar,bbar
    	if (wc==true)
    	  smatrixdxsxtepveuxgsx_(p12[0],&meout);
    	else
    	  smatrixdxsxtepveuxgsx_(p21[0],&meout);
      }
      else if (bptype==4) {  // a=sbar, b=c
    	if (wc==true) 
    	  smatrixudxtepvedgdx_(p12[0],&meout);
    	else 
    	  smatrixudxtepveugux_(p12[0],&meout);
      }
      else {  // a=sbar, b= u,d,t
    	if (wc==true) 
    	  smatrixusxtepvedgsx_(p12[0],&meout);
    	else 
    	  smatrixdxstepveuxgs_(p21[0],&meout);
      }
    }
    else if (aptype==-4) {    // a=cbar, b!=u
      if (bptype==-3) {  // b=sbar
    	if (wc==true)
    	  smatrixdxuxtepveuxgux_(p12[0],&meout);
    	else
    	  smatrixdxuxtepveuxgux_(p21[0],&meout);
    	meout*=2.;
      }
      else if (bptype<0) {  // b = ubar,dbar,bbar
    	if (wc==true)
    	  smatrixdxsxtepveuxgsx_(p12[0],&meout);
    	else
    	  smatrixdxsxtepveuxgsx_(p21[0],&meout);
      }
      else if (bptype==3) {  // a=cbar, b=s
    	if (wc==true) 
    	  smatrixudxtepvedgdx_(p12[0],&meout);
    	else 
    	  smatrixudxtepveugux_(p12[0],&meout);
      }
      else {  // a=cbar, b= u,d,t
    	if (wc==true) 
    	  smatrixusxtepvedgsx_(p12[0],&meout);
    	else 
    	  smatrixdxstepveuxgs_(p21[0],&meout);
      }
    }
    else if (aptype==-5) {    // a=bbar, b!=b
      if (bptype<0) {  // b = ubar,dbar,sbar, cbar
    	if (wc==true)
    	  smatrixdxsxtepveuxgsx_(p12[0],&meout);
    	else
    	  smatrixdxsxtepveuxgsx_(p21[0],&meout);
      }
      else {  // a=bbar, b= u,d,s,c
    	if (wc==true) 
    	  smatrixusxtepvedgsx_(p12[0],&meout);
    	else 
    	  smatrixdxstepveuxgs_(p21[0],&meout);
      }
    }
    else 
      std::cerr<<"Watch out - unanticipated parton configuration!"<<std::endl;
  }

  wt=meout;
  
  // double gs2=0.;
  // if (flags.scalesetting==0) //fixed scale
  //   gs2=4.*M_PI*AlphaS(flags.scale);
  // else if (flags.scalesetting==1)
  //   gs2=4.*M_PI*AlphaS(Ht);

  double gs2=4.*M_PI*Halphas(renscale);
  wt*=pow(gs2,3);
  if (flags.logcorrect==1)
    wt*=(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.at(1).perp()/renscale)/(2.*M_PI));

  return wt;
}

#if new2
double CWMultijet::Generate3jNonFKLflavourconf_and_eval(double pmg[7][4])
#else
double CWMultijet::Generate3jNonFKLflavourconf_and_eval()
#endif
{
  //  std::cout<<"Yep, 3j non FKL: ";
  double weight=1.;

  // p1 = b(+) a(-) --> e nu + 0 -,   p2 = a(-) b(+) --> e nu - 0 +
  // p3 = b(+) a(-) --> e nu 0 + -,   p4 = b(+) a(-) --> e nu + - 0
  // p5 = b(+) a(-) --> e nu - 0 +,   p6 = b(+) a(-) --> e nu 0 - +
  // p7 = b(+) a(-) --> e nu - + 0,   p8 = a(-) b(+) --> e nu 0 - +
  // p9 = a(-) b(+) --> e nu - + 0,   p10= a(-) b(+) --> e nu + 0 -
  // p11= a(-) b(+) --> e nu 0 + -,   p12= a(-) b(+) --> e nu + - 0

  // qi = pi with e and nu swapped (for e- nubar final states)

  // choose a matrix element to calculate and set the relevant flavours

  //  std::cout<<"Particles:  "<<aptype<<" and "<<bptype<<std::endl;

#if new2
  if (aptype==21) {// a=g 
    if (bptype==21) {
      static std::vector<sPMchoice> gg3Processes;
      if (gg3Processes.size()==0) 
	{
	  { // first gg process
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixggtepveuxgd_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=2;  // 2 flavours
	    PMtemp.MomShuffleVector.push_back(0);
	    for (int ii=2; ii<7; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    gg3Processes.push_back(PMtemp);

	    PMtemp.ptype=false; // process type p
	    gg3Processes.push_back(PMtemp);
	  }
	}
      weight*=PickAndEvalProcAndMom3j(gg3Processes,pmg,&lepton,&alepton);
    }
    else {    //a=g,  b!=g
      if (bptype==5)   {
	static std::vector<sPMchoice> gb3Processes;
	if (gb3Processes.size()==0) 
	  {
	    { // first gg process
	      struct sPMchoice PMtemp;
	      // tpmomshuffle tpmomtemp;
	      PMtemp.MGfunc=smatrixsgtepvesdux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;  // 2 flavours
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      
	      gb3Processes.push_back(PMtemp);
	      PMtemp.ptype=false; // process type p
	      gb3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(gb3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-5)   {
	static std::vector<sPMchoice> gbx3Processes;
	if (gbx3Processes.size()==0) 
	  {
	    { // first gg process
	      struct sPMchoice PMtemp;
	      // tpmomshuffle tpmomtemp;
	      PMtemp.MGfunc=smatrixdxgtepvedxcxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;  // 2 flavours
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      
	      gbx3Processes.push_back(PMtemp);
	      PMtemp.ptype=false; // process type p
	      gbx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(gbx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==2 || bptype==4) {   
	static std::vector<sPMchoice> gu3Processes;
	if (gu3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixugtepvedgg_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(4);
	      gu3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdgtepveddux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(4);
	      gu3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixsgtepvesdux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
	    	PMtemp.MomShuffleVector.push_back(ii);
	      gu3Processes.push_back(PMtemp);

	      //  Now for original proc 6
	      PMtemp.ptype=false; // process type p
	      gu3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixugtepveccxd_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
	    	PMtemp.MomShuffleVector.push_back(ii);
	      gu3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixugtepvedddx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(4);
	      gu3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixugtepveuuxd_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
	    	PMtemp.MomShuffleVector.push_back(ii);
	      gu3Processes.push_back(PMtemp);
	    }
	    // std::cout<<"gu3\n";
	  }
	weight*=PickAndEvalProcAndMom3j(gu3Processes,pmg,&lepton,&alepton);
      }  
      else if (bptype>0) {
	static std::vector<sPMchoice> gd3Processes;  
	if (gd3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixugtepvedgg_;
	      PMtemp.ptype=true; // process type q 
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(4);
	      gd3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdgtepveddux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(4);
	      gd3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixsgtepvesdux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      gd3Processes.push_back(PMtemp);

	      //  Now for original proc 6
	      PMtemp.ptype=false; // process type p
	      gd3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixugtepveccxd_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      gd3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixugtepvedddx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(4);
	      gd3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixugtepveuuxd_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      gd3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(gd3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-1 || bptype==-3) {  
	static std::vector<sPMchoice> gdx3Processes;
	if (gdx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxgtepveuxgg_;
	      PMtemp.ptype=false; // process type p 
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(4);
	      gdx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixuxgtepveuxuxd_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(4);
	      gdx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixuxgtepveuxcxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      gdx3Processes.push_back(PMtemp);

	      //  Now for original proc 6
	      PMtemp.ptype=false; // process type p
	      gdx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxgtepveuxssx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      gdx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxgtepveuxuxu_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(4);
	      gdx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxgtepvedxdux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      gdx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(gdx3Processes,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> gux3Processes;
	if (gux3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxgtepveuxgg_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(4);
	      gux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixuxgtepveuxuxd_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(4);
	      gux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixuxgtepveuxcxs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      gux3Processes.push_back(PMtemp);

	      //  Now for original proc 6
	      PMtemp.ptype=false; // process type p
	      gux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxgtepveuxssx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      gux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxgtepveuxuxu_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(4);
	      gux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxgtepvedxdux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      gux3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(gux3Processes,pmg,&lepton,&alepton);
      }
    }
  }
  else if (bptype==21) {  // b=g, a!=g
    if (aptype==5) {
      static std::vector<sPMchoice> bg3Processes;
      if (bg3Processes.size()==0) 
	{
	  { // first gg process
	    struct sPMchoice PMtemp;
	    // tpmomshuffle tpmomtemp;
	    PMtemp.MGfunc=smatrixsgtepvesdux_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=2;  // 2 flavours
	      
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    
	    bg3Processes.push_back(PMtemp);
	    PMtemp.ptype=false; // process type p
	    bg3Processes.push_back(PMtemp);
	  }
	}
      weight*=PickAndEvalProcAndMom3j(bg3Processes,pmg,&lepton,&alepton);
    }
    else if (aptype==-5) {
      static std::vector<sPMchoice> bxg3Processes;
      if (bxg3Processes.size()==0) 
	{
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxgtepvedxcxs_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=2;  // 2 flavours
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    
	    bxg3Processes.push_back(PMtemp);
	    PMtemp.ptype=false; // process type p
	    bxg3Processes.push_back(PMtemp);
	  }
	}
      weight*=PickAndEvalProcAndMom3j(bxg3Processes,pmg,&lepton,&alepton);
    }
    else if (aptype==2 || aptype==4) { 
      static std::vector<sPMchoice> ug3Processes;
      if (ug3Processes.size()==0) 
	{
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixugtepvedgg_;
	    PMtemp.ptype=false; // process type p
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    ug3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdgtepveddux_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(8);
	    PMtemp.MomShuffleVector.push_back(9);
	    ug3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixsgtepvesdux_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ug3Processes.push_back(PMtemp);
	    
	    //  Now for original proc 6
	    PMtemp.ptype=false; // process type p
	    ug3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixugtepveccxd_;
	    PMtemp.ptype=false; // process type p
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ug3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixugtepvedddx_;
	    PMtemp.ptype=false; // process type p
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(8);
	    PMtemp.MomShuffleVector.push_back(9);
	    ug3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixugtepveuuxd_;
	    PMtemp.ptype=false; // process type p
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ug3Processes.push_back(PMtemp);
	  }
	  //	  std::cout<<"ug3\n";
	}
      weight*=PickAndEvalProcAndMom3j(ug3Processes,pmg,&lepton,&alepton);
    }
    else if (aptype>0) {  
      static std::vector<sPMchoice> dg3Processes;
      if (dg3Processes.size()==0) 
	{
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixugtepvedgg_;
	    PMtemp.ptype=true; // process type q 
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    dg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdgtepveddux_;
	    PMtemp.ptype=false; // process type p
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(8);
	    PMtemp.MomShuffleVector.push_back(9);
	    dg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixsgtepvesdux_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dg3Processes.push_back(PMtemp);
	    
	    //  Now for original proc 6
	    PMtemp.ptype=false; // process type p
	    dg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixugtepveccxd_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixugtepvedddx_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(8);
	    PMtemp.MomShuffleVector.push_back(9);
	    dg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixugtepveuuxd_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dg3Processes.push_back(PMtemp);
	  }
	}
      weight*=PickAndEvalProcAndMom3j(dg3Processes,pmg,&lepton,&alepton);
    }
    else if (aptype==-1 || aptype==-3) {  
      static std::vector<sPMchoice> dxg3Processes;
      if (dxg3Processes.size()==0) 
	{
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxgtepveuxgg_;
	    PMtemp.ptype=false; // process type p 
	    PMtemp.idpartmult=2;
	      
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    dxg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixuxgtepveuxuxd_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(8);
	    PMtemp.MomShuffleVector.push_back(9);
	    dxg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixuxgtepveuxcxs_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dxg3Processes.push_back(PMtemp);
	    
	    //  Now for original proc 6
	    PMtemp.ptype=false; // process type p
	    dxg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxgtepveuxssx_;
	    PMtemp.ptype=false; // process type p
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dxg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxgtepveuxuxu_;
	    PMtemp.ptype=false; // process type p
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(8);
	    PMtemp.MomShuffleVector.push_back(9);
	    dxg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxgtepvedxdux_;
	    PMtemp.ptype=false; // process type p
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dxg3Processes.push_back(PMtemp);
	  }
	}
      weight*=PickAndEvalProcAndMom3j(dxg3Processes,pmg,&lepton,&alepton);
    }
    else { 
      static std::vector<sPMchoice> uxg3Processes;
      if (uxg3Processes.size()==0) 
	{
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxgtepveuxgg_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    uxg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixuxgtepveuxuxd_;
	    PMtemp.ptype=false; // process type p
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(8);
	    PMtemp.MomShuffleVector.push_back(9);
	    uxg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixuxgtepveuxcxs_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    uxg3Processes.push_back(PMtemp);
	    
	    //  Now for original proc 6
	    PMtemp.ptype=false; // process type p
	    uxg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxgtepveuxssx_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    uxg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxgtepveuxuxu_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=2;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(8);
	    PMtemp.MomShuffleVector.push_back(9);
	    uxg3Processes.push_back(PMtemp);
	  }
	  { 
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxgtepvedxdux_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=1;
	    
	    PMtemp.MomShuffleVector.push_back(1);
	    for (int ii=7; ii<12; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    uxg3Processes.push_back(PMtemp);
	  }
	}
      weight*=PickAndEvalProcAndMom3j(uxg3Processes,pmg,&lepton,&alepton);
    }
  }
  else {  //  MARKER - two (anti-)quarks to deal with! 
    if (aptype==bptype) {
      if (fabs(aptype)==5) {
	setstatus(24);
	return 0.;       // All configs are in FKL bit
      }
      else if (aptype==2 || aptype==4) {
	static std::vector<sPMchoice> uu3Processes;
	if (uu3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixuutepveugd_;
	      PMtemp.ptype=false; // process type p 
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(5);
	      PMtemp.MomShuffleVector.push_back(6);
	      uu3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(uu3Processes,pmg,&lepton,&alepton);
      }
      else if (aptype>0) {
	static std::vector<sPMchoice> dd3Processes;
	if (dd3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixuutepveugd_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(5);
	      PMtemp.MomShuffleVector.push_back(6);
	      dd3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(dd3Processes,pmg,&lepton,&alepton);
      }
      else if (aptype==-1 || aptype==-3) {
	static std::vector<sPMchoice> dxdx3Processes;
	if (dxdx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxdxtepveuxgdx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(5);
	      PMtemp.MomShuffleVector.push_back(6);
	      dxdx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(dxdx3Processes,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> uxux3Processes;
	if (uxux3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxdxtepveuxgdx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      
	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      PMtemp.MomShuffleVector.push_back(5);
	      PMtemp.MomShuffleVector.push_back(6);
	      uxux3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(uxux3Processes,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==-bptype) {    /// MARKER - q qbar same flavour /// 
      if (fabs(aptype)==5) {   //  ME is symmetric under s<-->sx
	static std::vector<sPMchoice> bbx3Processes;
	if (bbx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixssxtepveuxgd_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2; // 2 flavours (u dx, c sx) and (ux d, cx s)
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      bbx3Processes.push_back(PMtemp);
	      
	      PMtemp.ptype=true; // process type q
	      bbx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(bbx3Processes,pmg,&lepton,&alepton);
      }
      else {                    //  not b bbar so uux-->dux, udx, scx, csx
	if (aptype==2 || aptype==4) {
	  static std::vector<sPMchoice> uux3Processes;
	  if (uux3Processes.size()==0) 
	    {
	      // { 
	      // 	struct sPMchoice PMtemp;
	      // 	PMtemp.MGfunc=smatrixuxutepveuxgd_;
	      // 	PMtemp.ptype=false; // process type p
	      // 	PMtemp.idpartmult=1; 
	      // 	for (int ii=2; ii<7; ii++)
	      // 	  PMtemp.MomShuffleVector.push_back(ii);
	      // 	uux3Processes.push_back(PMtemp);
	      // }
	      { 
	      	struct sPMchoice PMtemp;
	      	PMtemp.MGfunc=smatrixdxdtepveuxgd_;
	      	PMtemp.ptype=true; // process type q
	      	PMtemp.idpartmult=1; 
	      	for (int ii=2; ii<7; ii++)
	      	  PMtemp.MomShuffleVector.push_back(ii);
	      	uux3Processes.push_back(PMtemp);
	      }
	      // { 
	      // 	struct sPMchoice PMtemp;
	      // 	PMtemp.MGfunc=smatrixssxtepveuxgd_;
	      // 	PMtemp.ptype=false; // process type p
	      // 	PMtemp.idpartmult=1; 
		
	      // 	PMtemp.MomShuffleVector.push_back(1);
	      // 	for (int ii=7; ii<12; ii++)
	      // 	  PMtemp.MomShuffleVector.push_back(ii);
	      // 	uux3Processes.push_back(PMtemp);

	      // 	PMtemp.ptype=true; // process type q
	      // 	uux3Processes.push_back(PMtemp);
	      // }
	      // std::cout<<"uux3\n";
	    }
	  weight*=PickAndEvalProcAndMom3j(uux3Processes,pmg,&lepton,&alepton);
	}
	else if (aptype==1 || aptype==3) {  
	  static std::vector<sPMchoice> ddx3Processes;
	  if (ddx3Processes.size()==0) 
	    {
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixuxutepveuxgd_;
		PMtemp.ptype=true; // process type q
		PMtemp.idpartmult=1; 
		for (int ii=2; ii<7; ii++)
		  PMtemp.MomShuffleVector.push_back(ii);
		ddx3Processes.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixdxdtepveuxgd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		for (int ii=2; ii<7; ii++)
		  PMtemp.MomShuffleVector.push_back(ii);
		ddx3Processes.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixssxtepveuxgd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(1);
		for (int ii=7; ii<12; ii++)
		  PMtemp.MomShuffleVector.push_back(ii);
		ddx3Processes.push_back(PMtemp);

		PMtemp.ptype=true; // process type q
		ddx3Processes.push_back(PMtemp);
	      }
	    }
	  weight*=PickAndEvalProcAndMom3j(ddx3Processes,pmg,&lepton,&alepton);
	}
	else if (aptype==-1 || aptype==-3) { 
	  static std::vector<sPMchoice> dxd3Processes;
	  if (dxd3Processes.size()==0) 
	    {
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixuxutepveuxgd_;
		PMtemp.ptype=true; // process type q
		PMtemp.idpartmult=1; 
		for (int ii=7; ii<12; ii++) 
		  PMtemp.MomShuffleVector.push_back(ii);
		dxd3Processes.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixdxdtepveuxgd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		for (int ii=7; ii<12; ii++)
		  PMtemp.MomShuffleVector.push_back(ii);
		dxd3Processes.push_back(PMtemp);
	      }
	      { 
		struct sPMchoice PMtemp;
		PMtemp.MGfunc=smatrixssxtepveuxgd_;
		PMtemp.ptype=false; // process type p
		PMtemp.idpartmult=1; 
		
		PMtemp.MomShuffleVector.push_back(0);
		for (int ii=2; ii<7; ii++)
		  PMtemp.MomShuffleVector.push_back(ii);
		dxd3Processes.push_back(PMtemp);

		PMtemp.ptype=true; // process type q
		dxd3Processes.push_back(PMtemp);
	      }
	    }
	  weight*=PickAndEvalProcAndMom3j(dxd3Processes,pmg,&lepton,&alepton);
	}
	else {  // aptype==-2, -4 
	  static std::vector<sPMchoice> uxu3Processes;
	  if (uxu3Processes.size()==0) 
	    {
	      { 
	      	struct sPMchoice PMtemp;
	      	PMtemp.MGfunc=smatrixuxutepveuxgd_;
	      	PMtemp.ptype=false; // process type p
	      	PMtemp.idpartmult=1; 
	      	for (int ii=7; ii<12; ii++)
	      	  PMtemp.MomShuffleVector.push_back(ii);
	      	uxu3Processes.push_back(PMtemp);
	      }
	      { 
	      	struct sPMchoice PMtemp;
	      	PMtemp.MGfunc=smatrixdxdtepveuxgd_;
	      	PMtemp.ptype=true; // process type q
	      	PMtemp.idpartmult=1; 
	      	for (int ii=7; ii<12; ii++)
	      	  PMtemp.MomShuffleVector.push_back(ii);
	      	uxu3Processes.push_back(PMtemp);
	      }
	      { 
	      	struct sPMchoice PMtemp;
	      	PMtemp.MGfunc=smatrixssxtepveuxgd_;
	      	PMtemp.ptype=false; // process type p
	      	PMtemp.idpartmult=1; 
		
	      	PMtemp.MomShuffleVector.push_back(0);
	      	for (int ii=2; ii<7; ii++)
	      	  PMtemp.MomShuffleVector.push_back(ii);
	      	uxu3Processes.push_back(PMtemp);

	      	PMtemp.ptype=true; // process type q
	      	uxu3Processes.push_back(PMtemp);
	      }
	      //  std::cout<<"uxu3\n";
	    }
	  weight*=PickAndEvalProcAndMom3j(uxu3Processes,pmg,&lepton,&alepton);
	}
      }
    }
    else if (aptype==5) { // a=b, b \ne b
      if (bptype==2 || bptype==4) {
	static std::vector<sPMchoice> bu3Processes;
	if (bu3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      bu3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(bu3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype >0 ) {
	static std::vector<sPMchoice> bd3Processes;
	if (bd3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      bd3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(bd3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-1 || bptype==-3) { 
	static std::vector<sPMchoice> bdx3Processes;
	if (bdx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      bdx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(bdx3Processes,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> bux3Processes;
	if (bux3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      bux3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(bux3Processes,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==-5) {    // a=bbar, b!=b 
      if (bptype==2 || bptype==4) {
	static std::vector<sPMchoice> bxu3Processes;
	if (bxu3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      bxu3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(bxu3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype >0 ) {
	static std::vector<sPMchoice> bxd3Processes;
	if (bxd3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++) 
		PMtemp.MomShuffleVector.push_back(ii);
	      bxd3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(bxd3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-1 || bptype==-3) {
	static std::vector<sPMchoice> bxdx3Processes;
	if (bxdx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      bxdx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(bxdx3Processes,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> bxux3Processes;
	if (bxux3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      bxux3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(bxux3Processes,pmg,&lepton,&alepton);
      }
    }
    else if (bptype==5) { // b=b, a \ne b
      if (aptype==2 || aptype==4) {
	static std::vector<sPMchoice> ub3Processes;
	if (ub3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      ub3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(ub3Processes,pmg,&lepton,&alepton);
      }
      else if (aptype >0 ) {
	static std::vector<sPMchoice> db3Processes;
	if (db3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      db3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(db3Processes,pmg,&lepton,&alepton);
      }
      else if (aptype==-1 || aptype==-3) { 
	static std::vector<sPMchoice> dxb3Processes;
	if (dxb3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxb3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(dxb3Processes,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> uxb3Processes;
	if (uxb3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxb3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(uxb3Processes,pmg,&lepton,&alepton);
      }
    }
    else if (bptype==-5) {    // b=bbar, a!=b 
      if (aptype==2 || aptype==4) {
	static std::vector<sPMchoice> ubx3Processes;
	if (ubx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      ubx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(ubx3Processes,pmg,&lepton,&alepton);
      }
      else if (aptype >0 ) {
	static std::vector<sPMchoice> dbx3Processes;
	if (dbx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++) 
		PMtemp.MomShuffleVector.push_back(ii);
	      dbx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(dbx3Processes,pmg,&lepton,&alepton);
      }
      else if (aptype==-1 || aptype==-3) {
	static std::vector<sPMchoice> dxbx3Processes;
	if (dxbx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxbx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(dxbx3Processes,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> uxbx3Processes;
	if (uxbx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxbx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickMomAndEval3j(uxbx3Processes,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==1) {    // a=d, b!=d,dbar   ///  MARKER a = down // 
      if (bptype==2) {  // b=u         
	static std::vector<sPMchoice> du3Processes;
	if (du3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudtepvedgd_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(7);
	      PMtemp.MomShuffleVector.push_back(8);
	      du3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudtepvedgd_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      du3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(du3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==3) {
	static std::vector<sPMchoice> ds3Processes;
	if (ds3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // ds --> us
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      ds3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  // ds --> dc
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      ds3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(ds3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==4) { 
	static std::vector<sPMchoice> dc3Processes;
	if (dc3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // dc --> uc
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dc3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  // dc --> ds
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dc3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(dc3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-2) {  // dux --> uux, ddx, ssx, gg 
	static std::vector<sPMchoice> dux3Processes;
	if (dux3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // dux --> uux
	      PMtemp.MGfunc=smatrixudxtepvedgdx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dux3Processes.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveugux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvesgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepveggg_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=6; 
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      dux3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(dux3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) {    // b is sx 
	static std::vector<sPMchoice> dsx3Processes;
	if (dsx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dsx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dsx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(dsx3Processes,pmg,&lepton,&alepton);
      }
      else {    // b is cx 
	static std::vector<sPMchoice> dcx3Processes;
	if (dcx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dcx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dcx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(dcx3Processes,pmg,&lepton,&alepton);
      }
    }   
    else if (aptype==2) { // MARKER a=u
      if (bptype==1) {  // b=d         
	static std::vector<sPMchoice> ud3Processes;
	if (ud3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudtepvedgd_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(7);
	      PMtemp.MomShuffleVector.push_back(8);
	      ud3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudtepvedgd_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      ud3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(ud3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==3) { 
	static std::vector<sPMchoice> us3Processes;
	if (us3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      us3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      us3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(us3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==4) { // emit from c
	static std::vector<sPMchoice> uc3Processes;
	if (uc3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uc3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uc3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(uc3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) {  // udx --> ddx, uux, ssx, gg 
	static std::vector<sPMchoice> udx3Processes;
	if (udx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepvedgdx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      udx3Processes.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveugux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      udx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvesgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      udx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepveggg_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=6; 
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      udx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(udx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) {    // b is sx 
	static std::vector<sPMchoice> usx3Processes;
	if (usx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      usx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      usx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(usx3Processes,pmg,&lepton,&alepton);
      }
      else {    // b is cx 
	static std::vector<sPMchoice> ucx3Processes;
	if (ucx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      ucx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      ucx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(ucx3Processes,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==3) { // a=s  
      if (bptype==4) {  // b=c         
	static std::vector<sPMchoice> sc3Processes;
	if (sc3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudtepvedgd_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(7);
	      PMtemp.MomShuffleVector.push_back(8);
	      sc3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudtepvedgd_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      sc3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(sc3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==1) {
	static std::vector<sPMchoice> sd3Processes;
	if (sd3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // sd --> cd
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sd3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  // sd --> su
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sd3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(sd3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype>0) { // bptype = 2
	static std::vector<sPMchoice> su3Processes;
	if (su3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // su --> cu
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      su3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  // su --> sd
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      su3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(su3Processes,pmg,&lepton,&alepton);
      }  
      else if (bptype==-4) {  // scx --> ccx, ssx, ddx, gg
	static std::vector<sPMchoice> scx3Processes;
	if (scx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  // dux --> uux
	      PMtemp.MGfunc=smatrixudxtepvedgdx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      scx3Processes.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveugux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      scx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvesgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      scx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepveggg_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=6; 
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      scx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(scx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) {    // b is dx 
	static std::vector<sPMchoice> sdx3Processes;
	if (sdx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sdx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sdx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(sdx3Processes,pmg,&lepton,&alepton);
      }
      else {    // b is ux 
	static std::vector<sPMchoice> sux3Processes;
	if (sux3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sux3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(sux3Processes,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==4) { // a=c 
      if (bptype==3) {  // b=s         
	static std::vector<sPMchoice> cs3Processes;
	if (cs3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudtepvedgd_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(7);
	      PMtemp.MomShuffleVector.push_back(8);
	      cs3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudtepvedgd_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      cs3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(cs3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==1) {
	static std::vector<sPMchoice> cd3Processes;
	if (cd3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cd3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cd3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(cd3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==2) { 
	static std::vector<sPMchoice> cu3Processes;
	if (cu3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cu3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixustepvedgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cu3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(cu3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) {  // csx --> ssx, ccx, ddx, gg
	static std::vector<sPMchoice> csx3Processes;
	if (csx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepvedgdx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      csx3Processes.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveugux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      csx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvesgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      csx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepveggg_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=6; 
	      
	      PMtemp.MomShuffleVector.push_back(0);
	      csx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(csx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) {    // b is dx 
	static std::vector<sPMchoice> cdx3Processes;
	if (cdx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cdx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cdx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(cdx3Processes,pmg,&lepton,&alepton);
      }
      else {    // b is ux 
	static std::vector<sPMchoice> cux3Processes;
	if (cux3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cux3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(cux3Processes,pmg,&lepton,&alepton);
      }
    }
    // ///  Now go for a = anti-quark   --- MARKER a=qbar  ///  
    else if (aptype==-1) {    // a=dbar, b!=d   
      if (bptype==-2) {  // b=ux         
	static std::vector<sPMchoice> dxux3Processes;
	if (dxux3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxuxtepveuxgux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(7);
	      PMtemp.MomShuffleVector.push_back(8);
	      dxux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxuxtepveuxgux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      dxux3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(dxux3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-4) { 
	static std::vector<sPMchoice> dxcx3Processes;
	if (dxcx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxcx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxcx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(dxcx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) { 
	static std::vector<sPMchoice> dxsx3Processes;
	if (dxsx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxsx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxsx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(dxsx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==2) {  // dxu --> ddx, uux, ssx, gg
	static std::vector<sPMchoice> dxu3Processes;
	if (dxu3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepvedgdx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxu3Processes.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveugux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxu3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvesgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxu3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepveggg_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=6; 
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      dxu3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(dxu3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==4) {    // b is c
	static std::vector<sPMchoice> dxc3Processes;
	if (dxc3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxc3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxc3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(dxc3Processes,pmg,&lepton,&alepton);
      }
      else {    // b is s 
	static std::vector<sPMchoice> dxs3Processes;
	if (dxs3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxs3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      dxs3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(dxs3Processes,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==-2) {    // a=ubar, b!=u  
      if (bptype==-1) {  // b=dx         
	static std::vector<sPMchoice> uxdx3Processes;
	if (uxdx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxuxtepveuxgux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(7);
	      PMtemp.MomShuffleVector.push_back(8);
	      uxdx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxuxtepveuxgux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      uxdx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(uxdx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-4) {
	static std::vector<sPMchoice> uxcx3Processes;
	if (uxcx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxcx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxcx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(uxcx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) { 
	static std::vector<sPMchoice> uxsx3Processes;
	if (uxsx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxsx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxsx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(uxsx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==1) {  // uxd --> uux, ddx, ssx, gg
	static std::vector<sPMchoice> uxd3Processes;
	if (uxd3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepvedgdx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxd3Processes.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveugux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxd3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvesgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxd3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepveggg_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=6; 
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      uxd3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(uxd3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==4) {    // b is c
	static std::vector<sPMchoice> uxc3Processes;
	if (uxc3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxc3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxc3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(uxc3Processes,pmg,&lepton,&alepton);
      }
      else {    // b is s 
	static std::vector<sPMchoice> uxs3Processes;
	if (uxs3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxs3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      uxs3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(uxs3Processes,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==-3) {    // a=sbar, b!=u   
      if (bptype==-4) {  // b=cx         
	static std::vector<sPMchoice> sxcx3Processes;
	if (sxcx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxuxtepveuxgux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(7);
	      PMtemp.MomShuffleVector.push_back(8);
	      sxcx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxuxtepveuxgux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      sxcx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(sxcx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-2) {
	static std::vector<sPMchoice> sxux3Processes;
	if (sxux3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sxux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sxux3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(sxux3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) { 
	static std::vector<sPMchoice> sxdx3Processes;
	if (sxdx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sxdx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sxdx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(sxdx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==4) {  // sxc --> ssx, ccx, ddx, gg
	static std::vector<sPMchoice> sxc3Processes;
	if (sxc3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepvedgdx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sxc3Processes.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveugux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sxc3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvesgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sxc3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepveggg_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=6; 
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      sxc3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(sxc3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==2) {    // b is u
	static std::vector<sPMchoice> sxu3Processes;
	if (sxu3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sxu3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sxu3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(sxu3Processes,pmg,&lepton,&alepton);
      }
      else {    // b is d 
	static std::vector<sPMchoice> sxd3Processes;
	if (sxd3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sxd3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      sxd3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(sxd3Processes,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==-4) {    // a=cbar, b!=u
      if (bptype==-3) {  // b=sx //Jenni
	static std::vector<sPMchoice> cxsx3Processes;
	if (cxsx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxuxtepveuxgux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(7);
	      PMtemp.MomShuffleVector.push_back(8);
	      cxsx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxuxtepveuxgux_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=2;

	      PMtemp.MomShuffleVector.push_back(2);
	      PMtemp.MomShuffleVector.push_back(3);
	      cxsx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(cxsx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-2) {
	static std::vector<sPMchoice> cxux3Processes;
	if (cxux3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cxux3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cxux3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(cxux3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) {  // b = dx
	static std::vector<sPMchoice> cxdx3Processes;
	if (cxdx3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cxdx3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixdxsxtepveuxgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cxdx3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(cxdx3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==3) {  // cxs --> ccx, ssx, ddx, gg
	static std::vector<sPMchoice> cxs3Processes;
	if (cxs3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;  
	      PMtemp.MGfunc=smatrixudxtepvedgdx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cxs3Processes.push_back(PMtemp);
	    }
 	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixudxtepveugux_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cxs3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepvesgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=3;  // 3 flavours
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cxs3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxutepveggg_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=6; 
	      
	      PMtemp.MomShuffleVector.push_back(1);
	      cxs3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(cxs3Processes,pmg,&lepton,&alepton);
      }
      else if (bptype==2) {    // b is u
	static std::vector<sPMchoice> cxu3Processes;
	if (cxu3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=false; // process type p
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cxu3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cxu3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(cxu3Processes,pmg,&lepton,&alepton);
      }
      else {    // b is d 
	static std::vector<sPMchoice> cxd3Processes;
	if (cxd3Processes.size()==0) 
	  {
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixusxtepvedgsx_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=2; ii<7; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cxd3Processes.push_back(PMtemp);
	    }
	    { 
	      struct sPMchoice PMtemp;
	      PMtemp.MGfunc=smatrixdxstepveuxgs_;
	      PMtemp.ptype=true; // process type q
	      PMtemp.idpartmult=1;
	      for (int ii=7; ii<12; ii++)
		PMtemp.MomShuffleVector.push_back(ii);
	      cxd3Processes.push_back(PMtemp);
	    }
	  }
	weight*=PickAndEvalProcAndMom3j(cxd3Processes,pmg,&lepton,&alepton);
      }
    }
    else  {
      std::cerr<<"Watch out - unanticipated parton configuration!"<<std::endl;
    }
  }
#else
  // setup arrays and LH information:
  FREAL meout(0.);

  double idpartfact=1.;     //  Factor to take account of identical final states

  //  Def3j1();
  //  std::cout<<"momentum1: "<<pmg3[4][3]<<"  ";
  //  smatrixggtepveuxd_(pmg3[0],&meout);

  if (aptype==21) {// a=g 
    if (bptype==21) {
      int momp=int(12.*ran.flat())+1;   // 12 momentum configs
      idpartfact*=12.;    // *12 for choice
      if (momp<7) {
	if (momp<4) {
	  if (momp<2)
	    Def3jq2();
	  else if (momp<3)
	    Def3jq8();
	  else
	    Def3jq9();
	} else {
	  if (momp<5)
	    Def3jq10();
	  else if (momp<6)
	    Def3jq11();
	  else
	    Def3jq12();
	}
      } else {
	if (momp<10) {
	  if (momp<8)
	    Def3jp2();
	  else if (momp<9)
	    Def3jp8();
	  else
	    Def3jp9();
	} else {
	  if (momp<11)
	    Def3jp10();
	  else if (momp<12)
	    Def3jp11();
	  else
	    Def3jp12();
	}
      }
      smatrixggtepveuxgd_(pmg3[0],&meout);
    }
    else if (bptype==5) {          //MARKER! a=g,  b!=g    
      int momp=int(12.*ran.flat())+1;
      idpartfact*=24.;   // *12 for choices  *2 for flavours (udx=csx, uxd=cxs)
      if (momp==1)
	Def3jq1();
      else if (momp==2)
	Def3jq3();
      else if (momp==3)
	Def3jq4();
      else if (momp==4)
	Def3jq5();
      else if (momp==5)
	Def3jq6();
      else if (momp==6)
	Def3jq7();
      else if (momp==7)
	Def3jp1();
      else if (momp==8)
	Def3jp3();
      else if (momp==9)
	Def3jp4();
      else if (momp==10)
	Def3jp5();
      else if (momp==11)
	Def3jp6();
      else 
	Def3jp7();
      smatrixsgtepvesdux_(pmg3[0],&meout);
      //	cout<<meout<<"\n";
    }
    else if (bptype==-5) {
      int momp=int(12.*ran.flat())+1;
      idpartfact*=24.;   // *12 for choices  *2 for flavours (udx=csx, uxd=cxs)
      if (momp==1)
	Def3jq1();
      else if (momp==2)
	Def3jq3();
      else if (momp==3)
	Def3jq4();
      else if (momp==4)
	Def3jq5();
      else if (momp==5)
	Def3jq6();
      else if (momp==6)
	Def3jq7();
      else if (momp==7)
	Def3jp1();
      else if (momp==8)
	Def3jp3();
      else if (momp==9)
	Def3jp4();
      else if (momp==10)
	Def3jp5();
      else if (momp==11)
	Def3jp6();
      else if (momp==12)
	Def3jp7();
      smatrixdxgtepvedxcxs_(pmg3[0],&meout);
    }
    else if (bptype==2 || bptype==4) { 
      int proc=int(7.*ran.flat())+1;
      idpartfact*=7.;
      if (proc==1) {
    	double momp=ran.flat();
    	idpartfact*=4.;   // *2 for momp choice, *2 for identical gluons
    	if (momp < 0.5) 
    	  Def3jp3();
    	else 
    	  Def3jp5();
    	smatrixugtepvedgg_(pmg3[0],&meout);
    	//	std::cout<<meout*2.<<"\n";
      } else if (proc==2) {
    	double momp=3.*ran.flat();
    	idpartfact*=6.;  //  *3 for choice, *2 for 2 'd's in final state
    	if (momp < 1.)
    	  Def3jq1();
    	else if (momp < 2.)
    	  Def3jq4();
    	else
    	  Def3jq5();
    	smatrixdgtepveddux_(pmg3[0],&meout);
      } else if (proc==3) {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;
    	if (momp < 1.) 
    	  Def3jq1();
    	else if (momp < 2.)
    	  Def3jq3();
    	else if (momp < 3.)
    	  Def3jq4();
    	else if (momp < 4.)
    	  Def3jq5();
    	else if (momp < 5.)
    	  Def3jq6();
    	else
    	  Def3jq7();
    	smatrixsgtepvesdux_(pmg3[0],&meout);
      } else if (proc==4) {
    	double momp=6.*ran.flat();
    	idpartfact*=12.;   // *6 for choice, *2 as two channels (ccbar or ssbar)
    	if (momp < 1.) 
    	  Def3jp1();
    	else if (momp < 2.)
    	  Def3jp3();
    	else if (momp < 3.)
    	  Def3jp4();
    	else if (momp < 4.)
    	  Def3jp5();
    	else if (momp < 5.)
    	  Def3jp6();
    	else
    	  Def3jp7();
    	smatrixugtepveccxd_(pmg3[0],&meout);
      } else if (proc==5) {
    	double momp=3.*ran.flat();
    	idpartfact*=6.;  // *3 for choice, *2 for 2 'd's in final state
    	if (momp<1.)
    	  Def3jp1();
    	else if (momp<2.)
    	  Def3jp4();
    	else 
    	  Def3jp5();
    	smatrixugtepvedddx_(pmg3[0],&meout);
      } else if (proc==6) {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;   // *6 for choice
    	if (momp < 1.) 
    	  Def3jp1();
    	else if (momp < 2.)
    	  Def3jp3();
    	else if (momp < 3.)
    	  Def3jp4();
    	else if (momp < 4.)
    	  Def3jp5();
    	else if (momp < 5.)
    	  Def3jp6();
    	else
    	  Def3jp7();
    	smatrixugtepveucxs_(pmg3[0],&meout);
      } else {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;   // *6 for choice
    	if (momp < 1.) 
    	  Def3jp1();
    	else if (momp < 2.)
    	  Def3jp3();
    	else if (momp < 3.)
    	  Def3jp4();
    	else if (momp < 4.)
    	  Def3jp5();
    	else if (momp < 5.)
    	  Def3jp6();
    	else
    	  Def3jp7();
    	smatrixugtepveuuxd_(pmg3[0],&meout);
      }
    }
    else if (bptype>0) {    // b = d,s
      int proc=int(7.*ran.flat())+1;
      idpartfact*=7.;
      if (proc==1) {
    	double momp=ran.flat();
    	idpartfact*=4.;   // *2 for momp choice, *2 for identical gluons
    	if (momp < 0.5) {
    	  Def3jq3();
    	  smatrixugtepvedgg_(pmg3[0],&meout);
    	} else {
    	  Def3jq5();
    	  smatrixugtepvedgg_(pmg3[0],&meout);
    	}
      } else if (proc==2) {
    	double momp=3.*ran.flat();
    	idpartfact*=6.;  //  *3 for choice, *2 for 2 'd's in final state
    	if (momp < 1.)
    	  Def3jp1();
    	else if (momp < 2.)
    	  Def3jp4();
    	else
    	  Def3jp5();
    	smatrixdgtepveddux_(pmg3[0],&meout);
      } else if (proc==3) {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;
    	if (momp < 1.) 
    	  Def3jp1();
    	else if (momp < 2.)
    	  Def3jp3();
    	else if (momp < 3.)
    	  Def3jp4();
    	else if (momp < 4.)
    	  Def3jp5();
    	else if (momp < 5.)
    	  Def3jp6();
    	else
    	  Def3jp7();
    	smatrixsgtepvesdux_(pmg3[0],&meout);
      } else if (proc==4) {
    	double momp=6.*ran.flat();
    	idpartfact*=12.;   // *6 for choice, *2 as two channels (ccbar or ssbar)
    	if (momp < 1.) 
    	  Def3jq1();
    	else if (momp < 2.)
    	  Def3jq3();
    	else if (momp < 3.)
    	  Def3jq4();
    	else if (momp < 4.)
    	  Def3jq5();
    	else if (momp < 5.)
    	  Def3jq6();
    	else
    	  Def3jq7();
    	smatrixugtepveccxd_(pmg3[0],&meout);
      } else if (proc==5) {
    	double momp=3.*ran.flat();
    	idpartfact*=6.;  // *3 for choice, *2 for 2 'd's in final state
    	if (momp<1.)
    	  Def3jq1();
    	else if (momp<2.)
    	  Def3jq4();
    	else 
    	  Def3jq5();
    	smatrixugtepvedddx_(pmg3[0],&meout);
      } else if (proc==6) {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;   // *6 for choice
    	if (momp < 1.) 
    	  Def3jq1();
    	else if (momp < 2.)
    	  Def3jq3();
    	else if (momp < 3.)
    	  Def3jq4();
    	else if (momp < 4.)
    	  Def3jq5();
    	else if (momp < 5.)
    	  Def3jq6();
    	else
    	  Def3jq7();
    	smatrixugtepveucxs_(pmg3[0],&meout);
      } else {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;   // *6 for choice
    	if (momp < 1.) 
    	  Def3jq1();
    	else if (momp < 2.)
    	  Def3jq3();
    	else if (momp < 3.)
    	  Def3jq4();
    	else if (momp < 4.)
    	  Def3jq5();
    	else if (momp < 5.)
    	  Def3jq6();
    	else
    	  Def3jq7();
    	smatrixugtepveuuxd_(pmg3[0],&meout);
      }
    }
    else if (bptype==-1 || bptype==-3) {
      int proc=int(7.*ran.flat())+1;
      idpartfact*=7.;
      if (proc==1) {
    	double momp=ran.flat();
    	idpartfact*=4.;   // *2 for momp choice, *2 for identical gluons
    	if (momp < 0.5) 
    	  Def3jp3();
    	else 
    	  Def3jp5();
    	smatrixdxgtepveuxgg_(pmg3[0],&meout);
    	//	std::cout<<meout*2.<<"\n";
      } else if (proc==2) {
    	double momp=3.*ran.flat();
    	idpartfact*=6.;  //  *3 for choice, *2 for 2 'd's in final state
    	if (momp < 1.)
    	  Def3jq1();
    	else if (momp < 2.)
    	  Def3jq4();
    	else
    	  Def3jq5();
    	smatrixuxgtepveuxuxd_(pmg3[0],&meout);
    	//	std::cout<<meout<<"\n";
      } else if (proc==3) {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;
    	if (momp < 1.) 
    	  Def3jq1();
    	else if (momp < 2.)
    	  Def3jq3();
    	else if (momp < 3.)
    	  Def3jq4();
    	else if (momp < 4.)
    	  Def3jq5();
    	else if (momp < 5.)
    	  Def3jq6();
    	else
    	  Def3jq7();
    	smatrixuxgtepveuxcxs_(pmg3[0],&meout);
      } else if (proc==4) {
    	double momp=6.*ran.flat();
    	idpartfact*=12.;   // *6 for choice, *2 as two channels (ccbar or ssbar)
    	if (momp < 1.) 
    	  Def3jp1();
    	else if (momp < 2.)
    	  Def3jp3();
    	else if (momp < 3.)
    	  Def3jp4();
    	else if (momp < 4.)
    	  Def3jp5();
    	else if (momp < 5.)
    	  Def3jp6();
    	else
    	  Def3jp7();
    	smatrixdxgtepveuxssx_(pmg3[0],&meout);
       } else if (proc==5) {
    	double momp=3.*ran.flat();
    	idpartfact*=6.;  // *3 for choice, *2 for 2 'ux's in final state
    	if (momp<1.)
    	  Def3jp1();
    	else if (momp<2.)
    	  Def3jp4();
    	else 
    	  Def3jp5();
    	smatrixdxgtepveuxuxu_(pmg3[0],&meout);
       } else if (proc==6) {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;   // *6 for choice
    	if (momp < 1.) 
    	  Def3jp1();
    	else if (momp < 2.)
    	  Def3jp3();
    	else if (momp < 3.)
    	  Def3jp4();
    	else if (momp < 4.)
    	  Def3jp5();
    	else if (momp < 5.)
    	  Def3jp6();
    	else
    	  Def3jp7();
    	smatrixdxgtepvedxcxs_(pmg3[0],&meout);
       } else {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;   // *6 for choice
    	if (momp < 1.) 
    	  Def3jp1();
    	else if (momp < 2.)
    	  Def3jp3();
    	else if (momp < 3.)
    	  Def3jp4();
    	else if (momp < 4.)
    	  Def3jp5();
    	else if (momp < 5.)
    	  Def3jp6();
    	else
    	  Def3jp7();
    	smatrixdxgtepvedxdux_(pmg3[0],&meout);
      }
    }
    else {
      int proc=int(7.*ran.flat())+1;
      idpartfact*=7.;
      if (proc==1) {
    	double momp=ran.flat();
    	idpartfact*=4.;   // *2 for momp choice, *2 for identical gluons
    	if (momp < 0.5) 
    	  Def3jq3();
    	else 
    	  Def3jq5();
    	smatrixdxgtepveuxgg_(pmg3[0],&meout);
    	  //	std::cout<<meout*2.<<"\n";
      } else if (proc==2) {
    	double momp=3.*ran.flat();
    	idpartfact*=6.;  //  *3 for choice, *2 for 2 'd's in final state
    	if (momp < 1.)
    	  Def3jp1();
    	else if (momp < 2.)
    	  Def3jp4();
    	else
    	  Def3jp5();
    	smatrixuxgtepveuxuxd_(pmg3[0],&meout);
      } else if (proc==3) {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;
    	if (momp < 1.) 
    	  Def3jp1();
    	else if (momp < 2.)
    	  Def3jp3();
    	else if (momp < 3.)
    	  Def3jp4();
    	else if (momp < 4.)
    	  Def3jp5();
    	else if (momp < 5.)
    	  Def3jp6();
    	else
    	  Def3jp7();
    	smatrixuxgtepveuxcxs_(pmg3[0],&meout);
      } else if (proc==4) {
    	double momp=6.*ran.flat();
    	idpartfact*=12.;   // *6 for choice, *2 as two channels (ccbar or ssbar)
    	if (momp < 1.) 
    	  Def3jq1();
    	else if (momp < 2.)
    	  Def3jq3();
    	else if (momp < 3.)
    	  Def3jq4();
    	else if (momp < 4.)
    	  Def3jq5();
    	else if (momp < 5.)
    	  Def3jq6();
    	else
    	  Def3jq7();
    	smatrixdxgtepveuxssx_(pmg3[0],&meout);
      } else if (proc==5) {
    	double momp=3.*ran.flat();
    	idpartfact*=6.;  // *3 for choice, *2 for 2 'ux's in final state
    	if (momp<1.)
    	  Def3jq1();
    	else if (momp<2.)
    	  Def3jq4();
    	else 
    	  Def3jq5();
    	smatrixdxgtepveuxuxu_(pmg3[0],&meout);
      } else if (proc==6) {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;   // *6 for choice
    	if (momp < 1.) 
    	  Def3jq1();
    	else if (momp < 2.)
    	  Def3jq3();
    	else if (momp < 3.)
    	  Def3jq4();
    	else if (momp < 4.)
    	  Def3jq5();
    	else if (momp < 5.)
    	  Def3jq6();
    	else
    	  Def3jq7();
    	smatrixdxgtepvedxcxs_(pmg3[0],&meout);
      } else {
    	double momp=6.*ran.flat();
    	idpartfact*=6.;   // *6 for choice
    	if (momp < 1.) 
    	  Def3jq1();
    	else if (momp < 2.)
    	  Def3jq3();
    	else if (momp < 3.)
    	  Def3jq4();
    	else if (momp < 4.)
    	  Def3jq5();
    	else if (momp < 5.)
    	  Def3jq6();
    	else
    	  Def3jq7();
    	smatrixdxgtepvedxdux_(pmg3[0],&meout);
      }
    }
  }
  else if (bptype==21) {  // b=g, a!=g
    if (aptype==5) {
      int momp=int(12.*ran.flat())+1;
      idpartfact*=24.;   // *12 for choices  *2 for flavours (udx=csx, uxd=cxs)
	if (momp==1)
	  Def3jq2();
	else if (momp==2)
	  Def3jq8();
	else if (momp==3)
	  Def3jq9();
	else if (momp==4)
	  Def3jq10();
	else if (momp==5)
	  Def3jq11();
	else if (momp==6)
	  Def3jq12();
	else if (momp==7)
	  Def3jp2();
	else if (momp==8)
	  Def3jp8();
	else if (momp==9)
	  Def3jp9();
	else if (momp==10)
	  Def3jp10();
	else if (momp==11)
	  Def3jp11();
	else 
	  Def3jp12();
	smatrixsgtepvesdux_(pmg3[0],&meout);
	//	cout<<meout<<"\n";
    }
    else if (aptype==-5) {
      int momp=int(12.*ran.flat())+1;
      idpartfact*=24.;   // *12 for choices  *2 for flavours (udx=csx, uxd=cxs)
	if (momp==1)
	  Def3jq2();
	else if (momp==2)
	  Def3jq8();
	else if (momp==3)
	  Def3jq9();
	else if (momp==4)
	  Def3jq10();
	else if (momp==5)
	  Def3jq11();
	else if (momp==6)
	  Def3jq12();
	else if (momp==7)
	  Def3jp2();
	else if (momp==8)
	  Def3jp8();
	else if (momp==9)
	  Def3jp9();
	else if (momp==10)
	  Def3jp10();
	else if (momp==11)
	  Def3jp11();
	else if (momp==12)
	  Def3jp12();
	smatrixdxgtepvedxcxs_(pmg3[0],&meout);
    }
    else if (aptype==2 || aptype==4) { 
      int proc=int(7.*ran.flat())+1;
      idpartfact*=7.;
      if (proc==1) {
	double momp=ran.flat();
	idpartfact*=4.;   // *2 for momp choice, *2 for identical gluons
	if (momp < 0.5) 
	  Def3jp8();
	else 
	  Def3jp10();
	smatrixugtepvedgg_(pmg3[0],&meout);
	//	std::cout<<meout*2.<<"\n";
      } else if (proc==2) {
	double momp=3.*ran.flat();
	idpartfact*=6.;  //  *3 for choice, *2 for 2 'd's in final state
	if (momp < 1.)
	  Def3jq2();
	else if (momp < 2.)
	  Def3jq9();
	else
	  Def3jq10();
	smatrixdgtepveddux_(pmg3[0],&meout);
      } else if (proc==3) {
	double momp=6.*ran.flat();
	idpartfact*=6.;
	if (momp < 1.) 
	  Def3jq2();
	else if (momp < 2.)
	  Def3jq8();
	else if (momp < 3.)
	  Def3jq9();
	else if (momp < 4.)
	  Def3jq10();
	else if (momp < 5.)
	  Def3jq11();
	else
	  Def3jq12();
	smatrixsgtepvesdux_(pmg3[0],&meout);
      } else if (proc==4) {
	double momp=6.*ran.flat();
	idpartfact*=12.;   // *6 for choice, *2 as two channels (ccbar or ssbar)
	if (momp < 1.) 
	  Def3jp2();
	else if (momp < 2.)
	  Def3jp8();
	else if (momp < 3.)
	  Def3jp9();
	else if (momp < 4.)
	  Def3jp10();
	else if (momp < 5.)
	  Def3jp11();
	else
	  Def3jp12();
	smatrixugtepveccxd_(pmg3[0],&meout);
      } else if (proc==5) {
	double momp=3.*ran.flat();
	idpartfact*=6.;  // *3 for choice, *2 for 2 'd's in final state
	if (momp<1.)
	  Def3jp2();
	else if (momp<2.)
	  Def3jp9();
	else 
	  Def3jp10();
	smatrixugtepvedddx_(pmg3[0],&meout);
      } else if (proc==6) {
	double momp=6.*ran.flat();
	idpartfact*=6.;   // *6 for choice
	if (momp < 1.) 
	  Def3jp2();
	else if (momp < 2.)
	  Def3jp8();
	else if (momp < 3.)
	  Def3jp9();
	else if (momp < 4.)
	  Def3jp10();
	else if (momp < 5.)
	  Def3jp11();
	else
	  Def3jp12();
	smatrixugtepveucxs_(pmg3[0],&meout);
      } else {
	double momp=6.*ran.flat();
	idpartfact*=6.;   // *6 for choice
	if (momp < 1.) 
	  Def3jp2();
	else if (momp < 2.)
	  Def3jp8();
	else if (momp < 3.)
	  Def3jp9();
	else if (momp < 4.)
	  Def3jp10();
	else if (momp < 5.)
	  Def3jp11();
	else
	  Def3jp12();
	smatrixugtepveuuxd_(pmg3[0],&meout);
      }
    }
    else if (aptype>0) {
      int proc=int(7.*ran.flat())+1;
      idpartfact*=7.;
      if (proc==1) {
	double momp=ran.flat();
	idpartfact*=4.;   // *2 for momp choice, *2 for identical gluons
	if (momp < 0.5) {
	  Def3jq8();
	  smatrixugtepvedgg_(pmg3[0],&meout);
	} else {
	  Def3jq10();
	  smatrixugtepvedgg_(pmg3[0],&meout);
	}
      } else if (proc==2) {
	double momp=3.*ran.flat();
	idpartfact*=6.;  //  *3 for choice, *2 for 2 'd's in final state
	if (momp < 1.)
	  Def3jp2();
	else if (momp < 2.)
	  Def3jp9();
	else
	  Def3jp10();
	smatrixdgtepveddux_(pmg3[0],&meout);
      } else if (proc==3) {
	double momp=6.*ran.flat();
	idpartfact*=6.;
	if (momp < 1.) 
	  Def3jp2();
	else if (momp < 2.)
	  Def3jp8();
	else if (momp < 3.)
	  Def3jp9();
	else if (momp < 4.)
	  Def3jp10();
	else if (momp < 5.)
	  Def3jp11();
	else
	  Def3jp12();
	smatrixsgtepvesdux_(pmg3[0],&meout);
      } else if (proc==4) {
	double momp=6.*ran.flat();
	idpartfact*=12.;   // *6 for choice, *2 as two channels (ccbar or ssbar)
	if (momp < 1.) 
	  Def3jq2();
	else if (momp < 2.)
	  Def3jq8();
	else if (momp < 3.)
	  Def3jq9();
	else if (momp < 4.)
	  Def3jq10();
	else if (momp < 5.)
	  Def3jq11();
	else
	  Def3jq12();
	smatrixugtepveccxd_(pmg3[0],&meout);
      } else if (proc==5) {
	double momp=3.*ran.flat();
	idpartfact*=6.;  // *3 for choice, *2 for 2 'd's in final state
	if (momp<1.)
	  Def3jq2();
	else if (momp<2.)
	  Def3jq9();
	else 
	  Def3jq10();
	smatrixugtepvedddx_(pmg3[0],&meout);
      } else if (proc==6) {
	double momp=6.*ran.flat();
	idpartfact*=6.;   // *6 for choice
	if (momp < 1.) 
	  Def3jq2();
	else if (momp < 2.)
	  Def3jq8();
	else if (momp < 3.)
	  Def3jq9();
	else if (momp < 4.)
	  Def3jq10();
	else if (momp < 5.)
	  Def3jq11();
	else
	  Def3jq12();
	smatrixugtepveucxs_(pmg3[0],&meout);
      } else {
	double momp=6.*ran.flat();
	idpartfact*=6.;   // *6 for choice
	if (momp < 1.) 
	  Def3jq2();
	else if (momp < 2.)
	  Def3jq8();
	else if (momp < 3.)
	  Def3jq9();
	else if (momp < 4.)
	  Def3jq10();
	else if (momp < 5.)
	  Def3jq11();
	else
	  Def3jq12();
	smatrixugtepveuuxd_(pmg3[0],&meout);
      }
    }
    else if (aptype==-1 || aptype==-3) {
      int proc=int(7.*ran.flat())+1;
      idpartfact*=7.;
      if (proc==1) {
	double momp=ran.flat();
  	idpartfact*=4.;   // *2 for momp choice, *2 for identical gluons
  	if (momp < 0.5) 
  	  Def3jp8();
  	else 
  	  Def3jp10();
  	smatrixdxgtepveuxgg_(pmg3[0],&meout);
	//	std::cout<<meout*2.<<"\n";
      } else if (proc==2) {
	double momp=3.*ran.flat();
	idpartfact*=6.;  //  *3 for choice, *2 for 2 'd's in final state
	if (momp < 1.)
	  Def3jq2();
	else if (momp < 2.)
	  Def3jq9();
	else
	  Def3jq10();
	smatrixuxgtepveuxuxd_(pmg3[0],&meout);
	//	std::cout<<meout<<"\n";
      } else if (proc==3) {
  	double momp=6.*ran.flat();
  	idpartfact*=6.;
	if (momp < 1.) 
  	  Def3jq2();
	else if (momp < 2.)
  	  Def3jq8();
	else if (momp < 3.)
  	  Def3jq9();
	else if (momp < 4.)
  	  Def3jq10();
	else if (momp < 5.)
  	  Def3jq11();
	else
	  Def3jq12();
	smatrixuxgtepveuxcxs_(pmg3[0],&meout);
      } else if (proc==4) {
  	double momp=6.*ran.flat();
  	idpartfact*=12.;   // *6 for choice, *2 as two channels (ccbar or ssbar)
  	if (momp < 1.) 
  	  Def3jp2();
  	else if (momp < 2.)
  	  Def3jp8();
  	else if (momp < 3.)
  	  Def3jp9();
  	else if (momp < 4.)
  	  Def3jp10();
  	else if (momp < 5.)
  	  Def3jp11();
  	else
  	  Def3jp12();
  	smatrixdxgtepveuxssx_(pmg3[0],&meout);
       } else if (proc==5) {
  	double momp=3.*ran.flat();
  	idpartfact*=6.;  // *3 for choice, *2 for 2 'ux's in final state
  	if (momp<1.)
  	  Def3jp2();
  	else if (momp<2.)
  	  Def3jp9();
  	else 
  	  Def3jp10();
  	smatrixdxgtepveuxuxu_(pmg3[0],&meout);
       } else if (proc==6) {
  	double momp=6.*ran.flat();
  	idpartfact*=6.;   // *6 for choice
  	if (momp < 1.) 
  	  Def3jp2();
  	else if (momp < 2.)
  	  Def3jp8();
  	else if (momp < 3.)
  	  Def3jp9();
  	else if (momp < 4.)
  	  Def3jp10();
  	else if (momp < 5.)
  	  Def3jp11();
  	else
  	  Def3jp12();
  	smatrixdxgtepvedxcxs_(pmg3[0],&meout);
       } else {
  	double momp=6.*ran.flat();
  	idpartfact*=6.;   // *6 for choice
  	if (momp < 1.) 
  	  Def3jp2();
  	else if (momp < 2.)
  	  Def3jp8();
  	else if (momp < 3.)
  	  Def3jp9();
  	else if (momp < 4.)
  	  Def3jp10();
  	else if (momp < 5.)
  	  Def3jp11();
  	else
  	  Def3jp12();
  	smatrixdxgtepvedxdux_(pmg3[0],&meout);
      }
    }
    else {
      int proc=int(7.*ran.flat())+1;
      idpartfact*=7.;
      if (proc==1) {
	double momp=ran.flat();
	idpartfact*=4.;   // *2 for momp choice, *2 for identical gluons
	if (momp < 0.5) 
	  Def3jq8();
	else 
	  Def3jq10();
	smatrixdxgtepveuxgg_(pmg3[0],&meout);
	  //	std::cout<<meout*2.<<"\n";
      } else if (proc==2) {
	double momp=3.*ran.flat();
	idpartfact*=6.;  //  *3 for choice, *2 for 2 'd's in final state
	if (momp < 1.)
	  Def3jp2();
	else if (momp < 2.)
	  Def3jp9();
	else
	  Def3jp10();
	smatrixuxgtepveuxuxd_(pmg3[0],&meout);
      } else if (proc==3) {
	double momp=6.*ran.flat();
	idpartfact*=6.;
	if (momp < 1.) 
	  Def3jp2();
	else if (momp < 2.)
	  Def3jp8();
	else if (momp < 3.)
	  Def3jp9();
	else if (momp < 4.)
	  Def3jp10();
	else if (momp < 5.)
	  Def3jp11();
	else
	  Def3jp12();
	smatrixuxgtepveuxcxs_(pmg3[0],&meout);
      } else if (proc==4) {
	double momp=6.*ran.flat();
	idpartfact*=12.;   // *6 for choice, *2 as two channels (ccbar or ssbar)
	if (momp < 1.) 
	  Def3jq2();
	else if (momp < 2.)
	  Def3jq8();
	else if (momp < 3.)
	  Def3jq9();
	else if (momp < 4.)
	  Def3jq10();
	else if (momp < 5.)
	  Def3jq11();
	else
	  Def3jq12();
	smatrixdxgtepveuxssx_(pmg3[0],&meout);
      } else if (proc==5) {
	double momp=3.*ran.flat();
	idpartfact*=6.;  // *3 for choice, *2 for 2 'ux's in final state
	if (momp<1.)
	  Def3jq2();
	else if (momp<2.)
	  Def3jq9();
	else 
	  Def3jq10();
	smatrixdxgtepveuxuxu_(pmg3[0],&meout);
      } else if (proc==6) {
	double momp=6.*ran.flat();
	idpartfact*=6.;   // *6 for choice
	if (momp < 1.) 
	  Def3jq2();
	else if (momp < 2.)
	  Def3jq8();
	else if (momp < 3.)
	  Def3jq9();
	else if (momp < 4.)
	  Def3jq10();
	else if (momp < 5.)
	  Def3jq11();
	else
	  Def3jq12();
	smatrixdxgtepvedxcxs_(pmg3[0],&meout);
      } else {
	double momp=6.*ran.flat();
	idpartfact*=6.;   // *6 for choice
	if (momp < 1.) 
	  Def3jq2();
	else if (momp < 2.)
	  Def3jq8();
	else if (momp < 3.)
	  Def3jq9();
	else if (momp < 4.)
	  Def3jq10();
	else if (momp < 5.)
	  Def3jq11();
	else
	  Def3jq12();
	smatrixdxgtepvedxdux_(pmg3[0],&meout);
      }
    }
  }
  else {  //  MARKER - two (anti-)quarks to deal with!
    if (aptype==bptype) {
      if (fabs(aptype)==5)
	return 0;     // No bb or bxbx processes without tops
      else if (aptype==2 || aptype==4) {   // uu or cc
	double momp=ran.flat();
	idpartfact*=4.;   // *4 for choice
	if (momp<0.5) {
	  if (momp<0.25)
	    Def3jp3();
	  else
	    Def3jp4();
	} else {
	  if (momp<0.75)
	    Def3jp6();
	  else
	    Def3jp7();
	}
	smatrixuutepveugd_(pmg3[0],&meout);
      }
      else if (aptype > 0) {   // dd or ss
	double momp=ran.flat();
	idpartfact*=4.;   // *4 for choice
	if (momp<0.5) {
	  if (momp<0.25)
	    Def3jq3();
	  else
	    Def3jq4();
	} else {
	  if (momp<0.75)
	    Def3jq6();
	  else
	    Def3jq7();
	}
	smatrixuutepveugd_(pmg3[0],&meout);
      }
      else if (aptype==-1 || aptype==-3) {   // dxdx or sxsx
	double momp=ran.flat();
	idpartfact*=4.;   // *4 for choice
	if (momp<0.5) {
	  if (momp<0.25)
	    Def3jp3();
	  else
	    Def3jp4();
	} else {
	  if (momp<0.75)
	    Def3jp6();
	  else
	    Def3jp7();
	}
	smatrixdxdxtepveuxgdx_(pmg3[0],&meout);
      }
      else {             // uxux or cxcx
	double momp=ran.flat();
	idpartfact*=4.;   // *4 for choice
	if (momp<0.5) {
	  if (momp<0.25)
	    Def3jq3();
	  else
	    Def3jq4();
	} else {
	  if (momp<0.75)
	    Def3jq6();
	  else
	    Def3jq7();
	}
	smatrixdxdxtepveuxgdx_(pmg3[0],&meout);
      }
    }
    else if (aptype==-bptype) {    /// MARKER - q qbar same flavour ///
      if (fabs(aptype)==5) {   //  ME is symmetric under s<-->sx
	int momp=int(12.*ran.flat())+1;   // 12 momentum configs
	idpartfact*=24.;    // *12 for choice *2 for flavours
	if (momp<7) {
	  if (momp<4) {
	    if (momp<2)
	      Def3jp1();
	    else if (momp<3)
	      Def3jp3();
	    else
	      Def3jp4();
	  } else {
	    if (momp<5)
	      Def3jp5();
	    else if (momp<6)
	      Def3jp6();
	    else
	      Def3jp7();
	  }
	} else {
	  if (momp<10) {
	    if (momp<8)
	      Def3jq1();
	    else if (momp<9)
	      Def3jq3();
	    else
	      Def3jq4();
	  } else {
	    if (momp<11)
	      Def3jq5();
	    else if (momp<12)
	      Def3jq6();
	    else
	      Def3jq7();
	  }
	}
	smatrixssxtepveuxgd_(pmg3[0],&meout);
      }
      else {                    //  not b bbar so uux-->dux, udx, scx, csx
    	double process=ran.flat();  
    	idpartfact*=4.;
    	if (process<0.5) {   // same flavour outgoing
    	  if (aptype==2 || aptype==4) {  // backward particle is q, forward qbar
    	    if (process<0.25) {
	      double momp=ran.flat();
	      idpartfact*=5.;
	      if (momp<0.2)
		Def3jp3();
	      else if (momp<0.4)
		Def3jp4();
	      else if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
    	      smatrixuxutepveuxgd_(pmg3[0],&meout);  // uux->e+ nu dux
	    }
    	    else {
	      double momp=ran.flat();
	      idpartfact*=5.;
	      if (momp<0.2)
		Def3jq3();
	      else if (momp<0.4)
		Def3jq4();
	      else if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
    	      smatrixdxdtepveuxgd_(pmg3[0],&meout);  // uux->e- nux udx
	    }
    	  }
	  else if (aptype==1 || aptype==3) {
    	    if (process<0.25) {
	      double momp=ran.flat();
	      idpartfact*=5.;
	      if (momp<0.2)
		Def3jq3();
	      else if (momp<0.4)
		Def3jq4();
	      else if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
    	      smatrixuxutepveuxgd_(pmg3[0],&meout);  // uux->e+ nu dux
	    }
    	    else {
	      double momp=ran.flat();
	      idpartfact*=5.;
	      if (momp<0.2)
		Def3jp3();
	      else if (momp<0.4)
		Def3jp4();
	      else if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
    	      smatrixdxdtepveuxgd_(pmg3[0],&meout);  // uux->e- nux udx
	    }
	  }
     	  else if (aptype==-1 || aptype==-3) {
    	    if (process<0.25) {
	      double momp=ran.flat();
	      idpartfact*=5.;
	      if (momp<0.2)
		Def3jq8();
	      else if (momp<0.4)
		Def3jq9();
	      else if (momp<0.6)
		Def3jq10();
	      else if (momp<0.8)
		Def3jq11();
	      else
		Def3jq12();
    	      smatrixuxutepveuxgd_(pmg3[0],&meout);  // uux->e+ nu dux
	    }
    	    else {
	      double momp=ran.flat();
	      idpartfact*=5.;
	      if (momp<0.2)
		Def3jp8();
	      else if (momp<0.4)
		Def3jp9();
	      else if (momp<0.6)
		Def3jp10();
	      else if (momp<0.8)
		Def3jp11();
	      else
		Def3jp12();
    	      smatrixdxdtepveuxgd_(pmg3[0],&meout);  // uux->e- nux udx
	    }
     	  }
     	  else {  // aptype==-2, -4
    	    if (process<0.25) {
	      double momp=ran.flat();
	      idpartfact*=5.;
	      if (momp<0.2)
		Def3jp8();
	      else if (momp<0.4)
		Def3jp9();
	      else if (momp<0.6)
		Def3jp10();
	      else if (momp<0.8)
		Def3jp11();
	      else
		Def3jp12();
    	      smatrixuxutepveuxgd_(pmg3[0],&meout);  // uux->e+ nu dux
	    }
    	    else {
	      double momp=ran.flat();
	      idpartfact*=5.;
	      if (momp<0.2)
		Def3jq8();
	      else if (momp<0.4)
		Def3jq9();
	      else if (momp<0.6)
		Def3jq10();
	      else if (momp<0.8)
		Def3jq11();
	      else
		Def3jq12();
    	      smatrixdxdtepveuxgd_(pmg3[0],&meout);  // uux->e- nux udx
	    }
     	  }
     	}
	else {
   	  int mom=int(6.*ran.flat())+1;
   	  idpartfact*=6.;
   	  if (process<0.75) { // uux->e+ nu scx    , ME is symmetric s<-->sx
  	    if (mom ==1) 
	      Def3jp1();
	    else if (mom==2)
	      Def3jp3();
	    else if (mom==3)
	      Def3jp4();
	    else if (mom==4)
	      Def3jp5();
	    else if (mom==5)
	      Def3jp6();
	    else
	      Def3jp7();
	    smatrixssxtepveuxgd_(pmg3[0],&meout);  
    	  }
    	  else {   // uux -> e- nux c sx
  	    if (mom ==1) 
	      Def3jq1();
	    else if (mom==2)
	      Def3jq3();
	    else if (mom==3)
	      Def3jq4();
	    else if (mom==4)
	      Def3jq5();
	    else if (mom==5)
	      Def3jq6();
	    else
	      Def3jq7();
	    smatrixssxtepveuxgd_(pmg3[0],&meout);  
	  }
     	}
      }
    }
    else if (aptype==1) {    // a=d, b!=d,dbar   ///  MARKER a = down //
      if (bptype==2) {  // b=u
	double momp=ran.flat();
	idpartfact*=8.; // *4 for choice, *2 for identical final state particles
	if (momp<0.5) {
	  if (momp<0.25) 
	    Def3jp3();
	  else 
	    Def3jp4();
	} else {
	  if (momp<0.75)
	    Def3jq8();
	  else
	    Def3jq9();
	}
	smatrixudtepvedgd_(pmg3[0],&meout);
      }
      else if (bptype==5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq8();
	  else
	    Def3jq9();
	} else {
	  if (momp<0.6)
	    Def3jq10();
	  else if (momp<0.8)
	    Def3jq11();
	  else
	    Def3jq12();
	}
    	smatrixustepvedgs_(pmg3[0],&meout);   // Don't emit from b
      }
      else if (bptype>2) {
    	double r1=ran.flat();
	double momp=ran.flat();
	idpartfact*=10.;
    	if (r1<0.5) {   // emit from d
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
	  smatrixustepvedgs_(pmg3[0],&meout);
    	}
    	else {   
    	  if (bptype==4) { // emit from c
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixustepvedgs_(pmg3[0],&meout);
  	  }
    	  else {   // emit from s 
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixustepvedgs_(pmg3[0],&meout);
  	  }
    	}
      }
      else if (bptype==-2) {
    	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4
    	idpartfact*=4.;
    	if (process==1) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
    	  smatrixudxtepvedgdx_(pmg3[0],&meout);
  	}
    	else if (process==2) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
    	  smatrixudxtepveugux_(pmg3[0],&meout);
  	}
    	else if (process==3) {
	  int proc=int(ran.flat()*6.)+1;
    	  idpartfact*=18.;    // *6 for choice, *3 flavours
	  if (proc<3.5) {
	    if (proc==1)
	      Def3jq1();
	    else if (proc==2)
	      Def3jq3();
	    else
	      Def3jq4();
	  } else {
	    if (proc==4)
	      Def3jq5();
	    else if (proc==5)
	      Def3jq6();
	    else
	      Def3jq7();
	  }
    	  smatrixdxutepvesgsx_(pmg3[0],&meout);
    	}
    	else {
    	  idpartfact*=6.;    // *6 for ggg final state
	  Def3jq1();
    	  smatrixdxutepveggg_(pmg3[0],&meout);
    	}
      }
      else if (bptype==-5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq8();
	  else
	    Def3jq9();
	} else {
	  if (momp<0.6)
	    Def3jq10();
	  else if (momp<0.8)
	    Def3jq11();
	  else
	    Def3jq12();
	}
    	smatrixusxtepvedgsx_(pmg3[0],&meout);
      }
      else {    // b is sx or cx
    	double r2=ran.flat();
	double momp=ran.flat();
    	idpartfact*=10.;
    	if (r2<0.5) {    // emit from d
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
	  smatrixusxtepvedgsx_(pmg3[0],&meout);
  	}
    	else {
    	  if (bptype==-3) {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixdxstepveuxgs_(pmg3[0],&meout);
  	  }
    	  else {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixdxstepveuxgs_(pmg3[0],&meout);
  	  }
    	}
      }
    }
    else if (aptype==2) { // MARKER a=u //
      if (bptype==1) {  // b=u
	double momp=ran.flat();
	idpartfact*=8.; // *4 for choice, *2 for identical final state particles
	if (momp<0.5) {
	  if (momp<0.25) 
	    Def3jp8();
	  else 
	    Def3jp9();
	} else {
	  if (momp<0.75) 
	    Def3jq3();
	  else
	    Def3jq4();
	}
	smatrixudtepvedgd_(pmg3[0],&meout);
      }
      else if (bptype==5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp8();
	  else
	    Def3jp9();
	} else {
	  if (momp<0.6)
	    Def3jp10();
	  else if (momp<0.8)
	    Def3jp11();
	  else
	    Def3jp12();
	}
    	smatrixustepvedgs_(pmg3[0],&meout);   // Don't emit from b
      }
      else if (bptype>2) {
    	double r1=ran.flat();
	double momp=ran.flat();
	idpartfact*=10.;
    	if (r1<0.5) {   // emit from u
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
	  smatrixustepvedgs_(pmg3[0],&meout);
    	}
    	else {   
    	  if (bptype==4) { // emit from c
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixustepvedgs_(pmg3[0],&meout);
  	  }
    	  else {   // emit from s 
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixustepvedgs_(pmg3[0],&meout);
  	  }
    	}
      }
      else if (bptype==-1) {
    	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4
    	idpartfact*=4.;
    	if (process==1) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
    	  smatrixudxtepvedgdx_(pmg3[0],&meout);
  	}
    	else if (process==2) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
    	  smatrixudxtepveugux_(pmg3[0],&meout);
  	}
    	else if (process==3) {
	  int proc=int(ran.flat()*6.)+1;
    	  idpartfact*=18.;    // *6 for choice, *3 flavours
	  if (proc<3.5) {
	    if (proc==1)
	      Def3jp1();
	    else if (proc==2)
	      Def3jp3();
	    else
	      Def3jp4();
	  } else {
	    if (proc==4)
	      Def3jp5();
	    else if (proc==5)
	      Def3jp6();
	    else
	      Def3jp7();
	  }
    	  smatrixdxutepvesgsx_(pmg3[0],&meout);
    	}
    	else {
    	  idpartfact*=6.;    // *6 for ggg final state
	  Def3jp1();
    	  smatrixdxutepveggg_(pmg3[0],&meout);
    	}
      }
      else if (bptype==-5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp8();
	  else
	    Def3jp9();
	} else {
	  if (momp<0.6)
	    Def3jp10();
	  else if (momp<0.8)
	    Def3jp11();
	  else
	    Def3jp12();
	}
    	smatrixusxtepvedgsx_(pmg3[0],&meout);
      }
      else {    // b is sx or cx
    	double r2=ran.flat();
	double momp=ran.flat();
    	idpartfact*=10.;
    	if (r2<0.5) {    // emit from d
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
	  smatrixusxtepvedgsx_(pmg3[0],&meout);
  	}
    	else {
    	  if (bptype==-3) {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixdxstepveuxgs_(pmg3[0],&meout);
  	  }
    	  else {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixdxstepveuxgs_(pmg3[0],&meout);
  	  }
    	}
      }
    }
    else if (aptype==3) { // a=s
      if (bptype==4) {  // b=c
	double momp=ran.flat();
	idpartfact*=8.; // *4 for choice, *2 for identical final state particles
	if (momp<0.5) {
	  if (momp<0.25)
	    Def3jp3();
	  else
	    Def3jp4();
	}
	else {
	  if (momp<0.75) 
	    Def3jq8();
	  else
	    Def3jq9();
	}
	smatrixudtepvedgd_(pmg3[0],&meout);
      }
      else if (bptype==5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq8();
	  else
	    Def3jq9();
	} else {
	  if (momp<0.6)
	    Def3jq10();
	  else if (momp<0.8)
	    Def3jq11();
	  else
	    Def3jq12();
	}
    	smatrixustepvedgs_(pmg3[0],&meout);   // Don't emit from b
      }
      else if (bptype>0) {
    	double r1=ran.flat();
	double momp=ran.flat();
	idpartfact*=10.;
    	if (r1<0.5) {   // emit from d
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
	  smatrixustepvedgs_(pmg3[0],&meout);
    	}
    	else {   
    	  if (bptype==2) { // emit from u
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixustepvedgs_(pmg3[0],&meout);
  	  }
    	  else {   // emit from d 
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixustepvedgs_(pmg3[0],&meout);
  	  }
    	}
      }
      else if (bptype==-4) {
    	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4
    	idpartfact*=4.;
    	if (process==1) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
    	  smatrixudxtepvedgdx_(pmg3[0],&meout);
  	}
    	else if (process==2) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
    	  smatrixudxtepveugux_(pmg3[0],&meout);
  	}
    	else if (process==3) {
	  int proc=int(ran.flat()*6.)+1;
    	  idpartfact*=18.;    // *6 for choice, *3 flavours
	  if (proc<3.5) {
	    if (proc==1)
	      Def3jq1();
	    else if (proc==2)
	      Def3jq3();
	    else
	      Def3jq4();
	  } else {
	    if (proc==4)
	      Def3jq5();
	    else if (proc==5)
	      Def3jq6();
	    else
	      Def3jq7();
	  }
    	  smatrixdxutepvesgsx_(pmg3[0],&meout);
    	}
    	else {
    	  idpartfact*=6.;    // *6 for ggg final state
	  Def3jq1();
    	  smatrixdxutepveggg_(pmg3[0],&meout);
    	}
      }
      else if (bptype==-5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq8();
	  else
	    Def3jq9();
	} else {
	  if (momp<0.6)
	    Def3jq10();
	  else if (momp<0.8)
	    Def3jq11();
	  else
	    Def3jq12();
	}
    	smatrixusxtepvedgsx_(pmg3[0],&meout);
      }
      else {    // b is sx or cx
    	double r2=ran.flat();
	double momp=ran.flat();
    	idpartfact*=10.;
    	if (r2<0.5) {    // emit from d
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
	  smatrixusxtepvedgsx_(pmg3[0],&meout);
  	}
    	else {
    	  if (bptype==-1) {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixdxstepveuxgs_(pmg3[0],&meout);
  	  }
    	  else {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixdxstepveuxgs_(pmg3[0],&meout);
  	  }
    	}
      }
    }
    else if (aptype==4) { // a=c
      if (bptype==3) {  // b=s
	double momp=ran.flat();
	idpartfact*=8.; // *5 for choice, *2 for identical final state particles
	if (momp<0.5) {
	  if (momp<0.25) 
	    Def3jp8();
	  else
	    Def3jp9();
	} else {
	  if (momp<0.75) 
	    Def3jq3();
	  else
	    Def3jq4();
	}
	smatrixudtepvedgd_(pmg3[0],&meout);
      }
      else if (bptype==5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp8();
	  else
	    Def3jp9();
	} else {
	  if (momp<0.6)
	    Def3jp10();
	  else if (momp<0.8)
	    Def3jp11();
	  else
	    Def3jp12();
	}
    	smatrixustepvedgs_(pmg3[0],&meout);   // Don't emit from b
      }
      else if (bptype>0) {
    	double r1=ran.flat();
	double momp=ran.flat();
	idpartfact*=10.;
    	if (r1<0.5) {   // emit from c
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
	  smatrixustepvedgs_(pmg3[0],&meout);
    	}
    	else {   
    	  if (bptype==2) { // emit from u
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixustepvedgs_(pmg3[0],&meout);
  	  }
    	  else {   // emit from s 
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixustepvedgs_(pmg3[0],&meout);
  	  }
    	}
      }
      else if (bptype==-3) {
    	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4
    	idpartfact*=4.;
    	if (process==1) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
    	  smatrixudxtepvedgdx_(pmg3[0],&meout);
  	}
    	else if (process==2) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
    	  smatrixudxtepveugux_(pmg3[0],&meout);
  	}
    	else if (process==3) {
	  int proc=int(ran.flat()*6.)+1;
    	  idpartfact*=18.;    // *6 for choice, *3 flavours
	  if (proc<3.5) {
	    if (proc==1)
	      Def3jp1();
	    else if (proc==2)
	      Def3jp3();
	    else
	      Def3jp4();
	  } else {
	    if (proc==4)
	      Def3jp5();
	    else if (proc==5)
	      Def3jp6();
	    else
	      Def3jp7();
	  }
    	  smatrixdxutepvesgsx_(pmg3[0],&meout);
    	}
    	else {
    	  idpartfact*=6.;    // *6 for ggg final state
	  Def3jp1();
    	  smatrixdxutepveggg_(pmg3[0],&meout);
    	}
      }
      else if (bptype==-5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp8();
	  else
	    Def3jp9();
	} else {
	  if (momp<0.6)
	    Def3jp10();
	  else if (momp<0.8)
	    Def3jp11();
	  else
	    Def3jp12();
	}
    	smatrixusxtepvedgsx_(pmg3[0],&meout);
      }
      else {    // b is sx or cx
    	double r2=ran.flat();
	double momp=ran.flat();
    	idpartfact*=10.;
    	if (r2<0.5) {    // emit from d
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
	  smatrixusxtepvedgsx_(pmg3[0],&meout);
  	}
    	else {
    	  if (bptype==-1) {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixdxstepveuxgs_(pmg3[0],&meout);
  	  }
    	  else {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixdxstepveuxgs_(pmg3[0],&meout);
  	  }
    	}
      }
    }
    else if (aptype>0) { // a=b
      if (bptype==2 || bptype==4) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp3();
	  else
	    Def3jp4();
	}
	else {
	  if (momp<0.6)
	    Def3jp5();
	  else if (momp<0.8)
	    Def3jp6();
	  else
	    Def3jp7();
	}
	smatrixustepvedgs_(pmg3[0],&meout);
      }
      else if (bptype>0) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq3();
	  else
	    Def3jq4();
	}
	else {
	  if (momp<0.6)
	    Def3jq5();
	  else if (momp<0.8)
	    Def3jq6();
	  else
	    Def3jq7();
	}
	smatrixustepvedgs_(pmg3[0],&meout);
      }
      else if (bptype==-1 || bptype==-3) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp3();
	  else
	    Def3jp4();
	}
	else {
	  if (momp<0.6)
	    Def3jp5();
	  else if (momp<0.8)
	    Def3jp6();
	  else
	    Def3jp7();
	}
	smatrixdxstepveuxgs_(pmg3[0],&meout);
      }
      else {   // bptype = ubar, cbar
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq3();
	  else
	    Def3jq4();
	}
	else {
	  if (momp<0.6)
	    Def3jq5();
	  else if (momp<0.8)
	    Def3jq6();
	  else
	    Def3jq7();
	}
	smatrixdxstepveuxgs_(pmg3[0],&meout);
      }
    }
  //   // ///  Now go for a = anti-quark   --- MARKER a=qbar  ///
    else if (aptype==-1) {    // a=dbar, b!=d
      if (bptype==-2) {  // b=u
	double momp=ran.flat();
	idpartfact*=8.; // *4 for choice, *2 for identical final state particles
	if (momp<0.5) {
	  if (momp<0.25) 
	    Def3jp8();
	  else 
	    Def3jp9();
	} else {
	  if (momp<0.75) 
	    Def3jq3();
	  else
	    Def3jq4();
	}
	smatrixdxuxtepveuxgux_(pmg3[0],&meout);
      }
      else if (bptype==-5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp8();
	  else
	    Def3jp9();
	} else {
	  if (momp<0.6)
	    Def3jp10();
	  else if (momp<0.8)
	    Def3jp11();
	  else
	    Def3jp12();
	}
    	smatrixdxsxtepveuxgsx_(pmg3[0],&meout);   // Don't emit from b
      }
      else if (bptype<0) {
    	double r1=ran.flat();
	double momp=ran.flat();
	idpartfact*=10.;
    	if (r1<0.5) {   // emit from dx
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
	  smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
    	}
    	else {   
    	  if (bptype==-3) { // emit from sx
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
  	  }
    	  else {   // emit from cx 
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
  	  }
    	}
      }
      else if (bptype==2) {
	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4
    	idpartfact*=4.;
    	if (process==1) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp3();
	    else
	      Def3jp4();
	  } else {
	    if (momp<0.6)
	      Def3jp5();
	    else if (momp<0.8)
	      Def3jp6();
	    else
	      Def3jp7();
	  }
    	  smatrixudxtepvedgdx_(pmg3[0],&meout);
  	}
    	else if (process==2) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp3();
	    else
	      Def3jp4();
	  } else {
	    if (momp<0.6)
	      Def3jp5();
	    else if (momp<0.8)
	      Def3jp6();
	    else
	      Def3jp7();
	  }
    	  smatrixudxtepveugux_(pmg3[0],&meout);
  	}
    	else if (process==3) {
	  int proc=int(ran.flat()*6.)+1;
    	  idpartfact*=18.;    // *6 for choice, *3 flavours
	  if (proc<3.5) {
	    if (proc==1)
	      Def3jp2();
	    else if (proc==2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (proc==4)
	      Def3jp10();
	    else if (proc==5)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
    	  smatrixdxutepvesgsx_(pmg3[0],&meout);
    	}
    	else {
    	  idpartfact*=6.;    // *6 for ggg final state
	  Def3jp2();
    	  smatrixdxutepveggg_(pmg3[0],&meout);
    	}
      }
      else if (bptype==5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp8();
	  else
	    Def3jp9();
	} else {
	  if (momp<0.6)
	    Def3jp10();
	  else if (momp<0.8)
	    Def3jp11();
	  else
	    Def3jp12();
	}
    	smatrixdxstepveuxgs_(pmg3[0],&meout);
      }
      else {   // b=s,c
    	double r2=ran.flat();
	double momp=ran.flat();
    	idpartfact*=10.;
    	if (r2<0.5) {    // emit from dx
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
	  smatrixdxstepveuxgs_(pmg3[0],&meout);
  	}
    	else {
    	  if (bptype==4) {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixusxtepvedgsx_(pmg3[0],&meout);
  	  }
    	  else {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixusxtepvedgsx_(pmg3[0],&meout);
  	  }
	}
      }
    }
    else if (aptype==-2) {    // a=ubar, b!=u
      if (bptype==-1) {  // b=dx
	double momp=ran.flat();
	idpartfact*=8.; // *4 for choice, *2 for identical final state particles
	if (momp<0.5) {
	  if (momp<0.25) 
	    Def3jq8();
	  else 
	    Def3jq9();
	} else {
	  if (momp<0.75) 
	    Def3jp3();
	  else
	    Def3jp4();
	}
	smatrixdxuxtepveuxgux_(pmg3[0],&meout);
      }
      else if (bptype==-5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq8();
	  else
	    Def3jq9();
	} else {
	  if (momp<0.6)
	    Def3jq10();
	  else if (momp<0.8)
	    Def3jq11();
	  else
	    Def3jq12();
	}
    	smatrixdxsxtepveuxgsx_(pmg3[0],&meout);   // Don't emit from b
      }
      else if (bptype<0) {
    	double r1=ran.flat();
	double momp=ran.flat();
	idpartfact*=10.;
    	if (r1<0.5) {   // emit from dx
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
	  smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
    	}
    	else {   
    	  if (bptype==-3) { // emit from sx
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
  	  }
    	  else {   // emit from cx 
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
  	  }
    	}
      }
      else if (bptype==1) {
	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4
    	idpartfact*=4.;
    	if (process==1) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq3();
	    else
	      Def3jq4();
	  } else {
	    if (momp<0.6)
	      Def3jq5();
	    else if (momp<0.8)
	      Def3jq6();
	    else
	      Def3jq7();
	  }
    	  smatrixudxtepvedgdx_(pmg3[0],&meout);
  	}
    	else if (process==2) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq3();
	    else
	      Def3jq4();
	  } else {
	    if (momp<0.6)
	      Def3jq5();
	    else if (momp<0.8)
	      Def3jq6();
	    else
	      Def3jq7();
	  }
    	  smatrixudxtepveugux_(pmg3[0],&meout);
  	}
    	else if (process==3) {
	  int proc=int(ran.flat()*6.)+1;
    	  idpartfact*=18.;    // *6 for choice, *3 flavours
	  if (proc<3.5) {
	    if (proc==1)
	      Def3jq2();
	    else if (proc==2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (proc==4)
	      Def3jq10();
	    else if (proc==5)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
    	  smatrixdxutepvesgsx_(pmg3[0],&meout);
    	}
    	else {
    	  idpartfact*=6.;    // *6 for ggg final state
	  Def3jq2();
    	  smatrixdxutepveggg_(pmg3[0],&meout);
    	}
      }
      else if (bptype==5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq8();
	  else
	    Def3jq9();
	} else {
	  if (momp<0.6)
	    Def3jq10();
	  else if (momp<0.8)
	    Def3jq11();
	  else
	    Def3jq12();
	}
    	smatrixdxstepveuxgs_(pmg3[0],&meout);
      }
      else {   // b=s,c
    	double r2=ran.flat();
	double momp=ran.flat();
    	idpartfact*=10.;
    	if (r2<0.5) {    // emit from ux
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
	  smatrixdxstepveuxgs_(pmg3[0],&meout);
  	}
    	else {
    	  if (bptype==4) {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixusxtepvedgsx_(pmg3[0],&meout);
  	  }
    	  else {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixusxtepvedgsx_(pmg3[0],&meout);
  	  }
	}
      }
    }
    else if (aptype==-3) {    // a=sbar, b!=u
      if (bptype==-4) {  // b=c
	double momp=ran.flat();
	idpartfact*=8.; // *4 for choice, *2 for identical final state particles
	if (momp<0.5) {
	  if (momp<0.25) 
	    Def3jp8();
	  else 
	    Def3jp9();
	} else {
	  if (momp<0.75) 
	    Def3jq3();
	  else
	    Def3jq4();
	}
	smatrixdxuxtepveuxgux_(pmg3[0],&meout);
      }
      else if (bptype==-5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp8();
	  else
	    Def3jp9();
	} else {
	  if (momp<0.6)
	    Def3jp10();
	  else if (momp<0.8)
	    Def3jp11();
	  else
	    Def3jp12();
	}
    	smatrixdxsxtepveuxgsx_(pmg3[0],&meout);   // Don't emit from b
      }
      else if (bptype<0) {
    	double r1=ran.flat();
	double momp=ran.flat();
	idpartfact*=10.;
    	if (r1<0.5) {   // emit from dx
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
	  smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
    	}
    	else {   
    	  if (bptype==-1) { // emit from dx
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
  	  }
    	  else {   // emit from sx 
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
  	  }
    	}
      }
      else if (bptype==4) {
	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4
    	idpartfact*=4.;
    	if (process==1) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp3();
	    else
	      Def3jp4();
	  } else {
	    if (momp<0.6)
	      Def3jp5();
	    else if (momp<0.8)
	      Def3jp6();
	    else
	      Def3jp7();
	  }
    	  smatrixudxtepvedgdx_(pmg3[0],&meout);
  	}
    	else if (process==2) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp3();
	    else
	      Def3jp4();
	  } else {
	    if (momp<0.6)
	      Def3jp5();
	    else if (momp<0.8)
	      Def3jp6();
	    else
	      Def3jp7();
	  }
    	  smatrixudxtepveugux_(pmg3[0],&meout);
  	}
    	else if (process==3) {
	  int proc=int(ran.flat()*6.)+1;
    	  idpartfact*=18.;    // *6 for choice, *3 flavours
	  if (proc<3.5) {
	    if (proc==1)
	      Def3jp2();
	    else if (proc==2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (proc==4)
	      Def3jp10();
	    else if (proc==5)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
    	  smatrixdxutepvesgsx_(pmg3[0],&meout);
    	}
    	else {
    	  idpartfact*=6.;    // *6 for ggg final state
	  Def3jp2();
    	  smatrixdxutepveggg_(pmg3[0],&meout);
    	}
      }
      else if (bptype==5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp8();
	  else
	    Def3jp9();
	} else {
	  if (momp<0.6)
	    Def3jp10();
	  else if (momp<0.8)
	    Def3jp11();
	  else
	    Def3jp12();
	}
    	smatrixdxstepveuxgs_(pmg3[0],&meout);
      }
      else {   // b=d,u
    	double r2=ran.flat();
	double momp=ran.flat();
    	idpartfact*=10.;
    	if (r2<0.5) {    // emit from dx
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jp8();
	    else
	      Def3jp9();
	  } else {
	    if (momp<0.6)
	      Def3jp10();
	    else if (momp<0.8)
	      Def3jp11();
	    else
	      Def3jp12();
	  }
	  smatrixdxstepveuxgs_(pmg3[0],&meout);
  	}
    	else {
    	  if (bptype==2) {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixusxtepvedgsx_(pmg3[0],&meout);
  	  }
    	  else {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixusxtepvedgsx_(pmg3[0],&meout);
  	  }
	}
      }
    }
    else if (aptype==-4) {    // a=cbar, b!=u
      if (bptype==-3) {  // b=sx
	double momp=ran.flat();
	idpartfact*=8.; // *4 for choice, *2 for identical final state particles
	if (momp<0.5) {
	  if (momp<0.25) 
	    Def3jq8();
	  else 
	    Def3jq9();
	} else {
	  if (momp<0.75) 
	    Def3jp3();
	  else
	    Def3jp4();
	}
	smatrixdxuxtepveuxgux_(pmg3[0],&meout);
      }
      else if (bptype==-5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq8();
	  else
	    Def3jq9();
	} else {
	  if (momp<0.6)
	    Def3jq10();
	  else if (momp<0.8)
	    Def3jq11();
	  else
	    Def3jq12();
	}
    	smatrixdxsxtepveuxgsx_(pmg3[0],&meout);   // Don't emit from b
      }
      else if (bptype<0) {
    	double r1=ran.flat();
	double momp=ran.flat();
	idpartfact*=10.;
    	if (r1<0.5) {   // emit from dx
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
	  smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
    	}
    	else {   
    	  if (bptype==-1) { // emit from dx
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
  	  }
    	  else {   // emit from cx 
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
  	  }
    	}
      }
      else if (bptype==3) {
	int process=int(4.*ran.flat())+1; // 1, 2, 3 or 4
    	idpartfact*=4.;
    	if (process==1) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq3();
	    else
	      Def3jq4();
	  } else {
	    if (momp<0.6)
	      Def3jq5();
	    else if (momp<0.8)
	      Def3jq6();
	    else
	      Def3jq7();
	  }
    	  smatrixudxtepvedgdx_(pmg3[0],&meout);
  	}
    	else if (process==2) {
	  double momp=ran.flat();
	  idpartfact*=5;
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq3();
	    else
	      Def3jq4();
	  } else {
	    if (momp<0.6)
	      Def3jq5();
	    else if (momp<0.8)
	      Def3jq6();
	    else
	      Def3jq7();
	  }
    	  smatrixudxtepveugux_(pmg3[0],&meout);
  	}
    	else if (process==3) {
	  int proc=int(ran.flat()*6.)+1;
    	  idpartfact*=18.;    // *6 for choice, *3 flavours
	  if (proc<3.5) {
	    if (proc==1)
	      Def3jq2();
	    else if (proc==2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (proc==4)
	      Def3jq10();
	    else if (proc==5)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
    	  smatrixdxutepvesgsx_(pmg3[0],&meout);
    	}
    	else {
    	  idpartfact*=6.;    // *6 for ggg final state
	  Def3jq2();
    	  smatrixdxutepveggg_(pmg3[0],&meout);
    	}
	//	std::cout<<process<<"  "<<meout<<"  "<<idpartfact<<"\n";
      }
      else if (bptype==5) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq8();
	  else
	    Def3jq9();
	} else {
	  if (momp<0.6)
	    Def3jq10();
	  else if (momp<0.8)
	    Def3jq11();
	  else
	    Def3jq12();
	}
    	smatrixdxstepveuxgs_(pmg3[0],&meout);
      }
      else {   // b=s,c
    	double r2=ran.flat();
	double momp=ran.flat();
    	idpartfact*=10.;
    	if (r2<0.5) {    // emit from cx
	  if (momp<0.4) {
	    if (momp<0.2)
	      Def3jq8();
	    else
	      Def3jq9();
	  } else {
	    if (momp<0.6)
	      Def3jq10();
	    else if (momp<0.8)
	      Def3jq11();
	    else
	      Def3jq12();
	  }
	  smatrixdxstepveuxgs_(pmg3[0],&meout);
  	}
    	else {
    	  if (bptype==2) {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jp3();
	      else
		Def3jp4();
	    } else {
	      if (momp<0.6)
		Def3jp5();
	      else if (momp<0.8)
		Def3jp6();
	      else
		Def3jp7();
	    }
    	    smatrixusxtepvedgsx_(pmg3[0],&meout);
  	  }
    	  else {
	    if (momp<0.4) {
	      if (momp<0.2)
		Def3jq3();
	      else
		Def3jq4();
	    } else {
	      if (momp<0.6)
		Def3jq5();
	      else if (momp<0.8)
		Def3jq6();
	      else
		Def3jq7();
	    }
    	    smatrixusxtepvedgsx_(pmg3[0],&meout);
  	  }
	}
      }
    }
    else if (aptype==-5) {    // a=bbar, b!=b
      if (bptype==-1 || bptype==-3) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp3();
	  else
	    Def3jp4();
	}
	else {
	  if (momp<0.6)
	    Def3jp5();
	  else if (momp<0.8)
	    Def3jp6();
	  else
	    Def3jp7();
	}
	smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
      }
      else if (bptype<0) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq3();
	  else
	    Def3jq4();
	}
	else {
	  if (momp<0.6)
	    Def3jq5();
	  else if (momp<0.8)
	    Def3jq6();
	  else
	    Def3jq7();
	}
	smatrixdxsxtepveuxgsx_(pmg3[0],&meout);
      }
      else if (bptype==2 || bptype==4) {
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jp3();
	  else
	    Def3jp4();
	}
	else {
	  if (momp<0.6)
	    Def3jp5();
	  else if (momp<0.8)
	    Def3jp6();
	  else
	    Def3jp7();
	}
	smatrixusxtepvedgsx_(pmg3[0],&meout);
      }
      else {   // bptype = d,s
	double momp=ran.flat();
	idpartfact*=5.;
	if (momp<0.4) {
	  if (momp<0.2)
	    Def3jq3();
	  else
	    Def3jq4();
	}
	else {
	  if (momp<0.6)
	    Def3jq5();
	  else if (momp<0.8)
	    Def3jq6();
	  else
	    Def3jq7();
	}
	smatrixusxtepvedgsx_(pmg3[0],&meout);
      }
    }
  //   else  {
  //     std::cerr<<"Watch out - unanticipated parton configuration!"<<std::endl;
  }

  //  std::cout<<"pointer gives: "<<meout<<std::endl;

  // if (isnan(meout)) {
  //   std::cout<<"Getting nan: "<<std::endl;
  // }

  //  std::cout<<"b: "<<meout<<"  "<<idpartfact<<"\n";
  weight*=meout*idpartfact;
  //  std::cout<<"s1 "<<idpartfact<<"  "<<meout<<"  "<<weight<<"  "<<pow(4.*M_PI*AlphaS(flags.scale),2)<<"  "<<aptype<<"  "<<bptype<<"  "<<status<<std::endl;
#endif

  // if (flags.scalesetting==0) //fixed scale
  //   weight*=pow(4.*M_PI*AlphaS(flags.scale),3);
  // else if (flags.scalesetting==1)
  //   weight*=pow(4.*M_PI*AlphaS(Ht),3);
  weight*=pow(4.*M_PI*Halphas(renscale),3);
  if (flags.logcorrect==1)
    weight*=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI));

  //  std::cout<<"s2: "<<weight<<"\n";
  return weight;
}

double CWMultijet::MG4(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb, CLHEP::HepLorentzVector aleptonp, CLHEP::HepLorentzVector leptonp)
{
  double wt(0.);
  FREAL meout(0.),p12[8][4],p21[8][4];
  // p12 = config where partons are ordered as in name
  // p21 = config where partons are reversed relative to name

  p12[0][0]=pb.e();    // forward parton
  p12[0][1]=pb.x();
  p12[0][2]=pb.y();
  p12[0][3]=pb.z();
  
  p12[1][0]=pa.e();   //  backward parton
  p12[1][1]=pa.x();
  p12[1][2]=pa.y();
  p12[1][3]=pa.z();
  
  p12[2][0]=aleptonp.e();   // anti-lepton
  p12[2][1]=aleptonp.x();
  p12[2][2]=aleptonp.y();
  p12[2][3]=aleptonp.z();
  
  p12[3][0]=leptonp.e();   // lepton
  p12[3][1]=leptonp.x();
  p12[3][2]=leptonp.y();
  p12[3][3]=leptonp.z();
  
  p12[4][0]=(partons.back()).e();   // most forward jet
  p12[4][1]=(partons.back()).x();
  p12[4][2]=(partons.back()).y();
  p12[4][3]=(partons.back()).z();
  
  p12[5][0]=(partons.at(2)).e();  // middle forward jet
  p12[5][1]=(partons.at(2)).x();
  p12[5][2]=(partons.at(2)).y();
  p12[5][3]=(partons.at(2)).z();

  p12[6][0]=(partons.at(1)).e();  // middle backward jet
  p12[6][1]=(partons.at(1)).x();
  p12[6][2]=(partons.at(1)).y();
  p12[6][3]=(partons.at(1)).z();

  p12[7][0]=(partons.front()).e();  // most backward jet
  p12[7][1]=(partons.front()).x();
  p12[7][2]=(partons.front()).y();
  p12[7][3]=(partons.front()).z();

  for (int ii=0; ii<4; ii++) {
    p21[0][ii]=p12[1][ii];
    p21[1][ii]=p12[0][ii];
    p21[2][ii]=p12[2][ii];
    p21[3][ii]=p12[3][ii];
    p21[4][ii]=p12[7][ii];
    p21[5][ii]=p12[6][ii];
    p21[6][ii]=p12[5][ii];
    p21[7][ii]=p12[4][ii];
  }

  if (aptype==21) {// a=g, we know they are not both gluons so b!= g
    if (bptype>0)
      smatrixugtepvedggg_(p12[0],&meout);
    //      std::cout<<"Yes!"<<std::endl;
    else
      smatrixdxgtepveuxggg_(p12[0],&meout);
    meout*=6.;
  }
  else if (bptype==21) {  // b=g, a!=g
    if (aptype>0) 
      smatrixugtepvedggg_(p21[0],&meout);
    else 
      smatrixdxgtepveuxggg_(p21[0],&meout);
    meout*=6.;
  }
  else {  //  two (anti-)quarks to deal with!
    if (aptype==bptype) {
      if (aptype > 0) {
  	if (wc==true)   // emit off b
  	  smatrixuutepveuggd_(p21[0],&meout);
  	else            // emit off a so swap momenta
  	  smatrixuutepveuggd_(p12[0],&meout);
	meout*=2.;
      }
      else {
  	if (wc==true)   // emit off b
  	  smatrixdxdxtepveuxggdx_(p12[0],&meout);
  	else            // emit off a so swap momenta
  	  smatrixdxdxtepveuxggdx_(p21[0],&meout);
	meout*=2.;
      }
    }
    else if (aptype==-bptype) {
      if (aptype > 0) {  // backward particle is q, forward qbar
  	if (wc==true) 
  	  smatrixdxdtepveuxggd_(p12[0],&meout);
  	else 
  	  smatrixuxutepveuxggd_(p12[0],&meout);
	meout*=2.;
      }
      else {            // forward (b) particle is q, backward (a) qbar
  	if (wc==true) 
  	  smatrixuxutepveuxggd_(p21[0],&meout);
  	else 
  	  smatrixdxdtepveuxggd_(p21[0],&meout);
	meout*=2.;
      }
    }
    else if (aptype==1) {    // a=d, b!=d,dbar
      if (bptype==2) {  // b=u
    	if (wc==true) 
    	  smatrixudtepvedggd_(p12[0],&meout);
    	else 
    	  smatrixudtepvedggd_(p21[0],&meout);
    	meout*=4.;
      }
      else if (bptype>2) {
    	if (wc==true)
    	  smatrixustepvedggs_(p12[0],&meout);
    	else
    	  smatrixustepvedggs_(p21[0],&meout);
	meout*=2.;
      }
      else if (bptype==-2) {
    	if (wc==true) 
    	  smatrixudxtepveuggux_(p21[0],&meout);
    	else 
    	  smatrixudxtepvedggdx_(p21[0],&meout);
	meout*=2.;
      }
      else {
    	if (wc==true)
    	  smatrixdxstepveuxggs_(p12[0],&meout);
    	else
    	  smatrixusxtepvedggsx_(p21[0],&meout);
	meout*=2.;
      }
    }
    else if (aptype==2) { // a=u
      if (bptype==1) {  // b=d
    	if (wc==true) 
    	  smatrixudtepvedggd_(p12[0],&meout);
    	else                 //  Not sure if this is necessary
    	  smatrixudtepvedggd_(p21[0],&meout);
    	meout*=4.;
      }
      else if (bptype>2) {    // b = s,c,b
    	if (wc==true)
    	  smatrixustepvedggs_(p12[0],&meout);
    	else
    	  smatrixustepvedggs_(p21[0],&meout);
	meout*=2.;
      }
      else if (bptype==-1) {  // b = dbar
    	if (wc==true) 
    	  smatrixudxtepveuggux_(p21[0],&meout);
    	else 
    	  smatrixudxtepvedggdx_(p21[0],&meout);
	meout*=2.;
      }
      else {            // b = sbar, cbar,bbar
    	if (wc==true)
    	  smatrixdxstepveuxggs_(p12[0],&meout);
    	else 
    	  smatrixusxtepvedggsx_(p21[0],&meout);
	meout*=2.;
      }
    }
    else if (aptype==3) { // a=s
      if (bptype==4) {  // b=c
    	if (wc==true) 
    	  smatrixudtepvedggd_(p12[0],&meout);
    	else
    	  smatrixudtepvedggd_(p21[0],&meout);
    	meout*=4.;
      }
      else if (bptype>0) {    // b = u,d,b
    	if (wc==true)
    	  smatrixustepvedggs_(p12[0],&meout);
    	else
    	  smatrixustepvedggs_(p21[0],&meout);
	meout*=2.;
      }
      else if (bptype==-4) {  // b = cbar
    	if (wc==true) 
    	  smatrixudxtepveuggux_(p21[0],&meout);
    	else 
    	  smatrixudxtepvedggdx_(p21[0],&meout);
	meout*=2.;
      }
      else {            // b = ubar, dbar,bbar
    	if (wc==true)
    	  smatrixdxstepveuxggs_(p12[0],&meout);
    	else
    	  smatrixusxtepvedggsx_(p21[0],&meout);
	meout*=2.;
      }
    }
    else if (aptype==4) { // a=c
      if (bptype==3) {  // b=s
    	if (wc==true) 
    	  smatrixudtepvedggd_(p12[0],&meout);
    	else                 //  Not sure if this is necessary
    	  smatrixudtepvedggd_(p21[0],&meout);
    	meout*=4.;
      }
      else if (bptype>0) {    // b = u,d,b
    	if (wc==true)
    	  smatrixustepvedggs_(p12[0],&meout);
    	else
    	  smatrixustepvedggs_(p21[0],&meout);
	meout*=2.;
      }
      else if (bptype==-3) {  // b = cbar
    	if (wc==true) 
    	  smatrixudxtepveuggux_(p21[0],&meout);
    	else 
    	  smatrixudxtepvedggdx_(p21[0],&meout);
	meout*=2.;
      }
      else {            // b = ubar, dbar,bbar
    	if (wc==true)
    	  smatrixdxstepveuxggs_(p12[0],&meout);
    	else
    	  smatrixusxtepvedggsx_(p21[0],&meout);
	meout*=2.;
      }
    }
    else if (aptype>0) { // a=b
      if (bptype>0) {    // b = u,d,c,s
    	if (wc==true)
    	  smatrixustepvedggs_(p12[0],&meout);
    	else
    	  smatrixustepvedggs_(p21[0],&meout);
	meout*=2.;
      }
      else {            // b = ubar, dbar, sbar, cbar
    	if (wc==true)
    	  smatrixdxstepveuxggs_(p12[0],&meout);
    	else
    	  smatrixusxtepvedggsx_(p21[0],&meout);
	meout*=2.;
      }
    }
    ///  Now go for a = anti-quark
    else if (aptype==-1) {    // a=dbar, b!=d
      if (bptype==-2) {  // b=ubar
    	if (wc==true)
    	  smatrixdxuxtepveuxggux_(p12[0],&meout);
    	else
    	  smatrixdxuxtepveuxggux_(p21[0],&meout);
    	meout*=4.;
      }
      else if (bptype<-2) {
    	if (wc==true)
    	  smatrixdxsxtepveuxggsx_(p12[0],&meout);
    	else
    	  smatrixdxsxtepveuxggsx_(p21[0],&meout);
	meout*=2.;
      }
      else if (bptype==2) {
    	if (wc==true) 
    	  smatrixudxtepvedggdx_(p12[0],&meout);
    	else 
    	  smatrixudxtepveuggux_(p12[0],&meout);
	meout*=2.;
      }
      else {  // a=dbar, b= s,c,t
    	if (wc==true)
    	  smatrixusxtepvedggsx_(p12[0],&meout);
    	else
    	  smatrixdxstepveuxggs_(p21[0],&meout);
	meout*=2.;
      }
    }
    else if (aptype==-2) {    // a=ubar, b!=u
      if (bptype==-1) {  // b=dbar
    	if (wc==true)
    	  smatrixdxuxtepveuxggux_(p12[0],&meout);
    	else
    	  smatrixdxuxtepveuxggux_(p21[0],&meout);
    	meout*=4.;
      }
      else if (bptype<-2) {  // b = sbar,cbar,bbar
    	if (wc==true)
    	  smatrixdxsxtepveuxggsx_(p12[0],&meout);
    	else
    	  smatrixdxsxtepveuxggsx_(p21[0],&meout);
	meout*=2.;
      }
      else if (bptype==1) {  // a=ubar, b=d
    	if (wc==true) 
    	  smatrixudxtepvedggdx_(p12[0],&meout);
    	else 
    	  smatrixudxtepveuggux_(p12[0],&meout);
	meout*=2.;
      }
      else {  // a=ubar, b= s,c,t
    	if (wc==true) 
    	  smatrixusxtepvedggsx_(p12[0],&meout);
    	else 
    	  smatrixdxstepveuxggs_(p21[0],&meout);
	meout*=2.;
      }
    }
    else if (aptype==-3) {    // a=sbar, b!=u
      if (bptype==-4) {  // b=cbar
    	if (wc==true)
    	  smatrixdxuxtepveuxggux_(p12[0],&meout);
    	else
    	  smatrixdxuxtepveuxggux_(p21[0],&meout);
    	meout*=4.;
      }
      else if (bptype<0) {  // b = ubar,dbar,bbar
    	if (wc==true)
    	  smatrixdxsxtepveuxggsx_(p12[0],&meout);
    	else
    	  smatrixdxsxtepveuxggsx_(p21[0],&meout);
	meout*=2.;
      }
      else if (bptype==4) {  // a=sbar, b=c
    	if (wc==true) 
    	  smatrixudxtepvedggdx_(p12[0],&meout);
    	else 
    	  smatrixudxtepveuggux_(p12[0],&meout);
	meout*=2.;
      }
      else {  // a=sbar, b= u,d,t
    	if (wc==true) 
    	  smatrixusxtepvedggsx_(p12[0],&meout);
    	else 
    	  smatrixdxstepveuxggs_(p21[0],&meout);
	meout*=2.;
      }
    }
    else if (aptype==-4) {    // a=cbar, b!=u
      if (bptype==-3) {  // b=sbar
    	if (wc==true)
    	  smatrixdxuxtepveuxggux_(p12[0],&meout);
    	else
    	  smatrixdxuxtepveuxggux_(p21[0],&meout);
    	meout*=4.;
      }
      else if (bptype<0) {  // b = ubar,dbar,bbar
    	if (wc==true)
    	  smatrixdxsxtepveuxggsx_(p12[0],&meout);
    	else
    	  smatrixdxsxtepveuxggsx_(p21[0],&meout);
	meout*=2.;
      }
      else if (bptype==3) {  // a=cbar, b=s
    	if (wc==true) 
    	  smatrixudxtepvedggdx_(p12[0],&meout);
    	else 
    	  smatrixudxtepveuggux_(p12[0],&meout);
	meout*=2.;
      }
      else {  // a=cbar, b= u,d,t
    	if (wc==true) 
    	  smatrixusxtepvedggsx_(p12[0],&meout);
    	else 
    	  smatrixdxstepveuxggs_(p21[0],&meout);
	meout*=2.;
      }
    }
    else if (aptype==-5) {    // a=bbar, b!=b
      if (bptype<0) {  // b = ubar,dbar,sbar, cbar
    	if (wc==true)
    	  smatrixdxsxtepveuxggsx_(p12[0],&meout);
    	else
    	  smatrixdxsxtepveuxggsx_(p21[0],&meout);
	meout*=2.;
      }
      else {  // a=bbar, b= u,d,s,c
    	if (wc==true) 
    	  smatrixusxtepvedggsx_(p12[0],&meout);
    	else 
    	  smatrixdxstepveuxggs_(p21[0],&meout);
	meout*=2.;
      }
    }
    else 
      std::cerr<<"Watch out - unanticipated parton configuration!"<<std::endl;
  }

  wt=meout;
  
  // double gs2=0.;
  // if (flags.scalesetting==0) //fixed scale
  //   gs2=4.*M_PI*AlphaS(flags.scale);
  // else if (flags.scalesetting==1)
  //   gs2=4.*M_PI*AlphaS(Ht);

  double gs2=4.*M_PI*Halphas(renscale);
  wt*=pow(gs2,4);
  if (flags.logcorrect==1)
    wt*=(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.at(1).perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.at(2).perp()/renscale)/(2.*M_PI));

  return wt;
}

double CWMultijet::Generate4jNonFKLflavourconf_and_eval(double pmg[8][4])
{
  //    std::cout<<"Yep, 4j non FKL: ";
  double weight=1.;

  if (aptype==21) {// a=g 
    if (bptype==21) {
      
      static std::vector<sPMchoice> ggProcesses;
      if (ggProcesses.size()==0) 
	{
	  { // first gg process
	    struct sPMchoice PMtemp;
	    // tpmomshuffle tpmomtemp;
	    PMtemp.MGfunc=smatrixggtepveuxdssx_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=6;
	    
	    for (int ii=0;ii<24;ii++) // lean and mean: Take them all!
	      PMtemp.MomShuffleVector.push_back(ii);
	    
	    ggProcesses.push_back(PMtemp);
	    PMtemp.ptype=false; // process type p
	    ggProcesses.push_back(PMtemp);
	  }
	  { // second gg process
	    struct sPMchoice PMtemp;
	    // tpmomshuffle tpmomtemp;
	    PMtemp.MGfunc=smatrixggtepveuxuxud_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=4;
	    
	    for (int ii=0;ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    
	    ggProcesses.push_back(PMtemp);
	    PMtemp.ptype=false; // process type p
	    ggProcesses.push_back(PMtemp);
	  }
	  { // third gg process
	    struct sPMchoice PMtemp;
	    // tpmomshuffle tpmomtemp;
	    PMtemp.MGfunc=smatrixggtepvedduxdx_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=4;
	    
	    for (int ii=0;ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    
	    ggProcesses.push_back(PMtemp);
	    PMtemp.ptype=false; // process type p
	    ggProcesses.push_back(PMtemp);
	  }
	  { // fourth gg process
	    struct sPMchoice PMtemp;
	    // tpmomshuffle tpmomtemp;
	    PMtemp.MGfunc=smatrixggtepvegguxd_;
	    PMtemp.ptype=true; // process type q
	    PMtemp.idpartmult=4;
	    
	    for (int ii=0;ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    
	    ggProcesses.push_back(PMtemp);
	    PMtemp.ptype=false; // process type p
	    ggProcesses.push_back(PMtemp);
	  }
      }
      weight*=PickAndEvalProcAndMom4j(ggProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==5) {          //MARKER! a=g,  b!=g    
      static std::vector<sPMchoice> gbProcesses;
      if (gbProcesses.size()==0) {
	struct sPMchoice PMtemp;
	PMtemp.MGfunc=smatrixugtepveucxsg_;
	PMtemp.ptype=false; // p
	PMtemp.idpartmult=2;
	for (int ii=0;ii<24;ii++)
	  PMtemp.MomShuffleVector.push_back(ii);
	gbProcesses.push_back(PMtemp);
	PMtemp.ptype=true; // q
	gbProcesses.push_back(PMtemp);
      }
      weight*=PickAndEvalProcAndMom4j(gbProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==-5) {
      static std::vector<sPMchoice> gbxProcesses;
      if (gbxProcesses.size()==0) {
	struct sPMchoice PMtemp;
	PMtemp.MGfunc=smatrixdxgtepvedxcxsg_;
	PMtemp.ptype=false; // p
	PMtemp.idpartmult=2;
	for (int ii=0;ii<24;ii++)
	  PMtemp.MomShuffleVector.push_back(ii);
	gbxProcesses.push_back(PMtemp);
	PMtemp.ptype=true; // q
	gbxProcesses.push_back(PMtemp);
      }
      weight*=PickAndEvalProcAndMom4j(gbxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==2 || bptype==4) { 
      static std::vector<sPMchoice> guProcesses;
      if (guProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; //p
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(18);
	  guProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedssxg_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=false; // p
	  for (int ii=0; ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  guProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixugtepveuuxdg_;
	  PMtemp.idpartmult=1;
	  guProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixugtepveucxsg_;
	  guProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=true; // q
	  guProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedddxg_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0; ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  guProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdgtepvedduxg_;
	  PMtemp.ptype=true; // q
	  guProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(guProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype>0) {    // b = d,s
      static std::vector<sPMchoice> gdProcesses;
      if (gdProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(18);
	  gdProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedssxg_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=true; // q
	  for (int ii=0; ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  gdProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixugtepveuuxdg_;
	  PMtemp.idpartmult=1;
	  gdProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixugtepveucxsg_;
	  gdProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=false; // p
	  gdProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedddxg_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=0; ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  gdProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdgtepvedduxg_;
	  PMtemp.ptype=false; // p
	  gdProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(gdProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==-1 || bptype==-3) {
      static std::vector<sPMchoice> gdxProcesses;
      if (gdxProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; //p
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(18);
	  gdxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxssxg_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=false; // p
	  for (int ii=0; ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  gdxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxgtepvedxuxdg_;
	  PMtemp.idpartmult=1;
	  gdxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxgtepvedxcxsg_;
	  gdxProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=true; // q
	  gdxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxuxug_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0; ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  gdxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuxgtepveuxuxdg_;
	  PMtemp.ptype=true; // q
	  gdxProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(gdxProcesses,pmg,&lepton,&alepton);
    }
    else {
      static std::vector<sPMchoice> guxProcesses;
      if (guxProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; //q
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(18);
	  guxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxssxg_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=true; // q
	  for (int ii=0; ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  guxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxgtepvedxuxdg_;
	  PMtemp.idpartmult=1;
	  guxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxgtepvedxcxsg_;
	  guxProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=false; // p
	  guxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxuxug_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=0; ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  guxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuxgtepveuxuxdg_;
	  PMtemp.ptype=false; // p
	  guxProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(guxProcesses,pmg,&lepton,&alepton);
    }
  }
  else if (bptype==21) {  // b=g, a!=g
    if (aptype==5) {
      static std::vector<sPMchoice> bgProcesses;
      if (bgProcesses.size()==0) {
	struct sPMchoice PMtemp;
	PMtemp.MGfunc=smatrixugtepveucxsg_;
	PMtemp.ptype=false; // p
	PMtemp.idpartmult=2;
	for (int ii=24;ii<48;ii++)
	  PMtemp.MomShuffleVector.push_back(ii);
	bgProcesses.push_back(PMtemp);
	PMtemp.ptype=true; // q
	bgProcesses.push_back(PMtemp);
      }
      weight*=PickAndEvalProcAndMom4j(bgProcesses,pmg,&lepton,&alepton);
    }
    else if (aptype==-5) {
      static std::vector<sPMchoice> bxgProcesses;
      if (bxgProcesses.size()==0) {
	struct sPMchoice PMtemp;
	PMtemp.MGfunc=smatrixdxgtepvedxcxsg_;
	PMtemp.ptype=false; // p
	PMtemp.idpartmult=2;
	for (int ii=24;ii<48;ii++)
	  PMtemp.MomShuffleVector.push_back(ii);
	bxgProcesses.push_back(PMtemp);
	PMtemp.ptype=true; // q
	bxgProcesses.push_back(PMtemp);
      }
      weight*=PickAndEvalProcAndMom4j(bxgProcesses,pmg,&lepton,&alepton);
    }
    else if (aptype==2 || aptype==4) { 
      static std::vector<sPMchoice> ugProcesses;
      if (ugProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; //p
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(42);
	  ugProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedssxg_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=false; // p
	  for (int ii=24; ii<48;ii++)
	    //	  for (int ii=24; ii<25;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ugProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixugtepveuuxdg_;
	  PMtemp.idpartmult=1;
	  ugProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixugtepveucxsg_;
	  ugProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=true; // q
	  ugProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedddxg_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=24; ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ugProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdgtepvedduxg_;
	  PMtemp.ptype=true; // q
	  ugProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(ugProcesses,pmg,&lepton,&alepton);
    }
    else if (aptype>0) {
      static std::vector<sPMchoice> dgProcesses;
      if (dgProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(42);
	  dgProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedssxg_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=true; // q
	  for (int ii=24; ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dgProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixugtepveuuxdg_;
	  PMtemp.idpartmult=1;
	  dgProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixugtepveucxsg_;
	  dgProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=false; // p
	  dgProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixugtepvedddxg_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=24; ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dgProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdgtepvedduxg_;
	  PMtemp.ptype=false; // p
	  dgProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dgProcesses,pmg,&lepton,&alepton);
    }
    else if (aptype==-1 || aptype==-3) {
      static std::vector<sPMchoice> dxgProcesses;
      if (dxgProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; //p
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(42);
	  dxgProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxssxg_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=false; // p
	  for (int ii=24; ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxgProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxgtepvedxuxdg_;
	  PMtemp.idpartmult=1;
	  dxgProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxgtepvedxcxsg_;
	  dxgProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=true; // q
	  dxgProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxuxug_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=24; ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxgProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuxgtepveuxuxdg_;
	  PMtemp.ptype=true; // q
	  dxgProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxgProcesses,pmg,&lepton,&alepton);
    }
    else {
      static std::vector<sPMchoice> uxgProcesses;
      if (uxgProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxggg_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; //q
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(42);
	  uxgProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxssxg_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=true; // q
	  for (int ii=24; ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  uxgProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxgtepvedxuxdg_;
	  PMtemp.idpartmult=1;
	  uxgProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxgtepvedxcxsg_;
	  uxgProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=false; // p
	  uxgProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxgtepveuxuxug_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=24; ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  uxgProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuxgtepveuxuxdg_;
	  PMtemp.ptype=false; // p
	  uxgProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(uxgProcesses,pmg,&lepton,&alepton);
    }
  }

  ////////////  These are identical lines to help svn ////////////
  ////////////  These are identical lines to help svn ////////////
  ////////////  These are identical lines to help svn ////////////
  ////////////  These are identical lines to help svn ////////////
  ////////////  These are identical lines to help svn ////////////
  ////////////  These are identical lines to help svn ////////////
  ////////////  These are identical lines to help svn ////////////
  ////////////  These are identical lines to help svn ////////////
  ////////////  These are identical lines to help svn ////////////

  else {  //  MARKER - two (anti-)quarks to deal with!
    if (aptype==bptype) {
      if (aptype==5) {
	static std::vector<sPMchoice> bbProcesses;
	if (bbProcesses.size()==0) {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixuutepveuucxs_;
	    PMtemp.idpartmult=4;
	    PMtemp.ptype=false; //p
	    for (int ii=0;ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    bbProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; //q
	    bbProcesses.push_back(PMtemp);
	}
	weight*=PickAndEvalProcAndMom4j(bbProcesses,pmg,&lepton,&alepton);
      }
      else if (aptype==-5) {
	static std::vector<sPMchoice> bxbxProcesses;
	if (bxbxProcesses.size()==0) {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixuutepveuucxs_;
	    PMtemp.idpartmult=4;
	    PMtemp.ptype=false; //p
	    for (int ii=0;ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    bxbxProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; //q
	    bxbxProcesses.push_back(PMtemp);
	}
	weight*=PickAndEvalProcAndMom4j(bxbxProcesses,pmg,&lepton,&alepton);
      }
      else if (aptype==2 || aptype==4) {   // uu or cc
	static std::vector<sPMchoice> uuProcesses;
	if (uuProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuutepveuggd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  uuProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuutepveddudx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  uuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuutepveuuuxd_;
	  uuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuutepveuucxs_;
	  uuProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // q
	  uuProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuutepvedussx_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  uuProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixddtepvedddux_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(9);
	  uuProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(uuProcesses,pmg,&lepton,&alepton);
      }
      else if (aptype > 0) {   // dd or ss
	static std::vector<sPMchoice> ddProcesses;
	if (ddProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuutepveuggd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  ddProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuutepveddudx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ddProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuutepveuuuxd_;
	  ddProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuutepveuucxs_;
	  ddProcesses.push_back(PMtemp);

	  PMtemp.ptype=false; // p
	  ddProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuutepvedussx_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ddProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixddtepvedddux_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(9);
	  ddProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(ddProcesses,pmg,&lepton,&alepton);
      }
      else if (aptype==-1 || aptype==-3) {   // dxdx or sxsx
	static std::vector<sPMchoice> dxdxProcesses;
	if (dxdxProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixdxdxtepveuxggdx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dxdxProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixdxdxtepvedxdxuxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxdxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxdxtepveuxuxudx_;
	  dxdxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxdxtepvedxdxcxs_;
	  dxdxProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // q
	  dxdxProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixdxdxtepveuxdxssx_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxdxProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuxuxtepveuxuxuxd_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(9);
	  dxdxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(dxdxProcesses,pmg,&lepton,&alepton);
      }
      else {             // uxux or cxcx
	static std::vector<sPMchoice> uxuxProcesses;
	if (uxuxProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixdxdxtepveuxggdx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  uxuxProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixdxdxtepvedxdxuxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  uxuxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxdxtepveuxuxudx_;
	  uxuxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxdxtepvedxdxcxs_;
	  uxuxProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // q
	  uxuxProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixdxdxtepveuxdxssx_;
	  PMtemp.idpartmult=3;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  uxuxProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuxuxtepveuxuxuxd_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(9);
	  uxuxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(uxuxProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==-bptype) {    /// MARKER - q qbar same flavour ///
      if (aptype==5) {   
	static std::vector<sPMchoice> bbxProcesses;
	if (bbxProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuuxtepveggscx_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // p
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  bbxProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // q
	  bbxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuuxtepvesssxcx_;
	  PMtemp.ptype=false; // p
	  bbxProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=true; // q
	  bbxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuuxtepvecxcxsc_;
	  PMtemp.ptype=false; // p
	  bbxProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // q
	  bbxProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuuxtepveuuxscx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  bbxProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // q
	  bbxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuuxtepvebbxscx_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // p
	  bbxProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // q
	  bbxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(bbxProcesses,pmg,&lepton,&alepton);
      }
      else if (aptype==-5) {                    
	static std::vector<sPMchoice> bxbProcesses;
	if (bxbProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuuxtepveggscx_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  bxbProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // q
	  bxbProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuuxtepvesssxcx_;
	  PMtemp.ptype=false; // p
	  bxbProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=true; // q
	  bxbProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuuxtepvecxcxsc_;
	  PMtemp.ptype=false; // p
	  bxbProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // q
	  bxbProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuuxtepveuuxscx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  bxbProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // q
	  bxbProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuuxtepvebbxscx_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // p
	  bxbProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // q
	  bxbProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(bxbProcesses,pmg,&lepton,&alepton);
      }
      else if (aptype==2 || aptype==4) {
	static std::vector<sPMchoice> uuxProcesses;
	if (uuxProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuxutepveuxggd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  uuxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxdtepveuxggd_;
	  PMtemp.ptype=true; // q
	  uuxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    
	    PMtemp.MGfunc=smatrixuuxtepveggscx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24; ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvedduxdx_;
	    PMtemp.ptype=false; // p
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepveuxuxdu_;
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepvedduxdx_;
	    PMtemp.ptype=true; // q
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepveuxuxdu_;
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvesssxcx_;
	    PMtemp.ptype=false; // p
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvecxcxsc_;
	    PMtemp.ptype=false; // p
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    uuxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    
	    PMtemp.MGfunc=smatrixuuxtepveduxssx_;
	    PMtemp.idpartmult=3;
	    PMtemp.ptype=false; // p
	    for (int ii=24; ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepveduxssx_;
	    PMtemp.ptype=true; // q
	    uuxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixuuxtepveuuxscx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=false; // p
	    uuxProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepveddxscx_;
	    PMtemp.ptype=false; // p
	    uuxProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    uuxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvebbxscx_;
	    PMtemp.ptype=false; // p
	    uuxProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    uuxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(uuxProcesses,pmg,&lepton,&alepton);
      }
      else if (aptype>0) {
	static std::vector<sPMchoice> ddxProcesses;
	if (ddxProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuxutepveuxggd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  ddxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxdtepveuxggd_;
	  PMtemp.ptype=false; // p
	  ddxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    
	    PMtemp.MGfunc=smatrixuuxtepveggscx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24; ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvedduxdx_;
	    PMtemp.ptype=true; // q
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepveuxuxdu_;
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepvedduxdx_;
	    PMtemp.ptype=false; // p
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepveuxuxdu_;
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvesssxcx_;
	    PMtemp.ptype=false; // p
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvecxcxsc_;
	    PMtemp.ptype=false; // p
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    ddxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    
	    PMtemp.MGfunc=smatrixuuxtepveduxssx_;
	    PMtemp.idpartmult=3;
	    PMtemp.ptype=true; // q
	    for (int ii=24; ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepveduxssx_;
	    PMtemp.ptype=false; // p
	    ddxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixuuxtepveuuxscx_;
	    PMtemp.idpartmult=1;
	    ddxProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepveddxscx_;
	    PMtemp.ptype=false; // p
	    ddxProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    ddxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvebbxscx_;
	    PMtemp.ptype=false; // p
	    ddxProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    ddxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(ddxProcesses,pmg,&lepton,&alepton);
      }
      else if (aptype==-1 || aptype==-3) {
	static std::vector<sPMchoice> dxdProcesses;
	if (dxdProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuxutepveuxggd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxdProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxdtepveuxggd_;
	  PMtemp.ptype=false; // p
	  dxdProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    
	    PMtemp.MGfunc=smatrixuuxtepveggscx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=0; ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvedduxdx_;
	    PMtemp.ptype=true; // q
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepveuxuxdu_;
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepvedduxdx_;
	    PMtemp.ptype=false; // p
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepveuxuxdu_;
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvesssxcx_;
	    PMtemp.ptype=false; // p
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvecxcxsc_;
	    PMtemp.ptype=false; // p
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    dxdProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    
	    PMtemp.MGfunc=smatrixuuxtepveduxssx_;
	    PMtemp.idpartmult=3;
	    PMtemp.ptype=true; // q
	    for (int ii=0; ii<24;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepveduxssx_;
	    PMtemp.ptype=false; // p
	    dxdProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixuuxtepveuuxscx_;
	    PMtemp.idpartmult=1;
	    dxdProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepveddxscx_;
	    PMtemp.ptype=false; // p
	    dxdProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    dxdProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvebbxscx_;
	    PMtemp.ptype=false; // p
	    dxdProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    dxdProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(dxdProcesses,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> uxuProcesses;
	if (uxuProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixuxutepveuxggd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  uxuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxdtepveuxggd_;
	  PMtemp.ptype=true; // q
	  uxuProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    
	    PMtemp.MGfunc=smatrixuuxtepveggscx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=0; ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvedduxdx_;
	    PMtemp.ptype=false; // p
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepveuxuxdu_;
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepvedduxdx_;
	    PMtemp.ptype=true; // q
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepveuxuxdu_;
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvesssxcx_;
	    PMtemp.ptype=false; // p
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvecxcxsc_;
	    PMtemp.ptype=false; // p
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.ptype=true; // q
	    uxuProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    
	    PMtemp.MGfunc=smatrixuuxtepveduxssx_;
	    PMtemp.idpartmult=3;
	    PMtemp.ptype=false; // p
	    for (int ii=0; ii<24;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixddxtepveduxssx_;
	    PMtemp.ptype=true; // q
	    uxuProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixuuxtepveuuxscx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=false; // p
	    uxuProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepveddxscx_;
	    PMtemp.ptype=false; // p
	    uxuProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    uxuProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixuuxtepvebbxscx_;
	    PMtemp.ptype=false; // p
	    uxuProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    uxuProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(uxuProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==1) {    // a=d, b!=d,dbar   ///  MARKER a = down //
      if (bptype==2) {  // b=u
	static std::vector<sPMchoice> duProcesses;
	if (duProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepvedggd_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(15);
	  duProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepvedggd_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(39);
	  duProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixudtepvedduux_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=0; ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=9;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    duProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudtepveddssx_;
	    PMtemp.idpartmult=6;
	    duProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixudtepvedduux_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24; ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    duProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudtepveddssx_;
	    PMtemp.idpartmult=6;
	    duProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepveudscx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=false; // p
	    for (int ii=0;ii<24;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    duProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepveudscx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    duProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepveddddx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(4);
	  PMtemp.MomShuffleVector.push_back(9);
	  duProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepveddddx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(24);
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(33);
	  duProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(duProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==5) {
	static std::vector<sPMchoice> dbProcesses;
	if (dbProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dbProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixustepvedddxs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dbProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvesssxd_;
	    dbProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvessucx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=0;ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dbProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvedsbbx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dbProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvedsuux_;
	    PMtemp.idpartmult=1;
	    dbProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixubtepveubscx_;
	    dbProcesses.push_back(PMtemp);

	    PMtemp.ptype=false; // p
	    dbProcesses.push_back(PMtemp);
	  }
	}	
	weight*=PickAndEvalProcAndMom4j(dbProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==3) {
	static std::vector<sPMchoice> dsProcesses;
	if (dsProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dsProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dsProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // q
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  dsProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // q
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  dsProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  PMtemp.ptype=false; // p
	  dsProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  PMtemp.ptype=false; // p
	  dsProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(dsProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype>0) {   
	static std::vector<sPMchoice> dcProcesses;
	if (dcProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dcProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dcProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  dcProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // q
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  dcProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  dcProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  dcProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(dcProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-2) {
	static std::vector<sPMchoice> duxProcesses;
	if (duxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepvedggdx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    duxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveuggux_;
	    duxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepvedddxdx_;
	    PMtemp.idpartmult=4;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(24);
	    PMtemp.MomShuffleVector.push_back(26);
	    PMtemp.MomShuffleVector.push_back(28);
	    PMtemp.MomShuffleVector.push_back(32);
	    PMtemp.MomShuffleVector.push_back(34);
	    PMtemp.MomShuffleVector.push_back(36);
	    duxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveuuuxux_;
	    duxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixduxtepvedduxux_;
	    PMtemp.ptype=false; // p
	    duxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepveddxuux_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    duxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveudxscx_;
	    duxProcesses.push_back(PMtemp);

	    PMtemp.ptype=false; // p
	    duxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveddxssx_;
	    PMtemp.idpartmult=3;
	    PMtemp.ptype=true; // q
	    duxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixudxtepveuuxssx_;
	    duxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(duxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-5) {
	static std::vector<sPMchoice> dbxProcesses;
	if (dbxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    dbxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dbxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    dbxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    PMtemp.ptype=false; // p
	    dbxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dbxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    dbxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixubxtepveubxscx_;
	    dbxProcesses.push_back(PMtemp);

	    PMtemp.ptype=false; // p
	    dbxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(dbxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) {    // b is sx
	static std::vector<sPMchoice> dsxProcesses;
	if (dsxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    dsxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxstepveuxggs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    PMtemp.MomShuffleVector.push_back(12);
	    PMtemp.MomShuffleVector.push_back(13);
	    PMtemp.MomShuffleVector.push_back(15);
	    PMtemp.MomShuffleVector.push_back(18);
	    PMtemp.MomShuffleVector.push_back(19);
	    PMtemp.MomShuffleVector.push_back(21);
	    dsxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	    PMtemp.ptype=false; // p
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    dsxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	    PMtemp.ptype=false; // p
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveussxcx_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	    dsxProcesses.push_back(PMtemp);
	  }
	}      
	weight*=PickAndEvalProcAndMom4j(dsxProcesses,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> dcxProcesses;
	if (dcxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    dcxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxstepveuxggs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    PMtemp.MomShuffleVector.push_back(12);
	    PMtemp.MomShuffleVector.push_back(13);
	    PMtemp.MomShuffleVector.push_back(15);
	    PMtemp.MomShuffleVector.push_back(18);
	    PMtemp.MomShuffleVector.push_back(19);
	    PMtemp.MomShuffleVector.push_back(21);
	    dcxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    PMtemp.ptype=false; // p
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	    dcxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveussxcx_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	    dcxProcesses.push_back(PMtemp);
	  }
	}      
	weight*=PickAndEvalProcAndMom4j(dcxProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==2) { // MARKER a=u //
      if (bptype==1) {  // b=d
	static std::vector<sPMchoice> udProcesses;
	if (udProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepvedggd_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(39);
	  udProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepvedggd_;
	    PMtemp.idpartmult=4;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(15);
	    udProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepvedduux_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24; ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    udProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudtepveddssx_;
	    PMtemp.idpartmult=6;
	    udProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixudtepvedduux_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=0; ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=9;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    udProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudtepveddssx_;
	    PMtemp.idpartmult=6;
	    udProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepveudscx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    udProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepveudscx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=0;ii<24;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    udProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepveddddx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(24);
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(33);
	  udProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepveddddx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(4);
	  PMtemp.MomShuffleVector.push_back(9);
	  udProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(udProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==5) {
	static std::vector<sPMchoice> ubProcesses;
	if (ubProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  ubProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixustepvedddxs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ubProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvesssxd_;
	    ubProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvessucx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=0;ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ubProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvedsbbx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ubProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvedsuux_;
	    PMtemp.idpartmult=1;
	    ubProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixubtepveubscx_;
	    ubProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    ubProcesses.push_back(PMtemp);
	  }
	}	
	weight*=PickAndEvalProcAndMom4j(ubProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==3) {
	static std::vector<sPMchoice> usProcesses;
	if (usProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  usProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  usProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // p
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  usProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // q
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  usProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  usProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  usProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(usProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype>0) {   
	static std::vector<sPMchoice> ucProcesses;
	if (ucProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  ucProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  ucProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // p
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  ucProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  ucProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  PMtemp.ptype=true; // q
	  ucProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  PMtemp.ptype=true; // q
	  ucProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(ucProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) {
	static std::vector<sPMchoice> udxPrcesses;
	if (udxPrcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepvedggdx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveuggux_;
	    udxPrcesses.push_back(PMtemp);

	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepvedddxdx_;
	    PMtemp.idpartmult=4;
	    PMtemp.ptype=false; // p
	    PMtemp.MomShuffleVector.push_back(24);
	    PMtemp.MomShuffleVector.push_back(26);
	    PMtemp.MomShuffleVector.push_back(28);
	    PMtemp.MomShuffleVector.push_back(32);
	    PMtemp.MomShuffleVector.push_back(34);
	    PMtemp.MomShuffleVector.push_back(36);
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveuuuxux_;
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixduxtepvedduxux_;
	    PMtemp.ptype=true; // q
	    udxPrcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepveddxuux_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveudxscx_;
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveddxssx_;
	    PMtemp.idpartmult=3;
	    PMtemp.ptype=false; // p
	    udxPrcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixudxtepveuuxssx_;
	    udxPrcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(udxPrcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-5) {
	static std::vector<sPMchoice> ubxProcesses;
	if (ubxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    ubxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ubxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    ubxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    PMtemp.ptype=true; // q
	    ubxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ubxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    ubxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixubxtepveubxscx_;
	    ubxProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    ubxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(ubxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) {    // b is sx
	static std::vector<sPMchoice> usxProcesses;
	if (usxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    usxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxstepveuxggs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    PMtemp.MomShuffleVector.push_back(12);
	    PMtemp.MomShuffleVector.push_back(13);
	    PMtemp.MomShuffleVector.push_back(15);
	    PMtemp.MomShuffleVector.push_back(18);
	    PMtemp.MomShuffleVector.push_back(19);
	    PMtemp.MomShuffleVector.push_back(21);
	    usxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	    PMtemp.ptype=false; // p
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    usxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	    PMtemp.ptype=false; // p
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveussxcx_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	    usxProcesses.push_back(PMtemp);
	  }
	}      
	weight*=PickAndEvalProcAndMom4j(usxProcesses,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> ucxProcesses;
	if (ucxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    ucxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxstepveuxggs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    PMtemp.MomShuffleVector.push_back(12);
	    PMtemp.MomShuffleVector.push_back(13);
	    PMtemp.MomShuffleVector.push_back(15);
	    PMtemp.MomShuffleVector.push_back(18);
	    PMtemp.MomShuffleVector.push_back(19);
	    PMtemp.MomShuffleVector.push_back(21);
	    ucxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    PMtemp.ptype=false; // p
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	    ucxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveussxcx_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	    ucxProcesses.push_back(PMtemp);
	  }
	}      
	weight*=PickAndEvalProcAndMom4j(ucxProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==3) { // a=s
      if (bptype==4) {  // b=c
	static std::vector<sPMchoice> duProcesses;
	if (duProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepvedggd_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(15);
	  duProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepvedggd_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(39);
	  duProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixudtepvedduux_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=0; ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=9;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    duProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudtepveddssx_;
	    PMtemp.idpartmult=6;
	    duProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixudtepvedduux_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24; ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    duProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudtepveddssx_;
	    PMtemp.idpartmult=6;
	    duProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepveudscx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=false; // p
	    for (int ii=0;ii<24;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    duProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepveudscx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    duProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepveddddx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(4);
	  PMtemp.MomShuffleVector.push_back(9);
	  duProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepveddddx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(24);
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(33);
	  duProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(duProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==5) {
	static std::vector<sPMchoice> dbProcesses;
	if (dbProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dbProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixustepvedddxs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dbProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvesssxd_;
	    dbProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvessucx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=0;ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dbProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvedsbbx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dbProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvedsuux_;
	    PMtemp.idpartmult=1;
	    dbProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixubtepveubscx_;
	    dbProcesses.push_back(PMtemp);

	    PMtemp.ptype=false; // p
	    dbProcesses.push_back(PMtemp);
	  }
	}	
	weight*=PickAndEvalProcAndMom4j(dbProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==1) {
	static std::vector<sPMchoice> dsProcesses;
	if (dsProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dsProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dsProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // q
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  dsProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // q
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  dsProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  PMtemp.ptype=false; // p
	  dsProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  dsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  PMtemp.ptype=false; // p
	  dsProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(dsProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype>0) {   // b=up
	static std::vector<sPMchoice> dcProcesses;
	if (dcProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dcProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dcProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  dcProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // q
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  dcProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  dcProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  dcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  dcProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(dcProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-4) {
	static std::vector<sPMchoice> duxProcesses;
	if (duxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepvedggdx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    duxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveuggux_;
	    duxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepvedddxdx_;
	    PMtemp.idpartmult=4;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(24);
	    PMtemp.MomShuffleVector.push_back(26);
	    PMtemp.MomShuffleVector.push_back(28);
	    PMtemp.MomShuffleVector.push_back(32);
	    PMtemp.MomShuffleVector.push_back(34);
	    PMtemp.MomShuffleVector.push_back(36);
	    duxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveuuuxux_;
	    duxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixduxtepvedduxux_;
	    PMtemp.ptype=false; // p
	    duxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepveddxuux_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    duxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveudxscx_;
	    duxProcesses.push_back(PMtemp);

	    PMtemp.ptype=false; // p
	    duxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveddxssx_;
	    PMtemp.idpartmult=3;
	    PMtemp.ptype=true; // q
	    duxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixudxtepveuuxssx_;
	    duxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(duxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-5) {
	static std::vector<sPMchoice> dbxProcesses;
	if (dbxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    dbxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dbxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    dbxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    PMtemp.ptype=false; // p
	    dbxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dbxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    dbxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixubxtepveubxscx_;
	    dbxProcesses.push_back(PMtemp);

	    PMtemp.ptype=false; // p
	    dbxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(dbxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) {    // b is sx
	static std::vector<sPMchoice> dsxProcesses;
	if (dsxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    dsxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxstepveuxggs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    PMtemp.MomShuffleVector.push_back(12);
	    PMtemp.MomShuffleVector.push_back(13);
	    PMtemp.MomShuffleVector.push_back(15);
	    PMtemp.MomShuffleVector.push_back(18);
	    PMtemp.MomShuffleVector.push_back(19);
	    PMtemp.MomShuffleVector.push_back(21);
	    dsxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	    PMtemp.ptype=false; // p
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    dsxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	    PMtemp.ptype=false; // p
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveussxcx_;
	    dsxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	    dsxProcesses.push_back(PMtemp);
	  }
	}      
	weight*=PickAndEvalProcAndMom4j(dsxProcesses,pmg,&lepton,&alepton);
      }
      else {  //  b is ubar
	static std::vector<sPMchoice> dcxProcesses;
	if (dcxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    dcxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxstepveuxggs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    PMtemp.MomShuffleVector.push_back(12);
	    PMtemp.MomShuffleVector.push_back(13);
	    PMtemp.MomShuffleVector.push_back(15);
	    PMtemp.MomShuffleVector.push_back(18);
	    PMtemp.MomShuffleVector.push_back(19);
	    PMtemp.MomShuffleVector.push_back(21);
	    dcxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    PMtemp.ptype=false; // p
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	    dcxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveussxcx_;
	    dcxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	    dcxProcesses.push_back(PMtemp);
	  }
	}      
	weight*=PickAndEvalProcAndMom4j(dcxProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype==4) { // a=c
      if (bptype==3) {  // b=s
	static std::vector<sPMchoice> udProcesses;
	if (udProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepvedggd_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(39);
	  udProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepvedggd_;
	    PMtemp.idpartmult=4;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(15);
	    udProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepvedduux_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24; ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    udProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudtepveddssx_;
	    PMtemp.idpartmult=6;
	    udProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixudtepvedduux_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=0; ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=9;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    udProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudtepveddssx_;
	    PMtemp.idpartmult=6;
	    udProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepveudscx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    udProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudtepveudscx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=0;ii<24;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    udProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepveddddx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(24);
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(33);
	  udProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixudtepveddddx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(4);
	  PMtemp.MomShuffleVector.push_back(9);
	  udProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(udProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==5) {
	static std::vector<sPMchoice> ubProcesses;
	if (ubProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  ubProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixustepvedddxs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ubProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvesssxd_;
	    ubProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvessucx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=0;ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ubProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvedsbbx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ubProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvedsuux_;
	    PMtemp.idpartmult=1;
	    ubProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixubtepveubscx_;
	    ubProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    ubProcesses.push_back(PMtemp);
	  }
	}	
	weight*=PickAndEvalProcAndMom4j(ubProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==1) {
	static std::vector<sPMchoice> usProcesses;
	if (usProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  usProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  usProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // p
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  usProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // q
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  usProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  usProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  usProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  usProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(usProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype>0) {   // b=up
	static std::vector<sPMchoice> ucProcesses;
	if (ucProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  ucProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  ucProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // p
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  ucProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedsuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsccx_;
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvedsbbx_;
	  ucProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  PMtemp.ptype=true; // q
	  ucProcesses.push_back(PMtemp);
	  }
	  {
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixustepvedddxs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvesssxd_;
	  ucProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixustepvessucx_;
	  PMtemp.ptype=true; // q
	  ucProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(ucProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-3) {
	static std::vector<sPMchoice> udxPrcesses;
	if (udxPrcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepvedggdx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveuggux_;
	    udxPrcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepvedddxdx_;
	    PMtemp.idpartmult=4;
	    PMtemp.ptype=false; // p
	    PMtemp.MomShuffleVector.push_back(24);
	    PMtemp.MomShuffleVector.push_back(26);
	    PMtemp.MomShuffleVector.push_back(28);
	    PMtemp.MomShuffleVector.push_back(32);
	    PMtemp.MomShuffleVector.push_back(34);
	    PMtemp.MomShuffleVector.push_back(36);
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveuuuxux_;
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixduxtepvedduxux_;
	    PMtemp.ptype=true; // q
	    udxPrcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixudxtepveddxuux_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveudxscx_;
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    udxPrcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixudxtepveddxssx_;
	    PMtemp.idpartmult=3;
	    PMtemp.ptype=false; // p
	    udxPrcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixudxtepveuuxssx_;
	    udxPrcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(udxPrcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-5) {
	static std::vector<sPMchoice> ubxProcesses;
	if (ubxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    ubxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ubxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    ubxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    PMtemp.ptype=true; // q
	    ubxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ubxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    ubxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixubxtepveubxscx_;
	    ubxProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    ubxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(ubxProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1) {    // b is dx
	static std::vector<sPMchoice> usxProcesses;
	if (usxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    usxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxstepveuxggs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    PMtemp.MomShuffleVector.push_back(12);
	    PMtemp.MomShuffleVector.push_back(13);
	    PMtemp.MomShuffleVector.push_back(15);
	    PMtemp.MomShuffleVector.push_back(18);
	    PMtemp.MomShuffleVector.push_back(19);
	    PMtemp.MomShuffleVector.push_back(21);
	    usxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	    PMtemp.ptype=false; // p
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    usxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	    PMtemp.ptype=false; // p
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveussxcx_;
	    usxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	    usxProcesses.push_back(PMtemp);
	  }
	}      
	weight*=PickAndEvalProcAndMom4j(usxProcesses,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> ucxProcesses;
	if (ucxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedggsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(25);
	    PMtemp.MomShuffleVector.push_back(27);
	    PMtemp.MomShuffleVector.push_back(30);
	    PMtemp.MomShuffleVector.push_back(31);
	    PMtemp.MomShuffleVector.push_back(33);
	    PMtemp.MomShuffleVector.push_back(36);
	    PMtemp.MomShuffleVector.push_back(37);
	    PMtemp.MomShuffleVector.push_back(39);
	    PMtemp.MomShuffleVector.push_back(42);
	    PMtemp.MomShuffleVector.push_back(43);
	    PMtemp.MomShuffleVector.push_back(45);
	    ucxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxstepveuxggs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    PMtemp.MomShuffleVector.push_back(12);
	    PMtemp.MomShuffleVector.push_back(13);
	    PMtemp.MomShuffleVector.push_back(15);
	    PMtemp.MomShuffleVector.push_back(18);
	    PMtemp.MomShuffleVector.push_back(19);
	    PMtemp.MomShuffleVector.push_back(21);
	    ucxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	    PMtemp.ptype=false; // p
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	    ucxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<48;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveussxcx_;
	    ucxProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	    ucxProcesses.push_back(PMtemp);
	  }
	}      
	weight*=PickAndEvalProcAndMom4j(ucxProcesses,pmg,&lepton,&alepton);
      }
    }
    else if (aptype>0) { // a=b
      if (bptype==2 || bptype==4) {
	static std::vector<sPMchoice> buProcesses;
	if (buProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  buProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixustepvedddxs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=0;ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    buProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvesssxd_;
	    buProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvessucx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    buProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvedsbbx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=0;ii<24;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    buProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvedsuux_;
	    PMtemp.idpartmult=1;
	    buProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixubtepveubscx_;
	    buProcesses.push_back(PMtemp);

	    PMtemp.ptype=true; // q
	    buProcesses.push_back(PMtemp);
	  }
	}	
	weight*=PickAndEvalProcAndMom4j(buProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype>0) {
	static std::vector<sPMchoice> bdProcesses;
	if (bdProcesses.size()==0) {
	  {
	  struct sPMchoice PMtemp;

	  PMtemp.MGfunc=smatrixustepvedggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  bdProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;

	    PMtemp.MGfunc=smatrixustepvedddxs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=0;ii<6;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=8;ii<14;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    bdProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvesssxd_;
	    bdProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvessucx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24;ii<30;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32;ii<38;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    bdProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixustepvedsbbx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=0;ii<24;ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    bdProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixustepvedsuux_;
	    PMtemp.idpartmult=1;
	    bdProcesses.push_back(PMtemp);

	    PMtemp.MGfunc=smatrixubtepveubscx_;
	    bdProcesses.push_back(PMtemp);

	    PMtemp.ptype=false; // p
	    bdProcesses.push_back(PMtemp);
	  }
	}	
	weight*=PickAndEvalProcAndMom4j(bdProcesses,pmg,&lepton,&alepton);
      }
      else if (bptype==-1 || bptype==-3) {
	static std::vector<sPMchoice> bdxProcesses;
	
	if (bdxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxstepveuxggs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // process type p
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    PMtemp.MomShuffleVector.push_back(12);
	    PMtemp.MomShuffleVector.push_back(13);
	    PMtemp.MomShuffleVector.push_back(15);
	    PMtemp.MomShuffleVector.push_back(18);
	    PMtemp.MomShuffleVector.push_back(19);
	    PMtemp.MomShuffleVector.push_back(21);
	    bdxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=false; // p
	    for (int ii=24; ii<30; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32; ii<38; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    bdxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	    bdxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	    PMtemp.ptype=true; // q
	    bdxProcesses.push_back(PMtemp);	  
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveussxcx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=false; // p
	    for (int ii=24; ii<48; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    bdxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	    PMtemp.idpartmult=2;
	    bdxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixbdxtepvebdxscx_;
	    PMtemp.idpartmult=1;
	    bdxProcesses.push_back(PMtemp);	  
	    
	    PMtemp.ptype=true; //q
	    bdxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(bdxProcesses,pmg,&lepton,&alepton);
      }
      else {
	static std::vector<sPMchoice> buxProcesses;

	if (buxProcesses.size()==0) {
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixdxstepveuxggs_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // process type q
	    PMtemp.MomShuffleVector.push_back(1);
	    PMtemp.MomShuffleVector.push_back(3);
	    PMtemp.MomShuffleVector.push_back(6);
	    PMtemp.MomShuffleVector.push_back(7);
	    PMtemp.MomShuffleVector.push_back(9);
	    PMtemp.MomShuffleVector.push_back(12);
	    PMtemp.MomShuffleVector.push_back(13);
	    PMtemp.MomShuffleVector.push_back(15);
	    PMtemp.MomShuffleVector.push_back(18);
	    PMtemp.MomShuffleVector.push_back(19);
	    PMtemp.MomShuffleVector.push_back(21);
	    buxProcesses.push_back(PMtemp);
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	    PMtemp.idpartmult=2;
	    PMtemp.ptype=true; // q
	    for (int ii=24; ii<30; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    for (int ii=32; ii<38; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    buxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	    buxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	    PMtemp.ptype=false; // p
	    buxProcesses.push_back(PMtemp);	  
	  }
	  {
	    struct sPMchoice PMtemp;
	    PMtemp.MGfunc=smatrixusxtepveussxcx_;
	    PMtemp.idpartmult=1;
	    PMtemp.ptype=true; // q
	    for (int ii=24; ii<48; ii++)
	      PMtemp.MomShuffleVector.push_back(ii);
	    buxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	    PMtemp.idpartmult=2;
	    buxProcesses.push_back(PMtemp);
	    
	    PMtemp.MGfunc=smatrixbdxtepvebdxscx_;
	    PMtemp.idpartmult=1;
	    buxProcesses.push_back(PMtemp);	  
	    
	    PMtemp.ptype=false; // p
	    buxProcesses.push_back(PMtemp);
	  }
	}
	weight*=PickAndEvalProcAndMom4j(buxProcesses,pmg,&lepton,&alepton);
      }
    }
  // //   // ///  Now go for a = anti-quark   --- MARKER a=qbar  ///
  else if (aptype==-1) {    // a=dbar, b!=d
    if (bptype==-2) {
      static std::vector<sPMchoice> dxuxProcesses;
      if (dxuxProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxggux_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(39);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxggux_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=true; // process type q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(15);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxuxu_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(24);
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(33);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxuxu_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // process type q
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(9);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxddx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxddx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxssx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxssx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // process type q
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepvedxuxcxs_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	  
	  PMtemp.MGfunc=smatrixdxuxtepvedxuxcxs_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // process type q
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxuxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==-5) {
      static std::vector<sPMchoice> dxbxProcesses;

      if (dxbxProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxbxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24; ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
	  dxbxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24; ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
	  PMtemp.idpartmult=1;
	  dxbxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=24; ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxbxtepvedxbxscx_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24; ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbxProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // process type q
	  dxbxProcesses.push_back(PMtemp);
	}
	
      }
      weight*=PickAndEvalProcAndMom4j(dxbxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==-3) {
      static std::vector<sPMchoice> dxsxProcesses;

      if (dxsxProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxsxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dxsxProcesses.push_back(PMtemp);
	}
      	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
	  dxsxProcesses.push_back(PMtemp);
	}
      	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
	  dxsxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24; ii<48; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);
	  //  Do ones with identical mom configs etc together
	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  dxsxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
	  dxsxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);
	  //  Do ones with identical mom configs etc together
	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  dxsxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
	  dxsxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=9; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxsxProcesses,pmg,&lepton,&alepton);;
    }
    else if (bptype<0) {   // b=cx
      static std::vector<sPMchoice> dxcxProcesses;

      if (dxcxProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxcxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dxcxProcesses.push_back(PMtemp);
	}
      	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
	  dxcxProcesses.push_back(PMtemp);
	}
      	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
	  dxcxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // process type q
	  for (int ii=24; ii<48; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);
	  //  Do ones with identical mom configs etc together
	  PMtemp.ptype=false; // process type p
	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  dxcxProcesses.push_back(PMtemp);
	  
	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
	  dxcxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);
	  //  Do ones with identical mom configs etc together
	  PMtemp.ptype=true; // process type q
	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  dxcxProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=false; // process type p
	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
	  dxcxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=9; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxcxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==2) {
      static std::vector<sPMchoice> dxuProcesses;

      if (dxuProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixudxtepvedggdx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixudxtepveuggux_;
	  dxuProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixudxtepvedddxdx_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(2);
	  PMtemp.MomShuffleVector.push_back(4);
	  PMtemp.MomShuffleVector.push_back(8);
	  PMtemp.MomShuffleVector.push_back(10);
	  PMtemp.MomShuffleVector.push_back(12);
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixudxtepveuuuxux_;
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixduxtepvedduxux_;
	  PMtemp.ptype=true; // process type q
	  dxuProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixudxtepveddxuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=0; ii< 24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuProcesses.push_back(PMtemp);
	  //  Include the other p1--p24 processes
	  PMtemp.MGfunc=smatrixudxtepveddxssx_;
	  PMtemp.idpartmult=3;
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixudxtepveuuxssx_;
	  PMtemp.idpartmult=3;
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixudxtepveudxscx_;
	  PMtemp.idpartmult=1;
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // process type q
	  dxuProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxuProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==3) {
      static std::vector<sPMchoice> dxsProcesses;

      if (dxsProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepvedggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dxsProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	  PMtemp.ptype=false; // p
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	  dxsProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	  PMtemp.ptype=true; // q
	  PMtemp.idpartmult=1;
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsProcesses.push_back(PMtemp);
	  
	  PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	  PMtemp.ptype=false; // p
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	  dxsProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxsProcesses.push_back(PMtemp);	  
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxsProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==4) {  // dx c
      static std::vector<sPMchoice> dxcProcesses;

      if (dxcProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepvedggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dxcProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcProcesses.push_back(PMtemp);
	  // other inverted 25-30, 33-38 processes
	  PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	  PMtemp.ptype=true; // q
	  dxcProcesses.push_back(PMtemp);
	  
	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	  dxcProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false;  // p
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcProcesses.push_back(PMtemp);
	  //  other inverted 25--48 processes
	  PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	  dxcProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxcProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxcProcesses,pmg,&lepton,&alepton);
    }
    else { // dx b
      static std::vector<sPMchoice> dxbProcesses;

      if (dxbProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxbProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	  dxbProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	  PMtemp.ptype=true; // q
	  dxbProcesses.push_back(PMtemp);	  
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // p
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	  PMtemp.idpartmult=2;
	  dxbProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixbdxtepvebdxscx_;
	  PMtemp.idpartmult=1;
	  dxbProcesses.push_back(PMtemp);	  

	  PMtemp.ptype=true; //q
	  dxbProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxbProcesses,pmg,&lepton,&alepton);
    }
  }
  else if (aptype==-2) {    // a=ubar, b!=u
    if (bptype==-1) {
      static std::vector<sPMchoice> uxdxProcesses;
      if (uxdxProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxggux_;
    	  PMtemp.idpartmult=4;
    	  PMtemp.ptype=false; // process type p
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxggux_;
    	  PMtemp.idpartmult=4;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxuxu_;
    	  PMtemp.idpartmult=6;
    	  PMtemp.ptype=false; // process type p
    	  PMtemp.MomShuffleVector.push_back(0);
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxuxu_;
    	  PMtemp.idpartmult=6;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(24);
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxddx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=0;ii<6;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8;ii<14;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxddx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24;ii<30;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=32;ii<38;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxssx_;
    	  PMtemp.idpartmult=6;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=0;ii<6;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8;ii<14;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxssx_;
    	  PMtemp.idpartmult=6;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24;ii<30;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=32;ii<38;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepvedxuxcxs_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=0;ii<24;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
	  
    	  PMtemp.MGfunc=smatrixdxuxtepvedxuxcxs_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24;ii<48;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxdxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==-5) {
      static std::vector<sPMchoice> uxbxProcesses;

      if (uxbxProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  PMtemp.MomShuffleVector.push_back(36);
    	  PMtemp.MomShuffleVector.push_back(37);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  PMtemp.MomShuffleVector.push_back(42);
    	  PMtemp.MomShuffleVector.push_back(43);
    	  PMtemp.MomShuffleVector.push_back(45);
    	  uxbxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24; ii<30;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=32; ii<38; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxbxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
    	  uxbxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=false; // p
	  uxbxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24; ii<48;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxbxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
    	  PMtemp.idpartmult=1;
    	  uxbxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixdxbxtepvedxbxscx_;
     	  PMtemp.ptype=false; // process type p
     	  uxbxProcesses.push_back(PMtemp);

     	  PMtemp.ptype=true; // process type q
     	  uxbxProcesses.push_back(PMtemp);
     	}
      }
      weight*=PickAndEvalProcAndMom4j(uxbxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==-3) {
      static std::vector<sPMchoice> uxsxProcesses;

      if (uxsxProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  PMtemp.MomShuffleVector.push_back(36);
    	  PMtemp.MomShuffleVector.push_back(37);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  PMtemp.MomShuffleVector.push_back(42);
    	  PMtemp.MomShuffleVector.push_back(43);
    	  PMtemp.MomShuffleVector.push_back(45);
    	  uxsxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // process type p
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxsxProcesses.push_back(PMtemp);
    	}
      	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24; ii<30; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=32; ii<38; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
    	  uxsxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=false; // p
	  uxsxProcesses.push_back(PMtemp);
    	}
      	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=0; ii<6; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
    	  uxsxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=true; // q
    	  uxsxProcesses.push_back(PMtemp);	  
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=24; ii<48; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsxProcesses.push_back(PMtemp);
    	  //  Do ones with identical mom configs etc together
    	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  PMtemp.ptype=true; // q
    	  uxsxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
    	  uxsxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii<24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsxProcesses.push_back(PMtemp);
    	  //  Do ones with identical mom configs etc together
    	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  PMtemp.ptype=false; // p
    	  uxsxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
    	  uxsxProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxsxProcesses,pmg,&lepton,&alepton);;
    }
    else if (bptype<0) {   // b=cx
      static std::vector<sPMchoice> uxcxProcesses;

      if (uxcxProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  PMtemp.MomShuffleVector.push_back(36);
    	  PMtemp.MomShuffleVector.push_back(37);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  PMtemp.MomShuffleVector.push_back(42);
    	  PMtemp.MomShuffleVector.push_back(43);
    	  PMtemp.MomShuffleVector.push_back(45);
    	  uxcxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxcxProcesses.push_back(PMtemp);
    	}
      	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24; ii<30; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=32; ii<38; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
    	  uxcxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=false; // p
	  uxcxProcesses.push_back(PMtemp);
    	}
      	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii<6; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
    	  uxcxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=false; // p
	  uxcxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24; ii<48; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcxProcesses.push_back(PMtemp);
    	  //  Do ones with identical mom configs etc together
    	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
    	  uxcxProcesses.push_back(PMtemp);
	  
    	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
    	  uxcxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii<24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcxProcesses.push_back(PMtemp);
    	  //  Do ones with identical mom configs etc together
    	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
    	  uxcxProcesses.push_back(PMtemp);
	  
    	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
    	  uxcxProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxcxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==1) {
      static std::vector<sPMchoice> uxdProcesses;

      if (uxdProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixudxtepvedggdx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixudxtepveuggux_;
    	  uxdProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixudxtepvedddxdx_;
    	  PMtemp.idpartmult=4;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(0);
    	  PMtemp.MomShuffleVector.push_back(2);
    	  PMtemp.MomShuffleVector.push_back(4);
    	  PMtemp.MomShuffleVector.push_back(8);
    	  PMtemp.MomShuffleVector.push_back(10);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixudxtepveuuuxux_;
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixduxtepvedduxux_;
    	  PMtemp.ptype=false; // process type p
    	  uxdProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixudxtepveddxuux_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii< 24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdProcesses.push_back(PMtemp);
    	  //  Include the other p1--p24 processes
    	  PMtemp.MGfunc=smatrixudxtepveddxssx_;
    	  PMtemp.idpartmult=3;
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixudxtepveuuxssx_;
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixudxtepveudxscx_;
    	  PMtemp.idpartmult=1;
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.ptype=false; // process type p
    	  uxdProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxdProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==3) {
      static std::vector<sPMchoice> uxsProcesses;
      
      if (uxsProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxsProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedddxsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // q
    	  for (int ii=0; ii<6; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
    	  PMtemp.ptype=false; // p
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
    	  uxsProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepveduuxsx_;
    	  PMtemp.ptype=true; // q
    	  PMtemp.idpartmult=1;
    	  for (int ii=0; ii<24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsProcesses.push_back(PMtemp);
	  
    	  PMtemp.MGfunc=smatrixusxtepvedsxccx_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveuddxcx_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
    	  uxsProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  PMtemp.MomShuffleVector.push_back(36);
    	  PMtemp.MomShuffleVector.push_back(37);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  PMtemp.MomShuffleVector.push_back(42);
    	  PMtemp.MomShuffleVector.push_back(43);
    	  PMtemp.MomShuffleVector.push_back(45);
    	  uxsProcesses.push_back(PMtemp);	  
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxsProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==4) {  // ux c
      static std::vector<sPMchoice> uxcProcesses;

      if (uxcProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // process type p
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxcProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedddxsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // p
    	  for (int ii=0; ii<6; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcProcesses.push_back(PMtemp);
    	  // other inverted 25-30, 33-38 processes
    	  PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
    	  uxcProcesses.push_back(PMtemp);
	  
    	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
    	  PMtemp.ptype=true; // q
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
    	  uxcProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepveduuxsx_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=false;  // p
    	  for (int ii=0; ii<24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcProcesses.push_back(PMtemp);
    	  //  other inverted 25--48 processes
    	  PMtemp.MGfunc=smatrixusxtepvedsxccx_;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	  PMtemp.ptype=true;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
    	  uxcProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  PMtemp.MomShuffleVector.push_back(36);
    	  PMtemp.MomShuffleVector.push_back(37);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  PMtemp.MomShuffleVector.push_back(42);
    	  PMtemp.MomShuffleVector.push_back(43);
    	  PMtemp.MomShuffleVector.push_back(45);
    	  uxcProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxcProcesses,pmg,&lepton,&alepton);
    }
    else { // ux b
      static std::vector<sPMchoice> uxbProcesses;

      if (uxbProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxbProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // q
    	  for (int ii=0; ii<6; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxbProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
    	  uxbProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
    	  PMtemp.ptype=false; // p
    	  uxbProcesses.push_back(PMtemp);	  
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // q
    	  for (int ii=0; ii<24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxbProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
    	  PMtemp.idpartmult=2;
    	  uxbProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixbdxtepvebdxscx_;
    	  PMtemp.idpartmult=1;
    	  uxbProcesses.push_back(PMtemp);	  

    	  PMtemp.ptype=false; // p
    	  uxbProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxbProcesses,pmg,&lepton,&alepton);
    }
  }
  else if (aptype==-3) {    // a=sbar, b!=u
    if (bptype==-4) {
      static std::vector<sPMchoice> dxuxProcesses;
      if (dxuxProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxggux_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(39);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxggux_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=true; // process type q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(15);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxuxu_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(24);
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(33);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxuxu_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // process type q
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(9);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxddx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxddx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxssx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24;ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32;ii<38;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepveuxuxssx_;
	  PMtemp.idpartmult=6;
	  PMtemp.ptype=true; // process type q
	  for (int ii=0;ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8;ii<14;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxuxtepvedxuxcxs_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	  
	  PMtemp.MGfunc=smatrixdxuxtepvedxuxcxs_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // process type q
	  for (int ii=24;ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxuxProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxuxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==-5) {
      static std::vector<sPMchoice> dxbxProcesses;

      if (dxbxProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxbxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24; ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
	  dxbxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24; ii<48;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
	  PMtemp.idpartmult=2;
	  dxbxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=24; ii<30;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxbxtepvedxbxscx_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=0; ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbxProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // process type q
	  dxbxProcesses.push_back(PMtemp);
	}
	
      }
      weight*=PickAndEvalProcAndMom4j(dxbxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==-1) {
      static std::vector<sPMchoice> dxsxProcesses;

      if (dxsxProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxsxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dxsxProcesses.push_back(PMtemp);
	}
      	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
	  dxsxProcesses.push_back(PMtemp);
	}
      	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
	  dxsxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24; ii<48; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);
	  //  Do ones with identical mom configs etc together
	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  dxsxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
	  dxsxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);
	  //  Do ones with identical mom configs etc together
	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  dxsxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
	  dxsxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=9; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsxProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxsxProcesses,pmg,&lepton,&alepton);;
    }
    else if (bptype<0) {   // b=ux
      static std::vector<sPMchoice> dxcxProcesses;

      if (dxcxProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxcxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dxcxProcesses.push_back(PMtemp);
	}
      	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
	  dxcxProcesses.push_back(PMtemp);
	}
      	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
	  dxcxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=true; // process type q
	  for (int ii=24; ii<48; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);
	  //  Do ones with identical mom configs etc together
	  PMtemp.ptype=false; // process type p
	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  dxcxProcesses.push_back(PMtemp);
	  
	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
	  dxcxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);
	  //  Do ones with identical mom configs etc together
	  PMtemp.ptype=true; // process type q
	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  dxcxProcesses.push_back(PMtemp);
	  
	  PMtemp.ptype=false; // process type p
	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
	  dxcxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=24; ii<30; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=32; ii<38; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=9; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcxProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxcxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==4) {
      static std::vector<sPMchoice> dxuProcesses;

      if (dxuProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixudxtepvedggdx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixudxtepveuggux_;
	  dxuProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixudxtepvedddxdx_;
	  PMtemp.idpartmult=4;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(0);
	  PMtemp.MomShuffleVector.push_back(2);
	  PMtemp.MomShuffleVector.push_back(4);
	  PMtemp.MomShuffleVector.push_back(8);
	  PMtemp.MomShuffleVector.push_back(10);
	  PMtemp.MomShuffleVector.push_back(12);
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixudxtepveuuuxux_;
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixduxtepvedduxux_;
	  PMtemp.ptype=true; // process type q
	  dxuProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixudxtepveddxuux_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // process type p
	  for (int ii=0; ii< 24; ii++)
	    PMtemp.MomShuffleVector.push_back(0);
	  dxuProcesses.push_back(PMtemp);
	  //  Include the other p1--p24 processes
	  PMtemp.MGfunc=smatrixudxtepveddxssx_;
	  PMtemp.idpartmult=3;
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixudxtepveuuxssx_;
	  PMtemp.idpartmult=3;
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixudxtepveudxscx_;
	  PMtemp.idpartmult=1;
	  dxuProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // process type q
	  dxuProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxuProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==1) {
      static std::vector<sPMchoice> dxsProcesses;

      if (dxsProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepvedggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // process type q
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dxsProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=true; // q
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	  PMtemp.ptype=false; // p
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	  dxsProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	  PMtemp.ptype=true; // q
	  PMtemp.idpartmult=1;
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxsProcesses.push_back(PMtemp);
	  
	  PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	  PMtemp.ptype=false; // p
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
	  dxsProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	  dxsProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxsProcesses.push_back(PMtemp);	  
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxsProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==2) {  // dx c
      static std::vector<sPMchoice> dxcProcesses;

      if (dxcProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepvedggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  dxcProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepvedddxsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcProcesses.push_back(PMtemp);
	  // other inverted 25-30, 33-38 processes
	  PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	  PMtemp.ptype=true; // q
	  dxcProcesses.push_back(PMtemp);
	  
	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	  dxcProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepveduuxsx_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false;  // p
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxcProcesses.push_back(PMtemp);
	  //  other inverted 25--48 processes
	  PMtemp.MGfunc=smatrixusxtepvedsxccx_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
	  dxcProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	  dxcProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxcProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxcProcesses,pmg,&lepton,&alepton);
    }
    else { // sx b
      static std::vector<sPMchoice> dxbProcesses;

      if (dxbProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(25);
	  PMtemp.MomShuffleVector.push_back(27);
	  PMtemp.MomShuffleVector.push_back(30);
	  PMtemp.MomShuffleVector.push_back(31);
	  PMtemp.MomShuffleVector.push_back(33);
	  PMtemp.MomShuffleVector.push_back(36);
	  PMtemp.MomShuffleVector.push_back(37);
	  PMtemp.MomShuffleVector.push_back(39);
	  PMtemp.MomShuffleVector.push_back(42);
	  PMtemp.MomShuffleVector.push_back(43);
	  PMtemp.MomShuffleVector.push_back(45);
	  dxbProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // p
	  for (int ii=0; ii<6; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
	  dxbProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
	  PMtemp.ptype=true; // q
	  dxbProcesses.push_back(PMtemp);	  
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
	  PMtemp.idpartmult=1;
	  PMtemp.ptype=false; // p
	  for (int ii=0; ii<24; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  dxbProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
	  PMtemp.idpartmult=2;
	  dxbProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixbdxtepvebdxscx_;
	  PMtemp.idpartmult=1;
	  dxbProcesses.push_back(PMtemp);	  

	  PMtemp.ptype=true; //q
	  dxbProcesses.push_back(PMtemp);
	}
      }
      weight*=PickAndEvalProcAndMom4j(dxbProcesses,pmg,&lepton,&alepton);
    }
  }
  else if (aptype==-4) {    // a=cbar, b!=c
    if (bptype==-3) {
      static std::vector<sPMchoice> uxdxProcesses;  // cxsx=uxdx
      if (uxdxProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxggux_;
    	  PMtemp.idpartmult=4;
    	  PMtemp.ptype=false; // process type p
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxggux_;
    	  PMtemp.idpartmult=4;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxuxu_;
    	  PMtemp.idpartmult=6;
    	  PMtemp.ptype=false; // process type p
    	  PMtemp.MomShuffleVector.push_back(0);
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxuxu_;
    	  PMtemp.idpartmult=6;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(24);
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxddx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=0;ii<6;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8;ii<14;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxddx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24;ii<30;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=32;ii<38;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxssx_;
    	  PMtemp.idpartmult=6;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=0;ii<6;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8;ii<14;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepveuxuxssx_;
    	  PMtemp.idpartmult=6;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24;ii<30;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=32;ii<38;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxuxtepvedxuxcxs_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=0;ii<24;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
	  
    	  PMtemp.MGfunc=smatrixdxuxtepvedxuxcxs_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24;ii<48;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdxProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxdxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==-5) {
      static std::vector<sPMchoice> uxbxProcesses;

      if (uxbxProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  PMtemp.MomShuffleVector.push_back(36);
    	  PMtemp.MomShuffleVector.push_back(37);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  PMtemp.MomShuffleVector.push_back(42);
    	  PMtemp.MomShuffleVector.push_back(43);
    	  PMtemp.MomShuffleVector.push_back(45);
    	  uxbxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24; ii<30;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=32; ii<38; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxbxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
    	  uxbxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=false; // p
	  uxbxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24; ii<48;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxbxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
    	  PMtemp.idpartmult=1;
    	  uxbxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixdxbxtepvedxbxscx_;
     	  PMtemp.ptype=false; // process type p
     	  uxbxProcesses.push_back(PMtemp);

     	  PMtemp.ptype=true; // process type q
     	  uxbxProcesses.push_back(PMtemp);
     	}
      }
      weight*=PickAndEvalProcAndMom4j(uxbxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==-1) {
      static std::vector<sPMchoice> uxsxProcesses;

      if (uxsxProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  PMtemp.MomShuffleVector.push_back(36);
    	  PMtemp.MomShuffleVector.push_back(37);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  PMtemp.MomShuffleVector.push_back(42);
    	  PMtemp.MomShuffleVector.push_back(43);
    	  PMtemp.MomShuffleVector.push_back(45);
    	  uxsxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // process type p
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxsxProcesses.push_back(PMtemp);
    	}
      	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24; ii<30; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=32; ii<38; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
    	  uxsxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=false; // p
	  uxsxProcesses.push_back(PMtemp);
    	}
      	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=0; ii<6; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
    	  uxsxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=true; // q
    	  uxsxProcesses.push_back(PMtemp);	  
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=24; ii<48; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsxProcesses.push_back(PMtemp);
    	  //  Do ones with identical mom configs etc together
    	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  PMtemp.ptype=true; // q
    	  uxsxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
    	  uxsxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii<24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsxProcesses.push_back(PMtemp);
    	  //  Do ones with identical mom configs etc together
    	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  PMtemp.ptype=false; // p
    	  uxsxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
    	  uxsxProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxsxProcesses,pmg,&lepton,&alepton);;
    }
    else if (bptype<0) {   // b=ux
      static std::vector<sPMchoice> uxcxProcesses;

      if (uxcxProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  PMtemp.MomShuffleVector.push_back(36);
    	  PMtemp.MomShuffleVector.push_back(37);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  PMtemp.MomShuffleVector.push_back(42);
    	  PMtemp.MomShuffleVector.push_back(43);
    	  PMtemp.MomShuffleVector.push_back(45);
    	  uxcxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxcxProcesses.push_back(PMtemp);
    	}
      	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24; ii<30; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=32; ii<38; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
    	  uxcxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=false; // p
	  uxcxProcesses.push_back(PMtemp);
    	}
      	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii<6; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
    	  uxcxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=false; // p
	  uxcxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=24; ii<48; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcxProcesses.push_back(PMtemp);
    	  //  Do ones with identical mom configs etc together
    	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
    	  uxcxProcesses.push_back(PMtemp);
	  
    	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
    	  uxcxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepvedxcxuux_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii<24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcxProcesses.push_back(PMtemp);
    	  //  Do ones with identical mom configs etc together
    	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
    	  uxcxProcesses.push_back(PMtemp);
	  
    	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
    	  uxcxProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxcxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==3) {
      static std::vector<sPMchoice> uxdProcesses;

      if (uxdProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixudxtepvedggdx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixudxtepveuggux_;
    	  uxdProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixudxtepvedddxdx_;
    	  PMtemp.idpartmult=4;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(0);
    	  PMtemp.MomShuffleVector.push_back(2);
    	  PMtemp.MomShuffleVector.push_back(4);
    	  PMtemp.MomShuffleVector.push_back(8);
    	  PMtemp.MomShuffleVector.push_back(10);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixudxtepveuuuxux_;
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixduxtepvedduxux_;
    	  PMtemp.ptype=false; // process type p
    	  uxdProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixudxtepveddxuux_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii< 24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxdProcesses.push_back(PMtemp);
    	  //  Include the other p1--p24 processes
    	  PMtemp.MGfunc=smatrixudxtepveddxssx_;
    	  PMtemp.idpartmult=3;
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixudxtepveuuxssx_;
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixudxtepveudxscx_;
    	  PMtemp.idpartmult=1;
    	  uxdProcesses.push_back(PMtemp);

    	  PMtemp.ptype=false; // process type p
    	  uxdProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxdProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==1) {
      static std::vector<sPMchoice> uxsProcesses;
      
      if (uxsProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxsProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedddxsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // q
    	  for (int ii=0; ii<6; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
    	  PMtemp.ptype=false; // p
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
    	  uxsProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepveduuxsx_;
    	  PMtemp.ptype=true; // q
    	  PMtemp.idpartmult=1;
    	  for (int ii=0; ii<24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxsProcesses.push_back(PMtemp);
	  
    	  PMtemp.MGfunc=smatrixusxtepvedsxccx_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveuddxcx_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
    	  uxsProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
    	  uxsProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  PMtemp.MomShuffleVector.push_back(36);
    	  PMtemp.MomShuffleVector.push_back(37);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  PMtemp.MomShuffleVector.push_back(42);
    	  PMtemp.MomShuffleVector.push_back(43);
    	  PMtemp.MomShuffleVector.push_back(45);
    	  uxsProcesses.push_back(PMtemp);	  
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxsProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==2) {  // cx u
      static std::vector<sPMchoice> uxcProcesses;

      if (uxcProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // process type p
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxcProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedddxsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // p
    	  for (int ii=0; ii<6; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcProcesses.push_back(PMtemp);
    	  // other inverted 25-30, 33-38 processes
    	  PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
    	  uxcProcesses.push_back(PMtemp);
	  
    	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
    	  PMtemp.ptype=true; // q
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
    	  uxcProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepveduuxsx_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=false;  // p
    	  for (int ii=0; ii<24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxcProcesses.push_back(PMtemp);
    	  //  other inverted 25--48 processes
    	  PMtemp.MGfunc=smatrixusxtepvedsxccx_;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveuddxcx_;
	  PMtemp.ptype=true;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
    	  uxcProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
    	  uxcProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // q
    	  PMtemp.MomShuffleVector.push_back(25);
    	  PMtemp.MomShuffleVector.push_back(27);
    	  PMtemp.MomShuffleVector.push_back(30);
    	  PMtemp.MomShuffleVector.push_back(31);
    	  PMtemp.MomShuffleVector.push_back(33);
    	  PMtemp.MomShuffleVector.push_back(36);
    	  PMtemp.MomShuffleVector.push_back(37);
    	  PMtemp.MomShuffleVector.push_back(39);
    	  PMtemp.MomShuffleVector.push_back(42);
    	  PMtemp.MomShuffleVector.push_back(43);
    	  PMtemp.MomShuffleVector.push_back(45);
    	  uxcProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxcProcesses,pmg,&lepton,&alepton);
    }
    else { // cx b
      static std::vector<sPMchoice> uxbProcesses;

      if (uxbProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxstepveuxggs_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  uxbProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepveuuuxcx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // q
    	  for (int ii=0; ii<6; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxbProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvecxcxcu_;
    	  uxbProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixucxtepvecxcxus_;
    	  PMtemp.ptype=false; // p
    	  uxbProcesses.push_back(PMtemp);	  
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepveussxcx_;
    	  PMtemp.idpartmult=1;
    	  PMtemp.ptype=true; // q
    	  for (int ii=0; ii<24; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  uxbProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveucxbbx_;
    	  PMtemp.idpartmult=2;
    	  uxbProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixbdxtepvebdxscx_;
    	  PMtemp.idpartmult=1;
    	  uxbProcesses.push_back(PMtemp);	  

    	  PMtemp.ptype=false; // p
    	  uxbProcesses.push_back(PMtemp);
    	}
      }
      weight*=PickAndEvalProcAndMom4j(uxbProcesses,pmg,&lepton,&alepton);
    }
  }
  else if (aptype==-5) {    // a=bbar, b!=b
    if (bptype==-1 || bptype==-3) {
      static std::vector<sPMchoice> bxdxProcesses;

      if (bxdxProcesses.size()==0) {
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  PMtemp.MomShuffleVector.push_back(1);
	  PMtemp.MomShuffleVector.push_back(3);
	  PMtemp.MomShuffleVector.push_back(6);
	  PMtemp.MomShuffleVector.push_back(7);
	  PMtemp.MomShuffleVector.push_back(9);
	  PMtemp.MomShuffleVector.push_back(12);
	  PMtemp.MomShuffleVector.push_back(13);
	  PMtemp.MomShuffleVector.push_back(15);
	  PMtemp.MomShuffleVector.push_back(18);
	  PMtemp.MomShuffleVector.push_back(19);
	  PMtemp.MomShuffleVector.push_back(21);
	  bxdxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=0; ii<6;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  for (int ii=8; ii<14; ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  bxdxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
	  bxdxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=true; // q
	  bxdxProcesses.push_back(PMtemp);
	}
	{
	  struct sPMchoice PMtemp;
	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
	  PMtemp.idpartmult=2;
	  PMtemp.ptype=false; // process type p
	  for (int ii=0; ii<24;ii++)
	    PMtemp.MomShuffleVector.push_back(ii);
	  bxdxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
	  PMtemp.idpartmult=1;
	  bxdxProcesses.push_back(PMtemp);

	  PMtemp.MGfunc=smatrixdxbxtepvedxbxscx_;
	  bxdxProcesses.push_back(PMtemp);

	  PMtemp.ptype=true; // process type q
	  bxdxProcesses.push_back(PMtemp);
	}
	
      }
      weight*=PickAndEvalProcAndMom4j(bxdxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype<0) {
      static std::vector<sPMchoice> bxuxProcesses;

      if (bxuxProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  bxuxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxuxusx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii<6;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  bxuxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvesxsxuxs_;
    	  bxuxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixuxcxtepveuxuxcxd_;
	  PMtemp.ptype=false; // p
	  bxuxProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixdxsxtepveuxsxbbx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii<24;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  bxuxProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixdxsxtepvedxsxdux_;
    	  PMtemp.idpartmult=1;
    	  bxuxProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixdxbxtepvedxbxscx_;
     	  PMtemp.ptype=false; // process type p
     	  bxuxProcesses.push_back(PMtemp);

     	  PMtemp.ptype=true; // process type q
     	  bxuxProcesses.push_back(PMtemp);
     	}
      }
      weight*=PickAndEvalProcAndMom4j(bxuxProcesses,pmg,&lepton,&alepton);
    }
    else if (bptype==2 || bptype==4) {
      static std::vector<sPMchoice> bxuProcesses;

      if (bxuProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // process type p
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  bxuProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=0; ii<6;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  bxuProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvedddxsx_;
    	  bxuProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	  PMtemp.ptype=true; // q
	  bxuProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=false; // process type p
    	  for (int ii=0; ii<24;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  bxuProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveduuxsx_;
    	  PMtemp.idpartmult=1;
    	  bxuProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixubxtepveubxscx_;
     	  bxuProcesses.push_back(PMtemp);

     	  PMtemp.ptype=true; // process type q
     	  bxuProcesses.push_back(PMtemp);
     	}
      }
      weight*=PickAndEvalProcAndMom4j(bxuProcesses,pmg,&lepton,&alepton);
    }
    else {
      static std::vector<sPMchoice> bxdProcesses;

      if (bxdProcesses.size()==0) {
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedggsx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  PMtemp.MomShuffleVector.push_back(1);
    	  PMtemp.MomShuffleVector.push_back(3);
    	  PMtemp.MomShuffleVector.push_back(6);
    	  PMtemp.MomShuffleVector.push_back(7);
    	  PMtemp.MomShuffleVector.push_back(9);
    	  PMtemp.MomShuffleVector.push_back(12);
    	  PMtemp.MomShuffleVector.push_back(13);
    	  PMtemp.MomShuffleVector.push_back(15);
    	  PMtemp.MomShuffleVector.push_back(18);
    	  PMtemp.MomShuffleVector.push_back(19);
    	  PMtemp.MomShuffleVector.push_back(21);
    	  bxdProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvesxsxsd_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii<6;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  for (int ii=8; ii<14; ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  bxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepvedddxsx_;
    	  bxdProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixdsxtepvedduxsx_;
	  PMtemp.ptype=false; // p
	  bxdProcesses.push_back(PMtemp);
    	}
    	{
    	  struct sPMchoice PMtemp;
    	  PMtemp.MGfunc=smatrixusxtepvedsxbbx_;
    	  PMtemp.idpartmult=2;
    	  PMtemp.ptype=true; // process type q
    	  for (int ii=0; ii<24;ii++)
    	    PMtemp.MomShuffleVector.push_back(ii);
    	  bxdProcesses.push_back(PMtemp);

    	  PMtemp.MGfunc=smatrixusxtepveduuxsx_;
    	  PMtemp.idpartmult=1;
    	  bxdProcesses.push_back(PMtemp);

     	  PMtemp.MGfunc=smatrixubxtepveubxscx_;
     	  bxdProcesses.push_back(PMtemp);

     	  PMtemp.ptype=false; // process type p
     	  bxdProcesses.push_back(PMtemp);
     	}
      }
      weight*=PickAndEvalProcAndMom4j(bxdProcesses,pmg,&lepton,&alepton);
    }
  }
  else  
    std::cerr<<"Watch out - unanticipated parton configuration!"<<std::endl;
  }

  if (weight==1)
    std::cout<<"weight: "<<weight<<"  "<<aptype<<"  "<<bptype<<std::endl;

  // if (isnan(meout)) {
  //   std::cout<<"Getting nan: "<<std::endl;
  // }

  //  std::cout<<"b: "<<meout<<"  "<<idpartfact<<"\n";
  // weight*=meout*idpartfact;
  //  std::cout<<"s1 "<<idpartfact<<"  "<<meout<<"  "<<weight<<"  "<<pow(4.*M_PI*AlphaS(flags.scale),2)<<"  "<<aptype<<"  "<<bptype<<"  "<<status<<std::endl;
  // if (flags.scalesetting==0) //fixed scale
  //   weight*=pow(4.*M_PI*AlphaS(flags.scale),4);
  // else if (flags.scalesetting==1)
  //   weight*=pow(4.*M_PI*AlphaS(Ht),4);
  weight*=pow(4.*M_PI*Halphas(renscale),4);  
  if (flags.logcorrect==1)
    weight*=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(2).perp()/renscale)/(2.*M_PI));
  //  std::cout<<"s2: "<<weight<<"\n";
  return weight;
}
